header {* Hume specialisation of TLA rules  *}

theory Hume
imports HumeSemantics 
begin

section "The Scheduler"

datatype SchState = Execute
                  | Super

consts
  sch :: "state \<Rightarrow> SchState"

(* hint/heuristic/plan \<longrightarrow> show that
   isExe \<longrightarrow> all wires are unchanged OR set to Bot *)

constdefs
  isExe :: "temporal"
  "isExe \<equiv> TEMP $sch = #Execute"
  
  isSup :: "temporal"
  "isSup \<equiv> TEMP $sch = #Super"

  S :: "temporal"
  "S \<equiv> TEMP sch$ = (if isExe then #Super else #Execute)"

lemma SchStatevals: fixes v :: "'a \<Rightarrow> SchState"
  shows "v w \<noteq> Execute \<Longrightarrow> v w = Super"
  by (induct "v w") auto

lemma SchStatevals2: fixes v :: "'a \<Rightarrow> SchState"
  shows "v w = Execute \<Longrightarrow> v w \<noteq>  Super"
  by (induct "v w") auto

lemma supexe: "\<turnstile> isSup \<or> isExe"
  apply (int_unlift)
  apply (unfold isSup_def isExe_def)
  apply auto
  apply (simp add: SchStatevals)
  done

lemma supexe2: "\<turnstile> isSup \<longrightarrow> \<not>isExe"
  apply (int_unlift)
  apply (unfold isSup_def isExe_def)
  apply auto
  done

lemma supexe3: "\<turnstile> isExe \<longrightarrow> \<not>isSup"
  apply (int_unlift)
  apply (unfold isSup_def isExe_def)
  apply auto
  done

lemma SchStatevals3: fixes v :: "'a \<Rightarrow> SchState"
  shows "v w = Super \<Longrightarrow> v w \<noteq> Execute"
  by (induct "v w") auto

lemma supexe_next: "|~ isSup \<and> S \<longrightarrow> \<circ>isExe"
  apply (pref_unlift)
  apply (unfold isSup_def isExe_def S_def)
  apply (simp only: next_all_eq)
  apply (unfold after_def)
  apply auto
  done

lemma exesup_next: "|~ isExe \<and> S \<longrightarrow> \<circ>isSup"
  apply (pref_unlift)
  apply (unfold isSup_def isExe_def S_def)
  apply (simp only: next_all_eq)
  apply (unfold after_def)
  apply auto
  done

lemma nexesup_next: "|~ \<not>isExe \<and> S \<longrightarrow> \<circ>isExe"
  apply (pref_unlift)
  apply (unfold isSup_def isExe_def S_def)
  apply (simp only: next_all_eq)
  apply (unfold after_def)
  apply auto
  done

lemma nsupexe_next: "|~ \<not>isSup \<and> S \<longrightarrow> \<circ>isSup"
proof (pref_unlift,auto)
  fix w
  assume a1: "\<not>(w \<Turnstile> isSup)"
    and a2: "w \<Turnstile> S"
  from supexe[THEN intD,of w] have "w \<Turnstile> \<not>isSup \<longrightarrow> isExe"
    by auto
  with a1 have "w \<Turnstile> isExe"
    by auto
  with a2 show "w \<Turnstile> \<circ>isSup"
    apply (unfold isSup_def isExe_def S_def)
    apply (simp only: next_all_eq)
    apply (unfold after_def)
    apply auto
    done
qed

lemma supexe_unl: "(w \<Turnstile> isSup) \<or> (w \<Turnstile> isExe)"
proof -
  have "w \<Turnstile> isSup \<or> isExe" by (rule supexe[THEN intD])
  thus ?thesis by auto
qed

lemma schcases: assumes H1: "(w \<Turnstile> isExe) \<Longrightarrow> P"
  and H2: "(w \<Turnstile> isSup) \<Longrightarrow> P"
  shows "P"
proof (cases "w \<Turnstile> isExe")
  assume "w \<Turnstile> isExe" 
  with H1 show ?thesis by simp
next
  assume "\<not>(w \<Turnstile> isExe)"  
  with supexe_unl[of w] have "w \<Turnstile> isSup" by simp
  with H2 show ?thesis by simp 
qed

lemma infoften_exe: "\<turnstile> \<box>\<diamond>\<langle>S\<rangle>_v \<longrightarrow> \<box>\<diamond>(isExe)"
proof -
  have "|~ S \<longrightarrow> \<diamond>(isExe)"
  proof (pref_unlift,rule impI)
    fix w
    assume a1: "w \<Turnstile> S"
    have "\<turnstile> isSup \<or> isExe" by (rule supexe)
    from this[THEN intD]
    have "(w \<Turnstile> isSup) \<or> (w \<Turnstile> isExe)"
      by auto
    thus "w \<Turnstile> \<diamond>(isExe)"
    proof
      assume a2: "w \<Turnstile> isSup"
      with a1 have "w \<Turnstile> sch$ = #Execute"
	by (auto simp add: isSup_def isExe_def S_def)
      hence "w \<Turnstile> (\<circ>$sch) = #Execute"
	by (unfold after_def)
      hence "w \<Turnstile> \<circ>($sch = #Execute)"
	by (simp only: next_all_eq)
      hence "w \<Turnstile> \<circ>isExe"
	by (unfold isExe_def)
      with E23[THEN prefD,of isExe]
      show ?thesis by simp
    next
      assume a2: "w \<Turnstile> isExe"
      with E3[THEN intD,of isExe]
      show ?thesis by auto
    qed
  qed
  hence "\<turnstile> \<diamond>\<langle>S\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>isExe\<rangle>_v"
    by (rule AA7)
  from this E25[of "LIFT \<diamond>isExe" v]
  have "\<turnstile> \<diamond>\<langle>S\<rangle>_v \<longrightarrow> \<diamond>\<diamond>(isExe)"
    by (rule lift_imp_trans)
  with E10[of isExe,THEN inteq_reflection]
  have "\<turnstile> \<diamond>\<langle>S\<rangle>_v \<longrightarrow> \<diamond>(isExe)"
    by auto
  thus ?thesis by (rule STL4)
qed
  
lemma infoften_sup: "\<turnstile> \<box>\<diamond>\<langle>S\<rangle>_v \<longrightarrow> \<box>\<diamond>(isSup)"
proof -
  have "|~ S \<longrightarrow> \<diamond>(isSup)"
  proof (pref_unlift,rule impI)
    fix w
    assume a1: "w \<Turnstile> S"
    have "\<turnstile> isSup \<or> isExe" by (rule supexe)
    from this[THEN intD]
    have "(w \<Turnstile> isSup) \<or> (w \<Turnstile> isExe)"
      by auto
    thus "w \<Turnstile> \<diamond>(isSup)"
    proof
      assume a2: "w \<Turnstile> isExe"
      with a1 have "w \<Turnstile> sch$ = #Super"
	by (auto simp add: isSup_def isExe_def S_def)
      hence "w \<Turnstile> (\<circ>$sch) = #Super"
	by (unfold after_def)
      hence "w \<Turnstile> \<circ>($sch = #Super)"
	by (simp only: next_all_eq)
      hence "w \<Turnstile> \<circ>isSup"
	by (unfold isSup_def)
      with E23[THEN prefD,of isSup]
      show ?thesis by simp
    next
      assume a2: "w \<Turnstile> isSup"
      with E3[THEN intD,of isSup]
      show ?thesis by auto
    qed
  qed
  hence "\<turnstile> \<diamond>\<langle>S\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>isSup\<rangle>_v"
    by (rule AA7)
  from this E25[of "LIFT \<diamond>isSup" v]
  have "\<turnstile> \<diamond>\<langle>S\<rangle>_v \<longrightarrow> \<diamond>\<diamond>(isSup)"
    by (rule lift_imp_trans)
  with E10[of isSup,THEN inteq_reflection]
  have "\<turnstile> \<diamond>\<langle>S\<rangle>_v \<longrightarrow> \<diamond>(isSup)"
    by auto
  thus ?thesis by (rule STL4)
qed

(* should be fixed once I have better way to compute the statespace *)
lemma Enabled_Sact: assumes h1: "basevars sch"
  shows "\<turnstile> Enabled S"
proof (int_unlift)
  fix w
  show "w \<Turnstile> Enabled S"
  proof (unfold S_def,rule base_enabled)
    show "basevars sch" by (rule h1)
  next 
    show "\<exists>c. \<forall>u. LIFT sch (first u) = c \<longrightarrow> 
      ((first w ## u) \<Turnstile> sch$ = (if isExe then #Super else #Execute))"
    proof (rule exI,rule allI,intro impI)
      fix u
      assume aa1: "LIFT sch (first u) = (w \<Turnstile> (if isExe then #Super else #Execute))"
      thus "(first w ## u) \<Turnstile> sch$ = (if isExe then #Super else #Execute)"
	by (auto simp add: app_def first_def
	  before_def after_def nexts_def tail_def suffix_def isExe_def) 
    qed
  qed
qed

lemma Enabled_Sact: assumes h1: "basevars sch"
  shows "\<turnstile> Enabled (\<langle>S\<rangle>_sch)"
proof (int_unlift)
  fix w
  show "w \<Turnstile> Enabled (\<langle>S\<rangle>_sch)"
  proof (unfold S_def,rule base_enabled)
    show "basevars sch" by (rule h1)
  next 
    show "\<exists>c. \<forall>u. LIFT sch (first u) = c \<longrightarrow> 
      ((first w ## u) \<Turnstile> \<langle>sch$ = (if isExe then #Super else #Execute)\<rangle>_sch)"
    proof (rule exI,rule allI,intro impI)
      fix u
      assume aa1: "LIFT sch (first u) = (w \<Turnstile> (if isExe then #Super else #Execute))"
      hence "(first w ## u) \<Turnstile> sch$ = (if isExe then #Super else #Execute)"
	by (auto simp add: app_def first_def
	  before_def after_def nexts_def tail_def suffix_def isExe_def) 
      thus "(first w ## u) \<Turnstile> \<langle>sch$ = (if isExe then #Super else #Execute)\<rangle>_sch"
	apply (unfold angle_actrans_def actrans_def) 
	apply (unfold unch_def) 
	apply (auto simp add: isExe_def)
	done
    qed
  qed
qed

lemma WFS_isExe: assumes h1: "basevars sch"
  shows "\<turnstile> WF(S)_sch \<longrightarrow> \<box>\<diamond>(isExe)"
proof (unfold WeakF_def,int_unlift,rule impI)
  fix w
  assume a1: "(w \<Turnstile> \<diamond>\<box>Enabled (\<langle>S\<rangle>_sch)) \<longrightarrow> (w \<Turnstile> \<box>\<diamond>\<langle>S\<rangle>_sch)"
  from h1 have "\<turnstile> Enabled (\<langle>S\<rangle>_sch)"
    by (rule Enabled_Sact)
  hence "\<turnstile> \<box>Enabled (\<langle>S\<rangle>_sch)" by (rule alw)
  from this[THEN intD,of w] E3[of "LIFT \<box>Enabled (\<langle>S\<rangle>_sch)",THEN intD,of w]
  have "w \<Turnstile> \<diamond>\<box>Enabled (\<langle>S\<rangle>_sch)" by auto
  with a1 have "w \<Turnstile> \<box>\<diamond>\<langle>S\<rangle>_sch" by auto
  with infoften_exe[THEN intD,of sch w]
  show "w \<Turnstile> \<box>\<diamond>(isExe)"
    by auto
qed

lemma WFS_isSup: assumes h1: "basevars sch"
  shows "\<turnstile> WF(S)_sch \<longrightarrow> \<box>\<diamond>(isSup)"
proof (unfold WeakF_def,int_unlift,rule impI)
  fix w
  assume a1: "(w \<Turnstile> \<diamond>\<box>Enabled (\<langle>S\<rangle>_sch)) \<longrightarrow> (w \<Turnstile> \<box>\<diamond>\<langle>S\<rangle>_sch)"
  from h1 have "\<turnstile> Enabled (\<langle>S\<rangle>_sch)"
    by (rule Enabled_Sact)
  hence "\<turnstile> \<box>Enabled (\<langle>S\<rangle>_sch)" by (rule alw)
  from this[THEN intD,of w] E3[of "LIFT \<box>Enabled (\<langle>S\<rangle>_sch)",THEN intD,of w]
  have "w \<Turnstile> \<diamond>\<box>Enabled (\<langle>S\<rangle>_sch)" by auto
  with a1 have "w \<Turnstile> \<box>\<diamond>\<langle>S\<rangle>_sch" by auto
  with infoften_sup[THEN intD,of sch w]
  show "w \<Turnstile> \<box>\<diamond>(isSup)"
    by auto
qed
  
lemma body_sch_cases:
  fixes A :: "temporal"
  and  B :: "temporal"
  and  C :: "temporal"
  assumes h1: "w \<Turnstile>  A \<longrightarrow> I" and h2: "w \<Turnstile>  B \<longrightarrow> I"
  shows "w \<Turnstile> (isExe \<longrightarrow> A) \<and> (isSup \<longrightarrow> B) \<longrightarrow> I"
  apply simp
  apply (rule impI,erule conjE,rule schcases[of w])
  using h1 h2 apply auto
done

lemma stut_isExe: "STUTINV isExe"
  by (unfold isExe_def)(simp only: humebasicstutinvs)

lemma stut_isSup: "STUTINV isSup"
  by (unfold isSup_def)(simp only: humebasicstutinvs)

lemma nstut_S: "NSTUTINV S"
  by (unfold S_def)(simp only: stut_isExe humebasicstutinvs)

lemmas humestutinvs =
  stut_isExe stut_isSup nstut_S humebasicstutinvs 


section "Boxes"

datatype BoxState = Runnable
                  | Blocked
                  | Matchfail

types
 boxstate = "state \<Rightarrow> BoxState"


end
