
theory HumeSemantics
imports "../State" 
begin

text "Values and projections"

constdefs

  trd :: "('a \<times> 'b \<times> 'c) \<Rightarrow> 'c"
  [simp]: "trd t \<equiv> snd(snd t)" 

  fourth :: "('a \<times> 'b \<times> 'c \<times> 'd) \<Rightarrow> 'd"
  [simp]: "fourth t \<equiv> snd(trd t)"

  fifth :: "('a \<times> 'b \<times> 'c \<times> 'd \<times> 'e) \<Rightarrow> 'e"
  [simp]: "fifth t \<equiv> snd(fourth t)"

consts
  fi :: "('a * 'b * 'c) \<Rightarrow> 'a" 
  se :: "('a * 'b * 'c) \<Rightarrow> 'b" 
  th :: "('a * 'b * 'c) \<Rightarrow> 'c" 

defs
 fi: "fi x \<equiv> fst x"
 se: "se x \<equiv> fst (snd x)"
 th: "th x \<equiv> snd (snd x)"


lemma f_s_t: "(x = (t1,t2,t3)) = ((fi x = t1) \<and> (se x = t2) \<and> (th x = t3))"
  by (auto simp add: fi se th)

lemma f_s_t_lift: "\<turnstile> (x = (t1,t2,t3)) = ((fi<x> = t1) \<and> (se<x> = t2) \<and> (th<x> = t3))"
  apply (int_unlift)
  apply (auto simp add: f_s_t)
  done

lemma f_s_t_unch: fixes x :: "'a * 'b * 'c" 
  shows "(x = y) = ((fi x = fi y) \<and> (se x = se y) \<and> (th x = th y))"
proof -
  have "\<exists> a b. x = (a,b)" by (rule surj_pair)
  then obtain a where "\<exists> b. x = (a,b)" ..
  then obtain b where g1: "x = (a,b)" ..
  have "\<exists> c d. b = (c,d)" by (rule surj_pair)
  then obtain c where "\<exists> d. b = (c,d)" ..
  then obtain d where "b = (c,d)" ..
  with g1 have a1: "x = (a,c,d)" by simp
  show ?thesis
  proof
    assume b1: "x = y"
    with a1 have "y = (a,c,d)" by simp
    with a1 show "(fi x = fi y) \<and> (se x = se y) \<and> (th x = th y)"
      by auto
  next
    assume b1: "(fi x = fi y) \<and> (se x = se y) \<and> (th x = th y)"
    from b1 a1 have g1: "fi y = fi (a,c,d)" by simp
    from b1 a1 have g2: "se y = se (a,c,d)" by simp
    from b1 a1 have g3: "th y = th (a,c,d)" by simp
    from  g1 g2 g3 have "y = (a,c,d)" by (simp add: fi se th Pair_fst_snd_eq)
    with a1 show "x=y" by auto
  qed
qed

lemmas f_s_t_eq = f_s_t_lift[THEN inteq_reflection]

types 'a HVal = "('a option) statefun"

(* need to work on projections *)
(* look at 'a option theories + schirmer work *)

consts
  toVal :: "'a option \<Rightarrow> 'a" 
  intoVal :: "'a \<Rightarrow> 'a option"
  isVal :: "'a option \<Rightarrow> bool"
  isEmpty :: "'a option \<Rightarrow> bool"
  exeWire :: "'a HVal \<Rightarrow> temporal"

primrec
  "toVal (Some n) = n"

primrec
  "isVal (Some n) = True"
  "isVal None = False"

defs
  isEmpty[simp]: "isEmpty v \<equiv> \<not>isVal(v)"


lemma not_isVal: "\<not>(isVal k) \<Longrightarrow> k = None"
  by (induct k) auto

lemma isEmpty_bot: "isEmpty k \<Longrightarrow> k = None"
  by (simp add: not_isVal)

lemma isEmpty_bot2: "isEmpty k = (k = None)"
  apply (simp add: not_isVal)
  apply (induct k)
  apply auto
  done

axioms
 shouldremove: "(\<exists> P. toVal x = P) = (x \<noteq> None)"
 maybebetter: "(\<exists> v. toVal None = v) \<Longrightarrow> False"

defs
  intoVal[simp]: "intoVal n \<equiv> Some n"
  exeWire: "exeWire w \<equiv> TEMP ((Unchanged w) \<or> w$ = #None)"

consts
  toBeforeVal :: "('a option) statefun \<Rightarrow> 'a tempfun"
  toAfterVal :: "('a option) statefun \<Rightarrow> 'a tempfun"
  intoVal2 :: "'a tempfun \<Rightarrow> ('a option) tempfun"

defs
  toBeforeVal: "toBeforeVal v \<equiv> LIFT (toVal<$v>)"
  toAfterVal: "toAfterVal v \<equiv> LIFT (toVal<v$>)"
  intoVal2: "intoVal2 v \<equiv> LIFT (intoVal<v>)"

syntax
 "_tobeforeval"    :: "lift \<Rightarrow> lift"  ("(@_)" [100] 99)
 "_toafterval"     :: "lift \<Rightarrow> lift"  ("(_@)" [100] 99)
 "_intoVal2"       :: "lift \<Rightarrow> lift"  ("(iV _)" [100] 99) (* reduce??? *)
 
translations
 "_tobeforeval"    \<rightleftharpoons> "toBeforeVal"
 "_toafterval"     \<rightleftharpoons> "toAfterVal" 
 "_intoVal2"     \<rightleftharpoons> "intoVal2"

(* move to axioms *) 
lemma t1: "|~ (\<circ>$x) = (x$)"
  by (simp add: after_def)


theorem stut_tobeforeval: "STUTINV @f"
proof -
  have "STUTINV $f" by (rule stut_before)
  hence "STUTINV toVal<$f>" by (rule stut_fun1)
  thus ?thesis
    by (simp add: toBeforeVal)
qed

theorem nstut_toafterval: "NSTUTINV f@"
proof -
  have "NSTUTINV f$" by (rule nstut_after)
  hence "NSTUTINV toVal<f$>" by (rule nstut_fun1)
  thus ?thesis
    by (simp add: toAfterVal)
qed

lemmas humebasicstutinvs = bothstutinvs stut_tobeforeval 
  stut_tobeforeval[THEN stutinv_strictly_stronger]
  nstut_toafterval livestutinv

lemma t2:  "|~ (\<circ>@x) = (x@)"
  by (simp add: next_fun1[THEN prefeq_reflection] toBeforeVal toAfterVal after_def)

lemmas after_before = t1[THEN prefeq_reflection] t2[THEN prefeq_reflection]

lemma "x \<noteq> None \<longrightarrow> (\<exists> n. n = toVal x)"
  by auto

section "Matches"

subsection "Matching Inputs"


datatype 'a Pattern = PIgnore
                    | PConsumeIgnore
                    | PConsume
                    | PConst "'a"

consts
  PMatch :: "'a Pattern \<Rightarrow> 'a option \<Rightarrow> bool" ("\<guillemotleft>_,_\<guillemotright>")
  mIgn :: "'a option \<Rightarrow> bool"
  mCon :: "'a option \<Rightarrow> bool"
  mConst :: "'a \<Rightarrow> 'a option \<Rightarrow> bool"

primrec
  "PMatch PIgnore iws = True"
  "PMatch PConsumeIgnore iws = True"
  "PMatch PConsume iws = isVal iws"
  "PMatch (PConst patt) iws = (isVal iws \<and> toVal iws = patt)"


defs
  mIgn[simp]: "mIgn iws \<equiv> \<guillemotleft>PIgnore,iws\<guillemotright>"
  mCon[simp]: "mCon iws \<equiv> \<guillemotleft>PConsume,iws\<guillemotright>"
  mConst[simp]: "mConst patt iws \<equiv> \<guillemotleft>PConst patt,iws\<guillemotright>"

subsection "Asserting Outputs"

consts 
  assertOut :: "'a option \<Rightarrow> 'a option \<Rightarrow> bool"

primrec
 "assertOut None ow = True"
 "assertOut (Some res) ow = (ow = None)"

lemma isEmpty_None: "isEmpty v \<Longrightarrow> v = None"
  apply (induct v)
  apply auto
  done

lemma assertout1: "isEmpty res \<Longrightarrow> assertOut res wire"
  by (induct res)(auto)

lemma assertout2:"isEmpty wire \<Longrightarrow> assertOut res wire"
  by (induct res)(auto simp add: isEmpty_None)

lemma pmatch1: assumes H1: "x = PConsume" and H2: "\<not>(PMatch x wire)"
  shows "isEmpty wire"
  using H1 H2 by (induct x)(auto simp add: split_if) (* if_def2 *)

lemma pmatch2: assumes H1: "\<not>(PMatch PConsume wire)"
  shows "isEmpty wire"
  using H1 by (auto simp add: split_if)

lemma pmatch3: assumes H1: "PMatch PConsume wire"
  shows "isVal wire"
proof -
  from H1 have "wire \<noteq> None"
    apply (induct wire)
    by auto
  thus ?thesis
    by (induct wire) auto
qed

consts
  nwire :: "'a option \<Rightarrow> 'a option \<Rightarrow> 'a option"

primrec
  "nwire None ow = ow"
  "nwire (Some v) ow = (Some v)"

lemma exe_bod2: 
  assumes h1: "(P \<longrightarrow> Q) \<longrightarrow> I"
  and h2: "(\<not>P \<longrightarrow> R) \<longrightarrow> I"
  shows "(if P then Q else R) \<longrightarrow> I"
using h1 h2 by auto

lemma nwire1: "(nwire a b = a) \<or> (nwire a b = b)"
  by (induct a) auto

lemma nwire2: "nwire x y \<noteq> x \<Longrightarrow> nwire x y = y"
  apply (induct x)
  apply auto
  done

lemma nwire1_lift: "(nwire (w \<Turnstile> $a)(w \<Turnstile> $b) = (w \<Turnstile> $a)) \<or> (nwire (w \<Turnstile> $a)(w \<Turnstile> $b) = (w \<Turnstile> $b))"
  apply (rule nwire1)
  done

lemma nwire1_lift_fun: "(nwire (w \<Turnstile> g<$a>)(w \<Turnstile> $b) = (w \<Turnstile> g<$a>)) \<or> (nwire (w \<Turnstile> g<$a>)(w \<Turnstile> $b) = (w \<Turnstile> $b))"
  apply (rule nwire1)
  done

lemma exe_bod: "((if P then Q else R) \<longrightarrow> I) = (((P \<and> Q) \<longrightarrow> I) \<and> ((\<not>P \<and> R) \<longrightarrow> I))"
  by auto

lemma exe_bod2: "((M \<and> (if P then Q else R)) \<longrightarrow> I) = ((( (M \<and> P) \<and> Q) \<longrightarrow> I) \<and> (((M \<and> \<not>P) \<and> R) \<longrightarrow> I))"
  by auto

lemma exe_bl: 
  assumes h1: "w \<Turnstile> P \<longrightarrow> I"
  and h2: "w \<Turnstile> Q \<longrightarrow> I"
  shows "w \<Turnstile> (if R then P else Q) \<longrightarrow> I"
  using h1 h2 by auto


lemma not_isVal: "(\<not>(isVal k)) = (k = None)"
by (induct k) auto

lemma inv_cas: "(A \<and> B) \<longrightarrow> I = ((A \<longrightarrow> I) \<and> (B \<longrightarrow> I))"
  by blast

text "Values"

lemma isVal_not_None: "isVal v = (v \<noteq> None)"
  apply (induct v)
  apply auto
  done

lemma isVal_some: "isVal v = (\<exists> y. v = Some y)"
  apply (induct v)
  apply auto
  done

lemma isVal_full: "isVal v = (v \<noteq> None \<and> (\<exists> y. v = Some y))"
  apply (induct v)
  apply auto
  done

lemma isVal_some: "isVal v \<longrightarrow> (toVal v = y) = (v = Some y)"
  apply (simp only: isVal_full)
  apply auto
  done

lemma some_isVal: "x = Some a \<Longrightarrow> isVal x"
  by simp

lemma valintoto: "isVal v \<longrightarrow>  (v = intoVal y) = (toVal v = y)"
  apply (simp only: isVal_full)
  apply auto
  done

lemma valintoto_lift: "\<turnstile> isVal<x$> \<longrightarrow> (x$ = intoVal <v>) = (x@ = v)"
  apply (unfold toAfterVal)
  apply int_unlift
  apply (rule valintoto)
  done

lemma  valintoto_unl: "isVal (w \<Turnstile> x$) \<longrightarrow> ((w \<Turnstile> x$) = (intoVal v)) = ((w \<Turnstile> x@)= v)"
  apply (unfold toAfterVal)
  apply int_unlift
  apply (rule valintoto)
  done

lemma  valintoto_unl2: "isVal (w \<Turnstile> $x) \<longrightarrow> ((w \<Turnstile> $x) = (intoVal v)) = ((w \<Turnstile> @x)= v)"
  apply (unfold toBeforeVal)
  apply int_unlift
  apply (rule valintoto)
  done

lemma  valintoto_unl_fun: "isVal (g(w \<Turnstile> x$)) \<longrightarrow> ((g(w \<Turnstile> x$)) = (intoVal v)) = ((w \<Turnstile> g<x>@) = v)"
  apply (unfold toAfterVal)
  apply (simp only: all_before_eq all_after_eq)
  apply int_unlift
  apply (rule valintoto)
  done

lemma  valintoto_unl_fun2: "isVal (g(w \<Turnstile> $x)) \<longrightarrow> ((g(w \<Turnstile> $x)) = (intoVal v)) = ((w \<Turnstile> @g<x>) = v)"
  apply (unfold toBeforeVal toAfterVal)
  apply (simp only: all_before_eq all_after_eq)
  apply int_unlift
  apply (rule valintoto)
  done

lemma condsym: "P \<longrightarrow> A = B \<Longrightarrow> P \<longrightarrow> B = A" 
  by auto

text "Matches"


lemma pmatch_const1: "(mConst A B) = (B = Some A)"
  apply (auto)
  apply (induct B)
  apply auto
  done

lemma pmatch_con1: "mCon A = (\<exists> v. Some v = A)"
  by (auto simp add: isVal_not_None)

lemma  pmatch_con2: "(mCon A)=(isVal A)"
  apply auto
  done

lemma lift_m1: "\<turnstile> (mConst<A,$B>)=(isVal<$B> \<and> @B = A)"
  apply int_unlift
  apply (unfold toBeforeVal)
  apply auto
  done

lemma lift_m2: "\<turnstile> mCon<$A> = (isVal<$A> \<and> (\<exists> v. @A = v))"
  apply int_unlift
  apply (unfold toBeforeVal)
  apply auto
  done

lemma lift_m3: "\<turnstile> (mConst<A,B$>)=(isVal<B$> \<and> B@ = A)"
  apply int_unlift
  apply (unfold toAfterVal)
  apply auto
  done

lemma lift_m4: "\<turnstile> mCon<A$> = (isVal<A$> \<and> (\<exists> v. A@ = v))"
  apply int_unlift
  apply (unfold toAfterVal)
  apply auto
  done

lemma unch_opt_val: "\<turnstile> x$ = $y \<longrightarrow> x@ = @y"
  apply int_unlift
  apply (unfold toBeforeVal toAfterVal)
  apply auto
  done

lemma after_opt_val: "\<turnstile> x$ = y$ \<longrightarrow> x@ = y@"
  apply int_unlift
  apply (unfold toBeforeVal toAfterVal)
  apply auto
  done

lemma before_opt_val: "\<turnstile> $x = $y \<longrightarrow> @x = @y"
  apply int_unlift
  apply (unfold toBeforeVal toAfterVal)
  apply auto
  done

lemma unch_opt_fun_unl: "g(w \<Turnstile> x$) = g(w \<Turnstile> $y) \<longrightarrow> (w \<Turnstile> g<x>@) = (w \<Turnstile> @g<y>)"
  apply int_unlift
  apply (unfold toBeforeVal toAfterVal)
  apply (simp only: all_before_eq all_after_eq)
  apply auto
  done

lemmas pmatches_lift = lift_m1 lift_m2 lift_m3 lift_m4

lemmas pmatches_eq = pmatches_lift[THEN inteq_reflection]



lemma isVal_valintoto: "isVal(w \<Turnstile> $w4) \<longrightarrow> (w \<Turnstile> $w4) = Some (w \<Turnstile> @w4)"
proof -
    have "isVal(w \<Turnstile> $w4) \<longrightarrow> ((w \<Turnstile> $w4) = intoVal (w \<Turnstile> @w4)) = ((w \<Turnstile> @w4) = (w \<Turnstile> @w4))"
      by (rule valintoto_unl2)
    thus ?thesis by auto
qed

end