header "Defines Liveness Constructs"

theory Liveness
imports Rules
begin

constdefs 

enabled :: "'a formula \<Rightarrow> 'a formula"
"enabled F \<equiv> \<lambda> s. \<exists> t. ((first s) ## t) \<Turnstile> F"

WeakF :: "'a formula \<Rightarrow> ('a,'b) stfun \<Rightarrow>  'a formula"
"WeakF F v \<equiv> LIFT (\<diamond>\<box>(enabled (angle_actrans F v)) \<longrightarrow> \<box>\<diamond>\<langle>F\<rangle>_v)"

StrongF :: "'a formula \<Rightarrow> ('a,'b) stfun \<Rightarrow> 'a formula"
"StrongF F v \<equiv> LIFT (\<box>\<diamond>(enabled (angle_actrans F v)) \<longrightarrow> \<box>\<diamond>\<langle>F\<rangle>_v)"

syntax
 "_WF" :: "[lift,lift] \<Rightarrow> lift" ("(WF(_)'_(_))" [40] 40)
 "_SF" :: "[lift,lift] \<Rightarrow> lift" ("(SF(_)'_(_))" [40] 40)
 "_Enabled" :: "lift \<Rightarrow> lift" ("(Enabled _)" [100] 100)

translations
 "_WF" \<rightleftharpoons> "WeakF"
 "_SF" \<rightleftharpoons> "StrongF"
 "_Enabled" \<rightleftharpoons> "enabled"

section "Properties of @{term enabled}"

lemma EnabledI: "\<turnstile> F \<longrightarrow> Enabled F"
proof int_unlift
  fix w :: "nat \<Rightarrow> 'a"
  have "w = (\<lambda>n. if n = 0 then first w else (tail w) (n - 1))"
    by (rule ext) (auto simp add: tail_def suffix_def first_def)
  hence "(w \<Turnstile> F) \<longrightarrow> ((\<lambda>n. if n = 0 then first w else (tail w) (n - 1)) \<Turnstile> F)"
    by auto
  thus "(w \<Turnstile> F) \<longrightarrow> (w \<Turnstile> Enabled F)"
    by (auto simp add:  enabled_def app_def)
qed

lemma Enabled_mono: assumes H1: "w \<Turnstile> Enabled F" and H2: "\<turnstile> F \<longrightarrow> G" 
   shows "w \<Turnstile> Enabled G"
proof (unfold enabled_def)
  from H1 have "\<exists> t. ((first w) ## t) \<Turnstile> F"
    by (auto simp add: first_def enabled_def)
  with H2[THEN intD] show "\<exists> t. ((first w) ## t) \<Turnstile> G"
    by auto
qed

lemma Enabled_disj1: "\<turnstile> (Enabled F \<longrightarrow> Enabled (F \<or> G))"
proof (int_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> Enabled F"
  have "\<turnstile> F \<longrightarrow> (F \<or> G)" by auto
  with a1 show "w \<Turnstile> Enabled (F \<or> G)" by (rule Enabled_mono)
qed

lemma  Enabled_disj2: "\<turnstile> (Enabled F \<longrightarrow> Enabled (G \<or> F))"
proof (int_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> Enabled F"
  have "\<turnstile> F \<longrightarrow> (G \<or> F)" by auto
  with a1 show "w \<Turnstile> Enabled (G \<or> F)" by (rule Enabled_mono)
qed

lemma  Enabled_conj1: "\<turnstile> (Enabled (F \<and> G) \<longrightarrow> Enabled F)"
proof (int_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> Enabled (F \<and> G)"
  have "\<turnstile> (F \<and> G) \<longrightarrow> F" by auto
  with a1 show "w \<Turnstile> Enabled F" by (rule Enabled_mono)
qed

lemma  Enabled_conj2: "\<turnstile> (Enabled (G \<and> F) \<longrightarrow> Enabled F)"
proof (int_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> Enabled (G \<and> F)"
  have "\<turnstile> (G \<and> F) \<longrightarrow> F" by auto
  with a1 show "w \<Turnstile> Enabled F" by (rule Enabled_mono)
qed

lemma Enabled_conjE:
     assumes H1: "w \<Turnstile> Enabled (F \<and> G)"
        and   H2: "\<lbrakk> w \<Turnstile> Enabled F ; w \<Turnstile> Enabled G \<rbrakk> \<Longrightarrow> Q"
       shows      "Q"
proof -
  from H1 Enabled_conj1[THEN intD] have a1: "w \<Turnstile> Enabled F" by auto
  from H1 Enabled_conj2[THEN intD] have "w \<Turnstile> Enabled G" by auto
  with a1 H2 show "Q" by auto
qed

lemma Enabled_disjD: "\<turnstile> Enabled (F \<or> G) \<longrightarrow> (Enabled F) \<or> (Enabled G)"
proof (int_unlift, rule impI)
  fix w
  assume a1: "w \<Turnstile> Enabled (F \<or> G)"
  thus "(w \<Turnstile> (Enabled F)) \<or> (w \<Turnstile> (Enabled G))" 
    by (auto simp add: enabled_def first_def)
qed

lemma Enabled_disj: "\<turnstile> Enabled (F \<or> G) = ((Enabled F) \<or> (Enabled G))"
proof (int_unlift,rule iffI)
  fix w 
  assume "w \<Turnstile> (Enabled (F \<or> G))"
  with intD[OF Enabled_disjD] show "(w \<Turnstile> Enabled F) \<or> (w \<Turnstile> Enabled G)" 
    by auto
next
  fix w
  assume "(w \<Turnstile> Enabled F) \<or> (w \<Turnstile> Enabled G)"
  thus "w \<Turnstile> (Enabled (F \<or> G))"
  proof
    assume "w \<Turnstile> Enabled F"
    with intD[OF Enabled_disj1] show ?thesis by auto
  next
    assume "w \<Turnstile> Enabled G"
    with intD[OF Enabled_disj2] show ?thesis by auto
  qed
qed

lemmas enab_refl = Enabled_disj[THEN inteq_reflection]

lemma Enabled_ex: "\<turnstile> Enabled (\<exists> x. F x) = (\<exists> x. Enabled (F x))"
  by (int_unlift) (auto simp add: enabled_def first_def)

lemma [simp]: "\<turnstile> \<not>\<not>Enabled F = Enabled F"
  by auto

section "Fairness Properties" 


lemma WF_alt: "\<turnstile> (WF (A)_v) = ((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)"
proof -
  have g1: "\<turnstile> (\<not>\<diamond>\<box> Enabled (\<langle>A\<rangle>_v)) = (\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v))" by (rule E12)
  have "\<turnstile> ((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) = ((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by auto
  from this[THEN intD] g1[THEN intD]
    have  "\<turnstile> ((\<not>\<diamond>\<box> Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) = ((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by auto
  from this[THEN intD]
    have "\<turnstile> ((\<diamond>\<box> Enabled (\<langle>A\<rangle>_v)) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_v) = ((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by auto
  thus ?thesis by (unfold WeakF_def)
qed

lemma SF_alt: "\<turnstile> (SF (A)_v) = ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)"
proof -
  have g1: "\<turnstile> (\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) = (\<not>\<box>\<diamond>Enabled (\<langle>A\<rangle>_v))" by (rule E13)
  have "\<turnstile> ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) = ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by auto
  from this[THEN intD] g1[THEN intD]
    have  "\<turnstile> ((\<not>\<box>\<diamond> Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) = ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by auto
  from this[THEN intD]
    have "\<turnstile> ((\<box>\<diamond> Enabled (\<langle>A\<rangle>_v)) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_v) = ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by auto
  thus ?thesis by (unfold StrongF_def)
qed

lemma AlwaysWFI: "\<turnstile> WF (A)_v \<longrightarrow> \<box>WF (A)_v"
proof -
  have "\<turnstile> (WF (A)_v) = ((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by (rule WF_alt)
  from this[THEN intD] have g1: "\<turnstile> WF (A)_v \<longrightarrow> (\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v"
    by auto
  have "\<turnstile> \<box>\<diamond>(\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<box>((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" 
    by (rule MM6)
  with g1 have g2: "\<turnstile> WF (A)_v \<longrightarrow> \<box>((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)"
    by (rule lift_imp_trans)
  have "\<turnstile> \<box>((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>WF (A)_v"
  proof (unfold WeakF_def)
    have "\<turnstile> (WF (A)_v) = ((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by (rule WF_alt)
    hence "\<turnstile> (\<diamond>\<box>(Enabled (\<langle>A\<rangle>_v)) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_v) = ((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" 
      by (unfold WeakF_def)
    from this[THEN intD] have "\<turnstile> ((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<diamond>\<box>(Enabled (\<langle>A\<rangle>_v)) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_v" 
      by auto
    thus  "\<turnstile> \<box>((\<box>\<diamond>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>(\<diamond>\<box>(Enabled (\<langle>A\<rangle>_v)) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_v)" by (rule STL4)
  qed
  with g2 show ?thesis by (rule lift_imp_trans) 
qed

theorem WF_always: "\<turnstile> (\<box>WF(A)_v) = (WF(A)_v)"
proof (int_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>WF(A)_v"
  have "\<turnstile> \<box>WF(A)_v \<longrightarrow> WF(A)_v" by (rule ax1)
  from this[THEN intD] a1 show "w \<Turnstile> WF(A)_v" 
    by auto
next
  fix w
  assume a1: "w \<Turnstile> WF(A)_v"
  have "\<turnstile> WF(A)_v \<longrightarrow> \<box>WF(A)_v" by (rule AlwaysWFI)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>WF(A)_v" 
    by auto  
qed

lemma AlwaysSFI: "\<turnstile> SF (A)_v \<longrightarrow> \<box>SF (A)_v"
proof -
  have "\<turnstile> (SF (A)_v) = ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by (rule SF_alt)
  from this[THEN intD] have g1: "\<turnstile> SF (A)_v \<longrightarrow> ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" 
    by auto
  have "\<turnstile> (\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) = (\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v))" by (rule E21)
  from this[THEN inteq_reflection] 
    have "\<turnstile> ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) \<longrightarrow> ((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)"
      by (auto simp add: Valid_def)
  with g1 have g1': "\<turnstile>  SF (A)_v \<longrightarrow> ((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)"
    by (rule lift_imp_trans)
  have "\<turnstile> ((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)"
    by (rule MM6)
  with g1' have g2: "\<turnstile> SF (A)_v \<longrightarrow> \<box>((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)"
    by (rule lift_imp_trans)
  have "\<turnstile> \<box>((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>SF (A)_v"
  proof (unfold StrongF_def)
    have g1: "\<turnstile> (SF (A)_v) = ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by (rule SF_alt)
    have "\<turnstile> (\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) = (\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v))" by (rule E21)
    from this[THEN inteq_reflection] 
    have "\<turnstile> ((\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) = ((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)"
      by (auto simp add: Valid_def)
    from this[THEN intD] g1[THEN intD]
      have g1': "\<turnstile> (SF (A)_v) = ((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)" by auto
      hence "\<turnstile> (\<box>\<diamond>Enabled (\<langle>A\<rangle>_v) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_v) = ((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v)"
	by (unfold StrongF_def)      
      from this[THEN intD] have "\<turnstile> ((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) \<longrightarrow> (\<box>\<diamond>Enabled (\<langle>A\<rangle>_v) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_v)"
	by auto
      thus  "\<turnstile> \<box>((\<box>\<diamond>\<box>\<not>Enabled (\<langle>A\<rangle>_v)) \<or> \<box>\<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>(\<box>\<diamond>Enabled (\<langle>A\<rangle>_v) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_v)" 
	by (rule STL4)
  qed
  with g2 show ?thesis by (rule lift_imp_trans) 
qed
  
theorem SF_always: "\<turnstile> (\<box>SF(A)_v) = (SF(A)_v)"
proof (int_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>SF(A)_v"
  have "\<turnstile> \<box>SF(A)_v \<longrightarrow> SF(A)_v" by (rule ax1)
  from this[THEN intD] a1 show "w \<Turnstile> SF(A)_v" 
    by auto
next
  fix w
  assume a1: "w \<Turnstile> SF(A)_v"
  have "\<turnstile> SF(A)_v \<longrightarrow> \<box>SF(A)_v" by (rule AlwaysSFI)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>SF(A)_v" 
    by auto  
qed

lemma Enabled_WFSF: "\<turnstile> \<box>Enabled (\<langle> F \<rangle>_v) \<longrightarrow> ((WF(F)_v) = (SF(F)_v))"
proof (unfold WeakF_def StrongF_def,int_unlift,intro impI,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>Enabled (\<langle> F \<rangle>_v)"
    and a2: "(w \<Turnstile> \<diamond>\<box>Enabled (\<langle>F\<rangle>_v)) \<longrightarrow> (w \<Turnstile> \<box>\<diamond>\<langle>F\<rangle>_v)"
  from E3[THEN intD] a1 have "w \<Turnstile> \<diamond>\<box>Enabled (\<langle> F \<rangle>_v)"
    by auto
  with a2 have "w \<Turnstile> \<box>\<diamond>\<langle>F\<rangle>_v" by auto
  thus "(w \<Turnstile> \<box>\<diamond>Enabled (\<langle> F \<rangle>_v)) \<longrightarrow> (w \<Turnstile> \<box>\<diamond>\<langle>F\<rangle>_v)" by auto
next
  fix w
  assume a1: "w \<Turnstile> \<box>Enabled (\<langle> F \<rangle>_v)"
    and a2: "(w \<Turnstile> \<box>\<diamond>Enabled (\<langle>F\<rangle>_v)) \<longrightarrow> (w \<Turnstile> \<box>\<diamond>\<langle>F\<rangle>_v)"
  have "\<turnstile> Enabled (\<langle> F \<rangle>_v) \<longrightarrow> \<diamond>Enabled (\<langle>F\<rangle>_v)"
    by (rule E3)
  hence "\<turnstile> \<box>Enabled (\<langle> F \<rangle>_v) \<longrightarrow> \<box>\<diamond>Enabled (\<langle>F\<rangle>_v)"
    by (rule STL4)
  from this[THEN intD] a1 have "w \<Turnstile> \<box>\<diamond>Enabled (\<langle>F\<rangle>_v)"
    by auto
  with a2 have "w \<Turnstile> \<box>\<diamond>\<langle>F\<rangle>_v" by auto
  thus "(w \<Turnstile> \<diamond>\<box>Enabled (\<langle> F \<rangle>_v)) \<longrightarrow> (w \<Turnstile> \<box>\<diamond>\<langle>F\<rangle>_v)" by auto 
qed

(* Based on Merz formalisation *)
(* NOTE: in h2 A should be replaced by \<langle>A\<rangle>_v *)
theorem WF1_old:  assumes h1: "|~ N \<and> P \<longrightarrow> \<circ>P \<or> \<circ>Q"
               and   h2: "|~ (N \<and> P) \<and> A \<longrightarrow> \<circ>Q"
               and   h3: "\<turnstile> P \<and> N \<longrightarrow> Enabled (\<langle>A\<rangle>_v)"
               and   h4: "|~ P \<and> Unchanged w \<longrightarrow> \<circ>P"
              shows     "\<turnstile> \<box>N \<and> (WF(A)_v) \<longrightarrow> (P \<leadsto> Q)"
proof (unfold leadsto_def)
  have "\<turnstile> \<box>N \<and> WF(A)_v \<longrightarrow> (P \<longrightarrow> \<diamond>Q)"
  proof (int_unlift, unfold WeakF_def, intro impI,elim conjE)
    fix w
    assume a1: "w \<Turnstile> P"
      and  a2: "w \<Turnstile> \<box>N" 
      and  a3: "w \<Turnstile> \<diamond>\<box>Enabled (\<langle>A\<rangle>_v) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_v"
    from h1 h4 have g1: "\<turnstile> \<box>N \<longrightarrow> P \<longrightarrow> \<box>P \<or> \<diamond>Q"
      by (rule LT31)
    from this[THEN intD] a1 a2 have "(w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q)"
      by auto
    thus "w \<Turnstile> \<diamond>Q"
    proof
      assume "w \<Turnstile> \<diamond>Q" thus ?thesis .
    next
      assume aa1: "w \<Turnstile> \<box>P"
      have "\<turnstile> (\<box>(N \<and> P)) = (\<box>N \<and> \<box>P)" by (rule STL5)
      from this[THEN intD] have "\<turnstile> \<box>N \<and> \<box>P \<longrightarrow> \<box>(N \<and> P)"
	by auto
      from this[THEN intD] a2 aa1 have aa2: "w \<Turnstile> \<box>(N \<and> P)"
	by auto
      from h3 have tg1: "\<turnstile> \<box>(P \<and> N) \<longrightarrow> \<box>(Enabled (\<langle>A\<rangle>_v))"
	by (rule STL4)
      have "\<turnstile> (\<box>(P \<and> N)) = (\<box>P \<and> \<box>N)" by (rule STL5)
      from this[THEN intD] have "\<turnstile> \<box>P \<and> \<box>N \<longrightarrow> \<box>(P \<and> N)"
	by auto
      from this tg1 have tg2: "\<turnstile> \<box>P \<and> \<box>N \<longrightarrow> \<box>(Enabled (\<langle>A\<rangle>_v))" 
	by (rule lift_imp_trans)
      have "\<turnstile> \<box>(Enabled (\<langle>A\<rangle>_v)) \<longrightarrow> \<diamond>\<box>(Enabled (\<langle>A\<rangle>_v))"
	by (rule E3)
      with tg2 have "\<turnstile> \<box>P \<and> \<box>N \<longrightarrow> \<diamond>\<box>(Enabled (\<langle>A\<rangle>_v))"
	by (rule lift_imp_trans)
      from this[THEN intD] a2 aa1 a3 have tg3: "w \<Turnstile> \<box>\<diamond>\<langle>A\<rangle>_v"
	by auto
      with ax1[THEN intD] have tg3': "w \<Turnstile> \<diamond>\<langle>A\<rangle>_v" by auto
      have tg4: "|~ \<circ>Q \<longrightarrow> \<diamond>Q" by (rule E23)
      have "\<turnstile> \<box>(N \<and> P) \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>((N \<and> P) \<and> A)\<rangle>_v" 
	by (rule AA19)
      from this[THEN intD] aa2 tg3'
        have tg5: "w \<Turnstile> \<diamond>\<langle>((N \<and> P) \<and> A)\<rangle>_v"
	by auto
      from h2[THEN prefD] tg4[THEN prefD]
	have "|~ (N \<and> P) \<and> A \<longrightarrow> \<diamond>Q" by auto
      hence tg6: "\<turnstile> \<box>[((N \<and> P) \<and> A \<longrightarrow> \<diamond>Q)]_v"
	by (rule sq)
      have "\<turnstile> \<box>[(((N \<and> P) \<and> A) \<longrightarrow> \<diamond>Q)]_v \<longrightarrow> \<diamond>\<langle>((N \<and> P) \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_v"
	by (rule AA5)
      with tg6 have "\<turnstile> \<diamond>\<langle>((N \<and> P) \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_v"
	by (rule fmp)
      from this[THEN intD] tg5 have tg3: "w \<Turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_v"
	by auto
      have "\<turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_v \<longrightarrow> \<diamond>\<diamond>Q" by (rule E25)
      from this[THEN intD] tg3 have tg4: "w \<Turnstile> \<diamond>\<diamond>Q"
	by auto
      with E10[of Q,THEN inteq_reflection]
       show "w \<Turnstile> \<diamond>Q" by auto
   qed
 qed
 hence g1: "\<turnstile> \<box>(\<box>N \<and> WF(A)_v) \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)"
   by (rule STL4)
 have "\<turnstile> (\<box>(\<box>N \<and> WF(A)_v)) = (\<box>\<box>N \<and> \<box>WF(A)_v)"
   by (rule STL5)
 from this[THEN intD] have g2: "\<turnstile> \<box>\<box>N \<and> \<box>WF(A)_v \<longrightarrow> \<box>(\<box>N \<and> WF(A)_v)"
   by auto
 thm STL3
 from STL3[of N,THEN intD] have g3: "\<turnstile> \<box>N \<longrightarrow> \<box>\<box>N" by auto
 have g4: "\<turnstile> WF(A)_v \<longrightarrow> \<box>WF(A)_v" by (rule AlwaysWFI)
 from g2[THEN intD] g3[THEN intD] g4[THEN intD]
   have "\<turnstile> \<box>N \<and> WF(A)_v \<longrightarrow> \<box>(\<box>N \<and> WF(A)_v)" by auto
 from this g1 show "\<turnstile> \<box>N \<and> WF(A)_v \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)" 
   by (rule lift_imp_trans)
qed

(* Based on Lamport's embedding *)


theorem WF1: assumes h1: "|~ P \<and> [N]_f \<longrightarrow> \<circ>P \<or> \<circ>Q"
                 and   h2: "|~ P \<and> \<langle>N \<and> A\<rangle>_f \<longrightarrow> \<circ>Q"
                 and   h3: "\<turnstile> P \<longrightarrow> (Enabled (\<langle>A\<rangle>_f))"
                 and   h4: "|~ P \<and> Unchanged f \<longrightarrow> \<circ>P"
              shows      "\<turnstile> \<box>[N]_f \<and> (WF(A)_f) \<longrightarrow> (P \<leadsto> Q)"
proof (unfold leadsto_def)
  have "\<turnstile> \<box>[N]_f \<and> WF(A)_f \<longrightarrow> (P \<longrightarrow> \<diamond>Q)"
  proof (int_unlift, unfold WeakF_def, intro impI,elim conjE)
    fix w
    assume a1: "w \<Turnstile> P"
      and  a2: "w \<Turnstile> \<box>[N]_f" 
      and  a3: "w \<Turnstile> \<diamond>\<box>Enabled (\<langle>A\<rangle>_f) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_f"
    from h1 h4 have g1: "\<turnstile> \<box>[N]_f \<longrightarrow> P \<longrightarrow> \<box>P \<or> \<diamond>Q" by (rule LT32)
    from this[THEN intD] a1 a2 have "(w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q)"
      by auto
    thus "w \<Turnstile> \<diamond>Q"
    proof
      assume "w \<Turnstile> \<diamond>Q" thus ?thesis .
    next
      assume aa1: "w \<Turnstile> \<box>P"
      from h3 have "\<turnstile> \<box>P \<longrightarrow> \<box>Enabled (\<langle>A\<rangle>_f)"
	by (rule STL4)
      from this[THEN intD] aa1 have g1: "w \<Turnstile> \<box>Enabled (\<langle>A\<rangle>_f)"
	by simp
      have "\<turnstile> \<box>Enabled (\<langle>A\<rangle>_f) \<longrightarrow> \<diamond>\<box>Enabled (\<langle>A\<rangle>_f)"
	by (rule E3)
      from this[THEN intD] g1 have g2: "w \<Turnstile> \<diamond>\<box>Enabled (\<langle>A\<rangle>_f)"
	by auto
      with a3 have g3: "w \<Turnstile> \<box>\<diamond>\<langle>A\<rangle>_f" by auto
      have "\<turnstile> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<langle>A\<rangle>_f" by (rule ax1)
      from this[THEN intD] g3 have g4: "w \<Turnstile> \<diamond>\<langle>A\<rangle>_f" by auto
      have "|~ \<circ>Q \<longrightarrow> \<diamond>Q" by (rule E23)
      from h2[THEN prefD] this[THEN prefD]
        have "|~ P \<and> \<langle>(N \<and> A)\<rangle>_f \<longrightarrow> \<diamond>Q" by auto
      hence g5: "\<turnstile> \<box>[(P \<and> \<langle>(N \<and> A)\<rangle>_f \<longrightarrow> \<diamond>Q)]_f"
	by (rule sq)
      have "\<turnstile> \<box>[(P \<and> \<langle>(N \<and> A)\<rangle>_f \<longrightarrow> \<diamond>Q)]_f \<longrightarrow> \<diamond>\<langle>(P \<and> \<langle>(N \<and> A)\<rangle>_f)\<rangle>_f \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_f"
	by (rule AA5)
      with g5 have g6: "\<turnstile> \<diamond>\<langle>(P \<and> \<langle>(N \<and> A)\<rangle>_f)\<rangle>_f \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_f"
	by (rule fmp)
      have "w \<Turnstile> \<diamond>\<langle>(P \<and> \<langle>(N \<and> A)\<rangle>_f)\<rangle>_f"
      proof -
	have "\<turnstile> \<box>[N]_f \<and> \<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<langle>(N \<and> A)\<rangle>_f" by (rule AA18)
	from this[THEN intD] a2 g4 have tg1: "w \<Turnstile> \<diamond>\<langle>(N \<and> A)\<rangle>_f"
	  by auto
	have "\<turnstile> (\<diamond>\<langle>\<langle>(N \<and> A)\<rangle>_f\<rangle>_f) = (\<diamond>\<langle>(N \<and> A)\<rangle>_f)" by (rule AA24)
	from this[THEN intD] tg1 have tg2: "w \<Turnstile> \<diamond>\<langle>\<langle>(N \<and> A)\<rangle>_f\<rangle>_f"
	  by auto
	have "\<turnstile> \<box>P \<and> \<diamond>\<langle>\<langle>(N \<and> A)\<rangle>_f\<rangle>_f \<longrightarrow> \<diamond>\<langle>(P \<and> \<langle>(N \<and> A)\<rangle>_f)\<rangle>_f"
	  by (rule AA19)
	from this[THEN intD] aa1 tg2
	  show ?thesis by auto
      qed
      with g6[THEN intD] have g7: "w \<Turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_f" by auto
      have "\<turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_f \<longrightarrow> \<diamond>\<diamond>Q" by (rule E25)
      from this[THEN intD] g7 have "w \<Turnstile> \<diamond>\<diamond>Q"
	by auto
      with E10[of Q,THEN inteq_reflection]
       show "w \<Turnstile> \<diamond>Q" by auto
   qed
 qed
  hence g1: "\<turnstile> \<box>(\<box>[N]_f \<and> WF(A)_f) \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)"
   by (rule STL4)
 have "\<turnstile> (\<box>(\<box>[N]_f \<and> WF(A)_f)) = (\<box>\<box>[N]_f \<and> \<box>WF(A)_f)"
   by (rule STL5)
 from this[THEN intD] have g2: "\<turnstile> \<box>\<box>[N]_f \<and> \<box>WF(A)_f \<longrightarrow> \<box>(\<box>[N]_f \<and> WF(A)_f)"
   by auto
 from T2[of N f,THEN intD] have g3: "\<turnstile> \<box>[N]_f \<longrightarrow> \<box>\<box>[N]_f" by auto
 have g4: "\<turnstile> WF(A)_f \<longrightarrow> \<box>WF(A)_f" by (rule AlwaysWFI)
 from g2[THEN intD] g3[THEN intD] g4[THEN intD]
   have "\<turnstile> \<box>[N]_f \<and> WF(A)_f \<longrightarrow> \<box>(\<box>[N]_f \<and> WF(A)_f)" by auto
 from this g1 show "\<turnstile> \<box>[N]_f \<and> WF(A)_f \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)" 
   by (rule lift_imp_trans)
qed


theorem SF1: assumes h1: "|~ (P \<and> [N]_f) \<longrightarrow> (\<circ>P \<or> \<circ>Q)"
               and   h2: "|~ (P \<and> \<langle>(N \<and> A)\<rangle>_f)  \<longrightarrow> \<circ>Q"
               and   h3: "\<turnstile> (\<box>P \<and> \<box>[N]_f \<and> \<box>F) \<longrightarrow> \<diamond>Enabled (\<langle>A\<rangle>_f)"
               and   h4: "|~ P \<and> Unchanged f \<longrightarrow> \<circ>P"
              shows      "\<turnstile> (\<box>[N]_f \<and> (SF(A)_f) \<and> \<box>F) \<longrightarrow> (P \<leadsto> Q)"
proof (unfold leadsto_def)
  have prem: "\<turnstile> \<box>P \<and> \<box>[N]_f \<and> \<box>F \<longrightarrow> \<box>(\<box>P \<and> \<box>[N]_f \<and> \<box>F)"
  proof -
    have g1: "\<turnstile> (\<box>P) = (\<box>\<box>P)" by (rule T1)
    from this[THEN intD] have g1': "\<turnstile> (\<box>P) \<longrightarrow> (\<box>\<box>P)"
      by auto
    have g2: "\<turnstile> (\<box>[N]_f) = (\<box>\<box>[N]_f)" by (rule T2)
    from this[THEN intD] have g2': "\<turnstile> (\<box>[N]_f) \<longrightarrow> (\<box>\<box>[N]_f)"
      by auto      
    have g3: "\<turnstile> (\<box>F) = (\<box>\<box>F)" by (rule T1)
    from this[THEN intD] have g3': "\<turnstile> (\<box>F) \<longrightarrow> (\<box>\<box>F)"
      by auto
    from g1'[THEN intD]
      g2'[THEN intD]
      g3'[THEN intD]
    have g4: "\<turnstile> (\<box>P \<and> \<box>[N]_f \<and> \<box>F) \<longrightarrow> (\<box>\<box>P \<and> \<box>\<box>[N]_f \<and> \<box>\<box>F)"
      by (auto simp add: Valid_def)
    have "\<turnstile> (\<box>(\<box>P \<and> \<box>[N]_f)) = (\<box>\<box>P \<and> \<box>\<box>[N]_f)"
      by (rule STL5)
    from this[THEN intD] have "\<turnstile> \<box>\<box>P \<and> \<box>\<box>[N]_f \<longrightarrow> \<box>(\<box>P \<and> \<box>[N]_f)"
      by auto
    from g4[THEN intD] this[THEN intD]
    have g5: "\<turnstile> (\<box>P \<and> \<box>[N]_f \<and> \<box>F) \<longrightarrow> (\<box>(\<box>P \<and> \<box>[N]_f) \<and> \<box>\<box>F)"
      by auto
    have "\<turnstile> (\<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>F)) = (\<box>(\<box>P \<and> \<box>[N]_f) \<and> \<box>\<box>F)"
      by (rule STL5)
    from this[THEN intD] have "\<turnstile> \<box>(\<box>P \<and> \<box>[N]_f) \<and> \<box>\<box>F \<longrightarrow> \<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>F)"
      by auto
    with g5 have g6: "\<turnstile> (\<box>P \<and> \<box>[N]_f \<and> \<box>F) \<longrightarrow> \<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>F)"
      by (rule lift_imp_trans)
    have "\<turnstile> (\<box>P \<and> \<box>[N]_f) \<and> \<box>F \<longrightarrow> (\<box>P \<and> \<box>[N]_f \<and> \<box>F)" by auto
    hence "\<turnstile> \<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>F) \<longrightarrow> \<box>(\<box>P \<and> \<box>[N]_f \<and> \<box>F)"
      by (rule STL4)
    with g6 show ?thesis by (rule lift_imp_trans)
  qed
  have "\<turnstile> \<box>[N]_f \<and> SF(A)_f \<and> \<box>F \<longrightarrow> (P \<longrightarrow> \<diamond>Q)"
  proof (int_unlift, unfold StrongF_def, intro impI,elim conjE)
    fix w
    assume a1: "w \<Turnstile> P"
      and  a2: "w \<Turnstile> \<box>[N]_f" 
      and  a3: "w \<Turnstile> \<box>\<diamond>Enabled (\<langle>A\<rangle>_f) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_f" 
      and  a4: "w \<Turnstile> \<box>F"
    from h1 h4 have g1: "\<turnstile> \<box>[N]_f \<longrightarrow> P \<longrightarrow> \<box>P \<or> \<diamond>Q" by (rule LT32)
    from this[THEN intD] a1 a2 have "(w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q)"
      by auto
    thus "w \<Turnstile> \<diamond>Q"
    proof
      assume "w \<Turnstile> \<diamond>Q" thus ?thesis .
    next
      assume aa1: "w \<Turnstile> \<box>P"
      from h3 have h3': "\<turnstile> \<box>(\<box>P \<and> \<box>[N]_f \<and> \<box>F) \<longrightarrow> \<box>\<diamond>Enabled (\<langle>A\<rangle>_f)"
	by (rule STL4)
      from prem h3' have g1: "\<turnstile> \<box>P \<and> \<box>[N]_f \<and> \<box>F \<longrightarrow> \<box>\<diamond>Enabled (\<langle>A\<rangle>_f)"
	by (rule lift_imp_trans)
      from this[THEN intD] aa1 a2 a4 have "w \<Turnstile> \<box>\<diamond>Enabled (\<langle>A\<rangle>_f)"
	by auto
      with a3 have g2: "w \<Turnstile> \<box>\<diamond>\<langle>A\<rangle>_f" by auto
      with ax1[THEN intD] have g2': "w \<Turnstile> \<diamond>\<langle>A\<rangle>_f" by auto
      
      have "|~ \<circ>Q \<longrightarrow> \<diamond>Q" by (rule E23)
      from h2[THEN prefD] this[THEN prefD]
      have "|~ P \<and> \<langle>(N \<and> A)\<rangle>_f \<longrightarrow> \<diamond>Q" by auto
      hence g5: "\<turnstile> \<box>[(P \<and> \<langle>(N \<and> A)\<rangle>_f \<longrightarrow> \<diamond>Q)]_f"
	by (rule sq)
      have "\<turnstile> \<box>[(P \<and> \<langle>(N \<and> A)\<rangle>_f \<longrightarrow> \<diamond>Q)]_f \<longrightarrow> \<diamond>\<langle>(P \<and> \<langle>(N \<and> A)\<rangle>_f)\<rangle>_f \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_f"
	by (rule AA5)
      with g5 have g6: "\<turnstile> \<diamond>\<langle>(P \<and> \<langle>(N \<and> A)\<rangle>_f)\<rangle>_f \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_f"
	by (rule fmp)
      have "w \<Turnstile> \<diamond>\<langle>(P \<and> \<langle>(N \<and> A)\<rangle>_f)\<rangle>_f"
      proof -
	have "\<turnstile> \<box>[N]_f \<and> \<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<langle>(N \<and> A)\<rangle>_f" by (rule AA18)
	from this[THEN intD] a2 g2' have tg1: "w \<Turnstile> \<diamond>\<langle>(N \<and> A)\<rangle>_f"
	  by auto
	have "\<turnstile> (\<diamond>\<langle>\<langle>(N \<and> A)\<rangle>_f\<rangle>_f) = (\<diamond>\<langle>(N \<and> A)\<rangle>_f)" by (rule AA24)
	from this[THEN intD] tg1 have tg2: "w \<Turnstile> \<diamond>\<langle>\<langle>(N \<and> A)\<rangle>_f\<rangle>_f"
	  by auto
	have "\<turnstile> \<box>P \<and> \<diamond>\<langle>\<langle>(N \<and> A)\<rangle>_f\<rangle>_f \<longrightarrow> \<diamond>\<langle>(P \<and> \<langle>(N \<and> A)\<rangle>_f)\<rangle>_f"
	  by (rule AA19)
	from this[THEN intD] aa1 tg2
	  show ?thesis by auto
      qed
      with g6[THEN intD] have g7: "w \<Turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_f" by auto
      have "\<turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_f \<longrightarrow> \<diamond>\<diamond>Q" by (rule E25)
      from this[THEN intD] g7 have "w \<Turnstile> \<diamond>\<diamond>Q"
	by auto
      with E10[of Q,THEN inteq_reflection]
       show "w \<Turnstile> \<diamond>Q" by auto
   qed
 qed
 hence g1: "\<turnstile> \<box>(\<box>[N]_f \<and> SF(A)_f \<and> \<box>F) \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)"
   by (rule STL4)
 have "\<turnstile> \<box>[N]_f \<and> SF(A)_f \<and> \<box>F \<longrightarrow> \<box>(\<box>[N]_f \<and> SF(A)_f \<and> \<box>F)"
 proof -
    have g1: "\<turnstile> (\<box>F) = (\<box>\<box>F)" by (rule T1)
    from this[THEN intD] have g1': "\<turnstile> (\<box>F) \<longrightarrow> (\<box>\<box>F)"
      by auto
    have g2: "\<turnstile> (\<box>[N]_f) = (\<box>\<box>[N]_f)" by (rule T2)
    from this[THEN intD] have g2': "\<turnstile> (\<box>[N]_f) \<longrightarrow> (\<box>\<box>[N]_f)"
      by auto      
    have g3: "\<turnstile> SF(A)_f \<longrightarrow> \<box>SF(A)_f" 
      by (rule AlwaysSFI)
    from g1'[THEN intD]
      g2'[THEN intD]
      g3[THEN intD]
    have g4: "\<turnstile> (\<box>[N]_f \<and> SF(A)_f \<and> \<box>F) \<longrightarrow> (\<box>\<box>[N]_f \<and> \<box>SF(A)_f \<and> \<box>\<box>F)"
      by (auto simp add: Valid_def)
    have "\<turnstile> (\<box>(\<box>[N]_f \<and> SF(A)_f)) = (\<box>\<box>[N]_f \<and> \<box>SF(A)_f)"
      by (rule STL5)
    from this[THEN intD] have "\<turnstile> \<box>\<box>[N]_f \<and> \<box>SF(A)_f \<longrightarrow> \<box>(\<box>[N]_f \<and> SF(A)_f)"
      by auto
    from g4[THEN intD] this[THEN intD]
    have g5: "\<turnstile> (\<box>[N]_f \<and> SF(A)_f \<and> \<box>F) \<longrightarrow> (\<box>(\<box>[N]_f \<and> SF(A)_f) \<and> \<box>\<box>F)"
      by auto
    have "\<turnstile> (\<box>((\<box>[N]_f \<and> SF(A)_f) \<and> \<box>F)) = (\<box>(\<box>[N]_f \<and> SF(A)_f) \<and> \<box>\<box>F)"
      by (rule STL5)
    from this[THEN intD] have "\<turnstile> \<box>(\<box>[N]_f \<and> SF(A)_f) \<and> \<box>\<box>F \<longrightarrow> \<box>((\<box>[N]_f \<and> SF(A)_f) \<and> \<box>F)"
      by auto
    with g5 have g6: "\<turnstile> (\<box>[N]_f \<and> SF(A)_f \<and> \<box>F) \<longrightarrow> \<box>((\<box>[N]_f \<and> SF(A)_f) \<and> \<box>F)"
      by (rule lift_imp_trans)
    have "\<turnstile> (\<box>[N]_f \<and> SF(A)_f) \<and> \<box>F \<longrightarrow> (\<box>[N]_f \<and> SF(A)_f \<and> \<box>F)" by auto
    hence "\<turnstile> \<box>((\<box>[N]_f \<and> SF(A)_f) \<and> \<box>F) \<longrightarrow> \<box>(\<box>[N]_f \<and> SF(A)_f \<and> \<box>F)"
      by (rule STL4)
    with g6 show ?thesis by (rule lift_imp_trans)
 qed
 from this g1 show "\<turnstile> \<box>[N]_f \<and> SF(A)_f \<and> \<box>F \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)"
   by (rule lift_imp_trans)
qed

theorem WF2:  assumes h1: "|~ \<langle>(N \<and> B)\<rangle>_f \<longrightarrow> \<langle>M\<rangle>_g"
               and   h2: "|~ P \<and> \<circ>P \<and> \<langle>(N \<and> A)\<rangle>_f \<longrightarrow> B"
               and   h3: "\<turnstile> P \<and> Enabled (\<langle>M\<rangle>_g) \<longrightarrow> Enabled (\<langle>A\<rangle>_f)"
               and   h4: "\<turnstile> \<box>[(N \<and> \<not>B)]_f \<and> (WF(A)_f) \<and> \<box>F \<and> 
                              (\<diamond>\<box>Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<diamond>\<box>P"
              shows     "\<turnstile> \<box>[N]_f \<and> (WF(A)_f) \<and> \<box>F \<longrightarrow> WF(M)_g"
proof -
  have "\<turnstile> \<box>[N]_f \<and> WF(A)_f \<and> \<box>F \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<box>\<diamond>\<langle>M\<rangle>_g"
  proof (int_unlift,intro impI,elim conjE, rule ccontr)
    fix w
    assume a1: "w \<Turnstile> \<box>[N]_f"
      and  a2: "w \<Turnstile> WF(A)_f"
      and  a3: "w \<Turnstile> \<box>F"
      and  a4: "w \<Turnstile> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g)"
      and  a5: "\<not>(w \<Turnstile> \<box>\<diamond>\<langle>M\<rangle>_g)"
    have a5': "w \<Turnstile> \<diamond>\<box>[\<not>M]_g"
    proof -
      from a5 have "w \<Turnstile> \<not>\<box>\<not>\<box>[\<not>M]_g" 
	by (auto simp add: angle_action_def)
      thus ?thesis by (simp add: eventually_def)
    qed
    from h1 have "\<turnstile> \<diamond>\<langle>(N \<and> B)\<rangle>_f \<longrightarrow> \<diamond>\<langle>M\<rangle>_g"
      by (rule AA25)
    hence "\<turnstile> \<box>\<diamond>\<langle>(N \<and> B)\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>M\<rangle>_g"
      by (rule STL4)
    from this[THEN intD] have "\<turnstile> \<not>\<box>\<diamond>\<langle>M\<rangle>_g \<longrightarrow> \<not>\<box>\<diamond>\<langle>(N \<and> B)\<rangle>_f"
      by (auto simp add: Valid_def)
    from this[THEN intD] a5 have g1: "w \<Turnstile> \<not>\<box>\<diamond>\<langle>(N \<and> B)\<rangle>_f"
      by auto
    hence "w \<Turnstile> \<not>\<box>\<not>\<box>[\<not>(N \<and> B)]_f" by (unfold angle_action_def)
    hence g1':"w \<Turnstile> \<diamond>\<box>[\<not>(N \<and> B)]_f" by (fold eventually_def)
    have "\<turnstile> \<box>[\<not>(N \<and> B)]_f \<longrightarrow> \<box>[N]_f \<longrightarrow> \<box>[(N \<and> \<not>B)]_f"
      by (rule MM11)
    hence g1'': "\<turnstile> \<diamond>\<box>[\<not>(N \<and> B)]_f \<longrightarrow> \<diamond>(\<box>[N]_f \<longrightarrow> \<box>[(N \<and> \<not>B)]_f)"
      by (rule E14)
    have "\<turnstile> \<diamond>(\<box>[N]_f \<longrightarrow> \<box>[(N \<and> \<not>B)]_f) \<longrightarrow> \<box>\<box>[N]_f \<longrightarrow> \<diamond>\<box>[(N \<and> \<not>B)]_f"
      by (rule E8)
    with g1'' have g1''': "\<turnstile> \<diamond>\<box>[\<not>(N \<and> B)]_f \<longrightarrow> \<box>\<box>[N]_f \<longrightarrow> \<diamond>\<box>[(N \<and> \<not>B)]_f"
      by (rule lift_imp_trans)
    have "\<turnstile> (\<box>[N]_f) = (\<box>\<box>[N]_f)" by (rule T2)
    from this[THEN intD] have "\<turnstile> \<box>[N]_f \<longrightarrow> \<box>\<box>[N]_f" by auto
    from this[THEN intD] a1 have a1': "w \<Turnstile> \<box>\<box>[N]_f" by auto
    with g1'''[THEN intD] g1' have g2: "w \<Turnstile> \<diamond>\<box>[(N \<and> \<not>B)]_f"
      by auto
    have g3: "(w \<Turnstile> \<diamond>\<box>[\<not>B]_f)"
    proof - 
      have "\<turnstile> \<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<box>[N]_f \<and> \<box>[\<not>B]_f" by (rule M6)
      hence tg1: "\<turnstile> \<diamond>\<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<diamond>(\<box>[N]_f \<and> \<box>[\<not>B]_f)"
	by (rule E14)
      have "\<turnstile> (\<diamond>\<box>[N]_f \<and> \<diamond>\<box>[\<not>B]_f) = (\<diamond>(\<box>[N]_f \<and> \<box>[\<not>B]_f))"
	by (rule STL6_act)
      from this[THEN intD] have "\<turnstile> \<diamond>(\<box>[N]_f \<and> \<box>[\<not>B]_f) \<longrightarrow> \<diamond>\<box>[\<not>B]_f"
	by (auto simp add: Valid_def)
      with tg1 have "\<turnstile> \<diamond>\<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<diamond>\<box>[\<not>B]_f"
	by (rule lift_imp_trans)
      from this[THEN intD] g2 show ?thesis by auto
    qed
    have "\<turnstile> (\<diamond>\<box>[\<not>B]_f) = (\<not>\<box>\<diamond>\<langle>B\<rangle>_f)" by (rule AA26)
    from this[THEN intD] g3 have g4: "w \<Turnstile> \<not>\<box>\<diamond>\<langle>B\<rangle>_f"
      by auto
    have g5: "w \<Turnstile> \<diamond>\<box>P"
    proof -
      have tg1: "\<turnstile> (\<diamond>\<box>\<box>[(N \<and> \<not>B)]_f \<and> \<diamond>\<box>((WF(A)_f) \<and> \<box>F \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g))) 
               = (\<diamond>\<box>(\<box>[(N \<and> \<not>B)]_f \<and> ((WF(A)_f) \<and> \<box>F \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g))))"
	by (rule STL6)
      have tg2: "\<turnstile> (\<diamond>\<box>(WF(A)_f) \<and> \<diamond>\<box>(\<box>F \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g)))=(\<diamond>\<box>((WF(A)_f) \<and> \<box>F \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g)))"
	by (rule STL6)	
      have tg3: "\<turnstile> (\<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<diamond>\<box>Enabled (\<langle>M\<rangle>_g)) =(\<diamond>\<box>(\<box>F \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g)))"
	by (rule STL6)
      from tg1[THEN inteq_reflection] tg2[THEN inteq_reflection] tg3[THEN inteq_reflection]
        have tg4: "\<turnstile> (\<diamond>\<box>\<box>[(N \<and> \<not>B)]_f \<and> \<diamond>\<box>(WF(A)_f) \<and> \<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<diamond>\<box>Enabled (\<langle>M\<rangle>_g))
	            \<longrightarrow> (\<diamond>\<box>(\<box>[(N \<and> \<not>B)]_f \<and> ((WF(A)_f) \<and> \<box>F \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g))))"
	by auto
      from h4 have "\<turnstile> \<box>(\<box>[(N \<and> \<not> B)]_f \<and> WF(A)_f \<and> \<box>F \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<box>\<diamond>\<box>P"
	by (rule STL4)
      hence "\<turnstile> \<diamond>\<box>(\<box>[(N \<and> \<not> B)]_f \<and> WF(A)_f \<and> \<box>F \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<diamond>\<box>\<diamond>\<box>P"
	by (rule E14)
      with tg4 have tg5: "\<turnstile> \<diamond>\<box>\<box>[(N \<and> \<not>B)]_f \<and> \<diamond>\<box>(WF(A)_f) \<and> \<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<diamond>\<box>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<diamond>\<box>\<diamond>\<box>P"
	by (rule lift_imp_trans)
      have  "\<turnstile> (\<box>[(N \<and> \<not>B)]_f) = (\<box>\<box>[(N \<and> \<not>B)]_f)" by (rule T2)
      from this[THEN intD] have "\<turnstile> \<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<box>\<box>[(N \<and> \<not>B)]_f"
	by auto
      hence mg1: "\<turnstile> \<diamond>\<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<diamond>\<box>\<box>[(N \<and> \<not>B)]_f" by (rule E14)
      have "\<turnstile> (\<box>WF(A)_f) = (WF(A)_f)" by (rule WF_always)
      from this[THEN intD] have mg2': "\<turnstile> WF(A)_f \<longrightarrow> \<box>WF(A)_f"
	by auto
      have  "\<turnstile> \<box>WF(A)_f \<longrightarrow> \<diamond>\<box>WF(A)_f" by (rule E3)
      with mg2' have mg2: "\<turnstile> WF(A)_f \<longrightarrow> \<diamond>\<box>WF(A)_f" 
	by (rule lift_imp_trans)
      have "\<turnstile> (\<box>F)=(\<box>\<box>F)" by (rule T1)
      from this[THEN intD] have mg3':"\<turnstile> \<box>F \<longrightarrow> \<box>\<box>F"
	by auto
      have "\<turnstile> \<box>\<box>F \<longrightarrow> \<diamond>\<box>\<box>F" by (rule E3)
      with mg3' have mg3: "\<turnstile> \<box>F \<longrightarrow> \<diamond>\<box>\<box>F" 
	by (rule lift_imp_trans)
      have mg4': "\<turnstile> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<box>\<diamond>\<box>Enabled (\<langle>M\<rangle>_g)" by (rule E19)
      have "\<turnstile> \<box>\<diamond>\<box>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<diamond>\<box>\<diamond>\<box>Enabled (\<langle>M\<rangle>_g)" by (rule E3)
      with mg4' have mg4: "\<turnstile> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<diamond>\<box>\<diamond>\<box>Enabled (\<langle>M\<rangle>_g)" 
	by (rule lift_imp_trans)
      from mg1[THEN intD] g2 mg2[THEN intD] a2 mg3[THEN intD] a3
      mg4[THEN intD] a4 tg5[THEN intD] have mg5: "w \<Turnstile> \<diamond>\<box>\<diamond>\<box>P"
	by auto
      have mg6': "\<turnstile> \<diamond>\<box>\<diamond>\<box>P \<longrightarrow> \<diamond>\<diamond>\<box>P" by (rule E16)
      have "\<turnstile> (\<diamond>\<box>P) = (\<diamond>\<diamond>\<box>P)" by (rule E10)
      from this[THEN intD] have "\<turnstile> \<diamond>\<diamond>\<box>P \<longrightarrow> \<diamond>\<box>P"
	by auto
      with mg6' have "\<turnstile> \<diamond>\<box>\<diamond>\<box>P \<longrightarrow> \<diamond>\<box>P"
	by (rule lift_imp_trans)
      from this[THEN intD] mg5 show ?thesis
	by auto
    qed
    from h3 have "\<turnstile> \<box>(P \<and> Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<box>Enabled (\<langle>A\<rangle>_f)"
      by (rule STL4)
    hence g6: "\<turnstile> \<diamond>\<box>(P \<and> Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<diamond>\<box>Enabled (\<langle>A\<rangle>_f)"
      by (rule E14)
    have "\<turnstile> (\<diamond>\<box>P \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g)) = (\<diamond>\<box>(P \<and> Enabled (\<langle>M\<rangle>_g)))"
      by (rule STL6)
    from this[THEN intD] have "\<turnstile> \<diamond>\<box>P \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<diamond>\<box>(P \<and> Enabled (\<langle>M\<rangle>_g))"
      by auto
    from this g6 have "\<turnstile> \<diamond>\<box>P \<and> \<diamond>\<box>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<diamond>\<box>Enabled (\<langle>A\<rangle>_f)"
      by (rule lift_imp_trans)
    from this[THEN intD] g5 a4 have g7: "w \<Turnstile> \<diamond>\<box>Enabled (\<langle>A\<rangle>_f)"
      by auto
    from a2 have "w \<Turnstile> \<diamond>\<box>Enabled (\<langle>A\<rangle>_f) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_f" by (simp add: WeakF_def)
    with g7 have g8: "w \<Turnstile> \<box>\<diamond>\<langle>A\<rangle>_f" by auto
    have "w \<Turnstile> \<box>\<diamond>\<langle>B\<rangle>_f"  
    proof -
      from h2[THEN prefD] have tg1: "|~ \<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f \<longrightarrow> \<langle>B\<rangle>_f"
	by (auto simp add: angle_actrans_def actrans_def)
      hence tg2: "\<turnstile> \<diamond>\<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f \<longrightarrow> \<diamond>\<langle>B\<rangle>_f"
	by (rule AA25)
      hence tg3: "\<turnstile> \<box>\<diamond>\<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule STL4)
      have tg4: "\<turnstile> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(N \<and> A)\<rangle>_f" by (rule AA29)
      have tg5: "\<turnstile> \<box>[\<circ>P]_f \<and> \<box>\<diamond>\<langle>(N \<and> A)\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(\<circ>P \<and> N \<and> A)\<rangle>_f" by (rule AA29)
      have tg6: "\<turnstile> \<box>[P]_f \<and> \<box>\<diamond>\<langle>(\<circ>P \<and> N \<and> A)\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f" by (rule AA29)
      from tg4[THEN intD] tg5[THEN intD] 
        have "\<turnstile> \<box>[\<circ>P]_f \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(\<circ>P \<and> N \<and> A)\<rangle>_f"
	by auto
      from this[THEN intD] tg6[THEN intD]
        have "\<turnstile> \<box>[P]_f \<and> \<box>[\<circ>P]_f \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f"
	by auto
      from this tg3 have tg7: "\<turnstile> \<box>[P]_f \<and> \<box>[\<circ>P]_f \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      have tg8: "\<turnstile> \<box>P \<longrightarrow> \<box>[P]_f" by (rule P4)
      have tg9: "\<turnstile> \<box>P \<longrightarrow> \<box>[\<circ>P]_f" by (rule T6)
      from tg7[THEN intD] tg8[THEN intD] tg9[THEN intD]
        have tg10: "\<turnstile> (\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by auto
      hence "\<turnstile> \<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f) \<longrightarrow> \<box>\<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule STL4)
      hence tg11: "\<turnstile> \<diamond>\<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f) \<longrightarrow> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule E14)
      have "\<turnstile> (\<diamond>\<box>\<box>P \<and> \<diamond>\<box>\<box>[N]_f) = (\<diamond>\<box>(\<box>P \<and> \<box>[N]_f))"
	by (rule STL6)
      from this[THEN intD]
        have tg12: "\<turnstile> \<diamond>\<box>\<box>P \<and> \<diamond>\<box>\<box>[N]_f \<longrightarrow> \<diamond>\<box>(\<box>P \<and> \<box>[N]_f)"
	by auto
      have "\<turnstile> (\<diamond>\<box>(\<box>P \<and> \<box>[N]_f) \<and> \<diamond>\<box>\<box>\<diamond>\<langle>A\<rangle>_f) = (\<diamond>\<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f))"
	by (rule STL6)
      from this[THEN intD] 
        have tg13: "\<turnstile> \<diamond>\<box>(\<box>P \<and> \<box>[N]_f) \<and> \<diamond>\<box>\<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f)"
	by auto
      from tg11[THEN intD] tg12[THEN intD] tg13[THEN intD]
      have tg14: "\<turnstile> \<diamond>\<box>\<box>P \<and> \<diamond>\<box>\<box>[N]_f \<and> \<diamond>\<box>\<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f"
	by auto
       from T1[of P,THEN intD] have "\<turnstile> \<box>P \<longrightarrow> \<box>\<box>P" 
	by auto
      hence tg15: "\<turnstile> \<diamond>\<box>P \<longrightarrow> \<diamond>\<box>\<box>P" by (rule E14)
      from T2[of N f,THEN intD] have "\<turnstile> \<box>[N]_f \<longrightarrow> \<box>\<box>[N]_f"
	by auto 
      with E3[THEN intD] have tg16: "\<turnstile> \<box>[N]_f \<longrightarrow> \<diamond>\<box>\<box>[N]_f"
	by (auto simp add: Valid_def)
      from T1[THEN intD] have "\<turnstile> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<box>\<diamond>\<langle>A\<rangle>_f"
 	by (auto simp add: Valid_def)
      with E3[THEN intD] have "\<turnstile> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<box>\<box>\<diamond>\<langle>A\<rangle>_f"
	by (auto simp add: Valid_def)      
      from tg14[THEN intD] tg15[THEN intD] tg16[THEN intD] this[THEN intD]
      have tg17: "\<turnstile> \<diamond>\<box>P \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f" by auto
      have "\<turnstile> \<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f" by (rule ax1)
      hence tg18: "\<turnstile> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<box>\<diamond>\<langle>B\<rangle>_f" by (rule E14)
      from AA28[of B f,THEN intD] have "\<turnstile> \<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<diamond>\<langle>B\<rangle>_f"
	by auto
      hence "\<turnstile> \<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<diamond>\<langle>B\<rangle>_f" by (rule STL4)
      hence "\<turnstile> \<diamond>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f" by (rule E14)
      with tg18 have tg19: "\<turnstile> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      have "\<turnstile> (\<diamond>\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f)=(\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f)" by (rule E27)
      from this[THEN intD] have "\<turnstile> \<diamond>\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<diamond>\<langle>B\<rangle>_f"
	by auto
      with tg19 have tg20: "\<turnstile> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      from AA28[of B f,THEN intD] have "\<turnstile> \<diamond>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<langle>B\<rangle>_f"
	by auto
      hence "\<turnstile> \<box>\<diamond>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f" by (rule STL4)
      with tg20 have "\<turnstile> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      with tg17 have "\<turnstile> \<diamond>\<box>P \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      from this[THEN intD] g5 a1 g8
        show ?thesis by auto
    qed
    with g3 g4 show False by simp 
  qed
  from this[THEN intD] show ?thesis by (auto simp add: WeakF_def)
qed

theorem SF2:  assumes h1: "|~ \<langle>(N \<and> B)\<rangle>_f \<longrightarrow> \<langle>M\<rangle>_g"
               and   h2: "|~ P \<and> \<circ>P \<and> \<langle>(N \<and> A)\<rangle>_f \<longrightarrow> B"
               and   h3: "\<turnstile> P \<and> Enabled (\<langle>M\<rangle>_g) \<longrightarrow> Enabled (\<langle>A\<rangle>_f)"
               and   h4: "\<turnstile> \<box>[(N \<and> \<not>B)]_f \<and> (SF(A)_f) \<and> \<box>F \<and> 
                              (\<box>\<diamond>Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<diamond>\<box>P"
              shows     "\<turnstile> \<box>[N]_f \<and> (SF(A)_f) \<and> \<box>F \<longrightarrow> SF(M)_g"
proof -
  have "\<turnstile> \<box>[N]_f \<and> SF(A)_f \<and> \<box>F \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<box>\<diamond>\<langle>M\<rangle>_g"
  proof (int_unlift,intro impI,elim conjE, rule ccontr)
    fix w
    assume a1: "w \<Turnstile> \<box>[N]_f"
      and  a2: "w \<Turnstile> SF(A)_f"
      and  a3: "w \<Turnstile> \<box>F"
      and  a4: "w \<Turnstile> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g)"
      and  a5: "\<not>(w \<Turnstile> \<box>\<diamond>\<langle>M\<rangle>_g)"
    have a5': "w \<Turnstile> \<diamond>\<box>[\<not>M]_g"
    proof -
      from a5 have "w \<Turnstile> \<not>\<box>\<not>\<box>[\<not>M]_g" 
	by (auto simp add: angle_action_def)
      thus ?thesis by (simp add: eventually_def)
    qed
    from h1 have "\<turnstile> \<diamond>\<langle>(N \<and> B)\<rangle>_f \<longrightarrow> \<diamond>\<langle>M\<rangle>_g"
      by (rule AA25)
    hence "\<turnstile> \<box>\<diamond>\<langle>(N \<and> B)\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>M\<rangle>_g"
      by (rule STL4)
    from this[THEN intD] have "\<turnstile> \<not>\<box>\<diamond>\<langle>M\<rangle>_g \<longrightarrow> \<not>\<box>\<diamond>\<langle>(N \<and> B)\<rangle>_f"
      by (auto simp add: Valid_def)
    from this[THEN intD] a5 have g1: "w \<Turnstile> \<not>\<box>\<diamond>\<langle>(N \<and> B)\<rangle>_f"
      by auto
    hence "w \<Turnstile> \<not>\<box>\<not>\<box>[\<not>(N \<and> B)]_f" by (unfold angle_action_def)
    hence g1':"w \<Turnstile> \<diamond>\<box>[\<not>(N \<and> B)]_f" by (fold eventually_def)
    have "\<turnstile> \<box>[\<not>(N \<and> B)]_f \<longrightarrow> \<box>[N]_f \<longrightarrow> \<box>[(N \<and> \<not>B)]_f"
      by (rule MM11)
    hence g1'': "\<turnstile> \<diamond>\<box>[\<not>(N \<and> B)]_f \<longrightarrow> \<diamond>(\<box>[N]_f \<longrightarrow> \<box>[(N \<and> \<not>B)]_f)"
      by (rule E14)
    have "\<turnstile> \<diamond>(\<box>[N]_f \<longrightarrow> \<box>[(N \<and> \<not>B)]_f) \<longrightarrow> \<box>\<box>[N]_f \<longrightarrow> \<diamond>\<box>[(N \<and> \<not>B)]_f"
      by (rule E8)
    with g1'' have g1''': "\<turnstile> \<diamond>\<box>[\<not>(N \<and> B)]_f \<longrightarrow> \<box>\<box>[N]_f \<longrightarrow> \<diamond>\<box>[(N \<and> \<not>B)]_f"
      by (rule lift_imp_trans)
    have "\<turnstile> (\<box>[N]_f) = (\<box>\<box>[N]_f)" by (rule T2)
    from this[THEN intD] have "\<turnstile> \<box>[N]_f \<longrightarrow> \<box>\<box>[N]_f" by auto
    from this[THEN intD] a1 have a1': "w \<Turnstile> \<box>\<box>[N]_f" by auto
    with g1'''[THEN intD] g1' have g2: "w \<Turnstile> \<diamond>\<box>[(N \<and> \<not>B)]_f"
      by auto
    have g3: "(w \<Turnstile> \<diamond>\<box>[\<not>B]_f)"
    proof - 
      have "\<turnstile> \<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<box>[N]_f \<and> \<box>[\<not>B]_f" by (rule M6)
      hence tg1: "\<turnstile> \<diamond>\<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<diamond>(\<box>[N]_f \<and> \<box>[\<not>B]_f)"
	by (rule E14)
      have "\<turnstile> (\<diamond>\<box>[N]_f \<and> \<diamond>\<box>[\<not>B]_f) = (\<diamond>(\<box>[N]_f \<and> \<box>[\<not>B]_f))"
	by (rule STL6_act)
      from this[THEN intD] have "\<turnstile> \<diamond>(\<box>[N]_f \<and> \<box>[\<not>B]_f) \<longrightarrow> \<diamond>\<box>[\<not>B]_f"
	by (auto simp add: Valid_def)
      with tg1 have "\<turnstile> \<diamond>\<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<diamond>\<box>[\<not>B]_f"
	by (rule lift_imp_trans)
      from this[THEN intD] g2 show ?thesis by auto
    qed
    have "\<turnstile> (\<diamond>\<box>[\<not>B]_f) = (\<not>\<box>\<diamond>\<langle>B\<rangle>_f)" by (rule AA26)
    from this[THEN intD] g3 have g4: "w \<Turnstile> \<not>\<box>\<diamond>\<langle>B\<rangle>_f"
      by auto
    have g5: "w \<Turnstile> \<diamond>\<box>P"
    proof -
      have tg1: "\<turnstile> (\<diamond>\<box>\<box>[(N \<and> \<not>B)]_f \<and> \<diamond>\<box>((SF(A)_f) \<and> \<box>F \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g))) 
               = (\<diamond>\<box>(\<box>[(N \<and> \<not>B)]_f \<and> ((SF(A)_f) \<and> \<box>F \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g))))"
	by (rule STL6)
      have tg2: "\<turnstile> (\<diamond>\<box>(SF(A)_f) \<and> \<diamond>\<box>(\<box>F \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g)))=(\<diamond>\<box>((SF(A)_f) \<and> \<box>F \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g)))"
	by (rule STL6)	
      have tg3: "\<turnstile> (\<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<box>\<diamond>Enabled (\<langle>M\<rangle>_g)) =(\<diamond>\<box>(\<box>F \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g)))"
	by (rule STL6)
      from tg1[THEN inteq_reflection] tg2[THEN inteq_reflection] tg3[THEN inteq_reflection]
        have tg4: "\<turnstile> (\<diamond>\<box>\<box>[(N \<and> \<not>B)]_f \<and> \<diamond>\<box>(SF(A)_f) \<and> \<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<box>\<diamond>Enabled (\<langle>M\<rangle>_g))
	            \<longrightarrow> (\<diamond>\<box>(\<box>[(N \<and> \<not>B)]_f \<and> ((SF(A)_f) \<and> \<box>F \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g))))"
	by auto
      from h4 have "\<turnstile> \<box>(\<box>[(N \<and> \<not> B)]_f \<and> SF(A)_f \<and> \<box>F \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<box>\<diamond>\<box>P"
	by (rule STL4)
      hence "\<turnstile> \<diamond>\<box>(\<box>[(N \<and> \<not> B)]_f \<and> SF(A)_f \<and> \<box>F \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<diamond>\<box>\<diamond>\<box>P"
	by (rule E14)
      with tg4 have tg5: "\<turnstile> \<diamond>\<box>\<box>[(N \<and> \<not>B)]_f \<and> \<diamond>\<box>(SF(A)_f) \<and> \<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<box>\<diamond>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<diamond>\<box>\<diamond>\<box>P"
	by (rule lift_imp_trans)
      have  "\<turnstile> (\<box>[(N \<and> \<not>B)]_f) = (\<box>\<box>[(N \<and> \<not>B)]_f)" by (rule T2)
      from this[THEN intD] have "\<turnstile> \<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<box>\<box>[(N \<and> \<not>B)]_f"
	by auto
      hence mg1: "\<turnstile> \<diamond>\<box>[(N \<and> \<not>B)]_f \<longrightarrow> \<diamond>\<box>\<box>[(N \<and> \<not>B)]_f" by (rule E14)
      have "\<turnstile> (\<box>SF(A)_f) = (SF(A)_f)" by (rule SF_always)
      from this[THEN intD] have mg2': "\<turnstile> SF(A)_f \<longrightarrow> \<box>SF(A)_f"
	by auto
      have  "\<turnstile> \<box>SF(A)_f \<longrightarrow> \<diamond>\<box>SF(A)_f" by (rule E3)
      with mg2' have mg2: "\<turnstile> SF(A)_f \<longrightarrow> \<diamond>\<box>SF(A)_f" 
	by (rule lift_imp_trans)
      have "\<turnstile> (\<box>F)=(\<box>\<box>F)" by (rule T1)
      from this[THEN intD] have mg3':"\<turnstile> \<box>F \<longrightarrow> \<box>\<box>F"
	by auto
      have "\<turnstile> \<box>\<box>F \<longrightarrow> \<diamond>\<box>\<box>F" by (rule E3)
      with mg3' have mg3: "\<turnstile> \<box>F \<longrightarrow> \<diamond>\<box>\<box>F" 
	by (rule lift_imp_trans)
      thm T1
      have "\<turnstile> (\<box>\<diamond>Enabled (\<langle>M\<rangle>_g))=(\<box>\<box>\<diamond>Enabled (\<langle>M\<rangle>_g))" by (rule T1)
      from this[THEN intD] have mg4': "\<turnstile> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<box>\<box>\<diamond>Enabled (\<langle>M\<rangle>_g)" 
	by auto
      have "\<turnstile> \<box>\<box>\<diamond>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<diamond>\<box>\<box>\<diamond>Enabled (\<langle>M\<rangle>_g)" by (rule E3)
      with mg4' have mg4: "\<turnstile> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<diamond>\<box>\<box>\<diamond>Enabled (\<langle>M\<rangle>_g)" 
	by (rule lift_imp_trans)
      from mg1[THEN intD] g2 mg2[THEN intD] a2 mg3[THEN intD] a3
      mg4[THEN intD] a4 tg5[THEN intD] have mg5: "w \<Turnstile> \<diamond>\<box>\<diamond>\<box>P"
	by auto
      have mg6': "\<turnstile> \<diamond>\<box>\<diamond>\<box>P \<longrightarrow> \<diamond>\<diamond>\<box>P" by (rule E16)
      have "\<turnstile> (\<diamond>\<box>P) = (\<diamond>\<diamond>\<box>P)" by (rule E10)
      from this[THEN intD] have "\<turnstile> \<diamond>\<diamond>\<box>P \<longrightarrow> \<diamond>\<box>P"
	by auto
      with mg6' have "\<turnstile> \<diamond>\<box>\<diamond>\<box>P \<longrightarrow> \<diamond>\<box>P"
	by (rule lift_imp_trans)
      from this[THEN intD] mg5 show ?thesis
	by auto
    qed
    from h3 have "\<turnstile> \<diamond>(P \<and> Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<diamond>Enabled (\<langle>A\<rangle>_f)"
      by (rule E14)
    hence g6: "\<turnstile> \<box>\<diamond>(P \<and> Enabled (\<langle>M\<rangle>_g)) \<longrightarrow> \<box>\<diamond>Enabled (\<langle>A\<rangle>_f)"
      by (rule STL4)
    have "\<turnstile> \<diamond>\<box>P \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<box>\<diamond>(P \<and> Enabled (\<langle>M\<rangle>_g))"
      by (rule E28)
    from this g6 have "\<turnstile> \<diamond>\<box>P \<and> \<box>\<diamond>Enabled (\<langle>M\<rangle>_g) \<longrightarrow> \<box>\<diamond>Enabled (\<langle>A\<rangle>_f)"
      by (rule lift_imp_trans)
    from this[THEN intD] g5 a4 have g7: "w \<Turnstile> \<box>\<diamond>Enabled (\<langle>A\<rangle>_f)"
      by auto
    from a2 have "w \<Turnstile> \<box>\<diamond>Enabled (\<langle>A\<rangle>_f) \<longrightarrow> \<box>\<diamond>\<langle>A\<rangle>_f" by (simp add: StrongF_def)
    with g7 have g8: "w \<Turnstile> \<box>\<diamond>\<langle>A\<rangle>_f" by auto
    have "w \<Turnstile> \<box>\<diamond>\<langle>B\<rangle>_f"  
    proof -
      from h2[THEN prefD] have tg1: "|~ \<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f \<longrightarrow> \<langle>B\<rangle>_f"
	by (auto simp add: angle_actrans_def actrans_def)
      hence tg2: "\<turnstile> \<diamond>\<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f \<longrightarrow> \<diamond>\<langle>B\<rangle>_f"
	by (rule AA25)
      hence tg3: "\<turnstile> \<box>\<diamond>\<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule STL4)
      have tg4: "\<turnstile> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(N \<and> A)\<rangle>_f" by (rule AA29)
      have tg5: "\<turnstile> \<box>[\<circ>P]_f \<and> \<box>\<diamond>\<langle>(N \<and> A)\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(\<circ>P \<and> N \<and> A)\<rangle>_f" by (rule AA29)
      have tg6: "\<turnstile> \<box>[P]_f \<and> \<box>\<diamond>\<langle>(\<circ>P \<and> N \<and> A)\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f" by (rule AA29)
      from tg4[THEN intD] tg5[THEN intD] 
        have "\<turnstile> \<box>[\<circ>P]_f \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(\<circ>P \<and> N \<and> A)\<rangle>_f"
	by auto
      from this[THEN intD] tg6[THEN intD]
        have "\<turnstile> \<box>[P]_f \<and> \<box>[\<circ>P]_f \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>(P \<and> \<circ>P \<and> N \<and> A)\<rangle>_f"
	by auto
      from this tg3 have tg7: "\<turnstile> \<box>[P]_f \<and> \<box>[\<circ>P]_f \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      have tg8: "\<turnstile> \<box>P \<longrightarrow> \<box>[P]_f" by (rule P4)
      have tg9: "\<turnstile> \<box>P \<longrightarrow> \<box>[\<circ>P]_f" by (rule T6)
      from tg7[THEN intD] tg8[THEN intD] tg9[THEN intD]
        have tg10: "\<turnstile> (\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by auto
      hence "\<turnstile> \<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f) \<longrightarrow> \<box>\<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule STL4)
      hence tg11: "\<turnstile> \<diamond>\<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f) \<longrightarrow> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule E14)
      have "\<turnstile> (\<diamond>\<box>\<box>P \<and> \<diamond>\<box>\<box>[N]_f) = (\<diamond>\<box>(\<box>P \<and> \<box>[N]_f))"
	by (rule STL6)
      from this[THEN intD]
        have tg12: "\<turnstile> \<diamond>\<box>\<box>P \<and> \<diamond>\<box>\<box>[N]_f \<longrightarrow> \<diamond>\<box>(\<box>P \<and> \<box>[N]_f)"
	by auto
      have "\<turnstile> (\<diamond>\<box>(\<box>P \<and> \<box>[N]_f) \<and> \<diamond>\<box>\<box>\<diamond>\<langle>A\<rangle>_f) = (\<diamond>\<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f))"
	by (rule STL6)
      from this[THEN intD] 
        have tg13: "\<turnstile> \<diamond>\<box>(\<box>P \<and> \<box>[N]_f) \<and> \<diamond>\<box>\<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<box>((\<box>P \<and> \<box>[N]_f) \<and> \<box>\<diamond>\<langle>A\<rangle>_f)"
	by auto
      from tg11[THEN intD] tg12[THEN intD] tg13[THEN intD]
      have tg14: "\<turnstile> \<diamond>\<box>\<box>P \<and> \<diamond>\<box>\<box>[N]_f \<and> \<diamond>\<box>\<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f"
	by auto
       from T1[of P,THEN intD] have "\<turnstile> \<box>P \<longrightarrow> \<box>\<box>P" 
	by auto
      hence tg15: "\<turnstile> \<diamond>\<box>P \<longrightarrow> \<diamond>\<box>\<box>P" by (rule E14)
      from T2[of N f,THEN intD] have "\<turnstile> \<box>[N]_f \<longrightarrow> \<box>\<box>[N]_f"
	by auto 
      with E3[THEN intD] have tg16: "\<turnstile> \<box>[N]_f \<longrightarrow> \<diamond>\<box>\<box>[N]_f"
	by (auto simp add: Valid_def)
      from T1[THEN intD] have "\<turnstile> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<box>\<diamond>\<langle>A\<rangle>_f"
 	by (auto simp add: Valid_def)
      with E3[THEN intD] have "\<turnstile> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<box>\<box>\<diamond>\<langle>A\<rangle>_f"
	by (auto simp add: Valid_def)      
      from tg14[THEN intD] tg15[THEN intD] tg16[THEN intD] this[THEN intD]
      have tg17: "\<turnstile> \<diamond>\<box>P \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f" by auto
      have "\<turnstile> \<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f" by (rule ax1)
      hence tg18: "\<turnstile> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<box>\<diamond>\<langle>B\<rangle>_f" by (rule E14)
      from AA28[of B f,THEN intD] have "\<turnstile> \<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<diamond>\<langle>B\<rangle>_f"
	by auto
      hence "\<turnstile> \<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<diamond>\<langle>B\<rangle>_f" by (rule STL4)
      hence "\<turnstile> \<diamond>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f" by (rule E14)
      with tg18 have tg19: "\<turnstile> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      have "\<turnstile> (\<diamond>\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f)=(\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f)" by (rule E27)
      from this[THEN intD] have "\<turnstile> \<diamond>\<box>\<diamond>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<diamond>\<langle>B\<rangle>_f"
	by auto
      with tg19 have tg20: "\<turnstile> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      from AA28[of B f,THEN intD] have "\<turnstile> \<diamond>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<diamond>\<langle>B\<rangle>_f"
	by auto
      hence "\<turnstile> \<box>\<diamond>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f" by (rule STL4)
      with tg20 have "\<turnstile> \<diamond>\<box>\<box>\<diamond>\<langle>B\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      with tg17 have "\<turnstile> \<diamond>\<box>P \<and> \<box>[N]_f \<and> \<box>\<diamond>\<langle>A\<rangle>_f \<longrightarrow> \<box>\<diamond>\<langle>B\<rangle>_f"
	by (rule lift_imp_trans)
      from this[THEN intD] g5 a1 g8
        show ?thesis by auto
    qed
    with g3 g4 show False by simp 
  qed
  from this[THEN intD] show ?thesis by (auto simp add: StrongF_def)
qed

text {* This is the lattice rule from TLA *}
theorem wf_leadsto:
  assumes h1: "wf r"
    and   h2: "\<forall> x. w \<Turnstile> F x \<leadsto> (G \<or> (\<exists> y. #((y,x) \<in> r) \<and> F y))"
  shows       "w \<Turnstile> F x \<leadsto> G"
using h1
proof (rule wf_induct)
  fix x
  assume a1: "\<forall> y. ((y, x) \<in> r) \<longrightarrow> (w \<Turnstile> (F y \<leadsto> G))"  
  from h2 have h2' :"w \<Turnstile> F x \<leadsto> (G \<or> (\<exists> y. #((y,x) \<in> r) \<and> F y))" ..
  have "\<turnstile> (F x \<leadsto> (G \<or> (\<exists> y. #((y,x) \<in> r) \<and> F y))) 
        \<and> ((\<exists> y. #((y,x) \<in> r) \<and> F y) \<leadsto> G) \<longrightarrow> (F x \<leadsto> G)"
    by (rule LT19)
  from this[THEN intD] h2'
    have g1: "w \<Turnstile> ((\<exists> y. #((y,x) \<in> r) \<and> F y) \<leadsto> G) \<longrightarrow> (F x \<leadsto> G)"
      by auto
  have "\<turnstile> ((\<exists> y. #((y,x) \<in> r) \<and> F y) \<leadsto> G) = (\<forall> y. ((#((y,x) \<in> r) \<and> F y) \<leadsto> G))"
    by (rule LT21)
  from this[THEN intD]
    have g2: "\<turnstile> (\<forall> y. ((#((y,x) \<in> r) \<and> F y) \<leadsto> G)) \<longrightarrow> ((\<exists> y. #((y,x) \<in> r) \<and> F y) \<leadsto> G)"
      by auto
  have "\<forall> y. w \<Turnstile> ((#((y,x) \<in> r) \<and> F y) \<leadsto> G)"
  proof (rule allI, unfold leadsto_def)
    fix y
    from a1 have "(y, x) \<in> r \<longrightarrow> (w \<Turnstile> (F y \<leadsto> G))" ..
    hence tg1: "(y, x) \<in> r \<longrightarrow> (w \<Turnstile> \<box>(F y \<longrightarrow> \<diamond>G))"
      by (unfold leadsto_def)
    show "(w \<Turnstile> \<box>((#((y,x) \<in> r) \<and> F y) \<longrightarrow> \<diamond>G))"
    proof (cases "(y, x) \<in> r")
      assume aa1: "(y, x) \<in> r"
      from aa1 tg1 have tg2: "(w \<Turnstile> \<box>(F y \<longrightarrow> \<diamond>G))" by simp
      have tg3: "(y, x) \<in> r \<longrightarrow> (w \<Turnstile> #((y, x) \<in> r))" by auto
      have "\<turnstile> (#((y, x) \<in> r)) = (\<box>(#((y, x) \<in> r)))" by (rule MM4)
      hence "w \<Turnstile> #((y, x) \<in> r) \<longrightarrow> \<box>(#((y, x) \<in> r))" by auto
      with tg3 have "(y, x) \<in> r \<longrightarrow> (w \<Turnstile> \<box>(#((y, x) \<in> r)))" by auto
      with aa1 have tg4: "w \<Turnstile> \<box>(#((y, x) \<in> r))" by auto
      have "\<turnstile> \<box>(#((y, x) \<in> r)) \<and> \<box>(F y \<longrightarrow> \<diamond>G) \<longrightarrow> \<box>((#((y,x) \<in> r) \<and> F y) \<longrightarrow> \<diamond>G)" 
	by (rule MM2)
      from this[THEN intD] tg2 tg4 show "w \<Turnstile>  \<box>((#((y,x) \<in> r) \<and> F y) \<longrightarrow> \<diamond>G)"
	by auto
    next
      assume aa1: "(y, x) \<notin> r"
      hence aa1': "\<not>((y, x) \<in> r)" by auto
      have tg2: "\<not>((y, x) \<in> r) \<longrightarrow> (w \<Turnstile> \<not>(#((y, x) \<in> r)))" by auto
      have "\<turnstile> (\<not>#((y, x) \<in> r)) = (\<box>(\<not>#((y, x) \<in> r)))" by (rule MM4b)
      hence "w \<Turnstile> (\<not>#((y, x) \<in> r)) \<longrightarrow> \<box>(\<not>#((y, x) \<in> r))" by auto
      with tg2 have "\<not>((y, x) \<in> r) \<longrightarrow> (w \<Turnstile> \<box>(\<not>#((y, x) \<in> r)))" by auto
      with aa1 have tg3: "w \<Turnstile> \<box>(\<not>#((y, x) \<in> r))" by auto
      have "\<turnstile> \<box>(\<not>#((y, x) \<in> r)) \<longrightarrow> \<box>((#((y,x) \<in> r) \<and> F y) \<longrightarrow> \<diamond>G)" by (rule MM3)
      from this[THEN intD] tg3 show "w \<Turnstile> \<box>((#((y,x) \<in> r) \<and> F y) \<longrightarrow> \<diamond>G)" by auto
    qed
  qed
  with g2[THEN intD] have "w \<Turnstile> ((\<exists> y. #((y,x) \<in> r) \<and> F y) \<leadsto> G)" by auto
  with g1 show "w \<Turnstile> (F x \<leadsto> G)" by auto
qed

section "Stuttering Invariance"

theorem stut_Enabled: "STUTINV (Enabled (\<langle>F\<rangle>_v))"
proof(unfold enabled_def stutinv_def,(rule allI)+,rule impI)
  fix s t :: "nat \<Rightarrow> 'a"
  assume a1: "s \<approx> t"
  hence g1: "first s = first t" by (rule sim_first)
  show "(\<exists> k. (first s ## k) \<Turnstile> \<langle>F\<rangle>_v)=(\<exists> k. (first t ## k) \<Turnstile> \<langle>F\<rangle>_v)"
  proof
    assume a1: "\<exists> k. (first s ## k) \<Turnstile> \<langle>F\<rangle>_v"
    with g1 show "\<exists> k. (first t ## k) \<Turnstile> \<langle>F\<rangle>_v"
      by auto
  next
    assume a1: "\<exists> k. (first t ## k) \<Turnstile> \<langle>F\<rangle>_v"
    with g1 show "\<exists> k. (first s ## k) \<Turnstile> \<langle>F\<rangle>_v"
      by auto
  qed
qed

theorem stut_WF: assumes H: "NSTUTINV F" shows "STUTINV WF(F)_v"
  using H
  by (unfold WeakF_def) (auto simp only: stut_Enabled bothstutinvs)

theorem stut_SF: assumes H: "NSTUTINV F" shows "STUTINV SF(F)_v"
  using H
  by (unfold StrongF_def) (auto simp only: stut_Enabled bothstutinvs)

lemmas livestutinv = stut_WF stut_SF stut_Enabled

section "Some Higher Level Proof Rules for Monolithic Specifications"

theorem live_invmono: assumes h1: "\<turnstile> I \<longrightarrow> P"
                   and   h2: "|~ P \<and> [N]_f \<longrightarrow> \<circ>P" 
                  shows      "\<turnstile> I \<and> \<box>[N]_f \<and> Live \<longrightarrow> \<box>P"
proof (int_unlift,rule impI,elim conjE)
  fix w
  assume a1: "w \<Turnstile> I"
    and a2: "w \<Turnstile> \<box>[N]_f"
  from h1 h2 have "\<turnstile> I \<and> \<box>[N]_f \<longrightarrow> \<box>P"
    by (rule invmono)
  from this[THEN intD] a1 a2 show "w \<Turnstile> \<box>P" by auto
qed

theorem live_refinementmono: assumes h1: "\<turnstile> P \<longrightarrow> Q"
  and h2: "|~ ([A]_f) \<longrightarrow> [B]_g"
  and h3: "\<turnstile> P \<and> \<box>[A]_f \<and> L1 \<longrightarrow> L2"
  shows "\<turnstile> P \<and> \<box>[A]_f \<and> L1 \<longrightarrow> Q \<and> \<box>[B]_g \<and> L2" 
proof -
  from h1 h2 have "\<turnstile> P \<and> \<box>[A]_f \<longrightarrow> Q \<and> \<box>[B]_g" 
    by (rule refinement1)
  from this[THEN intD] have
    "\<turnstile> P \<and> \<box>[A]_f \<and> L1 \<longrightarrow> Q \<and> \<box>[B]_g" by auto
  from this[THEN intD] h3[THEN intD]
  show ?thesis by auto
qed

theorem strenghen_live_invmono:
  assumes h1: "\<turnstile> I \<and> J \<longrightarrow> P"
  and h2: "|~ P \<and> [N \<and> J \<and> \<circ>J]_f \<longrightarrow> \<circ>P"
  and h3: "\<turnstile> I \<and> \<box>[N]_f \<and> Live \<longrightarrow> \<box>J"
  shows      "\<turnstile> I \<and> \<box>[N]_f \<and> Live \<longrightarrow> \<box>P"
proof (int_unlift,intro impI,elim conjE)
  fix w
  assume a1: "w \<Turnstile> I"
  and a2: "w \<Turnstile> \<box>[N]_f"
  and a3: "w \<Turnstile> Live"
  from h3[THEN intD,of w] a1 a2 a3 have g1: "w \<Turnstile> \<box>J"
    by simp
  with ax1[of J,THEN intD,of w] have "w \<Turnstile> J"
    by auto
  with h1[THEN intD] a1 have g2: "w \<Turnstile> P"
    by auto
  thm INV2
  have "\<turnstile> \<box>J \<longrightarrow> \<box>[N]_f = \<box>[N \<and> J \<and> \<circ>J]_f"
    by (rule INV2)
  from this[THEN intD,of w] g1 have g3: "(w \<Turnstile> \<box>[N]_f)=(w \<Turnstile> \<box>[N \<and> J \<and> \<circ>J]_f)"
    by auto
  with a2 have g4: "w \<Turnstile> \<box>[N \<and> J \<and> \<circ>J]_f" by simp
  from h2 have "\<turnstile> P \<and> \<box>[N \<and> J \<and> \<circ>J]_f \<longrightarrow> \<box>P"
    by (rule INV1)
  from this[THEN intD,of w] g2 g4 show "w \<Turnstile> \<box>P"
    by auto
qed

end
