header {* This is the 1 heap version which will be used in the paper *}

theory EvenOdd2
imports Integrated
begin

constdefs
 emptyHeap :: "Heap"
 "emptyHeap == fimap(empty)"

consts CONS :: ConName
consts NIL  :: ConName
    

consts mList::"(nat \<times> Locn \<times> (Locn set) \<times> Heap) set"
inductive mList intros
mListNIL: "\<lbrakk> fmap_lookup h a = Some (C NIL []) \<rbrakk> \<Longrightarrow> (0,a,{a},h) : mList"
mListCONS:"\<lbrakk> fmap_lookup h a = Some (C CONS [rx,rxs]) ; a \<notin> U; (Ref b) = rxs ; (n,b,U,h):mList\<rbrakk>
          \<Longrightarrow> (Suc n, a, U \<union> {a}, h): mList"
mListRef: "\<lbrakk> fmap_lookup h a = Some (R rx) ; a \<notin> U; (Ref b) = rx ; (n,b,U,h):mList\<rbrakk>
          \<Longrightarrow> (Suc n, a, U \<union> {a}, h): mList"

consts locs :: "Locn \<Rightarrow> Heap \<Rightarrow> Locn set"
defs
  locs: "locs l h \<equiv> \<Union>{x. \<exists> n. (n,l,x,h) \<in> mList}"

lemma "(n,l,y,h) \<in> mList \<and> (m,l,x,h) \<in> mList \<and> m < n \<longrightarrow> y \<subseteq> x"
oops

lemma "locs l h = (THE x. \<exists> m. \<forall> n y. (n,l,y,h) \<in> mList \<and> (m,l,x,h) \<in> mList \<longrightarrow> n <= m)"
oops

consts
  heap_restrict :: "Heap \<Rightarrow> (Locn set) \<Rightarrow> Heap"
  loc_restrict :: "Heap \<Rightarrow> Locn \<Rightarrow> Heap"
  HoVal_restrict :: "Heap \<Rightarrow> HoVal \<Rightarrow> Heap"

defs
  heap_restrict: "heap_restrict h ls \<equiv> fimap ((themap h) |` ls)"
  loc_restrict: "loc_restrict h l \<equiv> heap_restrict h (locs l h)"

consts simplecopy :: "HoVal \<Rightarrow> Heap \<Rightarrow> HoVal * Heap"

defs 
  simplecopy: "simplecopy v h \<equiv> (v,h)"

section "Coordination layer state"

datatype PC = peven | podd

consts
  pc :: "PC statefun" (* the program counter - sequential execution of boxes *)
  odd_st :: "BoxState statefun" (* state/mode of box *)
  even_st :: "BoxState statefun" (* state/mode of box *)
  (* the heap *)
  programheap :: "Heap statefun"
  (* Auxiliary tempory heaps *)
  aeheap :: "Heap statefun"
  aoheap :: "Heap statefun"
  even_res :: "HoVal statefun"
  odd_res :: "HoVal statefun"
  (* the wire heap *)
  w1 :: "HoVal statefun"
  w2 :: "HoVal statefun"
  rso1 :: "Rescomp statefun"
  rso2 :: "Rescomp statefun"

section "The Expession VDM layer"

syntax 
 x1  :: "VarName"
 xin :: "VarName"

translations
"x1" == "(VN ''x1'')"
"xin" == "(VN ''xin'')"

constdefs iplus :: "HoVal => HoVal => HoVal"
"iplus x y == I ((theInt x)+(theInt y))"
constdefs iinc :: "HoVal => HoVal => HoVal"
"iinc x y == I ((theInt x)+(1::int))"

(* heap version, ie. arg is a pointer to a heap cell containing an int *)
constdefs incbody :: "Exp"
"incbody == CaseOneExp xin (PVar x1) (RPrimBin iinc x1 x1) (ValueExp BOTTOM)"

(* stack version, ie. arg is an int *)
constdefs incbodyI :: "Exp"
"incbodyI == (PrimBin iinc xin xin)"

theorem incbody_spec: "{} \<rhd> incbody : (\<lambda> E h hh v p. 
  (\<forall> i r . (E xin = (R (Ref r)) \<and> (fmap_lookup h r = Some (I i)))  \<longrightarrow> v = (I (i+1))))"
apply (simp add: incbody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
apply (rule vdmRPrimBin)
apply (rule vdmValue)
apply clarsimp
apply (erule_tac x="E(x1:=(R (Ref r)))" in allE)
apply (erule_tac x="R (Ref r)" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPVar)
apply (simp add: iinc_def)
done

theorem incbodyI_spec: "{} \<rhd> incbodyI : (\<lambda> E h hh v p. 
  (\<forall> i r . (E xin = (I i))  \<longrightarrow> v = (I (i+1))))"
apply (simp add: incbodyI_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmPrimBin)
apply clarsimp
apply (simp add: iinc_def)
done

(* r1 and r2 must be disjoint!!!  *)
theorem incbody_spec_v2: "{} \<rhd> incbody : (\<lambda> E h hh v p. 
  (\<forall> i r1 r2. (E xin = (R (Ref r1)) \<and> (fmap_lookup h r1 = Some (I i))  \<longrightarrow> (v = (R (Ref r2))) 
\<and> (r2 \<notin> fmap_dom h) \<and> (fmap_lookup hh r2 = Some (I (i+1))) \<and> (\<forall> r3. r3 \<noteq> r2 \<and> r3 \<noteq> r1 \<longrightarrow> fmap_lookup h r3 = fmap_lookup hh r3)  )))"
sorry

theorem incbody_unch: "{} \<rhd> incbody : (\<lambda> E h hh v p. 
  (\<forall> r1 . (E xin = (R (Ref r1)) \<longrightarrow> (\<forall> r2 \<in> fmap_dom h. r2 \<noteq> r1 \<longrightarrow> fmap_loopup h r2 = fmap_lookup hh r2))))"
  sorry

theorem incbody_spec_v2_w1: "{} \<rhd> incbody : (\<lambda> E h hh v p. 
  (\<forall> i r1 r2 . (E xin = (R (Ref r1))  \<longrightarrow> (v = (R (Ref r2))) \<and>
  r1 \<noteq> r2)))"
sorry

section "The coordination layer"

constdefs
  S' :: "temporal"
  "S' \<equiv> TEMP (sch$ = (if isExe \<and> $pc = #podd then #Super else #Execute))
           \<and> (isExe \<longrightarrow> pc$ = #peven)"

subsection "Even Box"

consts
 evenoddExe :: "HoVal \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"

defs
  evenoddExe: "evenoddExe val heap \<equiv> exe (emptyEnv(xin := val)) heap incbody"

consts
  even :: "temporal"
  odd :: "temporal"
  super :: "temporal"
  init :: "temporal"
  program :: "temporal"

(* note: assumes that returned ref is different
  from existing !!! *)
(* as an alternative: the input wire (just ref) 
   can/should be consumed in the expression/VDM
  layer!!! *)

defs 
 even: "even \<equiv> TEMP
(if $pc \<noteq> #peven
 then Unchanged (w2,even_res,even_st,aeheap)
 else pc$ = #podd \<and> 
 (if $even_st = #Blocked
   then Unchanged (w2,even_res,even_st,programheap,aeheap)
   else 
 (if $w2 = #BOTTOM
   then Unchanged (w2,programheap,aeheap)
     \<and>  even_res$ = #BOTTOM 
     \<and>  even_st$ = #Matchfail
   else (even_res,programheap,rso1)$ = evenoddExe<$w2,$programheap>
    \<and> (w2,aeheap)$ = (#BOTTOM,$aeheap))))"
defs
 odd: "odd \<equiv> TEMP
(if $pc \<noteq> #podd
 then Unchanged (w1,odd_res,odd_st,aoheap)
 else pc$ = #peven \<and> 
 (if $odd_st = #Blocked
   then Unchanged (w1,odd_res,odd_st,aoheap,programheap)
   else 
 (if $w1 = #BOTTOM
   then Unchanged (w1,programheap,aoheap)
     \<and>  odd_res$ = #BOTTOM 
     \<and>  odd_st$ = #Matchfail
   else (odd_res,programheap,rso2)$ = evenoddExe<$w1,$programheap>
    \<and> (w1,aoheap)$ = (#BOTTOM,$aoheap))))"
(* add consume !!! *)

(* add tempheap??? *)
defs
 super: "super \<equiv> TEMP
    pc$ = #peven
 \<and> Unchanged (aeheap,aoheap,programheap)
 \<and> (if $w1 \<noteq> #BOTTOM \<and> $even_res \<noteq> #BOTTOM
    then even_st$ = #Blocked \<and> Unchanged (w1,even_res)  
    else (if $even_res = #BOTTOM
    then even_st$ = #Runnable \<and> Unchanged (w1,even_res)
    else even_st$ = #Runnable \<and> w1$ = $even_res \<and> even_res$ = #BOTTOM))
 \<and> (if $w2 \<noteq> #BOTTOM \<and> $odd_res \<noteq> #BOTTOM
    then odd_st$ = #Blocked \<and> Unchanged (w2,odd_res)  
    else (if $odd_res = #BOTTOM
    then odd_st$ = #Runnable \<and> Unchanged (w2,odd_res)
    else odd_st$ = #Runnable \<and> w2$ = $odd_res \<and> odd_res$ = #BOTTOM))"

defs
  init: "init \<equiv> TEMP
    $sch = #Execute \<and> $pc = #peven
  \<and> $programheap = fmap_upd<#emptyHeap,(freshloc<#{}>),#(I 0)>
  \<and> $(even_st,odd_st) = #(Runnable,Runnable)
  \<and> $(aeheap,aoheap) = #(emptyHeap,emptyHeap)
  \<and> $(even_res,odd_res) = #(BOTTOM,BOTTOM)
  \<and> $(w1,w2) = #(R (Ref (freshloc {})),BOTTOM)"


defs
  program: "program \<equiv> TEMP
   init \<and> \<box>[S' \<and> (isExe \<longrightarrow> even \<and> odd) \<and> (isSup \<longrightarrow> super)]_(pc,sch,odd_st,
                   even_st,aeheap,aoheap,even_res,odd_res,programheap,w1,w2)"

section "Properties"

consts EvenN :: "int set"
       OddN  :: "int set"

inductive EvenN OddN 
intros
 zero: "0 \<in> EvenN" 
 one:  "1 \<in> OddN"
 plus1: "n \<in> OddN \<Longrightarrow> (n+1) \<in> EvenN"
 plus2: "n \<in> EvenN \<Longrightarrow> (n+1) \<in> OddN"
 minus1: "n \<in> OddN \<Longrightarrow> (n-(1::int)) \<in> EvenN"
 minus2: "n \<in> EvenN \<Longrightarrow> (n-(1::int)) \<in> OddN"
 umin1: "n \<in> EvenN \<Longrightarrow> (-n) \<in> EvenN"
 umin2: "n \<in> OddN \<Longrightarrow> (-n) \<in> OddN"

lemmas hume_rews =
  intensional_rews all_after_eq all_before_eq
  after_before next_all_eq toBeforeVal toAfterVal
  isExe_def isSup_def

lemma vdmexe2: "\<lbrakk>{} \<rhd> e : P; (v,h',p) = exe E h e\<rbrakk> 
  \<Longrightarrow> P E h h' v p"
  by (simp add: vdmexe)

lemma vdmexe3: "\<forall> E h h' v p e P. ({} \<rhd> e : P) \<and> ((v,h',p) = exe E h e) \<longrightarrow>  P E h h' v p"
  by (simp add: vdmexe)

lemma evodd1:  "\<lbrakk>(v',h',rs) = evenoddExe v h; v = R (Ref r); fmap_lookup h r = Some (I i)\<rbrakk> \<Longrightarrow> v' =( R (Ref r2))"
proof -
  assume a1: "(v',h',rs) = evenoddExe v h"
    and  a2: " v = R (Ref r)"
    and  a3: "fmap_lookup h r = Some (I i)"
  from a1 a2 have a1': "(v', h', rs) = exe (emptyEnv(xin := R (Ref r))) h incbody"
     by (auto simp add: evenoddExe)
  show ?thesis
    ML "reset show_types"
    thm vdmexe2
    apply (insert vdmexe2[OF incbody_spec_v2 a1'])
    apply auto
    apply (erule_tac x="i" in allE)
    apply (subgoal_tac "(\<forall>r2. (v' = (R (Ref r2)) \<and>
          r \<noteq> r2 \<and> fmap_lookup h' r2 = (Some (I (i + 1)))))")
   apply simp
   apply (insert a3)
   apply (erule mp)
   apply simp
   done
qed

lemma evodd1_2:  "\<lbrakk>(v',h',rs) = evenoddExe v h; v = R (Ref r) \<and> fmap_lookup h r = Some (I i)\<rbrakk> \<Longrightarrow> v' =( R (Ref r2))"
  by (auto simp add: evodd1)

lemma evodd1_w1:  "\<lbrakk>(v',h',rs) = evenoddExe v h; v = R (Ref r); fmap_lookup h r = Some (I i)\<rbrakk> \<Longrightarrow> \<exists> ref. v' =( R (Ref ref))"
  apply (rule exI)
  apply (rule evodd1)
  apply auto
  done

lemma evodd2: "\<lbrakk>(v',h',rs) = evenoddExe v h; v = R (Ref r); fmap_lookup h r = Some (I i)\<rbrakk> \<Longrightarrow> 
  v' =( R (Ref r2)) \<and> (fmap_lookup h' r2 = Some (I (i+1)))"
proof -
  assume a1: "(v',h',rs) = evenoddExe v h"
    and  a2: " v = R (Ref r)"
    and  a3: "fmap_lookup h r = Some (I i)"
  from a1 a2 have a1': "(v', h', rs) = exe (emptyEnv(xin := R (Ref r))) h incbody"
     by (auto simp add: evenoddExe)
  show ?thesis
    ML "reset show_types"
    thm vdmexe2
    apply (insert vdmexe2[OF incbody_spec_v2 a1'])
    apply clarsimp
    apply (erule_tac x="i" in allE)
    apply (subgoal_tac "(\<forall>r2. (v' = (R (Ref r2)) \<and>
          r \<noteq> r2 \<and> fmap_lookup h' r2 = (Some (I (i + 1)))))")
   apply simp
   apply (insert a3)
   apply (erule mp)
   apply simp
   done
qed

lemma "v = freshloc x \<longrightarrow> (\<forall> y \<in> x. v \<noteq> y)"
  oops

lemma "fmap_lookup h r = Some k \<longrightarrow>  r \<in> fmap_dom h"
oops

lemma "\<turnstile> program \<longrightarrow> 
\<box>((\<forall> r1. ($w1 = #(R (Ref r1)) \<longrightarrow> (\<forall> r2. ($odd_res = #(R (Ref r2)))
     \<or> ($even_res = #(R (Ref r2)) \<or> ($w2 = #(R (Ref r2)) \<longrightarrow>  #r1 \<noteq> #r2)))))
 \<and> (\<forall> r1. ($w2 = #(R (Ref r1)) \<longrightarrow> (\<forall> r2. ($odd_res = #(R (Ref r2)))
     \<or> ($even_res = #(R (Ref r2)) \<or> ($w1 = #(R (Ref r2)) \<longrightarrow>  #r1 \<noteq> #r2)))))
 \<and> (\<forall> r1. ($odd_res = #(R (Ref r1)) \<longrightarrow> (\<forall> r2. ($w1 = #(R (Ref r2)))
     \<or> ($even_res = #(R (Ref r2)) \<or> ($w2 = #(R (Ref r2)) \<longrightarrow>  #r1 \<noteq> #r2)))))
 \<and> (\<forall> r1. ($even_res = #(R (Ref r1)) \<longrightarrow> (\<forall> r2. ($odd_res = #(R (Ref r2)))
     \<or> ($w1 = #(R (Ref r2)) \<or> ($w2 = #(R (Ref r2)) \<longrightarrow>  #r1 \<noteq> #r2))))))"
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
   apply simp
   oops

lemma eo1: "\<turnstile> program \<longrightarrow> \<box>(((\<exists> r i. $w1 = #(R (Ref r)) \<and> (fmap_lookup<$programheap,#r> = #(Some (I i)))) \<or> $w1 = #BOTTOM)
                     \<and> ((\<exists> r i. $w2 = #(R (Ref r)) \<and> (fmap_lookup<$programheap,#r> = #(Some (I i)))) \<or> $w2 = #BOTTOM)
                     \<and> ((\<exists> r i. $even_res = #(R (Ref r)) \<and> (fmap_lookup<$programheap,#r> = #(Some (I i)))) \<or> $even_res = #BOTTOM)
                     \<and> ((\<exists> r i. $odd_res = #(R (Ref r)) \<and> (fmap_lookup<$programheap,#r> = #(Some (I i)))) \<or> $odd_res = #BOTTOM))"
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
   apply clarsimp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
   apply (case_tac "w \<Turnstile> $sch") 
   prefer 2 
   apply (simp add: isExe_def isSup_def super) 
  apply (simp only: unch_eq1, simp add: unch_def) 
  apply (simp add: isExe_def isSup_def)
   apply (case_tac "w \<Turnstile> $pc") 
   (* pc = even step *)
   apply (simp add: odd, simp only: unch_eq1, simp add: unch_def) 
   apply (simp only: even) 
    apply (case_tac "(w \<Turnstile> $even_st) = Blocked")
   apply (simp only: unch_eq1, simp add: unch_def) 
    apply (case_tac "(w \<Turnstile> $w2) = BOTTOM") 
   apply (simp only: unch_eq1, simp add: unch_def) 
   apply simp_all
   apply (intro impI,elim conjE)
   apply (simp only: hume_rews)
   apply clarsimp
   apply (intro conjI)
   apply (rule disjI2)
   apply clarsimp
   apply (subgoal_tac "((w \<Turnstile> even_res$), (w \<Turnstile> programheap$), (w \<Turnstile> rso1$)) = evenoddExe (R (Ref r)) (w \<Turnstile> $programheap)")
   prefer 2
    apply assumption
   thm evodd1
   apply (rule evodd1)
   apply (intro conjI)
   apply (rule disjI1)
   thm evodd1_w1
   apply (rule evodd1_w1)
   apply simp
   apply simp
   apply simp

  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
   apply clarsimp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (rule impI)
  apply (elim conjE)
  apply (simp add: hume_rews)





  apply (elim exE)
  oops








lemma eo1: "\<turnstile> program \<longrightarrow> \<box>(((\<exists> r i. $w1 = #(R (Ref r))) \<or> $w1 = #BOTTOM)
                     \<and> ((\<exists> r i. $w2 = #(R (Ref r))) \<or> $w2 = #BOTTOM)
                     \<and> ((\<exists> r i. $even_res = #(R (Ref r))) \<or> $even_res = #BOTTOM)
                     \<and> ((\<exists> r i. $odd_res = #(R (Ref r))) \<or> $odd_res = #BOTTOM))"
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
   apply clarsimp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
   apply (case_tac "w \<Turnstile> $sch") 
   prefer 2 
   apply (simp add: isExe_def isSup_def super) 
  apply (simp only: unch_eq1, simp add: unch_def) 
  apply (simp add: isExe_def isSup_def) 
   apply (case_tac "w \<Turnstile> $pc") 
   (* pc = even step *)
   apply (simp add: odd, simp only: unch_eq1, simp add: unch_def) 
   apply (simp only: even) 
    apply (case_tac "(w \<Turnstile> $even_st) = Blocked")   
   apply (simp only: unch_eq1, simp add: unch_def) 
    apply (case_tac "(w \<Turnstile> $w2) = BOTTOM") 
   apply (simp only: unch_eq1, simp add: unch_def) 
   apply simp_all
   apply (intro impI,elim conjE)
   apply (simp only: consumeheap hume_rews)
   apply simp
   apply (elim exE)
   apply (rule disjI1)
   thm evodd1_w1
   apply (rule evodd1_w1)
   apply simp
   apply simp
   apply simp

   apply clarsimp


   apply (subgoal_tac "(w \<Turnstile> even_res$) = I (n+1)")
   apply clarsimp
   apply (simp add: plus1 plus2)
   apply simp
   apply (rule evod1a)
   apply assumption
   (* pc = podd step *)
   apply (simp add: even, simp only: unch_eq1, simp add: unch_def) 
   apply (simp only: odd) 
    apply (case_tac "(w \<Turnstile> $odd_st) = Blocked")   
   apply (simp only: unch_eq1, simp add: unch_def) 
    apply (case_tac "(w \<Turnstile> $w1) = BOTTOM") 
   apply (simp only: unch_eq1, simp add: unch_def) 
   apply simp_all
   apply (intro impI,elim conjE)
   apply (simp only: hume_rews)
   apply (elim exE)
   apply (subgoal_tac "(w \<Turnstile> odd_res$) = I (n+1)")
   apply clarsimp
   apply (simp add: plus1 plus2)
   apply simp
   apply (rule evod1a)
   apply assumption
   done




  oops



(* Main theorem *)
lemma main: "\<turnstile> program \<longrightarrow> \<box>((\<forall> r i. $w1 = #(R (Ref r)) \<and> (fmap_lookup<$programheap,#r> = #(Some (I i))) \<longrightarrow> #i \<in> #EvenN)
                     \<and> (\<forall> r i. $w2 = #(R (Ref r)) \<and> (fmap_lookup<$programheap,#r> = #(Some (I i))) \<longrightarrow> #i \<in> #OddN)
                     \<and> (\<forall> r i. $even_res = #(R (Ref r)) \<and> (fmap_lookup<$programheap,#r> = #(Some (I i))) \<longrightarrow> #i \<in> #EvenN) 
                     \<and> (\<forall> r i. $odd_res = #(R (Ref r)) \<and> (fmap_lookup<$programheap,#r> = #(Some (I i))) \<longrightarrow> #i \<in> #OddN))"
  apply (unfold program) 
  oops

end 
