header {* Hume specialisation of TLA rules  *}

theory EvenOdd3
imports Integrated
begin

constdefs
 emptyHeap :: "Heap"
 "emptyHeap == fimap(empty)"

consts CONS :: ConName
consts NIL  :: ConName
    

consts mList::"(nat \<times> Locn \<times> (Locn set) \<times> Heap) set"
inductive mList intros
mListNIL: "\<lbrakk> fmap_lookup h a = Some (C NIL []) \<rbrakk> \<Longrightarrow> (0,a,{a},h) : mList"
mListCONS:"\<lbrakk> fmap_lookup h a = Some (C CONS [rx,rxs]) ; a \<notin> U; (Ref b) = rxs ; (n,b,U,h):mList\<rbrakk>
          \<Longrightarrow> (Suc n, a, U \<union> {a}, h): mList"
mListRef: "\<lbrakk> fmap_lookup h a = Some (R rx) ; a \<notin> U; (Ref b) = rx ; (n,b,U,h):mList\<rbrakk>
          \<Longrightarrow> (Suc n, a, U \<union> {a}, h): mList"

consts locs :: "Locn \<Rightarrow> Heap \<Rightarrow> Locn set"
defs
  locs: "locs l h \<equiv> \<Union>{x. \<exists> n. (n,l,x,h) \<in> mList}"

lemma "(n,l,y,h) \<in> mList \<and> (m,l,x,h) \<in> mList \<and> m < n \<longrightarrow> y \<subseteq> x"
oops

lemma "locs l h = (THE x. \<exists> m. \<forall> n y. (n,l,y,h) \<in> mList \<and> (m,l,x,h) \<in> mList \<longrightarrow> n <= m)"
oops

consts
  heap_restrict :: "Heap \<Rightarrow> (Locn set) \<Rightarrow> Heap"
  loc_restrict :: "Heap \<Rightarrow> Locn \<Rightarrow> Heap"
  HoVal_restrict :: "Heap \<Rightarrow> HoVal \<Rightarrow> Heap"

defs
  heap_restrict: "heap_restrict h ls \<equiv> fimap ((themap h) |` ls)"
  loc_restrict: "loc_restrict h l \<equiv> heap_restrict h (locs l h)"

lemma "fmap_lookup h l = Some (I n) \<longrightarrow> fmap_dom (loc_restrict h l) = {}"
  oops

lemma "fmap_lookup h l = Some (R (Ref r)) \<longrightarrow> fmap_dom (loc_restrict h l) \<noteq> {}"
  oops

(* need costing for deepcopy - maybe change to use one heap
   with a ref copy instead.
*)


(* many heaps and deep copy *)
consts
  deepcopy :: "(HoVal*Heap*Heap) \<Rightarrow> (HoVal*Heap) option"

recdef deepcopy "measure (\<lambda> (v,h1,h2). card (fmap_dom h2))"
  "deepcopy (BOTTOM,h,h') = Some (BOTTOM,h')"
  "deepcopy ((I i),h,h') = Some ((I i),h')" 
  "deepcopy ((B b),h,h') = Some ((B b),h')"
  "deepcopy ((R ref),h,h') = 
     (case ref of
          Nullref \<Rightarrow> Some (R (Nullref),h')
       |  Ref loc \<Rightarrow> 
            let nh = loc_restrict h' loc in 
              (if card (fmap_dom nh) >= card (fmap_dom h') 
                  then None 
                  else (case (fmap_lookup h loc) of
                           None \<Rightarrow> Some (R (Nullref),h')
                         | Some n \<Rightarrow> 
                            (case deepcopy (n,h,nh) of
                                None \<Rightarrow> None
                              | Some (rv,rh) \<Rightarrow> let v = freshloc (fmap_dom rh) ; nrh = fmap_upd rh v rv 
                                                in Some (R (Ref v),nrh)))))"
  "deepcopy ((C n refs),h,h') = None"
  "deepcopy ((X fname n refs),h,h') = None"

section "Coordination layer state"

datatype PC = peven | podd

consts
  pc :: "PC statefun" (* the program counter - sequential execution of boxes *)
  odd_st :: "BoxState statefun" (* state/mode of box *)
  even_st :: "BoxState statefun" (* state/mode of box *)
  (* box heaps *)
  evenheap :: "Heap statefun"
  oddheap  :: "Heap statefun"
  (* box buffer: 
     Q: not sure if the have a set location or not - guess not since emptyheaps 
        between calls *)
  even_res :: "HoVal statefun"
  odd_res :: "HoVal statefun"
  (* the wire heap *)
  wireheap :: "Heap statefun"
  (* the wires: pointers to the wire heap - consumed by expression layer ??? *)
  (* Q: Assumes w1 = NullRef is used for empty wire *)
  w1 :: "HoVal statefun"
  w2 :: "HoVal statefun"
  (* enables reasoning about resources - Q: one for each heap? *)
  rso1 :: "Rescomp statefun"
  rso2 :: "Rescomp statefun"

section "The Expession VDM layer"

text {* Full body, including a pattern match against a variable *}

(* FIX ME: change to returning heap and location on heap - this will be the box heap
   Q1: should values be copied from wire heap to box heap, and the expression layer
       given the box heap - or should expression layer use wire heap directly ??
   Q2: matchfail - return NullRef ???
*)

syntax 
 x1  :: "VarName"
 xin :: "VarName"

translations
"x1" == "(VN ''x1'')"
"xin" == "(VN ''xin'')"

constdefs iplus :: "HoVal => HoVal => HoVal"
"iplus x y == I ((theInt x)+(theInt y))"
constdefs iinc :: "HoVal => HoVal => HoVal"
"iinc x y == I ((theInt x)+(1::int))"

(* heap version, ie. arg is a pointer to a heap cell containing an int *)
constdefs incbody :: "Exp"
"incbody == CaseOneExp xin (PVar x1) (RPrimBin iinc x1 x1) (ValueExp BOTTOM)"

(* stack version, ie. arg is an int *)
constdefs incbodyI :: "Exp"
"incbodyI == (PrimBin iinc xin xin)"

theorem incbody_spec: "{} \<rhd> incbody : (\<lambda> E h hh v p. 
  (\<forall> i r . (E xin = (R (Ref r)) \<and> (fmap_lookup h r = Some (I i)))  \<longrightarrow> v = (I (i+1))))"
apply (simp add: incbody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
apply (rule vdmRPrimBin)
apply (rule vdmValue)
apply clarsimp
apply (erule_tac x="E(x1:=(R (Ref r)))" in allE)
apply (erule_tac x="R (Ref r)" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPVar)
apply (simp add: iinc_def)
done

theorem incbodyI_spec: "{} \<rhd> incbodyI : (\<lambda> E h hh v p. 
  (\<forall> i r . (E xin = (I i))  \<longrightarrow> v = (I (i+1))))"
apply (simp add: incbodyI_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmPrimBin)
apply clarsimp
apply (simp add: iinc_def)
done

section "The coordination layer"

constdefs
  S' :: "temporal"
  "S' \<equiv> TEMP (sch$ = (if isExe \<and> $pc = #podd then #Super else #Execute))
           \<and> (isExe \<longrightarrow> pc$ = #peven)"

subsection "Even Box"

(* since the super-step is atomic, it has to be dealt with globally *)

consts
 evenoddExe :: "HoVal \<Rightarrow> Heap \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"

(* fixme: must show that deepcopy returns a value and not None!!! *)
defs
  evenoddExe: "evenoddExe v hwire hbox \<equiv>
    let (v',hbox') = the (deepcopy (v,hwire,hbox)) in exe (emptyEnv(xin := v')) hbox' incbody"


(* 1: lift vdm to property to relevant HOL property *)

consts
  even :: "temporal"
  odd :: "temporal"
  super :: "temporal"
  init :: "temporal"
  program :: "temporal"

defs 
 even: "even \<equiv> TEMP
(if $pc \<noteq> #peven
 then Unchanged (w2,even_res,evenheap,even_st)
 else pc$ = #podd \<and> 
 (if $even_st = #Blocked
   then Unchanged (w2,even_res,even_st,evenheap,wireheap)
   else 
 (if $w2 = #BOTTOM
   then Unchanged (w2,evenheap,wireheap)
     \<and>  even_res$ = #BOTTOM 
     \<and>  even_st$ = #Matchfail
   else Unchanged evenheap 
    \<and> w2$ = #BOTTOM
    \<and> (even_res,evenheap,rso1)$ = evenoddExe<$w2,$wireheap,$evenheap>)))"
 odd: "odd \<equiv> TEMP
(if $pc \<noteq> #podd
 then Unchanged (w1,odd_res,oddheap,odd_st)
 else pc$ = #peven \<and> 
 (if $odd_st = #Blocked
   then Unchanged (w1,odd_res,odd_st,oddheap,wireheap)
   else 
 (if $w1 = #BOTTOM
   then Unchanged (w1,oddheap,wireheap)
     \<and>  odd_res$ = #BOTTOM 
     \<and>  odd_st$ = #Matchfail
   else Unchanged oddheap 
    \<and> w1$ = #BOTTOM
    \<and> (odd_res,oddheap,rso2)$ = evenoddExe<$w1,$wireheap,$oddheap>)))"

defs
 super: "super \<equiv> TEMP
    pc$ = #peven
 \<and> Unchanged (wireheap,evenheap,oddheap)
 \<and> (if $w1 \<noteq> #BOTTOM \<and> $even_res \<noteq> #BOTTOM
    then even_st$ = #Blocked \<and> Unchanged (w1,even_res)  
    else (if $even_res = #BOTTOM
    then even_st$ = #Runnable \<and> Unchanged (w1,even_res)
    else even_st$ = #Runnable \<and> w1$ = $even_res \<and> even_res$ = #BOTTOM))
 \<and> (if $w2 \<noteq> #BOTTOM \<and> $odd_res \<noteq> #BOTTOM
    then odd_st$ = #Blocked \<and> Unchanged (w2,odd_res)  
    else (if $odd_res = #BOTTOM
    then odd_st$ = #Runnable \<and> Unchanged (w2,odd_res)
    else odd_st$ = #Runnable \<and> w2$ = $odd_res \<and> odd_res$ = #BOTTOM))"
  

defs
  init: "init \<equiv> TEMP
    $sch = #Execute \<and> $pc = #peven
  \<and> $wireheap = fmap_upd<#emptyHeap,(freshloc<#{}>),#(I 0)>
  \<and> $(even_st,odd_st) = #(Runnable,Runnable)
  \<and> $(evenheap,oddheap) = #(emptyHeap,emptyHeap)
  \<and> $(even_res,odd_res) = #(BOTTOM,BOTTOM)
  \<and> $(w1,w2) = #(R (Ref (freshloc {})),BOTTOM)"


defs
  program: "program \<equiv> TEMP
   init \<and> \<box>[S' \<and> (isExe \<longrightarrow> even \<and> odd) \<and> (isSup \<longrightarrow> super)]_(pc,sch,odd_st,
                   even_st,evenheap,oddheap,even_res,odd_res,wireheap,w1,w2)"

section "Properties"

consts EvenN :: "int set"
       OddN  :: "int set"

inductive EvenN OddN 
intros
 zero: "0 \<in> EvenN" 
 one:  "1 \<in> OddN"
 plus1: "n \<in> OddN \<Longrightarrow> (n+1) \<in> EvenN"
 plus2: "n \<in> EvenN \<Longrightarrow> (n+1) \<in> OddN"
 minus1: "n \<in> OddN \<Longrightarrow> (n-(1::int)) \<in> EvenN"
 minus2: "n \<in> EvenN \<Longrightarrow> (n-(1::int)) \<in> OddN"
 umin1: "n \<in> EvenN \<Longrightarrow> (-n) \<in> EvenN"
 umin2: "n \<in> OddN \<Longrightarrow> (-n) \<in> OddN"

lemmas hume_rews =
  intensional_rews all_after_eq all_before_eq
  after_before next_all_eq toBeforeVal toAfterVal
  isExe_def isSup_def

lemma vdmexe2: "\<lbrakk>{} \<rhd> e : P; (v,h',p) = exe E h e\<rbrakk> 
  \<Longrightarrow> P E h h' v p"
  by (simp add: vdmexe)

lemma vdmexe3: "\<forall> E h h' v p e P. ({} \<rhd> e : P) \<and> ((v,h',p) = exe E h e) \<longrightarrow>  P E h h' v p"
  by (simp add: vdmexe)


(* lemma evod1: "\<lbrakk>(v',h,r) = evenoddExe v h1 h2; v = I n \<rbrakk> \<Longrightarrow> v' = I (n+1)"
  apply (auto simp add: evenoddExe)
  apply (insert vdmexe2[OF incbodyI_spec])
  apply auto
  done

lemma evod1a: "(v',h',r) = evenoddExe (I n) h hh \<Longrightarrow> v' = I (n+1)"
  apply (auto simp add: evenoddExe)
  apply (insert vdmexe2[OF incbodyI_spec])
  apply auto
  done

lemma evod1a: "(I (n+1),h,r) = evenoddExe (I n) h1 h2"
  apply (auto simp add: evenoddExe)
  apply (insert vdmexe2[OF incbodyI_spec])
  oops

lemma evod2: "\<lbrakk>(v',h,r) = evenoddExe v h1 h2; v = I n; n \<in> EvenN\<rbrakk> \<Longrightarrow> (\<exists> k. v' = I k \<and> k \<in> OddN)"
  apply (auto simp add: evenoddExe)
  apply (insert vdmexe2[OF incbodyI_spec])
  apply (rule_tac x = "n+1" in exI)
  apply (insert evod1)
  apply auto
  apply (rule plus2)
  apply assumption
  done

lemma evod3: "\<lbrakk>(v',h',r) = evenoddExe (I n) h hh; n \<in> OddN\<rbrakk> \<Longrightarrow> (\<exists> k. v' = I k \<and> k \<in> EvenN)"
  apply (auto simp add: evenoddExe)
  apply (insert vdmexe2[OF incbodyI_spec])
  apply (rule_tac x = "n+1" in exI)
  apply (insert evod1)
  apply auto
  apply (rule plus1)
  apply assumption
  done
*)
(* Possible aux *)
lemma eo1: "\<turnstile> program \<longrightarrow> \<box>(((\<exists> r i. $w1 = #(R (Ref r)) \<and> (fmap_lookup<$wireheap,#r> = #(Some (I i)))) \<or> $w1 = #BOTTOM)
                     \<and> ((\<exists> r i. $w2 = #(R (Ref r)) \<and> (fmap_lookup<$wireheap,#r> = #(Some (I i)))) \<or> $w2 = #BOTTOM)
                     \<and> ((\<exists> r i. $even_res = #(R (Ref r)) \<and> (fmap_lookup<$evenheap,#r> = #(Some (I i)))) \<or> $even_res = #BOTTOM)
                     \<and> ((\<exists> r i. $odd_res = #(R (Ref r)) \<and> (fmap_lookup<$oddheap,#r> = #(Some (I i)))) \<or> $odd_res = #BOTTOM))"
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
   apply clarsimp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  oops

(* Main theorem *)
lemma main: "\<turnstile> program \<longrightarrow> \<box>((\<forall> r i. $w1 = #(R (Ref r)) \<and> (fmap_lookup<$wireheap,#r> = #(Some (I i))) \<longrightarrow> #i \<in> #EvenN)
                     \<and> (\<forall> r i. $w2 = #(R (Ref r)) \<and> (fmap_lookup<$wireheap,#r> = #(Some (I i))) \<longrightarrow> #i \<in> #OddN)
                     \<and> (\<forall> r i. $even_res = #(R (Ref r)) \<and> (fmap_lookup<$evenheap,#r> = #(Some (I i))) \<longrightarrow> #i \<in> #EvenN) 
                     \<and> (\<forall> r i. $odd_res = #(R (Ref r)) \<and> (fmap_lookup<$oddheap,#r> = #(Some (I i))) \<longrightarrow> #i \<in> #OddN))"
  apply (unfold program) 
  oops


end 
