header {* Hume specialisation of TLA rules  *}

theory EvenOdd_PhD
imports Integrated
begin

constdefs
 emptyHeap :: "Heap"
 "emptyHeap == fimap(empty)"

section "Coordination layer state"

datatype PC = peven | podd

consts
  pc :: "PC statefun" 
  odd_st :: "BoxState statefun" 
  even_st :: "BoxState statefun" 
  heap :: "Heap statefun"
  even_res :: "HoVal statefun"
  odd_res :: "HoVal statefun"
  w1 :: "HoVal statefun"
  w2 :: "HoVal statefun"
  rsrc :: "Rescomp statefun"

section "The Expession VDM layer"

syntax 
 x1  :: "VarName"
 xin :: "VarName"

translations
"x1" == "(VN ''x1'')"
"xin" == "(VN ''xin'')"

constdefs iplus :: "HoVal => HoVal => HoVal"
"iplus x y == I ((theInt x)+(theInt y))"
constdefs iinc :: "HoVal => HoVal => HoVal"
"iinc x y == I ((theInt x)+(1::int))"

(* heap version, ie. arg is a pointer to a heap cell containing an int *)
constdefs incbody :: "Exp"
"incbody == CaseOneExp xin (PVar x1) (RPrimBin iinc x1 x1) (ValueExp BOTTOM)"

(* stack version, ie. arg is an int *)
constdefs incbodyI :: "Exp"
"incbodyI == (PrimBin iinc xin xin)"

theorem body1: "{} \<rhd> incbodyI : (\<lambda> E h hh v p. 
  (\<forall> i r . (E xin = (I i))  \<longrightarrow> v = (I (i+1))))"
apply (simp add: incbodyI_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmPrimBin)
apply clarsimp
apply (simp add: iinc_def)
done

section "The coordination layer"

constdefs
  S' :: "temporal"
  "S' \<equiv> TEMP (sch$ = (if isExe \<and> $pc = #podd then #Super else #Execute))
           \<and> (isExe \<longrightarrow> pc$ = #peven)"

consts
 exebody :: "HoVal \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"

defs
  exebody: "exebody v h \<equiv> exe (emptyEnv(xin := v)) h incbodyI"

consts
  even :: "temporal"
  odd :: "temporal"
  super :: "temporal"
  init :: "temporal"
  program :: "temporal"

defs 
 even: "even \<equiv> TEMP
(if $pc \<noteq> #peven
 then Unchanged (w2,even_res,even_st)
 else pc$ = #podd \<and> 
 (if $even_st = #Blocked
   then Unchanged (w2,even_res,even_st,heap,rsrc)
   else 
 (if $w2 = #BOTTOM
   then Unchanged (w2,heap,rsrc)
     \<and>  even_res$ = #BOTTOM 
     \<and>  even_st$ = #Matchfail
   else w2$ = #BOTTOM
    \<and> even_st$ = #Runnable
    \<and> (even_res,heap,rsrc)$ = exebody<$w2,$heap>)))"
defs
 odd: "odd \<equiv> TEMP
(if $pc \<noteq> #podd
 then Unchanged (w1,odd_res,odd_st)
 else pc$ = #peven \<and> 
 (if $odd_st = #Blocked
   then Unchanged (w1,odd_res,odd_st,heap,rsrc)
   else 
 (if $w1 = #BOTTOM
   then Unchanged (w1,heap,rsrc)
     \<and>  odd_res$ = #BOTTOM 
     \<and>  odd_st$ = #Matchfail
   else w1$ = #BOTTOM
    \<and> odd_st$ = #Runnable
    \<and> (odd_res,heap,rsrc)$ = exebody<$w1,$heap>)))"

defs
 super: "super \<equiv> TEMP
    pc$ = #peven
 \<and> Unchanged (heap,rsrc)
 \<and> (if $w1 \<noteq> #BOTTOM \<and> $even_res \<noteq> #BOTTOM
    then even_st$ = #Blocked \<and> Unchanged (w1,even_res)  
    else (if $even_res = #BOTTOM
    then even_st$ = #Runnable \<and> Unchanged (w1,even_res)
    else even_st$ = #Runnable \<and> w1$ = $even_res \<and> even_res$ = #BOTTOM))
 \<and> (if $w2 \<noteq> #BOTTOM \<and> $odd_res \<noteq> #BOTTOM
    then odd_st$ = #Blocked \<and> Unchanged (w2,odd_res)  
    else (if $odd_res = #BOTTOM
    then odd_st$ = #Runnable \<and> Unchanged (w2,odd_res)
    else odd_st$ = #Runnable \<and> w2$ = $odd_res \<and> odd_res$ = #BOTTOM))"
  
defs (* add empty rsrc *)
  init: "init \<equiv> TEMP
    $sch = #Execute \<and> $pc = #peven
  \<and> $heap = #emptyHeap
  \<and> $(even_st,odd_st) = #(Runnable,Runnable)
  \<and> $(even_res,odd_res) = #(BOTTOM,BOTTOM)
  \<and> $(w1,w2) = #(I 0,BOTTOM)"


defs
  program: "program \<equiv> TEMP
   init \<and> \<box>[S' \<and> (isExe \<longrightarrow> even \<and> odd) \<and> (isSup \<longrightarrow> super)]_(pc,sch,odd_st,
                   even_st,even_res,odd_res,heap,w1,w2,rsrc)"

section "Properties"

consts EvenN :: "int set"
       OddN  :: "int set"

inductive EvenN OddN 
intros
 zero: "0 \<in> EvenN" 
 one:  "1 \<in> OddN"
 plus1: "n \<in> OddN \<Longrightarrow> (n+1) \<in> EvenN"
 plus2: "n \<in> EvenN \<Longrightarrow> (n+1) \<in> OddN"
 minus1: "n \<in> OddN \<Longrightarrow> (n-(1::int)) \<in> EvenN"
 minus2: "n \<in> EvenN \<Longrightarrow> (n-(1::int)) \<in> OddN"
 umin1: "n \<in> EvenN \<Longrightarrow> (-n) \<in> EvenN"
 umin2: "n \<in> OddN \<Longrightarrow> (-n) \<in> OddN"

lemmas hume_rews =
  intensional_rews all_after_eq all_before_eq
  after_before next_all_eq toBeforeVal toAfterVal
  isExe_def isSup_def

lemma vdmexe2: "\<lbrakk>{} \<rhd> e : P; (v,h',p) = exe E h e\<rbrakk> 
  \<Longrightarrow> P E h h' v p"
  by (simp add: vdmexe)

lemma vdmexe3: "\<forall> E h h' v p e P. ({} \<rhd> e : P) \<and> ((v,h',p) = exe E h e) \<longrightarrow>  P E h h' v p"
  by (simp add: vdmexe)

lemma exebody1: "(v',h',r) = exebody (I n) h \<Longrightarrow> v' = I (n+1)"
  apply (auto simp add: exebody)
  apply (insert vdmexe2[OF body1])
  apply auto
  done


(* Possible aux *)
lemma l1: "\<turnstile> program \<longrightarrow> \<box>(((\<exists> n. $w1 = #(I n)) \<or> $w1 = #BOTTOM)
                     \<and> ((\<exists> n. $w2 = #(I n)) \<or> $w2 = #BOTTOM)
                     \<and> ((\<exists> n. $even_res = #(I n)) \<or> $even_res = #BOTTOM)
                     \<and> ((\<exists> n. $odd_res = #(I n)) \<or> $odd_res = #BOTTOM))"
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
   apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
   apply (case_tac "w \<Turnstile> $sch")
   prefer 2
   apply (simp add: isExe_def isSup_def super)
  apply (simp only: unch_eq1, simp add: unch_def)
  apply (simp add: isExe_def isSup_def)
   apply (case_tac "w \<Turnstile> $pc")
   apply (simp add: odd, simp only: unch_eq1, simp add: unch_def)
   (* even step *)
   apply (simp only: even)
    apply (case_tac "(w \<Turnstile> $even_st) = Blocked")
   apply (simp only: unch_eq1, simp add: unch_def)
    apply (case_tac "(w \<Turnstile> $w2) = BOTTOM")
   apply (simp only: unch_eq1, simp add: unch_def)
   apply simp
   apply (intro impI,elim conjE)
   apply (simp only: exebody Let_def)
   apply (simp only: hume_rews)
   apply clarsimp
   apply (subgoal_tac "\<forall> i r . ((emptyEnv(xin := I n)) xin = (I i))  \<longrightarrow> (w \<Turnstile> even_res$) = (I (i+1))")
   prefer 2
   apply (insert vdmexe3)
   apply (drule_tac x = "emptyEnv(xin := I n)" in spec)
   apply (drule_tac x = "w \<Turnstile> $heap" in spec)
   apply (drule_tac x = "w \<Turnstile> heap$" in spec)
   apply (drule_tac x = "w \<Turnstile> even_res$" in spec)
   apply (drule_tac x = "w \<Turnstile> rsrc$" in spec)
   apply (drule_tac x = "incbodyI" in spec)
   apply (drule_tac x = "(\<lambda> E h hh v p. (\<forall> i r . (E xin = (I i))  \<longrightarrow> v = (I (i+1))))" in spec)
   apply (insert body1)
   apply simp
   apply simp
   (* odd step !!! *)
   apply (simp only: even odd)
    apply (case_tac "(w \<Turnstile> $odd_st) = Blocked")
   apply (simp only: unch_eq1, simp add: unch_def)
    apply (case_tac "(w \<Turnstile> $w1) = BOTTOM")
   apply (simp only: unch_eq1, simp add: unch_def)
   apply simp
   apply (intro impI,elim conjE)
   apply (simp only: exebody Let_def)
   apply (simp only: hume_rews)
   apply clarsimp
   apply (subgoal_tac "\<forall> i r . ((emptyEnv(xin := I n)) xin = (I i))  \<longrightarrow> (w \<Turnstile> odd_res$) = (I (i+1))")
   prefer 2
   apply (insert vdmexe3)
   apply (drule_tac x = "emptyEnv(xin := I n)" in spec)
   apply (drule_tac x = "w \<Turnstile> $heap" in spec)
   apply (drule_tac x = "w \<Turnstile> heap$" in spec)
   apply (drule_tac x = "w \<Turnstile> odd_res$" in spec)
   apply (drule_tac x = "w \<Turnstile> rsrc$" in spec)
   apply (drule_tac x = "incbodyI" in spec)
   apply (drule_tac x = "(\<lambda> E h hh v p. (\<forall> i r . (E xin = (I i))  \<longrightarrow> v = (I (i+1))))" in spec)
   apply (insert body1)
   apply simp
   apply (simp only: unch_eq1, simp add: unch_def)
   done

(* Main theorem *)
theorem main: "\<turnstile> program \<longrightarrow> \<box>( (\<forall> n. $w1 = #(I n) \<longrightarrow> #n \<in> #EvenN) 
                              \<and> (\<forall> n. $w2 = #(I n) \<longrightarrow> #n \<in> #OddN) 
                              \<and> (\<forall> n. $even_res = #(I n) \<longrightarrow> #n \<in> #EvenN) 
                              \<and> (\<forall> n. $odd_res = #(I n) \<longrightarrow> #n \<in> #OddN))" 
  apply (unfold program) 
  apply (rule spec_inv2_mono)
  apply (rule l1[unfolded program])
   apply (rule invmono) 
   apply (unfold init) 
   apply int_unlift 
   apply (simp add: all_before_eq) 
  apply (simp add: zero) 
   apply (rule preimpsplit) 
   apply (simp_all only: next_all_eq after_before) 
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def) 
  apply pref_unlift 
   apply (case_tac "w \<Turnstile> $sch") 
   prefer 2 
   apply (simp add: isExe_def isSup_def super) 
  apply (simp only: unch_eq1, simp add: unch_def) 
  apply (simp add: isExe_def isSup_def) 
   apply (case_tac "w \<Turnstile> $pc") 
   (* pc = even step *)
   apply (simp add: odd, simp only: unch_eq1, simp add: unch_def) 
   apply (simp only: even) 
    apply (case_tac "(w \<Turnstile> $even_st) = Blocked")   
   apply (simp only: unch_eq1, simp add: unch_def) 
    apply (case_tac "(w \<Turnstile> $w2) = BOTTOM") 
   apply (simp only: unch_eq1, simp add: unch_def) 
   apply simp_all
   apply (intro impI,elim conjE)
   apply (simp only: hume_rews)
   apply (elim exE)
   apply (subgoal_tac "(w \<Turnstile> even_res$) = I (n+1)")
   apply clarsimp
   apply (simp add: plus1 plus2)
   apply simp
   apply (rule exebody1)
   apply assumption
   (* pc = podd step *)
   apply (simp add: even, simp only: unch_eq1, simp add: unch_def) 
   apply (simp only: odd) 
    apply (case_tac "(w \<Turnstile> $odd_st) = Blocked")   
   apply (simp only: unch_eq1, simp add: unch_def) 
    apply (case_tac "(w \<Turnstile> $w1) = BOTTOM") 
   apply (simp only: unch_eq1, simp add: unch_def) 
   apply simp_all
   apply (intro impI,elim conjE)
   apply (simp only: hume_rews)
   apply (elim exE)
   apply (subgoal_tac "(w \<Turnstile> odd_res$) = I (n+1)")
   apply clarsimp
   apply (simp add: plus1 plus2)
   apply simp
   apply (rule exebody1)
   apply assumption
   done


end 
