(* Time-stamp: <Fri Dec 12 2008 15:42:05 Stardate: [-28]0478.06 hwloidl>
   Example: listcopy/fuse
   Top-level theorems (expression level): 
    . copy_resources_ok
    . fuse_resources_ok
   (these specify bounds on heap consumption and on result data structure).
*)

theory ListCopy = Integrated:

(* ----------------------------------------------------------------------------- *)

syntax 
 h9var :: "VarName"
 t9var :: "VarName"
 y9var :: "VarName"
 b9var :: "VarName"
 x9arg :: "VarName"
 x9in  :: "VarName"
 konst :: "FunName"

translations
"h9var" == "(VN ''h9var'')"
"t9var" == "(VN ''t9var'')"
"y9var" == "(VN ''y9var'')"
"b9var" == "(VN ''b9var'')"
"x9arg"  == "(VN ''x9arg'')"
"x9in"  == "(VN ''x9in'')"
"konst" == "(FN ''konst'')"

constdefs is_zero :: "HoVal => HoVal => HoVal"
"is_zero x y == B ((theInt x)=(0 :: int))"

lemma is_zero_tertium_non_datur: "is_zero x y = B True \<or> is_zero x y = B False"
by (simp add: is_zero_def)

constdefs inc :: "HoVal => HoVal => HoVal"
"inc x y == I ((theInt x)+(1 :: int))"

constdefs dec :: "HoVal => HoVal => HoVal"
"dec x y == I ((theInt x)-(1 :: int))"

constdefs konstBody :: Exp
"konstBody \<equiv> (LetExp b9var (PrimBin is_zero x9arg x9arg)
                (IfExp b9var 
                  (ValueExp (I 0))
                  (LetExp t9var (PrimBin dec x9arg x9arg)
                   (LetExp y9var (CallFunExp konst [t9var])
                    (PrimBin inc y9var y9var)))))"

axioms konst_funTabArgs: "fst (funTab konst) == [x9arg]"
axioms konst_funTabBody: "snd (funTab konst) == konstBody"

constdefs konstAss :: "[VarName list] => Assertion"
"konstAss zs == (\<lambda> E h h' v p. 
                                    (\<forall> n . (E (hd zs) = (I n)) \<longrightarrow> v = (I n)))"

axioms konst_funSpecTab: "the_funSpecTab konst xs == konstAss xs"

constdefs konst_Context :: "Context"
"konst_Context \<equiv> {(CallFunExp konst [t9var] ,  konstAss [t9var])}"


lemma "hd (x#xs) = x"
by simp

lemma konst_goodContext: "goodContext0 the_funSpecTab konst_Context"
apply (simp add: goodContext0_def konst_funTabBody konstBody_def konst_Context_def konst_funSpecTab konstAss_def)
apply clarsimp
apply (rule vdmConseq)
prefer 2
apply (rule vdmLet)
apply (rule vdmPrimBin)
apply (rule vdmIf)
apply (rule vdmValue)
apply (rule vdmLet)
(* defer 1  *)
apply (rule vdmPrimBin)
apply (rule vdmLet)
apply (rule vdmAx)
apply simp
apply (rule vdmPrimBin)
apply clarsimp?
apply (case_tac "is_zero (E x9arg) (E x9arg) = B True")
 apply clarsimp
 defer 1
 apply clarsimp
 apply (subgoal_tac "is_zero (E x9arg) (E x9arg) = B False")
  apply clarsimp
  apply (simp add: inc_def dec_def konst_funTabArgs)
  apply (erule FrameInd.cases)
  apply clarsimp
  apply (erule FrameInd.cases)
  apply clarsimp
  apply clarsimp
 apply (insert is_zero_tertium_non_datur)
 apply fastsimp
 apply (simp add: is_zero_def konst_funTabArgs)
 apply (erule FrameInd.cases)
 apply clarsimp
 apply (erule FrameInd.cases)
 apply clarsimp
 apply fastsimp
done

(* uses vdmAxAdapt_UNPROVEN; uses subgoal when finding rec call *)
theorem konstCallAdapt: "{} \<rhd> (CallFunExp konst [xin]) : konstAss [xin]"
apply (rule vdmCallFun)
apply (simp add: konst_funTabBody konstBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmLet)
apply (rule vdmPrimBin)
apply (rule vdmIf)
apply (rule vdmValue)
apply (rule vdmLet)
(* defer 1  *)
apply (rule vdmPrimBin)
apply (rule vdmLet)
(* apply (rule vdmAx) *)
apply (rule CtxtWeakSingleton)
apply (rule vdmAdapt)
apply (rule konst_goodContext)
apply (simp add: konst_Context_def)
apply (simp add: konst_Context_def konst_funSpecTab)
apply fastsimp
apply (rule vdmPrimBin)
apply clarsimp?
apply (case_tac "is_zero (E x9arg) (E x9arg) = B True")
 apply clarsimp
 defer 1
 apply clarsimp
 apply (subgoal_tac "is_zero (E x9arg) (E x9arg) = B False")
  apply clarsimp
  apply (simp add: inc_def dec_def konst_funTabArgs konstAss_def konst_funSpecTab)
  apply clarsimp
  apply (erule FrameInd.cases)
  apply clarsimp
  apply (erule FrameInd.cases)
  apply clarsimp
  apply clarsimp
 apply (insert is_zero_tertium_non_datur)
 apply fastsimp
 apply (simp add: is_zero_def konst_funTabArgs konstAss_def konst_funSpecTab)
 apply (erule FrameInd.cases)
 apply clarsimp
 apply (erule FrameInd.cases)
 apply clarsimp
 apply fastsimp
done

(* uses vdmAxAdapt_UNPROVEN; obvious proof-strategy: first all syntax-directed rules, then only simps *)
(*
theorem konstCallAdapt1: "{}
    \<rhd> (CallFunExp konst [xin]) : konstAss [xin]"
apply (rule vdmCallFun)
apply (simp add: konst_funTabBody konstBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmLet)
apply (rule vdmPrimBin)
apply (rule vdmIf)
apply (rule vdmValue)
apply (rule vdmLet)
apply (rule vdmPrimBin)
apply (rule vdmLet)
apply (rule vdmAxAdapt_UNPROVEN)
apply fastsimp
apply (rule vdmPrimBin)
apply clarsimp
apply (simp add: konstAss_def)
apply (case_tac "is_zero (E x9arg) (E x9arg) = B True")
 apply clarsimp
 defer 1
 apply clarsimp
 apply (subgoal_tac "is_zero (E x9arg) (E x9arg) = B False")
  apply clarsimp
  apply (simp add: inc_def dec_def konst_funTabArgs)
  apply (erule FrameInd.cases)
  apply clarsimp
  apply (erule FrameInd.cases)
  apply clarsimp
  apply clarsimp
 prefer 2
 apply (simp add: is_zero_def konst_funTabArgs)
 apply (erule FrameInd.cases)
 apply clarsimp
 apply (erule FrameInd.cases)
 apply clarsimp
 apply fastsimp
 apply (insert is_zero_tertium_non_datur)
 apply fastsimp
done
*)

(* -------------------------------------------------- *)
(* modelling LISTS *)

syntax 
 CONS :: ConName
 NIL  :: ConName
translations
 "CONS" == "(CN ''CONS'')"
 "NIL" == "(CN ''NIL'')"

consts mList::"(nat \<times> Locn \<times> (Locn set) \<times> Heap) set"
inductive mList intros
mListNIL: "\<lbrakk> fmap_lookup h a = Some (C NIL []) \<rbrakk> \<Longrightarrow> (0,a,{a},h) : mList"
mListCONS:"\<lbrakk> fmap_lookup h a = Some (C CONS [rx,rxs]) ; a \<notin> U; (Ref b) = rxs ; (n,b,U,h):mList\<rbrakk>
          \<Longrightarrow> (Suc n, a, U \<union> {a}, h): mList"


lemma mListDom: "\<lbrakk>(n, a, U, h) \<in> mList\<rbrakk> \<Longrightarrow>  U \<subseteq> fmap_dom h"
apply (erule mList.induct)
apply (simp_all add:fmap_lookup_def fmap_dom_def dom_def)
done

lemma mListFresh: "(L, a, U, h) \<in> mList \<Longrightarrow> freshloc (fmap_dom h) \<notin> U"
apply (subgoal_tac "U \<subseteq> fmap_dom h")
apply (subgoal_tac "finite (fmap_dom h)")
apply (drule freshloc)
apply fastsimp
apply fastsimp
apply (erule mListDom)
done

(* TODO: move this to Basics.thy or thereabouts *)
constdefs sameH::"Locn set \<Rightarrow> Heap \<Rightarrow> Heap \<Rightarrow> bool"
"sameH U h hh \<equiv> (\<forall> l. (l \<in> U \<longrightarrow> (fmap_lookup h l = fmap_lookup hh l)))"


lemma mListSameH[rule_format]:
"(n,a,U,h) \<in> mList \<Longrightarrow> (\<forall> hh. sameH U h hh \<longrightarrow> (n,a,U,hh) \<in> mList)"
apply (erule mList.induct, safe)
apply (rule mListNIL)
apply (simp add: sameH_def)
apply (simp add: sameH_def)
apply (erule_tac x="hh" in allE)
apply (erule impE)
apply clarsimp
apply (subgoal_tac "(Suc n, a, U \<union> {a}, hh) \<in> mList")
apply simp
apply (rule mListCONS)
apply (erule_tac x="a" in allE)
apply ( simp add: sameH_def)
apply (subgoal_tac "fmap_lookup hh a = Some (C CONS [rx, Ref b])") 
apply (rotate_tac -2)
apply (erule thin_rl)
apply ( simp add: sameH_def)
apply (rule conjI)
apply simp+
done

(* -------------------------------------------------- *)
(* Example: list-copy *)


syntax 
 hvar :: "VarName"
 tvar :: "VarName"
 yvar :: "VarName"
 xarg  :: "VarName"
 xin  :: "VarName"
 copy :: "FunName"

translations
"hvar" == "(VN ''hvar'')"
"tvar" == "(VN ''tvar'')"
"yvar" == "(VN ''yvar'')"
"xarg"  == "(VN ''xarg'')"
"xin"  == "(VN ''xin'')"
"copy" == "(FN ''copy'')"

constdefs copyBody :: Exp
"copyBody \<equiv> CaseOneExp xarg 
                    (PCon CONS [hvar,tvar]) (LetExp yvar (CallFunExp copy [tvar]) 
                                              (ConstrExp CONS [hvar,yvar])) 
                    (ConstrExp NIL [])" 

constdefs copyAss :: "[VarName list] => Assertion"
"copyAss zs == (\<lambda> E h h' v p. 
                                 (\<forall> n r U . (E (hd zs) = (R (Ref r)) \<and> (n,r,U,h):mList  
                                             \<longrightarrow> (card (fmap_dom h') \<le> card (fmap_dom h) + n + 1) \<and> (\<exists> rv Uv. v = R (Ref rv) \<and> (n,rv,Uv,h'):mList))))"

axioms copy_funTabArgs: "fst (funTab copy) == [xarg]"
axioms copy_funTabBody: "snd (funTab copy) == copyBody"
axioms copy_funSpecTab: "the_funSpecTab copy xs = copyAss xs"

constdefs copy_Context :: Context
"copy_Context == {(CallFunExp copy [tvar], copyAss [tvar])}"

lemma a11: "fmap_dom (h((freshloc (fmap_dom h)) \<mapsto>\<^sub>f v)) = insert (freshloc (fmap_dom h)) (fmap_dom h)"
apply (simp add: fmap_dom_update)
done

lemma a12: "\<lbrakk> x \<notin> SS ; finite SS \<rbrakk> \<Longrightarrow> card (insert x SS) = Suc (card SS)"
apply (frule card_insert)
apply (case_tac "SS = {}")
 apply simp
 apply fastsimp
done

lemma copy_goodContext: "goodContext0 the_funSpecTab copy_Context "
apply (simp add: goodContext0_def copy_funTabBody copyBody_def)
apply clarsimp
apply (simp add: copy_Context_def)
apply clarsimp
apply (rule conjI)
apply (simp add: copy_funSpecTab)
apply clarsimp
apply (simp add: copy_funTabBody copyBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
apply (rule vdmLet)
(* defer 1  *)
(* instantiate expected pre-cond for recursive call *)
apply (rule vdmAx)
apply simp
(*
apply (subgoal_tac "{(CallFunExp copy [xin], copyAss [xin])} \<rhd> CallFunExp copy [tvar] : copyAss [tvar]")
apply simp
defer 1
*)
(*
apply (rule vdmAxAdapt)
apply fastsimp
*)
apply (rule vdmConstr)
apply (rule vdmConstr)
apply clarsimp
apply (simp add: copy_funTabArgs) 
apply (simp only: copyAss_def)
apply clarsimp
apply (simp add: copy_funSpecTab copyAss_def)
apply clarsimp
apply (erule_tac x="r" in allE)
apply (erule impE)
defer 1 (*  R (Ref r) = E xarg by unwrapping FrameInd *)
apply clarsimp?
apply (erule mList.cases)
apply simp_all
(* CASE: NIL *)
apply clarsimp
apply (rename_tac r h0)
apply (subgoal_tac " \<not> ( \<exists> rs0 . fmap_lookup h0 r = Some (C CONS rs0) )")
 prefer 2 apply clarsimp
apply clarsimp?
apply (erule_tac x="E" in allE)
apply (erule_tac x="BOTTOM" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPConFail)
apply clarsimp
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (subgoal_tac "h0 = h'a")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply (simp add: a11)
apply (subgoal_tac "freshloc (fmap_dom h'a) \<notin> (fmap_dom h'a)")
 prefer 2 apply (rule freshloc) apply (rule finite_dom_fmap)
apply (frule a12)
apply (rule finite_dom_fmap)
apply simp
(* clause on bounded result *)
apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
apply (rule mListNIL)
apply clarsimp
apply (erule ConstrInd.cases)
apply simp_all
(* CASE: CONS *)
apply (rename_tac l'x l't h'0 n't r'h r't)
apply (subgoal_tac "\<exists> E0'. ([hvar, tvar], [r'h, r't], E, E0'):PConInd")
apply clarsimp
apply (erule_tac x="E0'" in allE)
apply (erule_tac x="R (Ref l'x)" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPCon)
apply (rule_tac x="[r'h, Ref l't]" in exI)
apply clarsimp
(*
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
*)
apply clarsimp
apply (subgoal_tac "h'a = h'b")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply clarsimp?
apply (rename_tac ha')
apply (erule_tac x="n't" in allE)
apply (erule impE)
apply (rule_tac x="l't" in exI)
apply clarsimp?
apply (subgoal_tac "E0' tvar = R (Ref l't)")
 (* should follow from def of PConInd *)
apply (rule conjI)
apply clarsimp?
(* defer 1 *)
apply (rule_tac x="Ua" in exI)
apply clarsimp?
defer 1  (* should follow from def of PConInd *)
apply (rule conjI)
apply (subgoal_tac "card (insert (freshloc (fmap_dom ha')) (fmap_dom ha')) = Suc (card (fmap_dom ha'))")
 prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
apply simp 
(* main inequiv solved! *)
(* clause on bounded result *)
apply clarsimp?
apply (rule_tac x="Uv \<union> {(freshloc (fmap_dom ha'))}" in exI)
apply (rotate_tac -1)
(*
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply clarsimp
*)
apply (subgoal_tac "(n't, rv, Uv,  ha'(freshloc (fmap_dom ha')\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
apply (rule mListCONS)
apply simp+
(*
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply simp
*)
defer 1 (* rs = ... *)
apply (rule mListFresh)
apply simp+
defer 1
(*  \<exists>E0' ... *)
apply (rule_tac x="(E(tvar:=R r't))(hvar:=R r'h)" in exI)
apply (rule PConInd_CONS)
apply (rule PConInd_CONS)
apply (rule PConInd_NIL)
apply simp
apply fastsimp
(*  R (Ref r) = E xarg *)
apply (erule FrameInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply fastsimp
defer 1
apply (rule mListSameH)
apply simp
apply (simp add: sameH_def)
apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom ha')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom ha'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom ha')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
(* rs = ... *)
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
done

(* using goodContext and corresponding adaptation rule to prove it in an empty context *)
theorem copy_resources_ok: "{}  \<rhd> (CallFunExp copy [xin]) : copyAss [xin]"
apply (rule vdmCallFun)
apply (simp add: copy_funTabBody copyBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
apply (rule vdmLet)
(* defer 1  *)
(* instantiate expected pre-cond for recursive call *)
(*
apply (subgoal_tac "{(CallFunExp copy [xin], copyAss [xin])}  \<rhd> CallFunExp copy [tvar] : copyAss [tvar]")
apply simp
defer 1
*)
apply (rule CtxtWeakSingleton)
apply (rule vdmAdapt)
apply (rule copy_goodContext)
apply (simp add: copy_Context_def)
apply (simp add: copy_Context_def copy_funSpecTab)
apply fastsimp
apply (rule vdmConstr)
apply (rule vdmConstr)
apply clarsimp
apply (simp add: copy_funTabArgs) 
apply (simp only: copyAss_def)
apply clarsimp
(*
apply (simp add: copy_funSpecTab copyAss_def)
apply clarsimp
*)
apply (erule_tac x="r" in allE)
apply (erule impE)
defer 1 (*  R (Ref r) = E xarg by unwrapping FrameInd *)
apply clarsimp?
apply (erule mList.cases)
apply simp_all
(* CASE: NIL *)
apply clarsimp
apply (rename_tac r h0)
apply (subgoal_tac " \<not> ( \<exists> rs0 . fmap_lookup h0 r = Some (C CONS rs0) )")
 prefer 2 apply clarsimp
apply clarsimp?
apply (erule_tac x="E" in allE)
apply (erule_tac x="BOTTOM" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPConFail)
apply clarsimp
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (subgoal_tac "h0 = h'a")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply (simp add: a11)
apply (subgoal_tac "freshloc (fmap_dom h'a) \<notin> (fmap_dom h'a)")
 prefer 2 apply (rule freshloc) apply (rule finite_dom_fmap)
apply (frule a12)
apply (rule finite_dom_fmap)
apply simp
(* clause on bounded result *)
apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
apply (rule mListNIL)
apply clarsimp
apply (erule ConstrInd.cases)
apply simp_all
(* CASE: CONS *)
apply (rename_tac l'x l't h'0 n't r'h r't)
apply (subgoal_tac "\<exists> E0'. ([hvar, tvar], [r'h, r't], E, E0'):PConInd")
apply clarsimp
apply (erule_tac x="E0'" in allE)
apply (erule_tac x="R (Ref l'x)" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPCon)
apply (rule_tac x="[r'h, Ref l't]" in exI)
apply clarsimp
(*
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
*)
apply clarsimp
apply (subgoal_tac "h'a = h'b")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply clarsimp?
apply (simp add: copy_funSpecTab copyAss_def)
apply clarsimp?
apply (rename_tac ha')
apply (erule_tac x="n't" in allE)
apply (erule impE)
apply (rule_tac x="l't" in exI)
apply clarsimp?
apply (subgoal_tac "E0' tvar = R (Ref l't)")
 (* should follow from def of PConInd *)
apply (rule conjI)
apply clarsimp?
(* defer 1 *)
apply (rule_tac x="Ua" in exI)
apply clarsimp?
defer 1  (* should follow from def of PConInd *)
apply (rule conjI)
apply (subgoal_tac "card (insert (freshloc (fmap_dom ha')) (fmap_dom ha')) = Suc (card (fmap_dom ha'))")
 prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
apply simp 
(* main inequiv solved! *)
(* clause on bounded result *)
apply clarsimp?
apply (rule_tac x="Uv \<union> {(freshloc (fmap_dom ha'))}" in exI)
apply (rotate_tac -1)
(*
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply clarsimp
*)
apply (subgoal_tac "(n't, rv, Uv,  ha'(freshloc (fmap_dom ha')\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
apply (rule mListCONS)
apply simp+
(*
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply simp
*)
defer 1 (* rs = ... *)
apply (rule mListFresh)
apply simp+
defer 1
(*  \<exists>E0' ... *)
apply (rule_tac x="(E(tvar:=R r't))(hvar:=R r'h)" in exI)
apply (rule PConInd_CONS)
apply (rule PConInd_CONS)
apply (rule PConInd_NIL)
apply simp
apply fastsimp
(*  R (Ref r) = E xarg *)
apply (erule FrameInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply fastsimp
defer 1 (* rs = ... *)
apply (rule mListSameH)
apply simp
apply (simp add: sameH_def)
apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom ha')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom ha'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom ha')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
(* rs = ... *)
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
done

(* old proof of deferred rec call 
apply (subgoal_tac "{} \<union> {(CallFunExp copy [xin], copyAss [xin])} \<rhd> CallFunExp copy [tvar] : copyAss [tvar]")
apply simp
apply (rule ctxtWeak)
apply (rule vdmConseq)
prefer 2
apply (rule vdmAdapt)
apply (subgoal_tac "goodContext0 the_funSpecTab copy_Context")
apply simp
apply (rule copy_goodContext)
apply (simp add: copy_Context_def)
apply (simp add: copy_Context_def copy_funSpecTab copy_funTabArgs)
apply fastsimp
apply clarsimp
apply (simp add: copy_Context_def copy_funSpecTab copy_funTabArgs)
done
*)

(*  -----------------------------------------------------------------------------  *)
(* Example: fuse1 (same as fuse but ignore differing elements *)

syntax 
 h0var :: "VarName"
 t0var :: "VarName"
 h1var :: "VarName"
 t1var :: "VarName"
 y0var :: "VarName"
 x0arg  :: "VarName"
 x1arg  :: "VarName"
 x0in  :: "VarName"
 x1in  :: "VarName"
 b0var :: "VarName"
 fuse  :: "FunName"

translations
"h0var" == "(VN ''h0var'')"
"t0var" == "(VN ''t0var'')"
"h1var" == "(VN ''h1var'')"
"t1var" == "(VN ''t1var'')"
"y0var" == "(VN ''y0var'')"
"x0arg"  == "(VN ''x0arg'')"
"x1arg"  == "(VN ''x1arg'')"
"x0in"  == "(VN ''x0in'')"
"x1in"  == "(VN ''x1in'')"
"b0var"  == "(VN ''b0var'')"
"fuse" == "(FN ''fuse'')"

constdefs ieq :: "HoVal => HoVal => HoVal"
"ieq x y == B ((theInt x)=(theInt y))"

lemma ieq_tertium_non_datur: "ieq x y = B True \<or> ieq x y = B False"
by (simp add: ieq_def)

axioms list_only_heap: "\<forall> r h. r : fmap_dom h \<longrightarrow> fmap_lookup h r = Some (C NIL []) \<or> (\<exists> rh rt. fmap_lookup h r = Some (C CONS [Ref rh, Ref rt]))"

lemma bonzo11[rule_format]: "\<forall> n r U h. (n, r, U, h) \<in> mList \<longrightarrow> r : fmap_dom h"
apply clarsimp
apply (frule mListDom)
apply (erule mList.cases)
apply simp_all
done

(* only difference to fuse:
    (VarExp y0var)
   in else branch
*)

constdefs fuseBody :: Exp
"fuseBody \<equiv> (CaseOneExp x0arg 
                    (PCon CONS [h0var,t0var]) (CaseOneExp x1arg 
                                                 (PCon CONS [h1var,t1var]) (LetExp y0var (CallFunExp fuse [t0var,t1var]) 
                                                                            (LetExp b0var (RPrimBin ieq h0var h1var)
                                                                              (IfExp b0var 
                                                                                    (ConstrExp CONS [h0var,y0var])
                                                                                    (VarExp y0var)))) 
                                                 (ConstrExp NIL  [] ))
                    (ConstrExp NIL []))" 

constdefs fuseAss :: "[VarName list] => Assertion"
"fuseAss zs == \<lambda>E h h' v p.
          			   \<forall>n0 r0 U0 n1.
          			      E (hd zs) = R (Ref r0) \<and> (n0, r0, U0, h) \<in> mList \<and> (\<exists>r1. E (hd (tl zs)) = R (Ref r1) \<and> (\<exists>U1. (n1, r1, U1, h) \<in> mList) \<and> n0 = n1) \<longrightarrow>
          			      card (fmap_dom h') \<le> card (fmap_dom h) + n0 + 1 \<and> (\<exists> nv rv Uv. v = R (Ref rv) \<and> (nv, rv, Uv, h') \<in> mList \<and> nv \<le> n0 ) "

axioms fuse_funTabArgs: "fst (funTab fuse) == [x0arg,x1arg]"
axioms fuse_funTabBody: "snd (funTab fuse) == fuseBody"
axioms fuse_funSpecTab: "the_funSpecTab fuse xs == fuseAss xs"

constdefs fuse_Context :: Context
"fuse_Context == {(CallFunExp fuse [t0var, t1var], fuseAss [t0var,t1var])}"

lemma fuse_goodContext: "goodContext0 the_funSpecTab fuse_Context"
apply (simp add:  fuse_Context_def fuse_funSpecTab fuse_funTabBody fuseBody_def goodContext0_def )
apply clarsimp
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
apply (rule vdmAx)
apply fastsimp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmVar)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (simp only: fuseAss_def fuse_funTabArgs)
apply clarsimp
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)
apply (case_tac "hd ys \<noteq> x0arg \<and> hd ys \<noteq> x1arg")
 prefer 2
 apply clarsimp
 apply (erule disjE)
 apply clarsimp
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
 (* subgoal proven *)
apply (subgoal_tac " E x0arg = E' (hd ys)")
 prefer 2 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* CLAUSE: bounded output *)
 apply (rule_tac x="0" in exI)
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
 (* CLAUSE: bounded output *)
 apply (rule_tac x="0" in exI)
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
  (*
  apply (case_tac "vb' = \<bottom>") (* show contradiction *)
    apply clarsimp
    defer 1
  *)
    (* OTW *)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule_tac x="r't0" in exI)
    apply (rule conjI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule conjI)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4) (* (n0, r1, U1, h) \<in> mList *)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ieq vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (* CLAUSE: bounded output *)
     apply (rule_tac x="Suc nv" in exI)
     apply (rule conjI)
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
     apply (subgoal_tac "(nv, rv, Uv, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
     apply (rotate_tac -1)
     apply (rule mListCONS)
     prefer 4
     apply simp
     apply clarsimp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
     apply simp
     apply (rule mListSameH)
     apply simp
     apply (simp add: sameH_def)
     apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -5)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
     defer 1(*  Suc nv \<le> n0 \<dots>. *)
     (* ELSE branch: *)
     apply (subgoal_tac "ieq vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp
     (*
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp
     *)
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
    (* CLAUSE: bounded output *)
     apply (rule_tac x="nv" in exI)
     apply (rule conjI)
     apply (rule_tac x="Uv" in exI)
     apply simp
     apply arith
(* 8 mostly trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
   rs = ?...   
   freshloc (fmap_dom h'a) \<notin> Uv
   Suc nv \<le> n0
*)
 apply (insert ieq_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E0' t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* left-over subgoal from CLAUSE: bounded output *)
 apply (subgoal_tac "rs = [Ref r'h0, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 defer 1
 apply (rotate_tac 9)
 apply (frule mListDom)
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
   apply (rotate_tac -2)
   apply (drule subsetD)
   apply simp
   apply (subgoal_tac "finite (fmap_dom h'a)")
    prefer 2 apply (rule finite_dom_fmap)
   apply (frule freshloc)
   apply simp
   (* OTW *)
   apply assumption
 (* Suc nv \<le> n0 *)
 apply (subgoal_tac "(n0 - Suc 0) + Suc 0 \<le> n0")
  apply simp
  apply simp
 (* rs ... *)
 apply (subgoal_tac "E1' h0var = E0' h0var \<and> E0' h0var = R (Ref r'h0)")
 apply clarsimp
 apply (rule conjI)
 prefer 2
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp?
 apply (rotate_tac 9)
 apply (drule env_env_notin)
 prefer 2
 apply (subgoal_tac "E0' h0var = E1' h0var")
 apply simp
 apply simp
 apply simp
done

(* prove the assertion in an empty context, using adaptation rule *)
theorem fuse_resources_ok: "{}  \<rhd> (CallFunExp fuse [x0in,x1in]) : fuseAss [x0in, x1in]"
apply (rule vdmCallFun)
apply (simp add: fuse_funTabBody fuseBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
(* clumsy way of proving rec call 
apply (subgoal_tac "{(CallFunExp fuse [x0in, x1in], fuseAss [x0in, x1in])} \<rhd> CallFunExp fuse [t0var, t1var] : fuseAss [t0var, t1var]")
apply assumption
defer 1 (* must use fuse_goodContext *)
*)
apply (rule CtxtWeakSingleton)
apply (rule vdmAdapt)
apply (rule fuse_goodContext)
apply (simp add: fuse_Context_def)
apply (simp add: fuse_Context_def fuse_funSpecTab)
apply fastsimp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmVar)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (simp only: fuseAss_def fuse_funTabArgs)
apply clarsimp
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)
(*
apply (case_tac "hd ys \<noteq> x0arg \<and> hd ys \<noteq> x1arg")
 prefer 2
 apply clarsimp
 apply (erule disjE)
*)
 apply clarsimp?
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* CLAUSE: bounded output *)
 apply (rule_tac x="0" in exI)
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
 (* CLAUSE: bounded output *)
 apply (rule_tac x="0" in exI)
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
    (* OTW *)
    apply clarsimp?
    apply (simp add: fuse_Context_def fuse_funSpecTab fuseAss_def)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule_tac x="r't0" in exI)
    apply (rule conjI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule conjI)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4) (* (n0, r1, U1, h) \<in> mList *)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ieq vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (* CLAUSE: bounded output *)
     apply (rule_tac x="Suc nv" in exI)
     apply (rule conjI)
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
     apply (subgoal_tac "(nv, rv, Uv, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
     apply (rotate_tac -1)
     apply (rule mListCONS)
     prefer 4
     apply simp
     apply clarsimp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
     apply simp
     apply (rule mListSameH)
     apply simp
     apply (simp add: sameH_def)
     apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -5)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
     defer 1(*  Suc nv \<le> n0 \<dots>. *)
     (* ELSE branch: *)
     apply (subgoal_tac "ieq vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp
     (*
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp
     *)
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
    (* CLAUSE: bounded output *)
     apply (rule_tac x="nv" in exI)
     apply (rule conjI)
     apply (rule_tac x="Uv" in exI)
     apply simp
     apply arith
(* 8 mostly trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
   rs = ?...   
   freshloc (fmap_dom h'a) \<notin> Uv
   Suc nv \<le> n0
*)
 apply (insert ieq_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E0' t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* left-over subgoal from CLAUSE: bounded output *)
 apply (subgoal_tac "rs = [Ref r'h0, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 defer 1
 apply (rotate_tac 9)
 apply (frule mListDom)
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
   apply (rotate_tac -2)
   apply (drule subsetD)
   apply simp
   apply (subgoal_tac "finite (fmap_dom h'a)")
    prefer 2 apply (rule finite_dom_fmap)
   apply (frule freshloc)
   apply simp
   (* OTW *)
   apply assumption
 (* Suc nv \<le> n0 *)
 apply (subgoal_tac "(n0 - Suc 0) + Suc 0 \<le> n0")
  apply simp
  apply simp
 defer 1 (* rec call *)
 (* rs ... *)
 apply (subgoal_tac "E1' h0var = E0' h0var \<and> E0' h0var = R (Ref r'h0)")
 apply clarsimp
 apply (rule conjI)
 prefer 2
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp?
 apply (rotate_tac 9)
 apply (drule env_env_notin)
 prefer 2
 apply (subgoal_tac "E0' h0var = E1' h0var")
 apply simp
 apply simp
 apply simp
done

(* Example finished *)


(* old proof of deferred rec call 
apply (subgoal_tac "{} \<union> {(CallFunExp fuse [x0in, x1in], fuseAss [x0in, x1in])} \<rhd> CallFunExp fuse [t0var, t1var] : fuseAss [t0var, t1var]")
apply simp
apply (rule ctxtWeak)
apply (rule vdmConseq)
prefer 2
apply (rule vdmAdapt)
apply (subgoal_tac "goodContext0 the_funSpecTab fuse_Context")
apply simp
apply (rule fuse_goodContext)
apply (simp add: fuse_Context_def)
apply (simp add: fuse_Context_def fuse_funSpecTab fuse_funTabArgs)
apply fastsimp
apply clarsimp
apply (simp add: fuse_Context_def fuse_funSpecTab fuse_funTabArgs)
done
*)

end

(* ============================================================================= *)
(*  -----------------------------------------------------------------------------  *)
(* OUTDATED: this returns NIL in the else case, despite a recursive call before; doesn't make much sense *)
(* Example: fuse *)

syntax 
 h0var :: "VarName"
 t0var :: "VarName"
 h1var :: "VarName"
 t1var :: "VarName"
 y0var :: "VarName"
 x0arg  :: "VarName"
 x1arg  :: "VarName"
 x0in  :: "VarName"
 x1in  :: "VarName"
 b0var :: "VarName"
 fuse  :: "FunName"

translations
"h0var" == "(VN ''h0var'')"
"t0var" == "(VN ''t0var'')"
"h1var" == "(VN ''h1var'')"
"t1var" == "(VN ''t1var'')"
"y0var" == "(VN ''y0var'')"
"x0arg"  == "(VN ''x0arg'')"
"x1arg"  == "(VN ''x1arg'')"
"x0in"  == "(VN ''x0in'')"
"x1in"  == "(VN ''x1in'')"
"b0var"  == "(VN ''b0var'')"
"fuse" == "(FN ''fuse'')"

constdefs ieq :: "HoVal => HoVal => HoVal"
"ieq x y == B ((theInt x)=(theInt y))"

lemma ieq_tertium_non_datur: "ieq x y = B True \<or> ieq x y = B False"
by (simp add: ieq_def)

axioms list_only_heap: "\<forall> r h. r : fmap_dom h \<longrightarrow> fmap_lookup h r = Some (C NIL []) \<or> (\<exists> rh rt. fmap_lookup h r = Some (C CONS [Ref rh, Ref rt]))"

lemma bonzo11[rule_format]: "\<forall> n r U h. (n, r, U, h) \<in> mList \<longrightarrow> r : fmap_dom h"
apply clarsimp
apply (frule mListDom)
apply (erule mList.cases)
apply simp_all
done

constdefs fuseBody :: Exp
"fuseBody \<equiv> (CaseOneExp x0arg 
                    (PCon CONS [h0var,t0var]) (CaseOneExp x1arg 
                                                 (PCon CONS [h1var,t1var]) (LetExp y0var (CallFunExp fuse [t0var,t1var]) 
                                                                            (LetExp b0var (RPrimBin ieq h0var h1var)
                                                                              (IfExp b0var 
                                                                                    (ConstrExp CONS [h0var,y0var])
                                                                                    (ConstrExp NIL  [] )))) 
                                                 (ConstrExp NIL  [] ))
                    (ConstrExp NIL []))" 

constdefs fuseAss :: "[VarName list] => Assertion"
"fuseAss zs == \<lambda>E h h' v p.
          			   \<forall>n0 r0 U0 n1.
          			      E (hd zs) = R (Ref r0) \<and> (n0, r0, U0, h) \<in> mList \<and> (\<exists>r1. E (hd (tl zs)) = R (Ref r1) \<and> (\<exists>U1. (n1, r1, U1, h) \<in> mList) \<and> n0 = n1) \<longrightarrow>
          			      card (fmap_dom h') \<le> card (fmap_dom h) + n0 + 1"

axioms fuse_funTabArgs: "fst (funTab fuse) == [x0arg,x1arg]"
axioms fuse_funTabBody: "snd (funTab fuse) == fuseBody"
axioms fuse_funSpecTab: "the_funSpecTab fuse xs == fuseAss xs"

constdefs fuse_Context :: Context
"fuse_Context == {(CallFunExp fuse [t0var, t1var], fuseAss [t0var,t1var])}"

lemma fuse_goodContext: "goodContext0 the_funSpecTab fuse_Context"
apply (simp add:  fuse_Context_def fuse_funSpecTab fuse_funTabBody fuseBody_def goodContext0_def )
apply clarsimp
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
apply (rule vdmAx)
apply fastsimp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmConstr)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (simp only: fuseAss_def fuse_funTabArgs)
apply clarsimp
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)

apply (case_tac "hd ys \<noteq> x0arg \<and> hd ys \<noteq> x1arg")
 prefer 2
 apply clarsimp
 apply (erule disjE)
 apply clarsimp
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
 (* subgoal proven *)
apply (subgoal_tac " E x0arg = E' (hd ys)")
 prefer 2 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* defer 1 *)
 (* main inequiv solved (for NIL case)! *)
 (*
 apply fastsimp
 apply (erule mList.cases)
 apply simp_all
 apply clarsimp
 apply (subgoal_tac "finite (fmap_dom h) \<and> freshloc (fmap_dom h) \<notin> fmap_dom h")
  prefer 2 apply (rule conjI) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap) 
 apply (simp add: a12)
 apply (rule ConstrInd_h_same) apply simp
 *)
 (*
 apply (erule FrameInd.cases)
 apply simp_all
 *)
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
  (*
  apply (case_tac "vb' = \<bottom>") (* show contradiction *)
    apply clarsimp
    defer 1
  *)
    (* OTW *)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule_tac x="r't0" in exI)
    apply (rule conjI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule conjI)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4) (* (n0, r1, U1, h) \<in> mList *)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ieq vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (*
     apply (subgoal_tac "Suc (card (fmap_dom h'a)) \<le> card (fmap_dom h) + n0") 
     apply clarsimp
     apply arith
     *)
     (* main ineq solved for TRUE branch *)
     (* ELSE branch: *)
     apply (subgoal_tac "ieq vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (*
     apply (subgoal_tac "card (fmap_dom h) + (n0 - Suc 0)  \<le> Suc (card (fmap_dom h) + n0)") 
     apply clarsimp
     apply arith
     apply clarsimp
     apply arith
     *)
     (* main ineq solved for ELSE branch *)
 (* 5 trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
*)
 apply (insert ieq_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd emptyEnv t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd emptyEnv t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* hd ys \<noteq> x0arg \<and> hd ys \<noteq> x1arg *)
done

(* prove the assertion in an empty context, using adaptation rule *)
theorem fuse_resources_ok: "{}  \<rhd> (CallFunExp fuse [x0in,x1in]) : fuseAss [x0in, x1in]"
apply (rule vdmCallFun)
apply (simp add: fuse_funTabBody fuseBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
(* apply (rule vdmAxAdapt) *)
apply (subgoal_tac "{(CallFunExp fuse [x0in, x1in], fuseAss [x0in, x1in])} \<rhd> CallFunExp fuse [t0var, t1var] : fuseAss [t0var, t1var]")
apply assumption
defer 1 (* must use fuse_goodContext *)
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmConstr)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (simp only: fuseAss_def fuse_funTabArgs)
apply clarsimp
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)
apply (subgoal_tac " E x0arg = E' x0in")
 prefer 2 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* defer 1 *)
 (* main inequiv solved (for NIL case)! *)
 (*
 apply fastsimp
 apply (erule mList.cases)
 apply simp_all
 apply clarsimp
 apply (subgoal_tac "finite (fmap_dom h) \<and> freshloc (fmap_dom h) \<notin> fmap_dom h")
  prefer 2 apply (rule conjI) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap) 
 apply (simp add: a12)
 apply (rule ConstrInd_h_same) apply simp
 *)
 (*
 apply (erule FrameInd.cases)
 apply simp_all
 *)
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
  (*
  apply (case_tac "vb' = \<bottom>") (* show contradiction *)
    apply clarsimp
    defer 1
  *)
    (* OTW *)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule_tac x="r't0" in exI)
    apply (rule conjI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule conjI)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ieq vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (*
     apply (subgoal_tac "Suc (card (fmap_dom h'a)) \<le> card (fmap_dom h) + n0") 
     apply clarsimp
     apply arith
     *)
     (* main ineq solved for TRUE branch *)
     (* ELSE branch: *)
     apply (subgoal_tac "ieq vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (*
     apply (subgoal_tac "card (fmap_dom h) + (n0 - Suc 0)  \<le> Suc (card (fmap_dom h) + n0)") 
     apply clarsimp
     apply arith
     apply clarsimp
     apply arith
     *)
     (* main ineq solved for ELSE branch *)
 (* 5 trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
*)
 apply (insert ieq_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd emptyEnv t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd emptyEnv t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 defer 1 (* fuse_goodContext *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
(* finally, prove the assertion on the recursive call, using the context *)
 apply (subgoal_tac "{} \<union> {(CallFunExp fuse [x0in, x1in], fuseAss [x0in, x1in])} \<rhd> CallFunExp fuse [t0var, t1var] : fuseAss [t0var, t1var]")
 apply simp
 apply (rule ctxtWeak)
 apply (rule vdmConseq)
 prefer 2
 apply (rule vdmAdapt)
 apply (subgoal_tac "goodContext0 the_funSpecTab fuse_Context")
 apply simp
 apply (rule fuse_goodContext)
 apply (simp add: fuse_Context_def)
 apply (simp add: fuse_Context_def fuse_funSpecTab fuse_funTabArgs)
 apply fastsimp
 apply clarsimp
 apply (simp add: fuse_funSpecTab)
done

(* prove the assertion in an empty context, using UNPROVEN adaptation rule *)
theorem fuseCallAdapt: "{} \<rhd> (CallFunExp fuse [x0in,x1in]) : fuseAss [x0in, x1in]"
apply (rule vdmCallFun)
apply (simp add: fuse_funTabBody fuseBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
apply (rule vdmAxAdapt_UNPROVEN)
apply fastsimp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmConstr)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (simp only: fuseAss_def fuse_funTabArgs)
apply clarsimp
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)
apply (subgoal_tac " E x0arg = E' x0in")
 prefer 2 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* defer 1 *)
 (* main inequiv solved (for NIL case)! *)
 (*
 apply fastsimp
 apply (erule mList.cases)
 apply simp_all
 apply clarsimp
 apply (subgoal_tac "finite (fmap_dom h) \<and> freshloc (fmap_dom h) \<notin> fmap_dom h")
  prefer 2 apply (rule conjI) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap) 
 apply (simp add: a12)
 apply (rule ConstrInd_h_same) apply simp
 *)
 (*
 apply (erule FrameInd.cases)
 apply simp_all
 *)
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
  (*
  apply (case_tac "vb' = \<bottom>") (* show contradiction *)
    apply clarsimp
    defer 1
  *)
    (* OTW *)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule_tac x="r't0" in exI)
    apply (rule conjI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule conjI)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ieq vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (*
     apply (subgoal_tac "Suc (card (fmap_dom h'a)) \<le> card (fmap_dom h) + n0") 
     apply clarsimp
     apply arith
     *)
     (* main ineq solved for TRUE branch *)
     (* ELSE branch: *)
     apply (subgoal_tac "ieq vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (*
     apply (subgoal_tac "card (fmap_dom h) + (n0 - Suc 0)  \<le> Suc (card (fmap_dom h) + n0)") 
     apply clarsimp
     apply arith
     apply clarsimp
     apply arith
     *)
     (* main ineq solved for ELSE branch *)
 (* 5 trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
*)
 apply (insert ieq_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd emptyEnv t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd emptyEnv t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
done

(* uses a context containing necessary info for the recursive call *)
theorem fuseCall1: "{(CallFunExp fuse [t0var, t1var], fuseAss [t0var, t1var])}
         \<rhd> (CallFunExp fuse [x0in,x1in]) : (\<lambda> E h h' v p. 
                                 (\<forall> n0 r0 U0 n1 r1 U1 . (E x0in = (R (Ref r0)) \<and> (n0,r0,U0,h):mList \<and> E x1in = (R (Ref r1)) \<and> (n1,r1,U1,h):mList \<and> n0=n1)
                                                        \<longrightarrow> card (fmap_dom h') \<le> Suc (card (fmap_dom h) + n0)))"
apply (rule vdmCallFun)
apply (simp add: fuse_funTabBody fuseBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
apply (rule vdmAx)
apply simp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmConstr)
(* code has been completely unfolded; 1 huge subgoal left *)
apply clarsimp
apply (simp only: fuse_funTabArgs)
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)
apply (subgoal_tac " E x0arg = E' x0in")
 prefer 2 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* main inequiv solved (for NIL case)! *)
 (*
 apply fastsimp
 apply (erule mList.cases)
 apply simp_all
 apply clarsimp
 apply (subgoal_tac "finite (fmap_dom h) \<and> freshloc (fmap_dom h) \<notin> fmap_dom h")
  prefer 2 apply (rule conjI) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap) 
 apply (simp add: a12)
 apply (rule ConstrInd_h_same) apply simp
 *)
 (*
 apply (erule FrameInd.cases)
 apply simp_all
 *)
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
  (*
  apply (case_tac "vb' = \<bottom>") (* show contradiction *)
    apply clarsimp
    defer 1
  *)
    (* OTW *)
    apply (simp add: fuseAss_def)    
    apply clarsimp?
    apply (erule_tac x="n1 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule_tac x="r't0" in exI)
    apply (rule conjI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule conjI)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4) (* (n1, r1, U1, h) \<in> mList *)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ieq vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "card (fmap_dom h) + (n1 - Suc 0)  \<le> Suc (card (fmap_dom h) + n1)") 
     apply clarsimp
     apply (subgoal_tac "n1>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     apply arith
     (* main ineq solved for TRUE branch *)
     (* ELSE branch: *)
     apply (subgoal_tac "ieq vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp
     apply (subgoal_tac "card (fmap_dom h) + (n1 - Suc 0)  \<le> Suc (card (fmap_dom h) + n1)") 
     apply (subgoal_tac "n1>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     apply arith
     (* main ineq solved for ELSE branch *)
 (* 5 trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
*)
 apply (insert ieq_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd emptyEnv t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule bonzo11)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd emptyEnv t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
done

(* Example fuse completed *)

end