(* Time-stamp: <Sat Jan 24 2009 00:18:24 Stardate: [-28]0689.85 hwloidl>
   Example: listcopy/fuse
   Top-level theorems (expression level): 
    . copy_resources_ok
    . fuse_resources_ok
   (these specify bounds on heap consumption and on result data structure).
*)

theory ListCopy = Integrated:

(* ----------------------------------------------------------------------------- *)

syntax 
 h9var :: "VarName"
 t9var :: "VarName"
 y9var :: "VarName"
 b9var :: "VarName"
 x9arg :: "VarName"
 x9in  :: "VarName"
 konst :: "FunName"

translations
"h9var" == "(VN ''h9var'')"
"t9var" == "(VN ''t9var'')"
"y9var" == "(VN ''y9var'')"
"b9var" == "(VN ''b9var'')"
"x9arg"  == "(VN ''x9arg'')"
"x9in"  == "(VN ''x9in'')"
"konst" == "(FN ''konst'')"

constdefs is_zero :: "HoVal => HoVal => HoVal"
"is_zero x y == B ((theInt x)=(0 :: int))"

lemma is_zero_tertium_non_datur: "is_zero x y = B True \<or> is_zero x y = B False"
by (simp add: is_zero_def)

constdefs inc :: "HoVal => HoVal => HoVal"
"inc x y == I ((theInt x)+(1 :: int))"

constdefs dec :: "HoVal => HoVal => HoVal"
"dec x y == I ((theInt x)-(1 :: int))"

constdefs konstBody :: Exp
"konstBody \<equiv> (LetExp b9var (PrimBin is_zero x9arg x9arg)
                (IfExp b9var 
                  (ValueExp (I 0))
                  (LetExp t9var (PrimBin dec x9arg x9arg)
                   (LetExp y9var (CallFunExp konst [t9var])
                    (PrimBin inc y9var y9var)))))"

axioms konst_funTabArgs: "fst (funTab konst) == [x9arg]"
axioms konst_funTabBody: "snd (funTab konst) == konstBody"

constdefs konstAss :: "[VarName list] => Assertion"
"konstAss zs == (\<lambda> E h h' v p. 
                                    (\<forall> n . (E (hd zs) = (I n)) \<longrightarrow> v = (I n)))"

axioms konst_funSpecTab: "the_funSpecTab konst xs == konstAss xs"

constdefs konst_Context :: "Context"
"konst_Context \<equiv> {(CallFunExp konst [t9var] ,  konstAss [t9var])}"


lemma "hd (x#xs) = x"
by simp

lemma konst_goodContext: "goodContext0 the_funSpecTab konst_Context"
apply (simp add: goodContext0_def konst_funTabBody konstBody_def konst_Context_def konst_funSpecTab konstAss_def)
apply clarsimp
apply (rule vdmConseq)
prefer 2
apply (rule vdmLet)
apply (rule vdmPrimBin)
apply (rule vdmIf)
apply (rule vdmValue)
apply (rule vdmLet)
(* defer 1  *)
apply (rule vdmPrimBin)
apply (rule vdmLet)
apply (rule vdmAx)
apply simp
apply (rule vdmPrimBin)
apply clarsimp?
apply (case_tac "is_zero (E x9arg) (E x9arg) = B True")
 apply clarsimp
 defer 1
 apply clarsimp
 apply (subgoal_tac "is_zero (E x9arg) (E x9arg) = B False")
  apply clarsimp
  apply (simp add: inc_def dec_def konst_funTabArgs)
  apply (erule FrameInd.cases)
  apply clarsimp
  apply (erule FrameInd.cases)
  apply clarsimp
  apply clarsimp
 apply (insert is_zero_tertium_non_datur)
 apply fastsimp
 apply (simp add: is_zero_def konst_funTabArgs)
 apply (erule FrameInd.cases)
 apply clarsimp
 apply (erule FrameInd.cases)
 apply clarsimp
 apply fastsimp
done

(* uses vdmAxAdapt_UNPROVEN; uses subgoal when finding rec call *)
theorem konstCallAdapt: "{} \<rhd> (CallFunExp konst [xin]) : konstAss [xin]"
apply (rule vdmCallFun)
apply (simp add: konst_funTabBody konstBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmLet)
apply (rule vdmPrimBin)
apply (rule vdmIf)
apply (rule vdmValue)
apply (rule vdmLet)
(* defer 1  *)
apply (rule vdmPrimBin)
apply (rule vdmLet)
(* apply (rule vdmAx) *)
apply (rule CtxtWeakSingleton)
apply (rule vdmAdapt)
apply (rule konst_goodContext)
apply (simp add: konst_Context_def)
apply (simp add: konst_Context_def konst_funSpecTab)
apply fastsimp
apply (rule vdmPrimBin)
apply clarsimp?
apply (case_tac "is_zero (E x9arg) (E x9arg) = B True")
 apply clarsimp
 defer 1
 apply clarsimp
 apply (subgoal_tac "is_zero (E x9arg) (E x9arg) = B False")
  apply clarsimp
  apply (simp add: inc_def dec_def konst_funTabArgs konstAss_def konst_funSpecTab)
  apply clarsimp
  apply (erule FrameInd.cases)
  apply clarsimp
  apply (erule FrameInd.cases)
  apply clarsimp
  apply clarsimp
 apply (insert is_zero_tertium_non_datur)
 apply fastsimp
 apply (simp add: is_zero_def konst_funTabArgs konstAss_def konst_funSpecTab)
 apply (erule FrameInd.cases)
 apply clarsimp
 apply (erule FrameInd.cases)
 apply clarsimp
 apply fastsimp
done

(* uses vdmAxAdapt_UNPROVEN; obvious proof-strategy: first all syntax-directed rules, then only simps *)
(*
theorem konstCallAdapt1: "{}
    \<rhd> (CallFunExp konst [xin]) : konstAss [xin]"
apply (rule vdmCallFun)
apply (simp add: konst_funTabBody konstBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmLet)
apply (rule vdmPrimBin)
apply (rule vdmIf)
apply (rule vdmValue)
apply (rule vdmLet)
apply (rule vdmPrimBin)
apply (rule vdmLet)
apply (rule vdmAxAdapt_UNPROVEN)
apply fastsimp
apply (rule vdmPrimBin)
apply clarsimp
apply (simp add: konstAss_def)
apply (case_tac "is_zero (E x9arg) (E x9arg) = B True")
 apply clarsimp
 defer 1
 apply clarsimp
 apply (subgoal_tac "is_zero (E x9arg) (E x9arg) = B False")
  apply clarsimp
  apply (simp add: inc_def dec_def konst_funTabArgs)
  apply (erule FrameInd.cases)
  apply clarsimp
  apply (erule FrameInd.cases)
  apply clarsimp
  apply clarsimp
 prefer 2
 apply (simp add: is_zero_def konst_funTabArgs)
 apply (erule FrameInd.cases)
 apply clarsimp
 apply (erule FrameInd.cases)
 apply clarsimp
 apply fastsimp
 apply (insert is_zero_tertium_non_datur)
 apply fastsimp
done
*)

(* -------------------------------------------------- *)
(* modelling LISTS *)

syntax 
 CONS :: ConName
 NIL  :: ConName
translations
 "CONS" == "(CN ''CONS'')"
 "NIL" == "(CN ''NIL'')"

consts mList::"(nat \<times> Locn \<times> (Locn set) \<times> Heap) set"
inductive mList intros
mListNIL: "\<lbrakk> fmap_lookup h a = Some (C NIL []) \<rbrakk> \<Longrightarrow> (0,a,{a},h) : mList"
mListCONS:"\<lbrakk> fmap_lookup h a = Some (C CONS [rx,rxs]) ; a \<notin> U; (Ref b) = rxs ; (n,b,U,h):mList\<rbrakk>
          \<Longrightarrow> (Suc n, a, U \<union> {a}, h): mList"


lemma mListDom: "\<lbrakk>(n, a, U, h) \<in> mList\<rbrakk> \<Longrightarrow>  U \<subseteq> fmap_dom h"
apply (erule mList.induct)
apply (simp_all add:fmap_lookup_def fmap_dom_def dom_def)
done

lemma mListFresh: "(L, a, U, h) \<in> mList \<Longrightarrow> freshloc (fmap_dom h) \<notin> U"
apply (subgoal_tac "U \<subseteq> fmap_dom h")
apply (subgoal_tac "finite (fmap_dom h)")
apply (drule freshloc)
apply fastsimp
apply fastsimp
apply (erule mListDom)
done

(* TODO: move this to Basics.thy or thereabouts *)
constdefs sameH::"Locn set \<Rightarrow> Heap \<Rightarrow> Heap \<Rightarrow> bool"
"sameH U h hh \<equiv> (\<forall> l. (l \<in> U \<longrightarrow> (fmap_lookup h l = fmap_lookup hh l)))"


lemma mListSameH[rule_format]:
"(n,a,U,h) \<in> mList \<Longrightarrow> (\<forall> hh. sameH U h hh \<longrightarrow> (n,a,U,hh) \<in> mList)"
apply (erule mList.induct, safe)
apply (rule mListNIL)
apply (simp add: sameH_def)
apply (simp add: sameH_def)
apply (erule_tac x="hh" in allE)
apply (erule impE)
apply clarsimp
apply (subgoal_tac "(Suc n, a, U \<union> {a}, hh) \<in> mList")
apply simp
apply (rule mListCONS)
apply (erule_tac x="a" in allE)
apply ( simp add: sameH_def)
apply (subgoal_tac "fmap_lookup hh a = Some (C CONS [rx, Ref b])") 
apply (rotate_tac -2)
apply (erule thin_rl)
apply ( simp add: sameH_def)
apply (rule conjI)
apply simp+
done

lemma mListRefInDom[rule_format]: "\<forall> n r U h. (n, r, U, h) \<in> mList \<longrightarrow> r : fmap_dom h"
apply clarsimp
apply (frule mListDom)
apply (erule mList.cases)
apply simp_all
done

lemma mListSameRef[rule_format]:
"(n,a,U,h) \<in> mList \<Longrightarrow> (\<forall> n' U'. (n',a,U',h) \<in> mList \<longrightarrow> (n=n' \<and> U=U'))"
apply (erule mList.induct)
apply clarsimp
apply (erule mList.cases)
apply clarsimp
apply clarsimp
apply (rename_tac l'hd l'tl h n rx rxs)
apply (rule allI)+
apply (rule impI)
apply (rotate_tac -1)
apply (erule mList.cases)
defer 1
apply (subgoal_tac "l'hd = a \<and> n' = Suc na \<and> U' = Ua \<union> {a} \<and> h = ha")
apply (erule_tac x="na" in allE)
apply (erule_tac x="Ua" in allE)
apply (erule impE)
apply fastsimp
apply fastsimp
apply fastsimp
apply (subgoal_tac "l'hd = a \<and> n' = 0 \<and> U' = {a} \<and> h = ha")
apply (rotate_tac 4)
apply (erule thin_rl)
apply fastsimp
apply (rotate_tac 4)
apply (erule thin_rl)
apply fastsimp
done

(* -------------------------------------------------- *)
(* Example: list-copy *)


syntax 
 hvar :: "VarName"
 tvar :: "VarName"
 yvar :: "VarName"
 xarg  :: "VarName"
 xin  :: "VarName"
 copy :: "FunName"

translations
"hvar" == "(VN ''hvar'')"
"tvar" == "(VN ''tvar'')"
"yvar" == "(VN ''yvar'')"
"xarg"  == "(VN ''xarg'')"
"xin"  == "(VN ''xin'')"
"copy" == "(FN ''copy'')"

constdefs copyBody :: Exp
"copyBody \<equiv> CaseOneExp xarg 
                    (PCon CONS [hvar,tvar]) (LetExp yvar (CallFunExp copy [tvar]) 
                                              (ConstrExp CONS [hvar,yvar])) 
                    (ConstrExp NIL [])" 

constdefs copyAss :: "[VarName list] => Assertion"
"copyAss zs == (\<lambda> E h h' v p. 
                                 (\<forall> n r U . (E (hd zs) = (R (Ref r)) \<and> (n,r,U,h):mList  
                                             \<longrightarrow> (card (fmap_dom h') \<le> card (fmap_dom h) + n + 1) \<and> (\<exists> rv Uv. v = R (Ref rv) \<and> (n,rv,Uv,h'):mList))))"

constdefs copy0Ass :: "[VarName list] => Assertion"
"copy0Ass zs == (\<lambda> E h h' v p. 
                                 (\<forall> n r U . (E (hd zs) = (R (Ref r)) \<and> (n,r,U,h):mList  
                                             \<longrightarrow> (card (fmap_dom h') \<le> card (fmap_dom h) + n + 1) \<and> (\<exists> rv Uv. v = R (Ref rv) \<and> (n,rv,Uv,h'):mList \<and> sameH ((fmap_dom h)-Uv) h h' \<and> Uv \<inter> (fmap_dom h) = {}) \<and> (n,r,U,h'):mList)))"

axioms copy_funTabArgs: "fst (funTab copy) == [xarg]"
axioms copy_funTabBody: "snd (funTab copy) == copyBody"
axioms copy_funSpecTab: "the_funSpecTab copy xs = copyAss xs"
axioms copy0_funSpecTab: "the_funSpecTab copy xs = copy0Ass xs"

constdefs copy_Context :: Context
"copy_Context == {(CallFunExp copy [tvar], copyAss [tvar])}"

lemma a11: "fmap_dom (h((freshloc (fmap_dom h)) \<mapsto>\<^sub>f v)) = insert (freshloc (fmap_dom h)) (fmap_dom h)"
apply (simp add: fmap_dom_update)
done

lemma a12: "\<lbrakk> x \<notin> SS ; finite SS \<rbrakk> \<Longrightarrow> card (insert x SS) = Suc (card SS)"
apply (frule card_insert)
apply (case_tac "SS = {}")
 apply simp
 apply fastsimp
done

lemma copy_goodContext: "goodContext0 the_funSpecTab copy_Context "
apply (simp add: goodContext0_def copy_funTabBody copyBody_def)
apply clarsimp
apply (simp add: copy_Context_def)
apply clarsimp
apply (rule conjI)
apply (simp add: copy_funSpecTab)
apply clarsimp
apply (simp add: copy_funTabBody copyBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
apply (rule vdmLet)
(* defer 1  *)
(* instantiate expected pre-cond for recursive call *)
apply (rule vdmAx)
apply simp
(*
apply (subgoal_tac "{(CallFunExp copy [xin], copyAss [xin])} \<rhd> CallFunExp copy [tvar] : copyAss [tvar]")
apply simp
defer 1
*)
(*
apply (rule vdmAxAdapt)
apply fastsimp
*)
apply (rule vdmConstr)
apply (rule vdmConstr)
apply clarsimp
apply (simp add: copy_funTabArgs) 
apply (simp only: copyAss_def)
apply clarsimp
apply (simp add: copy_funSpecTab copyAss_def)
apply clarsimp
apply (erule_tac x="r" in allE)
apply (erule impE)
defer 1 (*  R (Ref r) = E xarg by unwrapping FrameInd *)
apply clarsimp?
apply (erule mList.cases)
apply simp_all
(* CASE: NIL *)
apply clarsimp
apply (rename_tac r h0)
apply (subgoal_tac " \<not> ( \<exists> rs0 . fmap_lookup h0 r = Some (C CONS rs0) )")
 prefer 2 apply clarsimp
apply clarsimp?
apply (erule_tac x="E" in allE)
apply (erule_tac x="BOTTOM" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPConFail)
apply clarsimp
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (subgoal_tac "h0 = h'a")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply (simp add: a11)
apply (subgoal_tac "freshloc (fmap_dom h'a) \<notin> (fmap_dom h'a)")
 prefer 2 apply (rule freshloc) apply (rule finite_dom_fmap)
apply (frule a12)
apply (rule finite_dom_fmap)
apply simp
(* clause on bounded result *)
apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
apply (rule mListNIL)
apply clarsimp
apply (erule ConstrInd.cases)
apply simp_all
(* CASE: CONS *)
apply (rename_tac l'x l't h'0 n't r'h r't)
apply (subgoal_tac "\<exists> E0'. ([hvar, tvar], [r'h, r't], E, E0'):PConInd")
apply clarsimp
apply (erule_tac x="E0'" in allE)
apply (erule_tac x="R (Ref l'x)" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPCon)
apply (rule_tac x="[r'h, Ref l't]" in exI)
apply clarsimp
(*
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
*)
apply clarsimp
apply (subgoal_tac "h'a = h'b")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply clarsimp?
apply (rename_tac ha')
apply (erule_tac x="n't" in allE)
apply (erule impE)
apply (rule_tac x="l't" in exI)
apply clarsimp?
apply (subgoal_tac "E0' tvar = R (Ref l't)")
 (* should follow from def of PConInd *)
apply (rule conjI)
apply clarsimp?
(* defer 1 *)
apply (rule_tac x="Ua" in exI)
apply clarsimp?
defer 1  (* should follow from def of PConInd *)
apply (rule conjI)
apply (subgoal_tac "card (insert (freshloc (fmap_dom ha')) (fmap_dom ha')) = Suc (card (fmap_dom ha'))")
 prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
apply simp 
(* main inequiv solved! *)
(* clause on bounded result *)
apply clarsimp?
apply (rule_tac x="Uv \<union> {(freshloc (fmap_dom ha'))}" in exI)
apply (rotate_tac -1)
(*
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply clarsimp
*)
apply (subgoal_tac "(n't, rv, Uv,  ha'(freshloc (fmap_dom ha')\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
apply (rule mListCONS)
apply simp+
(*
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply simp
*)
defer 1 (* rs = ... *)
apply (rule mListFresh)
apply simp+
defer 1
(*  \<exists>E0' ... *)
apply (rule_tac x="(E(tvar:=R r't))(hvar:=R r'h)" in exI)
apply (rule PConInd_CONS)
apply (rule PConInd_CONS)
apply (rule PConInd_NIL)
apply simp
apply fastsimp
(*  R (Ref r) = E xarg *)
apply (erule FrameInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply fastsimp
defer 1
apply (rule mListSameH)
apply simp
apply (simp add: sameH_def)
apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom ha')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom ha'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom ha')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
(* rs = ... *)
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
done

(* using goodContext and corresponding adaptation rule to prove it in an empty context *)
theorem copy_resources_ok: "{}  \<rhd> (CallFunExp copy [xin]) : copyAss [xin]"
apply (rule vdmCallFun)
apply (simp add: copy_funTabBody copyBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
apply (rule vdmLet)
(* defer 1  *)
(* instantiate expected pre-cond for recursive call *)
(*
apply (subgoal_tac "{(CallFunExp copy [xin], copyAss [xin])}  \<rhd> CallFunExp copy [tvar] : copyAss [tvar]")
apply simp
defer 1
*)
apply (rule CtxtWeakSingleton)
apply (rule vdmAdapt)
apply (rule copy_goodContext)
apply (simp add: copy_Context_def)
apply (simp add: copy_Context_def copy_funSpecTab)
apply fastsimp
apply (rule vdmConstr)
apply (rule vdmConstr)
apply clarsimp
apply (simp add: copy_funTabArgs) 
apply (simp only: copyAss_def)
apply clarsimp
(*
apply (simp add: copy_funSpecTab copyAss_def)
apply clarsimp
*)
apply (erule_tac x="r" in allE)
apply (erule impE)
defer 1 (*  R (Ref r) = E xarg by unwrapping FrameInd *)
apply clarsimp?
apply (erule mList.cases)
apply simp_all
(* CASE: NIL *)
apply clarsimp
apply (rename_tac r h0)
apply (subgoal_tac " \<not> ( \<exists> rs0 . fmap_lookup h0 r = Some (C CONS rs0) )")
 prefer 2 apply clarsimp
apply clarsimp?
apply (erule_tac x="E" in allE)
apply (erule_tac x="BOTTOM" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPConFail)
apply clarsimp
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (subgoal_tac "h0 = h'a")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply (simp add: a11)
apply (subgoal_tac "freshloc (fmap_dom h'a) \<notin> (fmap_dom h'a)")
 prefer 2 apply (rule freshloc) apply (rule finite_dom_fmap)
apply (frule a12)
apply (rule finite_dom_fmap)
apply simp
(* clause on bounded result *)
apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
apply (rule mListNIL)
apply clarsimp
apply (erule ConstrInd.cases)
apply simp_all
(* CASE: CONS *)
apply (rename_tac l'x l't h'0 n't r'h r't)
apply (subgoal_tac "\<exists> E0'. ([hvar, tvar], [r'h, r't], E, E0'):PConInd")
apply clarsimp
apply (erule_tac x="E0'" in allE)
apply (erule_tac x="R (Ref l'x)" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPCon)
apply (rule_tac x="[r'h, Ref l't]" in exI)
apply clarsimp
(*
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
*)
apply clarsimp
apply (subgoal_tac "h'a = h'b")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply clarsimp?
apply (simp add: copy_funSpecTab copyAss_def)
apply clarsimp?
apply (rename_tac ha')
apply (erule_tac x="n't" in allE)
apply (erule impE)
apply (rule_tac x="l't" in exI)
apply clarsimp?
apply (subgoal_tac "E0' tvar = R (Ref l't)")
 (* should follow from def of PConInd *)
apply (rule conjI)
apply clarsimp?
(* defer 1 *)
apply (rule_tac x="Ua" in exI)
apply clarsimp?
defer 1  (* should follow from def of PConInd *)
apply (rule conjI)
apply (subgoal_tac "card (insert (freshloc (fmap_dom ha')) (fmap_dom ha')) = Suc (card (fmap_dom ha'))")
 prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
apply simp 
(* main inequiv solved! *)
(* clause on bounded result *)
apply clarsimp?
apply (rule_tac x="Uv \<union> {(freshloc (fmap_dom ha'))}" in exI)
apply (rotate_tac -1)
(*
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply clarsimp
*)
apply (subgoal_tac "(n't, rv, Uv,  ha'(freshloc (fmap_dom ha')\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
apply (rule mListCONS)
apply simp+
(*
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply simp
*)
defer 1 (* rs = ... *)
apply (rule mListFresh)
apply simp+
defer 1
(*  \<exists>E0' ... *)
apply (rule_tac x="(E(tvar:=R r't))(hvar:=R r'h)" in exI)
apply (rule PConInd_CONS)
apply (rule PConInd_CONS)
apply (rule PConInd_NIL)
apply simp
apply fastsimp
(*  R (Ref r) = E xarg *)
apply (erule FrameInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply fastsimp
defer 1 (* rs = ... *)
apply (rule mListSameH)
apply simp
apply (simp add: sameH_def)
apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom ha')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom ha'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom ha')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
(* rs = ... *)
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
done

text {* aux lemmas on Finmaps etc *}

lemma GETrSome_DOM:"fmap_lookup G x = Some T \<Longrightarrow> x \<in> fmap_dom G"
by (simp add: fmap_dom_def dom_def fmap_lookup_def)

lemma GETrNone_DOM:"fmap_lookup G x = None \<Longrightarrow> x \<notin> fmap_dom G"
by (simp add: fmap_dom_def dom_def fmap_lookup_def)

lemma DOM_GETrSome:"x \<in> fmap_dom G \<Longrightarrow> (\<exists> t . fmap_lookup G x = Some t)"
by (simp add: fmap_dom_def dom_def fmap_lookup_def)

lemma DOM_GETrNone:"x \<notin> fmap_dom G \<Longrightarrow> fmap_lookup G x = None"
by (simp add: fmap_dom_def dom_def fmap_lookup_def)

(*
lemma DOM_GETr_DOM: "\<lbrakk>x \<in> fmap_dom D; x \<in> U; fmap_lookup C x = fmap_lookup D x\<rbrakk> \<Longrightarrow> x \<in> fmap_dom C"
apply (subgoal_tac "\<exists> T. fmap_lookup D x = Some T", clarsimp)
apply (rule GETrSome_DOM, clarsimp, simp)
apply (erule DOM_GETrSome)
done

lemma GetNONE1: "x \<notin> fmap_dom C \<Longrightarrow> None = fmap_lookup C x"
by (simp add: fmap_dom_def  dom_def fmap_lookup_def)

lemma GetNONE2: "fmap_lookup C x = None \<Longrightarrow> x \<notin> fmap_dom C"
by (simp add:  fmap_dom_def dom_def fmap_lookup_def)

lemma DOM_Update: "{x} \<union> (fmap_dom b) = fmap_dom (b(x\<mapsto>\<^sub>fT))" by (simp)

lemma DOM_Update1: "\<lbrakk>y \<in> fmap_dom (C(x\<mapsto>\<^sub>fT)); x \<noteq> y\<rbrakk> \<Longrightarrow> y:fmap_dom C"
by (subgoal_tac "y \<noteq> x", insert DOM_Update, fast,fast)
lemma GETr_Update1: "x \<noteq> y \<Longrightarrow> fmap_lookup (G(x\<mapsto>\<^sub>fk)) y = fmap_lookup G y"
apply (rule FMAPlookup1)
apply fastsimp
done

lemma GETr_Update1a: "x \<noteq> y \<Longrightarrow> fmap_lookup G y = fmap_lookup (G(x\<mapsto>\<^sub>fk)) y"
apply (rule FMAPlookup01)
apply fastsimp
done

lemma GETr_Update2: "\<lbrakk>fmap_lookup G xa = Some T; x \<noteq> xa\<rbrakk> \<Longrightarrow> fmap_lookup (G(x\<mapsto>\<^sub>fk)) xa = Some T"
by (subgoal_tac "fmap_lookup (G(x\<mapsto>\<^sub>fk)) xa = fmap_lookup G xa", clarsimp, erule GETr_Update1)

lemma GETr_Update2a: "\<lbrakk> x \<noteq> x'; fmap_lookup (C(x\<mapsto>\<^sub>fty)) x' = Some ty' \<rbrakk> \<Longrightarrow> fmap_lookup C x' = Some ty'"
apply (subgoal_tac "fmap_lookup (C(x\<mapsto>\<^sub>fty)) x' = fmap_lookup C x'")
apply simp
apply (rule GETr_Update1)
apply simp
done
*)

lemma sameHinsert1: "\<lbrakk> sameH U h h' ; fmap_lookup h l = fmap_lookup h' l \<rbrakk>
                    \<Longrightarrow> sameH (insert l U) h h'"
apply (simp add: sameH_def)
done

text {* Another predicate for removing one object from a heap.*}
constdefs HpMinus ::"Heap \<Rightarrow> Locn \<Rightarrow> Heap \<Rightarrow> bool"
"HpMinus h l hh == (sameH (fmap_dom hh) hh h \<and> fmap_dom hh = fmap_dom h - {l} \<and> l : fmap_dom h)"

subsection {* Lemmas about the same predicate over heaps *}

lemma sameH_sym: "sameH L h' h ==> sameH L h h'"
(*<*)by (simp add: sameH_def) (*>*)

lemma SameHTriv[simp]: "sameH (Dom h) h h" (*<*)by (simp add: sameH_def)(*>*)

lemma sameHCommutative: "sameH U h hh \<Longrightarrow> sameH U hh h"
(*<*)by (simp add: sameH_def)(*>*)

lemma SameHSubset: "\<lbrakk>sameH U h hh; U' \<subseteq> U\<rbrakk> \<Longrightarrow> sameH U' h hh" (*<*)by (simp add: sameH_def, fast)(*>*)

lemma SameHTransitive: "\<lbrakk>sameH U h h1; sameH U' h1 h2; U \<subseteq> U'\<rbrakk> \<Longrightarrow> sameH U h h2"
(*<*)by (simp add: sameH_def , fast)(*>*)

(*
lemma HpMinusSameH: "\<lbrakk>HpMinus h l hh; X \<subseteq> fmap_dom h; l \<notin> X\<rbrakk> \<Longrightarrow> sameH X h hh"
apply (simp add: HpMinus_def sameH_def)
apply clarsimp
apply (erule_tac x=la in allE) apply (erule impE)
apply (rule conjI)
apply fastsimp
apply fastsimp
apply simp
done
*)

constdefs copy0_Context :: Context
"copy0_Context == {(CallFunExp copy [tvar], copy0Ass [tvar])}"

lemma copy0_goodContext: "goodContext0 the_funSpecTab copy0_Context "
apply (simp add: goodContext0_def copy_funTabBody copyBody_def)
apply clarsimp
apply (simp add: copy0_Context_def)
apply clarsimp
apply (rule conjI)
apply (simp add: copy0_funSpecTab)
apply clarsimp
apply (simp add: copy_funTabBody copyBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
apply (rule vdmLet)
(* defer 1  *)
(* instantiate expected pre-cond for recursive call *)
apply (rule vdmAx)
apply simp
(*
apply (subgoal_tac "{(CallFunExp copy [xin], copyAss [xin])} \<rhd> CallFunExp copy [tvar] : copyAss [tvar]")
apply simp
defer 1
*)
(*
apply (rule vdmAxAdapt)
apply fastsimp
*)
apply (rule vdmConstr)
apply (rule vdmConstr)
apply clarsimp
apply (simp add: copy_funTabArgs copy0_funSpecTab copy0Ass_def) 
apply clarsimp
apply clarsimp?
apply (erule_tac x="r" in allE)
apply (erule impE)
defer 1 (*  R (Ref r) = E xarg by unwrapping FrameInd *)
apply clarsimp?
apply (erule mList.cases)
apply simp_all
(* CASE: NIL *)
apply clarsimp
apply (rename_tac r h0)
apply (subgoal_tac " \<not> ( \<exists> rs0 . fmap_lookup h0 r = Some (C CONS rs0) )")
 prefer 2 apply clarsimp
apply clarsimp?
apply (erule_tac x="E" in allE)
apply (erule_tac x="BOTTOM" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPConFail)
apply clarsimp
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (subgoal_tac "h0 = h'a")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply (simp add: a11)
apply (subgoal_tac "freshloc (fmap_dom h'a) \<notin> (fmap_dom h'a)")
 prefer 2 apply (rule freshloc) apply (rule finite_dom_fmap)
apply (frule a12)
apply (rule finite_dom_fmap)
apply simp
(* clause on bounded result *)
apply (rule conjI)
apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
apply (rule conjI)
apply (rule mListNIL)
apply clarsimp
apply (erule ConstrInd.cases)
apply simp_all
apply (simp add: sameH_def)
apply (subgoal_tac "fmap_lookup (h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC NIL rs)) r = fmap_lookup h'a r ")
apply clarsimp
defer 1
defer 1
defer 1
(*
apply (rule FMAPlookup1) 
defer 1  
apply (rule mListNIL)
apply clarsimp
apply (subgoal_tac "fmap_lookup (h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC NIL rs)) r = fmap_lookup h'a r ")
apply clarsimp
apply (rule FMAPlookup1)
defer 1  
*)
(*
apply (erule ConstrInd.cases)
apply simp_all
*)
(* input unchanged *)
(*
apply (rule conjI)
apply simp
apply (rule mListNIL)
apply (subgoal_tac "fmap_lookup (h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC NIL rs)) r = fmap_lookup h'a r")
apply simp
apply (rule FMAPlookup1)
apply (case_tac "r = freshloc (fmap_dom h'a)")
 apply (drule GETrSome_DOM) apply (insert freshloc) apply simp
apply simp
apply (rotate_tac -1)
apply (erule thin_rl)
*)
(* CASE: CONS *)
apply (rename_tac l'x l't h'0 n't r'h r't)
apply (subgoal_tac "\<exists> E0'. ([hvar, tvar], [r'h, r't], E, E0'):PConInd")
apply clarsimp
apply (erule_tac x="E0'" in allE)
apply (erule_tac x="R (Ref l'x)" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPCon)
apply (rule_tac x="[r'h, Ref l't]" in exI)
apply clarsimp
(*
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
*)
apply clarsimp
apply (subgoal_tac "h'a = h'b")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply clarsimp?
apply (rename_tac ha')
apply (erule_tac x="n't" in allE)
apply (erule_tac x="l't" in allE)
apply (erule_tac x="Ua" in allE)
apply (erule impE)
apply clarsimp?
(* apply (subgoal_tac "E0' tvar = R (Ref l't)") *)
defer 1 (* should follow from def of PConInd *)
apply (rule conjI)
apply (subgoal_tac "card (insert (freshloc (fmap_dom ha')) (fmap_dom ha')) = Suc (card (fmap_dom ha'))")
 prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
apply simp 
(* main inequiv solved! *)
(* clause on bounded result *)
apply clarsimp?
apply (rule conjI)
apply (rule_tac x="Uv \<union> {(freshloc (fmap_dom ha'))}" in exI)
apply (rotate_tac -1)
(*
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply clarsimp
*)
apply (rule conjI)
apply (subgoal_tac "(n't, rv, Uv,  ha'(freshloc (fmap_dom ha')\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
apply (rule mListCONS)
apply simp+
(*
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply simp
*)
defer 1 (* rs = ... *)
apply (rule mListFresh)
apply simp+
apply (rule mListSameH)
apply simp
apply (simp add: sameH_def)
apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom ha')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (rotate_tac -4)
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom ha'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom ha')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
(* rs = ... *)
apply (rule conjI)?
(* disjointness condition *)
apply (rename_tac h n't r'h E0' rs h' rv Uv)
apply (subgoal_tac "sameH (fmap_dom h - (Uv \<union> {freshloc (fmap_dom h')})) h h'")
 prefer 2 apply (rule SameHSubset) apply simp apply fastsimp
apply (rotate_tac -1)
apply (subgoal_tac "sameH (fmap_dom h - (Uv \<union> {freshloc (fmap_dom h')})) h' (h'(freshloc (fmap_dom h')\<mapsto>\<^sub>fC CONS rs))")
 prefer 2 apply (simp add: sameH_def) apply clarsimp
 apply (case_tac "l \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
apply (rule SameHTransitive)
apply simp
apply simp
apply clarsimp?
defer 1 
apply (rename_tac h n't r'h E0' rs h' rv Uv)
apply (rule mListSameH)
apply (subgoal_tac "(Suc n't, l'x, Ua \<union> {l'x}, h') \<in> mList")
apply simp
apply (rule mListCONS)
prefer 4
apply simp
prefer 2
apply simp
apply (frule GETrSome_DOM)
apply (subgoal_tac "l'x \<notin> Uv")
 prefer 2 apply fastsimp
apply (subgoal_tac "fmap_lookup h l'x = fmap_lookup h' l'x")
apply fastsimp
apply (simp add: sameH_def)
apply fastsimp
apply (rule sameHinsert1)
apply (case_tac "freshloc (fmap_dom h') \<notin> Ua")
 apply (simp add: sameH_def) apply clarsimp
 defer 1
 (* OTW *)
 apply clarsimp
 apply (subgoal_tac " freshloc (fmap_dom h') \<notin> Uv")

 apply (subgoal_tac "fmap_lookup h (freshloc (fmap_dom h')) = fmap_lookup h (freshloc (fmap_dom h'))")
 apply (simp add: sameH_def) apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (rotate_tac -8)
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom h'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
apply simp
 apply (rotate_tac -4)
 apply (frule mListDom)
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (case_tac "freshloc (fmap_dom h') \<in> Uv")
   apply fastsimp
   apply simp
apply (case_tac "l'x = freshloc (fmap_dom h')")
 apply (subgoal_tac "fmap_lookup h l'x = fmap_lookup h' l'x")
  prefer 2
  apply (subgoal_tac "l'x : (fmap_dom h - Uv)")
   prefer 2
   apply (frule GETrSome_DOM)
   apply (subgoal_tac "l'x \<notin> Uv")
    prefer 2 apply fastsimp
  apply simp
  apply (simp add: sameH_def) apply clarsimp?
 apply (subgoal_tac "rs = [r'h, Ref l't]")
 apply fastsimp
defer 1 (* rs = ... *)
apply (case_tac "l'x \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
(*  \<exists>E0' ... *)
apply (rule_tac x="(E(tvar:=R r't))(hvar:=R r'h)" in exI)
apply (rule PConInd_CONS)
apply (rule PConInd_CONS)
apply (rule PConInd_NIL)
apply simp
apply fastsimp
(*  R (Ref r) = E xarg *)
apply (erule FrameInd.cases)
apply simp_all
apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
apply (case_tac "r \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply (frule GETrSome_DOM)
 apply simp
apply (rule mListNIL)
apply (case_tac "r \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply (frule GETrSome_DOM)
 apply simp
(* E0' tvar = R (Ref l't) *)
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply fastsimp
(* rs = ... *)
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
apply (rename_tac h n't rs h' rv Uv r Eb Eb')
apply (case_tac "freshloc (fmap_dom h') : Uv") 
 apply (subgoal_tac "fmap_lookup h (freshloc (fmap_dom h')) = fmap_lookup h' (freshloc (fmap_dom h'))")
  prefer 2 apply (simp add: sameH_def) apply (erule_tac x="(freshloc (fmap_dom h'))" in allE) apply simp
 apply (rotate_tac -4)
 apply (frule mListDom)
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (subgoal_tac "freshloc (fmap_dom h') \<in> fmap_dom h'")
  prefer 2 apply (rule rev_subsetD) 
  apply (rotate_tac 3) (* freshloc (fmap_dom h') \<in> Uv *)
  apply simp
apply simp
apply simp
apply fastsimp
(* as above *)
apply (case_tac "freshloc (fmap_dom h') : Uv") 
 apply simp
 apply simp
 apply (subgoal_tac "fmap_lookup h (freshloc (fmap_dom h')) = fmap_lookup h' (freshloc (fmap_dom h'))")
  prefer 2 apply (simp add: sameH_def) 
 apply (rotate_tac -3) (* freshloc (fmap_dom h') \<in> fmap_dom h *)
 apply (frule DOM_GETrSome)
 apply clarsimp
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (rotate_tac -1) (* freshloc (fmap_dom h') \<notin> fmap_dom h' *)
 apply (frule DOM_GETrNone)
 apply simp
apply (case_tac "l \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
defer 1
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
(* rv = l't; proven by contradiction in lookups *)
apply (subgoal_tac "finite (fmap_dom ha)")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (rotate_tac -1) (* freshloc (fmap_dom h') \<notin> fmap_dom h' *)
 apply (frule DOM_GETrNone)
 apply simp
done

(* using goodContext and corresponding adaptation rule to prove it in an empty context *)
theorem copy0_resources_ok: "{}  \<rhd> (CallFunExp copy [xin]) : copy0Ass [xin]"
apply (rule vdmCallFun)
apply (simp add: copy_funTabBody copyBody_def)
apply (simp add: copy0_funSpecTab copy0Ass_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
apply (rule vdmLet)

apply (rule CtxtWeakSingleton)
apply (rule vdmAdapt)
apply (rule copy0_goodContext)
apply (simp add: copy0_Context_def copy0_funSpecTab)
apply (simp add: copy0_Context_def copy0_funSpecTab)
apply fastsimp

(* defer 1  *)
(* instantiate expected pre-cond for recursive call *)

apply (rule vdmConstr)
apply (rule vdmConstr)
apply clarsimp
apply (simp add: copy_funTabArgs copy0_funSpecTab copy0Ass_def) 
apply clarsimp?
apply clarsimp?
apply (erule_tac x="r" in allE)
apply (erule impE)
defer 1 (*  R (Ref r) = E xarg by unwrapping FrameInd *)
apply clarsimp?
apply (erule mList.cases)
apply simp_all
(* CASE: NIL *)
apply clarsimp
apply (rename_tac r h0)
apply (subgoal_tac " \<not> ( \<exists> rs0 . fmap_lookup h0 r = Some (C CONS rs0) )")
 prefer 2 apply clarsimp
apply clarsimp?
apply (erule_tac x="E" in allE)
apply (erule_tac x="BOTTOM" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPConFail)
apply clarsimp
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (subgoal_tac "h0 = h'a")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply (simp add: a11)
apply (subgoal_tac "freshloc (fmap_dom h'a) \<notin> (fmap_dom h'a)")
 prefer 2 apply (rule freshloc) apply (rule finite_dom_fmap)
apply (frule a12)
apply (rule finite_dom_fmap)
apply simp
(* clause on bounded result *)
apply (rule conjI)
apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
apply (rule conjI)
apply (rule mListNIL)
apply clarsimp
apply (erule ConstrInd.cases)
apply simp_all
apply (simp add: sameH_def)
apply (subgoal_tac "fmap_lookup (h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC NIL rs)) r = fmap_lookup h'a r ")
apply clarsimp
defer 1
defer 1
defer 1
(*
apply (rule FMAPlookup1) 
defer 1  
apply (rule mListNIL)
apply clarsimp
apply (subgoal_tac "fmap_lookup (h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC NIL rs)) r = fmap_lookup h'a r ")
apply clarsimp
apply (rule FMAPlookup1)
defer 1  
*)
(*
apply (erule ConstrInd.cases)
apply simp_all
*)
(* input unchanged *)
(*
apply (rule conjI)
apply simp
apply (rule mListNIL)
apply (subgoal_tac "fmap_lookup (h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC NIL rs)) r = fmap_lookup h'a r")
apply simp
apply (rule FMAPlookup1)
apply (case_tac "r = freshloc (fmap_dom h'a)")
 apply (drule GETrSome_DOM) apply (insert freshloc) apply simp
apply simp
apply (rotate_tac -1)
apply (erule thin_rl)
*)
(* CASE: CONS *)
apply (rename_tac l'x l't h'0 n't r'h r't)
apply (subgoal_tac "\<exists> E0'. ([hvar, tvar], [r'h, r't], E, E0'):PConInd")
apply clarsimp
apply (erule_tac x="E0'" in allE)
apply (erule_tac x="R (Ref l'x)" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPCon)
apply (rule_tac x="[r'h, Ref l't]" in exI)
apply clarsimp
(*
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
*)
apply clarsimp
apply (subgoal_tac "h'a = h'b")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply clarsimp?
apply (rename_tac ha')
apply (erule_tac x="n't" in allE)
apply (erule_tac x="l't" in allE)
apply (erule_tac x="Ua" in allE)
apply (erule impE)
apply clarsimp?
(* apply (subgoal_tac "E0' tvar = R (Ref l't)") *)
defer 1 (* should follow from def of PConInd *)
apply (rule conjI)
apply (subgoal_tac "card (insert (freshloc (fmap_dom ha')) (fmap_dom ha')) = Suc (card (fmap_dom ha'))")
 prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
apply simp 
(* main inequiv solved! *)
(* clause on bounded result *)
apply clarsimp?
apply (rule conjI)
apply (rule_tac x="Uv \<union> {(freshloc (fmap_dom ha'))}" in exI)
apply (rotate_tac -1)
(*
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply clarsimp
*)
apply (rule conjI)
apply (subgoal_tac "(n't, rv, Uv,  ha'(freshloc (fmap_dom ha')\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
apply (rule mListCONS)
apply simp+
(*
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply simp
*)
defer 1 (* rs = ... *)
apply (rule mListFresh)
apply simp+
apply (rule mListSameH)
apply simp
apply (simp add: sameH_def)
apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom ha')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (rotate_tac -4)
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom ha'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom ha')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
(* rs = ... *)
apply (rule conjI)?
(* disjointness condition *)
apply (rename_tac h n't r'h E0' rs h' rv Uv)
apply (subgoal_tac "sameH (fmap_dom h - (Uv \<union> {freshloc (fmap_dom h')})) h h'")
 prefer 2 apply (rule SameHSubset) apply simp apply fastsimp
apply (rotate_tac -1)
apply (subgoal_tac "sameH (fmap_dom h - (Uv \<union> {freshloc (fmap_dom h')})) h' (h'(freshloc (fmap_dom h')\<mapsto>\<^sub>fC CONS rs))")
 prefer 2 apply (simp add: sameH_def) apply clarsimp
 apply (case_tac "l \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
apply (rule SameHTransitive)
apply simp
apply simp
apply clarsimp?
defer 1 
apply (rename_tac h n't r'h E0' rs h' rv Uv)
apply (rule mListSameH)
apply (subgoal_tac "(Suc n't, l'x, Ua \<union> {l'x}, h') \<in> mList")
apply simp
apply (rule mListCONS)
prefer 4
apply simp
prefer 2
apply simp
apply (frule GETrSome_DOM)
apply (subgoal_tac "l'x \<notin> Uv")
 prefer 2 apply fastsimp
apply (subgoal_tac "fmap_lookup h l'x = fmap_lookup h' l'x")
apply fastsimp
apply (simp add: sameH_def)
apply fastsimp
apply (rule sameHinsert1)
apply (case_tac "freshloc (fmap_dom h') \<notin> Ua")
 apply (simp add: sameH_def) apply clarsimp
 defer 1
 (* OTW *)
 apply clarsimp
 apply (subgoal_tac " freshloc (fmap_dom h') \<notin> Uv")

 apply (subgoal_tac "fmap_lookup h (freshloc (fmap_dom h')) = fmap_lookup h (freshloc (fmap_dom h'))")
 apply (simp add: sameH_def) apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (rotate_tac -8)
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom h'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
apply simp
 apply (rotate_tac -4)
 apply (frule mListDom)
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (case_tac "freshloc (fmap_dom h') \<in> Uv")
   apply fastsimp
   apply simp
apply (case_tac "l'x = freshloc (fmap_dom h')")
 apply (subgoal_tac "fmap_lookup h l'x = fmap_lookup h' l'x")
  prefer 2
  apply (subgoal_tac "l'x : (fmap_dom h - Uv)")
   prefer 2
   apply (frule GETrSome_DOM)
   apply (subgoal_tac "l'x \<notin> Uv")
    prefer 2 apply fastsimp
  apply simp
  apply (simp add: sameH_def) apply clarsimp?
 apply (subgoal_tac "rs = [r'h, Ref l't]")
 apply fastsimp
defer 1 (* rs = ... *)
apply (case_tac "l'x \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
(*  \<exists>E0' ... *)
apply (rule_tac x="(E(tvar:=R r't))(hvar:=R r'h)" in exI)
apply (rule PConInd_CONS)
apply (rule PConInd_CONS)
apply (rule PConInd_NIL)
apply simp
apply fastsimp
(*  R (Ref r) = E xarg *)
apply (erule FrameInd.cases)
apply simp_all
apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
apply (case_tac "r \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply (frule GETrSome_DOM)
 apply simp
apply (rule mListNIL)
apply (case_tac "r \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply (frule GETrSome_DOM)
 apply simp
(* E0' tvar = R (Ref l't) *)
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply fastsimp
(* rs = ... *)
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
apply (rename_tac h n't rs h' rv Uv r Eb Eb')
apply (case_tac "freshloc (fmap_dom h') : Uv") 
 apply (subgoal_tac "fmap_lookup h (freshloc (fmap_dom h')) = fmap_lookup h' (freshloc (fmap_dom h'))")
  prefer 2 apply (simp add: sameH_def) apply (erule_tac x="(freshloc (fmap_dom h'))" in allE) apply simp
 apply (rotate_tac -4)
 apply (frule mListDom)
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (subgoal_tac "freshloc (fmap_dom h') \<in> fmap_dom h'")
  prefer 2 apply (rule rev_subsetD) 
  apply (rotate_tac 3) (* freshloc (fmap_dom h') \<in> Uv *)
  apply simp
apply simp
apply simp
apply fastsimp
(* as above *)
apply (case_tac "freshloc (fmap_dom h') : Uv") 
 apply simp
 apply simp
 apply (subgoal_tac "fmap_lookup h (freshloc (fmap_dom h')) = fmap_lookup h' (freshloc (fmap_dom h'))")
  prefer 2 apply (simp add: sameH_def) 
 apply (rotate_tac -3) (* freshloc (fmap_dom h') \<in> fmap_dom h *)
 apply (frule DOM_GETrSome)
 apply clarsimp
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (rotate_tac -1) (* freshloc (fmap_dom h') \<notin> fmap_dom h' *)
 apply (frule DOM_GETrNone)
 apply simp
apply (case_tac "l \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
defer 1
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
(* rv = l't; proven by contradiction in lookups *)
apply (subgoal_tac "finite (fmap_dom ha)")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (rotate_tac -1) (* freshloc (fmap_dom h') \<notin> fmap_dom h' *)
 apply (frule DOM_GETrNone)
 apply simp
done

theorem copy0Body_resources_ok: "{}  \<rhd> (copyBody) : copy0Ass [xarg]"
apply (simp add: copy_funTabBody copyBody_def)
apply (simp add: copy0_funSpecTab copy0Ass_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
apply (rule vdmLet)

(* apply (rule CtxtWeakSingleton) *)
apply (rule vdmAdapt)
apply (rule copy0_goodContext)
apply (simp add: copy0_Context_def copy0_funSpecTab)
apply (simp add: copy0_Context_def copy0_funSpecTab)
apply fastsimp

(* defer 1  *)
(* instantiate expected pre-cond for recursive call *)

apply (rule vdmConstr)
apply (rule vdmConstr)
apply clarsimp
apply (simp add: copy_funTabArgs copy0_funSpecTab copy0Ass_def) 
apply clarsimp?
apply clarsimp?
(*
apply (erule_tac x="r" in allE)
apply (erule impE)
defer 1 
*)
(* keep??? *)
apply clarsimp?
apply (erule mList.cases)
apply simp_all

(* CASE: NIL *)
apply clarsimp
apply (rename_tac r h0)
apply (subgoal_tac " \<not> ( \<exists> rs0 . fmap_lookup h0 r = Some (C CONS rs0) )")
 prefer 2 apply clarsimp
apply clarsimp?
apply (erule_tac x="E" in allE)
apply (erule_tac x="BOTTOM" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPConFail)
apply clarsimp
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (subgoal_tac "h0 = h'a")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply (simp add: a11)
apply (subgoal_tac "freshloc (fmap_dom h'a) \<notin> (fmap_dom h'a)")
 prefer 2 apply (rule freshloc) apply (rule finite_dom_fmap)
apply (frule a12)
apply (rule finite_dom_fmap)
apply simp
(* clause on bounded result *)
apply (rule conjI)
apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
apply (rule conjI)
apply (rule mListNIL)
apply clarsimp
apply (erule ConstrInd.cases)
apply simp_all
apply (simp add: sameH_def)
apply (subgoal_tac "fmap_lookup (h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC NIL rs)) r = fmap_lookup h'a r ")
apply clarsimp
defer 1
defer 1
defer 1
(*
apply (rule FMAPlookup1) 
defer 1  
apply (rule mListNIL)
apply clarsimp
apply (subgoal_tac "fmap_lookup (h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC NIL rs)) r = fmap_lookup h'a r ")
apply clarsimp
apply (rule FMAPlookup1)
defer 1  
*)
(*
apply (erule ConstrInd.cases)
apply simp_all
*)
(* input unchanged *)
(*
apply (rule conjI)
apply simp
apply (rule mListNIL)
apply (subgoal_tac "fmap_lookup (h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC NIL rs)) r = fmap_lookup h'a r")
apply simp
apply (rule FMAPlookup1)
apply (case_tac "r = freshloc (fmap_dom h'a)")
 apply (drule GETrSome_DOM) apply (insert freshloc) apply simp
apply simp
apply (rotate_tac -1)
apply (erule thin_rl)
*)
(* CASE: CONS *)
apply (rename_tac l'x l't h'0 n't r'h r't)
apply (subgoal_tac "\<exists> E0'. ([hvar, tvar], [r'h, r't], E, E0'):PConInd")
apply clarsimp
apply (erule_tac x="E0'" in allE)
apply (erule_tac x="R (Ref l'x)" in allE)
apply clarsimp
apply (erule impE)
apply (rule semPCon)
apply (rule_tac x="[r'h, Ref l't]" in exI)
apply clarsimp
(*
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
apply (erule exE)+
apply (erule conjE)+
*)
apply clarsimp
apply (subgoal_tac "h'a = h'b")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply clarsimp?
apply (rename_tac ha')
apply (erule_tac x="n't" in allE)
apply (erule_tac x="l't" in allE)
apply (erule_tac x="Ua" in allE)
apply (erule impE)
apply clarsimp?
(* apply (subgoal_tac "E0' tvar = R (Ref l't)") *)
defer 1 (* should follow from def of PConInd *)
apply (rule conjI)
apply (subgoal_tac "card (insert (freshloc (fmap_dom ha')) (fmap_dom ha')) = Suc (card (fmap_dom ha'))")
 prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
apply simp 
(* main inequiv solved! *)
(* clause on bounded result *)
apply clarsimp?
apply (rule conjI)
apply (rule_tac x="Uv \<union> {(freshloc (fmap_dom ha'))}" in exI)
apply (rotate_tac -1)
(*
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply clarsimp
*)
apply (rule conjI)
apply (subgoal_tac "(n't, rv, Uv,  ha'(freshloc (fmap_dom ha')\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
apply (rule mListCONS)
apply simp+
(*
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply simp
*)
defer 1 (* rs = ... *)
apply (rule mListFresh)
apply simp+
apply (rule mListSameH)
apply simp
apply (simp add: sameH_def)
apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom ha')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (rotate_tac -4)
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom ha'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom ha')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
(* rs = ... *)
apply (rule conjI)?
(* disjointness condition *)
apply (rename_tac h n't r'h E0' rs h' rv Uv)
apply (subgoal_tac "sameH (fmap_dom h - (Uv \<union> {freshloc (fmap_dom h')})) h h'")
 prefer 2 apply (rule SameHSubset) apply simp apply fastsimp
apply (rotate_tac -1)
apply (subgoal_tac "sameH (fmap_dom h - (Uv \<union> {freshloc (fmap_dom h')})) h' (h'(freshloc (fmap_dom h')\<mapsto>\<^sub>fC CONS rs))")
 prefer 2 apply (simp add: sameH_def) apply clarsimp
 apply (case_tac "l \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
apply (rule SameHTransitive)
apply simp
apply simp
apply clarsimp?
defer 1 
apply (rename_tac h n't r'h E0' rs h' rv Uv)
apply (rule mListSameH)
apply (subgoal_tac "(Suc n't, l'x, Ua \<union> {l'x}, h') \<in> mList")
apply simp
apply (rule mListCONS)
prefer 4
apply simp
prefer 2
apply simp
apply (frule GETrSome_DOM)
apply (subgoal_tac "l'x \<notin> Uv")
 prefer 2 apply fastsimp
apply (subgoal_tac "fmap_lookup h l'x = fmap_lookup h' l'x")
apply fastsimp
apply (simp add: sameH_def)
apply fastsimp
apply (rule sameHinsert1)
apply (case_tac "freshloc (fmap_dom h') \<notin> Ua")
 apply (simp add: sameH_def) apply clarsimp
 defer 1
 (* OTW *)
 apply clarsimp
 apply (subgoal_tac " freshloc (fmap_dom h') \<notin> Uv")

 apply (subgoal_tac "fmap_lookup h (freshloc (fmap_dom h')) = fmap_lookup h (freshloc (fmap_dom h'))")
 apply (simp add: sameH_def) apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (rotate_tac -8)
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom h'")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
apply simp
 apply (rotate_tac -4)
 apply (frule mListDom)
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (case_tac "freshloc (fmap_dom h') \<in> Uv")
   apply fastsimp
   apply simp
apply (case_tac "l'x = freshloc (fmap_dom h')")
 apply (subgoal_tac "fmap_lookup h l'x = fmap_lookup h' l'x")
  prefer 2
  apply (subgoal_tac "l'x : (fmap_dom h - Uv)")
   prefer 2
   apply (frule GETrSome_DOM)
   apply (subgoal_tac "l'x \<notin> Uv")
    prefer 2 apply fastsimp
  apply simp
  apply (simp add: sameH_def) apply clarsimp?
 apply (subgoal_tac "rs = [r'h, Ref l't]")
 apply fastsimp
defer 1 (* rs = ... *)
apply (case_tac "l'x \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
(*  \<exists>E0' ... *)
apply (rule_tac x="(E(tvar:=R r't))(hvar:=R r'h)" in exI)
apply (rule PConInd_CONS)
apply (rule PConInd_CONS)
apply (rule PConInd_NIL)
apply simp
apply fastsimp
(*  R (Ref r) = E xarg *)
(*
apply (erule FrameInd.cases)
apply simp_all
*)
apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
apply (case_tac "r \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply (frule GETrSome_DOM)
 apply simp
apply (rule mListNIL)
apply (case_tac "r \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply (frule GETrSome_DOM)
 apply simp
(* E0' tvar = R (Ref l't) *)
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply fastsimp
(* rs = ... *)
apply (subgoal_tac "rs = [r'h, Ref rv]")
apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
apply (rename_tac h n't rs h' rv Uv r Eb Eb')
apply (case_tac "freshloc (fmap_dom h') : Uv") 
 apply (subgoal_tac "fmap_lookup h (freshloc (fmap_dom h')) = fmap_lookup h' (freshloc (fmap_dom h'))")
  prefer 2 apply (simp add: sameH_def) apply (erule_tac x="(freshloc (fmap_dom h'))" in allE) apply simp
 apply (rotate_tac -4)
 apply (frule mListDom)
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (subgoal_tac "freshloc (fmap_dom h') \<in> fmap_dom h'")
  prefer 2 apply (rule rev_subsetD) 
  apply (rotate_tac 3) (* freshloc (fmap_dom h') \<in> Uv *)
  apply simp
apply simp
apply simp
apply fastsimp
(* as above *)
apply (case_tac "freshloc (fmap_dom h') : Uv") 
 apply simp
 apply simp
 apply (subgoal_tac "fmap_lookup h (freshloc (fmap_dom h')) = fmap_lookup h' (freshloc (fmap_dom h'))")
  prefer 2 apply (simp add: sameH_def) 
 apply (rotate_tac -3) (* freshloc (fmap_dom h') \<in> fmap_dom h *)
 apply (frule DOM_GETrSome)
 apply clarsimp
 apply (subgoal_tac "finite (fmap_dom h')")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (rotate_tac -1) (* freshloc (fmap_dom h') \<notin> fmap_dom h' *)
 apply (frule DOM_GETrNone)
 apply simp
apply (case_tac "l \<noteq> freshloc (fmap_dom h')")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (erule ConstrInd.cases) 
apply simp_all
apply clarsimp
apply (subgoal_tac "E0' hvar = R r'h")
apply simp
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
defer 1
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
(* rv = l't; proven by contradiction in lookups *)
apply (subgoal_tac "finite (fmap_dom ha)")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply (rotate_tac -1) (* freshloc (fmap_dom h') \<notin> fmap_dom h' *)
 apply (frule DOM_GETrNone)
 apply simp
done

(* check that the important unproven lemma can be proven with copy0Ass 
   see T.thy for the full proofs 

constdefs list_in :: "Heap \<Rightarrow> HoVal \<Rightarrow> nat \<Rightarrow> bool"
"list_in h w n \<equiv> (\<exists> r U . w = (R (Ref r)) \<and> (n,r,U,h) \<in> mList)"

lemma proven1_as_in_toProve1: "\<lbrakk>list_in h v n; (a,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' v n"
apply (insert copy0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
apply (erule_tac x="n" in allE)
apply (erule_tac x="U" in allE)
apply (erule impE)
apply simp
apply clarsimp
apply (rule_tac x="U" in exI)
apply clarsimp
done

lemma proven2_as_in_toProve2: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' nv n"
apply (insert copy0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
apply (erule_tac x="n" in allE)
apply (erule_tac x="U" in allE)
apply (erule impE)
apply simp
apply clarsimp
apply (rule_tac x="Uv" in exI)
apply clarsimp
done

lemma proven3_as_in_toProve3: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> nv \<noteq> BOTTOM"
apply (insert copy0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
done
*)



(* old proof of deferred rec call 
apply (subgoal_tac "{} \<union> {(CallFunExp copy [xin], copyAss [xin])} \<rhd> CallFunExp copy [tvar] : copyAss [tvar]")
apply simp
apply (rule ctxtWeak)
apply (rule vdmConseq)
prefer 2
apply (rule vdmAdapt)
apply (subgoal_tac "goodContext0 the_funSpecTab copy_Context")
apply simp
apply (rule copy_goodContext)
apply (simp add: copy_Context_def)
apply (simp add: copy_Context_def copy_funSpecTab copy_funTabArgs)
apply fastsimp
apply clarsimp
apply (simp add: copy_Context_def copy_funSpecTab copy_funTabArgs)
done
*)

(*  -----------------------------------------------------------------------------  *)
(* Example: fuse1 (same as fuse but ignore differing elements *)

syntax 
 h0var :: "VarName"
 t0var :: "VarName"
 h1var :: "VarName"
 t1var :: "VarName"
 y0var :: "VarName"
 x0arg  :: "VarName"
 x1arg  :: "VarName"
 x0in  :: "VarName"
 x1in  :: "VarName"
 b0var :: "VarName"
 fuse  :: "FunName"

translations
"h0var" == "(VN ''h0var'')"
"t0var" == "(VN ''t0var'')"
"h1var" == "(VN ''h1var'')"
"t1var" == "(VN ''t1var'')"
"y0var" == "(VN ''y0var'')"
"x0arg"  == "(VN ''x0arg'')"
"x1arg"  == "(VN ''x1arg'')"
"x0in"  == "(VN ''x0in'')"
"x1in"  == "(VN ''x1in'')"
"b0var"  == "(VN ''b0var'')"
"fuse" == "(FN ''fuse'')"

constdefs ieq :: "HoVal => HoVal => HoVal"
"ieq x y == B ((theInt x)=(theInt y))"

lemma ieq_tertium_non_datur: "ieq x y = B True \<or> ieq x y = B False"
by (simp add: ieq_def)

axioms list_only_heap: "\<forall> r h. r : fmap_dom h \<longrightarrow> fmap_lookup h r = Some (C NIL []) \<or> (\<exists> rh rt. fmap_lookup h r = Some (C CONS [Ref rh, Ref rt]))"

(* only difference to fuse:
    (VarExp y0var)
   in else branch
*)

constdefs fuseBody :: Exp
"fuseBody \<equiv> (CaseOneExp x0arg 
                    (PCon CONS [h0var,t0var]) (CaseOneExp x1arg 
                                                 (PCon CONS [h1var,t1var]) (LetExp y0var (CallFunExp fuse [t0var,t1var]) 
                                                                            (LetExp b0var (RPrimBin ieq h0var h1var)
                                                                              (IfExp b0var 
                                                                                    (ConstrExp CONS [h0var,y0var])
                                                                                    (VarExp y0var)))) 
                                                 (ConstrExp NIL  [] ))
                    (ConstrExp NIL []))" 

constdefs fuseAss :: "[VarName list] => Assertion"
"fuseAss zs == \<lambda>E h h' v p.
          			   \<forall>n0 r0 U0 n1.
          			      E (hd zs) = R (Ref r0) \<and> (n0, r0, U0, h) \<in> mList \<and> (\<exists>r1. E (hd (tl zs)) = R (Ref r1) \<and> (\<exists>U1. (n1, r1, U1, h) \<in> mList) \<and> n0 = n1) \<longrightarrow>
          			      card (fmap_dom h') \<le> card (fmap_dom h) + n0 + 1 \<and> (\<exists> nv rv Uv. v = R (Ref rv) \<and> (nv, rv, Uv, h') \<in> mList \<and> nv \<le> n0 ) "

axioms fuse_funTabArgs: "fst (funTab fuse) == [x0arg,x1arg]"
axioms fuse_funTabBody: "snd (funTab fuse) == fuseBody"
axioms fuse_funSpecTab: "the_funSpecTab fuse xs == fuseAss xs"

constdefs fuse_Context :: Context
"fuse_Context == {(CallFunExp fuse [t0var, t1var], fuseAss [t0var,t1var])}"

lemma fuse_goodContext: "goodContext0 the_funSpecTab fuse_Context"
apply (simp add:  fuse_Context_def fuse_funSpecTab fuse_funTabBody fuseBody_def goodContext0_def )
apply clarsimp
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
apply (rule vdmAx)
apply fastsimp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmVar)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (simp only: fuseAss_def fuse_funTabArgs)
apply clarsimp
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)
apply (case_tac "hd ys \<noteq> x0arg \<and> hd ys \<noteq> x1arg")
 prefer 2
 apply clarsimp
 apply (erule disjE)
 apply clarsimp
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
 (* subgoal proven *)
apply (subgoal_tac " E x0arg = E' (hd ys)")
 prefer 2 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* CLAUSE: bounded output *)
 apply (rule_tac x="0" in exI)
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
 (* CLAUSE: bounded output *)
 apply (rule_tac x="0" in exI)
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
  (*
  apply (case_tac "vb' = \<bottom>")
    apply clarsimp
    defer 1
  *)
    (* OTW *)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule_tac x="r't0" in exI)
    apply (rule conjI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule conjI)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4) (* (n0, r1, U1, h) \<in> mList *)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ieq vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (* CLAUSE: bounded output *)
     apply (rule_tac x="Suc nv" in exI)
     apply (rule conjI)
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
     apply (subgoal_tac "(nv, rv, Uv, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
     apply (rotate_tac -1)
     apply (rule mListCONS)
     prefer 4
     apply simp
     apply clarsimp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
     apply simp
     apply (rule mListSameH)
     apply simp
     apply (simp add: sameH_def)
     apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -5)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
     defer 1(*  Suc nv \<le> n0 \<dots>. *)
     (* ELSE branch: *)
     apply (subgoal_tac "ieq vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp
     (*
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp
     *)
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
    (* CLAUSE: bounded output *)
     apply (rule_tac x="nv" in exI)
     apply (rule conjI)
     apply (rule_tac x="Uv" in exI)
     apply simp
     apply arith
(* 8 mostly trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
   rs = ?...   
   freshloc (fmap_dom h'a) \<notin> Uv
   Suc nv \<le> n0
*)
 apply (insert ieq_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E0' t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* left-over subgoal from CLAUSE: bounded output *)
 apply (subgoal_tac "rs = [Ref r'h0, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 defer 1
 apply (rotate_tac 9)
 apply (frule mListDom)
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
   apply (rotate_tac -2)
   apply (drule subsetD)
   apply simp
   apply (subgoal_tac "finite (fmap_dom h'a)")
    prefer 2 apply (rule finite_dom_fmap)
   apply (frule freshloc)
   apply simp
   (* OTW *)
   apply assumption
 (* Suc nv \<le> n0 *)
 apply (subgoal_tac "(n0 - Suc 0) + Suc 0 \<le> n0")
  apply simp
  apply simp
 (* rs ... *)
 apply (subgoal_tac "E1' h0var = E0' h0var \<and> E0' h0var = R (Ref r'h0)")
 apply clarsimp
 apply (rule conjI)
 prefer 2
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp?
 apply (rotate_tac 9)
 apply (drule env_env_notin)
 prefer 2
 apply (subgoal_tac "E0' h0var = E1' h0var")
 apply simp
 apply simp
 apply simp
done

(* prove the assertion in an empty context, using adaptation rule *)
theorem fuse_resources_ok: "{}  \<rhd> (CallFunExp fuse [x0in,x1in]) : fuseAss [x0in, x1in]"
apply (rule vdmCallFun)
apply (simp add: fuse_funTabBody fuseBody_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
(* clumsy way of proving rec call 
apply (subgoal_tac "{(CallFunExp fuse [x0in, x1in], fuseAss [x0in, x1in])} \<rhd> CallFunExp fuse [t0var, t1var] : fuseAss [t0var, t1var]")
apply assumption
defer 1 
*)
apply (rule CtxtWeakSingleton)
apply (rule vdmAdapt)
apply (rule fuse_goodContext)
apply (simp add: fuse_Context_def)
apply (simp add: fuse_Context_def fuse_funSpecTab)
apply fastsimp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmVar)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (simp only: fuseAss_def fuse_funTabArgs)
apply clarsimp
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)
(*
apply (case_tac "hd ys \<noteq> x0arg \<and> hd ys \<noteq> x1arg")
 prefer 2
 apply clarsimp
 apply (erule disjE)
*)
 apply clarsimp?
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* CLAUSE: bounded output *)
 apply (rule_tac x="0" in exI)
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
 (* CLAUSE: bounded output *)
 apply (rule_tac x="0" in exI)
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
    (* OTW *)
    apply clarsimp?
    apply (simp add: fuse_Context_def fuse_funSpecTab fuseAss_def)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule_tac x="r't0" in exI)
    apply (rule conjI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule conjI)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4) (* (n0, r1, U1, h) \<in> mList *)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ieq vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (* CLAUSE: bounded output *)
     apply (rule_tac x="Suc nv" in exI)
     apply (rule conjI)
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
     apply (subgoal_tac "(nv, rv, Uv, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
     apply (rotate_tac -1)
     apply (rule mListCONS)
     prefer 4
     apply simp
     apply clarsimp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
     apply simp
     apply (rule mListSameH)
     apply simp
     apply (simp add: sameH_def)
     apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -5)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
     defer 1(*  Suc nv \<le> n0 \<dots>. *)
     (* ELSE branch: *)
     apply (subgoal_tac "ieq vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp
     (*
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp
     *)
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
    (* CLAUSE: bounded output *)
     apply (rule_tac x="nv" in exI)
     apply (rule conjI)
     apply (rule_tac x="Uv" in exI)
     apply simp
     apply arith
(* 8 mostly trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
   rs = ?...   
   freshloc (fmap_dom h'a) \<notin> Uv
   Suc nv \<le> n0
*)
 apply (insert ieq_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E0' t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* left-over subgoal from CLAUSE: bounded output *)
 apply (subgoal_tac "rs = [Ref r'h0, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 defer 1
 apply (rotate_tac 9)
 apply (frule mListDom)
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
   apply (rotate_tac -2)
   apply (drule subsetD)
   apply simp
   apply (subgoal_tac "finite (fmap_dom h'a)")
    prefer 2 apply (rule finite_dom_fmap)
   apply (frule freshloc)
   apply simp
   (* OTW *)
   apply assumption
 (* Suc nv \<le> n0 *)
 apply (subgoal_tac "(n0 - Suc 0) + Suc 0 \<le> n0")
  apply simp
  apply simp
 defer 1 (* rec call *)
 (* rs ... *)
 apply (subgoal_tac "E1' h0var = E0' h0var \<and> E0' h0var = R (Ref r'h0)")
 apply clarsimp
 apply (rule conjI)
 prefer 2
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp?
 apply (rotate_tac 9)
 apply (drule env_env_notin)
 prefer 2
 apply (subgoal_tac "E0' h0var = E1' h0var")
 apply simp
 apply simp
 apply simp
done

(* slight variant: generates list of exactly the same size --------------------------------- *)
(* only difference to fuse:
       (ConstrExp CONS [h0var,y0var])
   in the else branch, to make sure that the list size
   of the fused list is the same as the one in the input lists
*)

syntax
 fuse0  :: "FunName"
translations
"fuse0" == "(FN ''fuse0'')"

constdefs ilt :: "HoVal => HoVal => HoVal"
"ilt x y == B ((theInt x)<(theInt y))"

lemma ilt_tertium_non_datur: "ilt x y = B True \<or> ilt x y = B False"
by (simp add: ilt_def)


constdefs fuse0Body :: Exp
"fuse0Body \<equiv> (CaseOneExp x0arg 
                    (PCon CONS [h0var,t0var]) (CaseOneExp x1arg 
                                                 (PCon CONS [h1var,t1var]) (LetExp y0var (CallFunExp fuse0 [t0var,t1var]) 
                                                                            (LetExp b0var (RPrimBin ilt h0var h1var)
                                                                              (IfExp b0var 
                                                                                    (ConstrExp CONS [h0var,y0var])
                                                                                    (ConstrExp CONS [h1var,y0var]))))
                                                 (ConstrExp NIL  [] ))
                    (ConstrExp NIL []))" 

constdefs fuse0Ass :: "[VarName list] => Assertion"
"fuse0Ass zs == \<lambda>E h h' v p.
          		\<forall>n0 n1.
          		   ((\<exists> r0 U0. E (hd zs) = R (Ref r0) \<and> (n0, r0, U0, h) \<in> mList) \<and> (\<exists> r1 U1. E (hd (tl zs)) = R (Ref r1) \<and> (n1, r1, U1, h) \<in> mList) \<and> n0 = n1) \<longrightarrow>
          		    card (fmap_dom h') \<le> card (fmap_dom h) + n0 + 1 \<and> (\<exists> nv rv Uv. v = R (Ref rv) \<and> (nv, rv, Uv, h') \<in> mList \<and> nv = n0 ) "

axioms fuse0_funTabArgs: "fst (funTab fuse0) == [x0arg,x1arg]"
axioms fuse0_funTabBody: "snd (funTab fuse0) == fuse0Body"
axioms fuse0_funSpecTab: "the_funSpecTab fuse0 xs == fuse0Ass xs"

constdefs fuse0_Context :: Context
"fuse0_Context == {(CallFunExp fuse0 [t0var, t1var], fuse0Ass [t0var,t1var])}"

lemma fuse0_goodContext: "goodContext0 the_funSpecTab fuse0_Context"
apply (simp add:  fuse0_Context_def fuse0_funSpecTab fuse0_funTabBody fuse0Body_def goodContext0_def )
apply clarsimp
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
apply (rule vdmAx)
apply fastsimp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmConstr)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (simp only: fuse0Ass_def fuse0_funTabArgs)
apply clarsimp
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)
apply (case_tac "hd ys \<noteq> x0arg \<and> hd ys \<noteq> x1arg")
 prefer 2
 apply clarsimp
 apply (erule disjE)
 apply clarsimp
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
 (* subgoal proven *)
apply (subgoal_tac " E x0arg = E' (hd ys)")
 prefer 2 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* CLAUSE: bounded output *)
 (*
 apply (rule_tac x="0" in exI)
 apply clarsimp
 *)
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (subgoal_tac "n0=0")
 apply clarsimp
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* n0=0: by r0 -> NIL and mListSameRef *)
 apply (frule mListNIL)
 apply (rotate_tac -1)
 apply (frule mListSameRef)
 apply (rotate_tac 1)
 apply simp
 apply clarsimp
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
 (* CLAUSE: bounded output *)
 (* apply (rule_tac x="0" in exI) *)
 apply (subgoal_tac "n0=0")
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* n0=0: by r0 -> NIL and mListSameRef *)
 apply (frule mListNIL)
 apply (rotate_tac -1)
 apply (frule mListSameRef)
 apply (rotate_tac 1)
 apply simp
 apply clarsimp
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
  (*
  apply (case_tac "vb' = \<bottom>") 
    apply clarsimp
    defer 1
  *)
    (* OTW *)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule conjI)
    apply (rule_tac x="r't0" in exI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    apply (rule conjI)
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4) (* (n0, r1, U1, h) \<in> mList *)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ilt vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (* CLAUSE: bounded output *)
     (* apply (rule_tac x="Suc nv" in exI) *)
     (* apply (rule conjI) *)
     apply (subgoal_tac "\<exists> n'. Suc n' = n0")
     apply clarsimp
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
     apply (subgoal_tac "(n', rv, Uv, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
     apply (rotate_tac -1)
     apply (rule mListCONS)
     prefer 4
     apply simp
     apply clarsimp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
     apply simp
     apply (rule mListSameH)
     apply simp
     apply (simp add: sameH_def)
     apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -5)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
     defer 1(*  Suc nv \<le> n0 \<dots>. *)
     (* ELSE branch: *)
     apply (subgoal_tac "ilt vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp

     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp

     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
    (* CLAUSE: bounded output *)
     (*
     apply (rule conjI)
     apply (rule_tac x="nv" in exI)
     apply simp?
     apply arith
     *)
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
(* 8 mostly trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
   rs = ?...   
   freshloc (fmap_dom h'a) \<notin> Uv
   Suc nv \<le> n0
*)
 apply (subgoal_tac "(Suc (n0 - Suc 0), freshloc (fmap_dom h'a), Uv  \<union> {freshloc (fmap_dom h'a)}, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
 apply clarsimp
 apply (rule mListCONS)
 prefer 4
 apply (rule mListSameH)
 apply simp
 apply (simp add: sameH_def)
 apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -8)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
 apply simp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
 apply simp
 apply (insert ilt_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E0' t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* left-over subgoal from CLAUSE: bounded output *)
 apply (subgoal_tac "rs = [Ref r'h0, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 defer 1
 apply (rotate_tac 9)
 apply (frule mListDom)
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
   apply (rotate_tac -2)
   apply (drule subsetD)
   apply simp
   apply (subgoal_tac "finite (fmap_dom h'a)")
    prefer 2 apply (rule finite_dom_fmap)
   apply (frule freshloc)
   apply simp
   (* OTW *)
   apply assumption
 (* Suc nv \<le> n0 *)
 apply (simp add: gr0_conv_Suc)
 apply clarsimp
 (*
 apply (subgoal_tac "(n0 - Suc 0) + Suc 0 \<le> n0")
  apply simp
  apply simp
 *)
 (* rs ... *)
 apply (subgoal_tac "rs = [Ref r'h1, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply (subgoal_tac "E1' h1var = R (Ref r'h1)")
 apply clarsimp
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
      apply (rotate_tac -9)
      apply (frule mListDom)
      apply (subgoal_tac " freshloc (fmap_dom h'a) \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
   apply simp
 (*
 apply (subgoal_tac "E1' h0var = R (Ref r'h1)")
 apply clarsimp
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 *)
 apply (subgoal_tac "E1' h0var = E0' h0var \<and> E0' h0var = R (Ref r'h0)")
 apply clarsimp
 apply (rule conjI)
 apply (rotate_tac 9)
 apply (drule env_env_notin)
 prefer 2
 apply (subgoal_tac " E0' h0var =  E1' h0var")
 apply simp
 apply fastsimp
 apply simp
 apply (erule PConInd.cases)
 apply simp_all
done

(* prove the assertion in an empty context, using adaptation rule *)
theorem fuse0_resources_ok: "{}  \<rhd> (CallFunExp fuse0 [x0in,x1in]) : fuse0Ass [x0in, x1in]"
apply (rule vdmCallFun)
apply (simp add: fuse0_funTabBody fuse0Body_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
(* rec call *)
(* clumsy way of proving rec call 
apply (subgoal_tac "{(CallFunExp fuse0 [x0in, x1in], fuse0Ass [x0in, x1in])} \<rhd> CallFunExp fuse0 [t0var, t1var] : fuse0Ass [t0var, t1var]")
apply assumption
defer 1 
*)
apply (rule CtxtWeakSingleton)
apply (rule vdmAdapt)
apply (rule fuse0_goodContext)
apply (simp add: fuse0_Context_def)
apply (simp add: fuse0_Context_def fuse0_funSpecTab)
apply fastsimp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmConstr)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (simp only: fuse0Ass_def fuse0_funTabArgs)
apply clarsimp
apply (erule_tac x=r0 in allE)
apply (erule impE)
(* need to unfold parameter passing *)
apply (case_tac "x0in \<noteq> x0arg \<and> x1in \<noteq> x1arg")
 prefer 2
 apply clarsimp
 (*
 apply (erule disjE)
 apply clarsimp
 *)
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
(*
 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
apply (subgoal_tac " E x0arg = E' (hd ys)")
 prefer 2 apply (erule FrameInd.cases) apply clarsimp  apply (erule FrameInd.cases) apply clarsimp apply clarsimp
apply clarsimp
*)
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* CLAUSE: bounded output *)
 (*
 apply (rule_tac x="0" in exI)
 apply clarsimp
 *)
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (subgoal_tac "n0=0")
 apply clarsimp
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* n0=0: by r0 -> NIL and mListSameRef *)
 apply (frule mListNIL)
 apply (rotate_tac -1)
 apply (frule mListSameRef)
 apply (rotate_tac 1)
 apply simp
 apply clarsimp
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
 (* CLAUSE: bounded output *)
 (* apply (rule_tac x="0" in exI) *)
 apply (subgoal_tac "n0=0")
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* n0=0: by r0 -> NIL and mListSameRef *)
 apply (frule mListNIL)
 apply (rotate_tac -1)
 apply (frule mListSameRef)
 apply (rotate_tac 1)
 apply simp
 apply clarsimp
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
  (*
  apply (case_tac "vb' = \<bottom>")
    apply clarsimp
    defer 1
  *)
    (* OTW *)
    apply (simp add: fuse0_funSpecTab fuse0Ass_def)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule conjI)
    apply (rule_tac x="r't0" in exI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    apply (rule conjI)
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 4) (* (n0, r1, U1, h) \<in> mList *)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ilt vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (* CLAUSE: bounded output *)
     (* apply (rule_tac x="Suc nv" in exI) *)
     (* apply (rule conjI) *)
     apply (subgoal_tac "\<exists> n'. Suc n' = n0")
     apply clarsimp
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
     apply (subgoal_tac "(n', rv, Uv, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
     apply (rotate_tac -1)
     apply (rule mListCONS)
     prefer 4
     apply simp
     apply clarsimp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
     apply simp
     apply (rule mListSameH)
     apply simp
     apply (simp add: sameH_def)
     apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -5)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
     defer 1(*  Suc nv \<le> n0 \<dots>. *)
     (* ELSE branch: *)
     apply (subgoal_tac "ilt vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp

     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp

     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
    (* CLAUSE: bounded output *)
     (*
     apply (rule conjI)
     apply (rule_tac x="nv" in exI)
     apply simp?
     apply arith
     *)
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
(* 8 mostly trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
   rs = ?...   
   freshloc (fmap_dom h'a) \<notin> Uv
   Suc nv \<le> n0
*)
 apply (subgoal_tac "(Suc (n0 - Suc 0), freshloc (fmap_dom h'a), Uv  \<union> {freshloc (fmap_dom h'a)}, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
 apply clarsimp
 apply (rule mListCONS)
 prefer 4
 apply (rule mListSameH)
 apply simp
 apply (simp add: sameH_def)
 apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -8)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
 apply simp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
 apply simp
 apply (insert ilt_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E0' t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 6)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* left-over subgoal from CLAUSE: bounded output *)
 apply (subgoal_tac "rs = [Ref r'h0, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 defer 1
 apply (rotate_tac 9)
 apply (frule mListDom)
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
   apply (rotate_tac -2)
   apply (drule subsetD)
   apply simp
   apply (subgoal_tac "finite (fmap_dom h'a)")
    prefer 2 apply (rule finite_dom_fmap)
   apply (frule freshloc)
   apply simp
   (* OTW *)
   apply assumption
 (* Suc nv \<le> n0 *)
 apply (simp add: gr0_conv_Suc)
 apply clarsimp
 (*
 apply (subgoal_tac "(n0 - Suc 0) + Suc 0 \<le> n0")
  apply simp
  apply simp
 *)
 (* rs ... *)
 apply (subgoal_tac "rs = [Ref r'h1, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply (subgoal_tac "E1' h1var = R (Ref r'h1)")
 apply clarsimp
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
      apply (rotate_tac -9)
      apply (frule mListDom)
      apply (subgoal_tac " freshloc (fmap_dom h'a) \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
   apply simp
 (*
 apply (subgoal_tac "E1' h0var = R (Ref r'h1)")
 apply clarsimp
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 *)
 apply (subgoal_tac "E1' h0var = E0' h0var \<and> E0' h0var = R (Ref r'h0)")
 apply clarsimp
 apply (rule conjI)
 apply (rotate_tac 9)
 apply (drule env_env_notin)
 prefer 2
 apply (subgoal_tac " E0' h0var =  E1' h0var")
 apply simp
 apply fastsimp
 apply simp
 apply (erule PConInd.cases)
 apply simp_all
done


theorem fuse0Body_resources_ok: "{}  \<rhd> fuse0Body : fuse0Ass [x0arg,x1arg]"
apply (simp add: fuse0_funTabBody fuse0Body_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmCaseOne)
prefer 2
apply (rule vdmConstr)
apply (rule vdmLet)
(* rec call *)
(* clumsy way of proving rec call 
apply (subgoal_tac "{(CallFunExp fuse0 [x0in, x1in], fuse0Ass [x0in, x1in])} \<rhd> CallFunExp fuse0 [t0var, t1var] : fuse0Ass [t0var, t1var]")
apply assumption
defer 1
*)
apply (insert fuse0_resources_ok)
apply (rule vdmAdapt)
apply (rule fuse0_goodContext)
apply (simp add: fuse0_Context_def)
apply (simp add: fuse0_Context_def fuse0_funSpecTab)
apply fastsimp
apply (rule vdmLet)
apply (rule vdmRPrimBin)
apply (rule vdmIf)
apply (rule vdmConstr)
apply (rule vdmConstr)
(* code has been completely unfolded; 1 huge subgoal left *)
apply (erule thin_rl)
apply (simp only: fuse0Ass_def fuse0_funTabArgs)
apply clarsimp
(* outer case on x0arg *)
(* CASE NIL *)
apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
 apply (erule_tac x=E in allE)
 apply (erule_tac x="\<bottom>" in allE)
 apply (erule impE)
 apply (rule semPConFail)
 apply clarsimp
 apply clarsimp
 apply (subgoal_tac "h = h'a")
  prefer 2 apply (rule ConstrInd_h_same) apply assumption
 apply clarsimp
 apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
  prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
 apply simp 
 (* CLAUSE: bounded output *)
 (*
 apply (rule_tac x="0" in exI)
 apply clarsimp
 *)
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (subgoal_tac "n0=0")
 apply clarsimp
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* n0=0: by r0 -> NIL and mListSameRef *)
 apply (frule mListNIL)
 apply (rotate_tac -1)
 apply (frule mListSameRef)
 apply (rotate_tac 1)
 apply simp
 apply clarsimp
 (* CASE CONS ie, fmap_lookup h r0 \<noteq> Some (C NIL []) *)
 apply (subgoal_tac "\<exists> r'h0 r't0 E0'. fmap_lookup h r0 = Some (C CONS [Ref r'h0, Ref r't0]) \<and> ([h0var, t0var], [Ref r'h0, Ref r't0], E, E0'):PConInd")
 apply clarsimp
 apply (erule_tac x=E0' in allE)
 apply (erule_tac x="R (Ref r0)" in allE)
 apply (erule impE)
 apply (rule semPCon)
 apply (rule_tac x="[Ref r'h0, Ref r't0]" in exI)
 apply clarsimp
 apply clarsimp 
 apply (erule_tac x="r1" in allE)
 apply (erule impE)
 apply (subgoal_tac "E x1arg = E0' x1arg")
  prefer 2 apply (rule env_env_notin) apply assumption apply clarsimp
 apply clarsimp
 (*
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply (erule FrameInd.cases)
 apply simp_all
 apply clarsimp
 *)
 (* inner case: NIL *)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply (erule_tac x=E0' in allE)
  apply (erule_tac x="\<bottom>" in allE)
  apply (erule impE)
  apply (rule semPConFail)
  apply simp
  apply clarsimp 
  apply (subgoal_tac "h=h'a")
   prefer 2 apply (rule ConstrInd_h_same) apply assumption
  apply clarsimp
  apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
   prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)
  apply simp 
 (* CLAUSE: bounded output *)
 (* apply (rule_tac x="0" in exI) *)
 apply (subgoal_tac "n0=0")
 apply clarsimp
 apply (rule_tac x="{freshloc (fmap_dom h'a)}" in exI)
 apply (rule mListNIL)
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 (* n0=0: by r0 -> NIL and mListSameRef *)
 apply (frule mListNIL)
 apply (rotate_tac -1)
 apply (frule mListSameRef)
 apply (rotate_tac 1)
 apply simp
 apply clarsimp
  (* inner case: CONS *)
  (* CASE: both lists non-empty i.e., fmap_lookup h r0 \<noteq> Some (C NIL []) \<and> fmap_lookup h r1 \<noteq> Some (C NIL []) *)
  apply (subgoal_tac "\<exists> r'h1 r't1 E1'. fmap_lookup h r1 = Some (C CONS [Ref r'h1, Ref r't1]) \<and> ([h1var, t1var], [Ref r'h1, Ref r't1], E0', E1'):PConInd")
  apply clarsimp
  apply (erule_tac x="E1'" in allE)
  apply (erule_tac x="R (Ref r1)" in allE)
  apply (erule impE)
  apply (rule semPCon)
  apply (rule_tac x="[Ref r'h1, Ref r't1]" in exI)
  apply clarsimp
  apply clarsimp
  (*
  apply (case_tac "vb' = \<bottom>")
    apply clarsimp
    defer 1
  *)
    (* OTW *)
    apply (simp add: fuse0_funSpecTab fuse0Ass_def)
    apply clarsimp?
    apply (erule_tac x="n0 - 1" in allE)
    apply (rotate_tac -1)
    apply (erule impE)
    apply (rule conjI)
    apply (rule_tac x="r't0" in exI)
    apply (subgoal_tac "E0' t0var = E1' t0var")
       prefer 2 apply (rule env_env_notin) apply simp apply clarsimp
     apply clarsimp?
    apply (rule conjI)
    defer 1 (* E1' t0var = R (Ref r't0): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U0-{r0}" in exI)
    apply (erule mList.cases)
     apply simp_all
    (* defer 1 *)
    apply (rule_tac x="r't1" in exI)
    apply (rule conjI)
    defer 1 (* E1' t1var = R (Ref r't1): unwrap PConInd E0' -> E1' *)
    apply (rule_tac x="U1-{r1}" in exI)
    apply (rotate_tac 3) (* (n0, r1, U1, h) \<in> mList *)
    apply (erule mList.cases)
     apply simp_all
    (* case analysis over effect of conditional *)
    apply (case_tac  "ilt vx vy = B True") 
     apply clarsimp
     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply simp
     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     (* CLAUSE: bounded output *)
     (* apply (rule_tac x="Suc nv" in exI) *)
     (* apply (rule conjI) *)
     apply (subgoal_tac "\<exists> n'. Suc n' = n0")
     apply clarsimp
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
     apply (subgoal_tac "(n', rv, Uv, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
     apply (rotate_tac -1)
     apply (rule mListCONS)
     prefer 4
     apply simp
     apply clarsimp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
     apply simp
     apply (rule mListSameH)
     apply simp
     apply (simp add: sameH_def)
     apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -5)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
     defer 1(*  Suc nv \<le> n0 \<dots>. *)
     (* ELSE branch: *)
     apply (subgoal_tac "ilt vx vy = B False") (* follows by tertium non datur *)
     apply clarsimp

     apply (subgoal_tac "h'b=h'a")
      prefer 2 apply (rule ConstrInd_h_same) apply assumption
     apply clarsimp
     apply (subgoal_tac "card (insert (freshloc (fmap_dom h'a)) (fmap_dom h'a)) = Suc (card (fmap_dom h'a))")
       prefer 2 apply (rule card_insert_disjoint) apply (rule finite_dom_fmap) apply (rule freshloc) apply (rule finite_dom_fmap)  
     apply clarsimp

     apply (subgoal_tac "n0>0") 
       prefer 2   apply (erule mList.cases)   apply simp_all
     apply (rule_tac x="Uv \<union> {freshloc (fmap_dom h'a)}" in exI)
(* 8 mostly trivial subgoals left:
   ieq vx vy = B False ... by tertium non datur
                       ... if r1 not NIL, r1 must be a CONS 
                       ... if r0 not NIL, r0 must be a CONS 
   E1' t0var = R (Ref r't0) ... by unwrapping PConInd
   E1' t1var = R (Ref r't1) ... by unwrapping PConInd
   rs = ?...   
   freshloc (fmap_dom h'a) \<notin> Uv
   Suc nv \<le> n0
*)
 apply (subgoal_tac "(Suc (n0 - Suc 0), freshloc (fmap_dom h'a), Uv  \<union> {freshloc (fmap_dom h'a)}, h'a(freshloc (fmap_dom h'a)\<mapsto>\<^sub>fC CONS rs)) \<in> mList")
 apply clarsimp
 apply (rule mListCONS)
 prefer 4
 apply (rule mListSameH)
 apply simp
 apply (simp add: sameH_def)
 apply clarsimp
     apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
      apply (simp add: FMAPlookup1)
      (* OTW *)
      apply simp
      apply (rotate_tac -8)
      apply (frule mListDom)
      apply (subgoal_tac "l \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
 apply simp
     defer 1 (* rs = ... *)
     defer 1 (* freshloc (fmap_dom h'a) \<notin> Uv *)
 apply simp
 apply (insert ilt_tertium_non_datur)
 apply fastsimp
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r1" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r1 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E0' t1var (R (Ref rt))) h1var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* again, list_only_heap *)
 apply (insert list_only_heap)
 apply (rotate_tac -1)
 apply (erule_tac x="r0" in allE)
 apply (rotate_tac -1)
 apply (erule_tac x="h" in allE)
 apply (rotate_tac -1)
 apply (erule impE)
 apply (rule mListRefInDom)
 apply simp
 apply (rotate_tac -1)
 apply (case_tac "fmap_lookup h r0 = Some (C NIL [])")
  apply clarsimp
  apply clarsimp
  apply (rule_tac x="fun_upd (fun_upd E t0var (R (Ref rt))) h0var (R (Ref rh))" in exI)
  apply (rule PConInd_CONS)
  apply (rule PConInd_CONS)
  apply (rule PConInd_NIL)
  apply simp
  apply simp
 (* subgoal: E1' t0var = R (Ref r't0) *)
 apply (rotate_tac 5)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* subgoal:  E1' t1var = R (Ref r't1) *)
 apply (rotate_tac 7)
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (rotate_tac -1)
 apply (erule PConInd.cases)
 apply simp_all
 (* left-over subgoal from CLAUSE: bounded output *)
 apply (subgoal_tac "rs = [Ref r'h0, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 defer 1
 apply (rotate_tac 9)
 apply (frule mListDom)
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
   apply (rotate_tac -2)
   apply (drule subsetD)
   apply simp
   apply (subgoal_tac "finite (fmap_dom h'a)")
    prefer 2 apply (rule finite_dom_fmap)
   apply (frule freshloc)
   apply simp
   (* OTW *)
   apply assumption
 (* Suc nv \<le> n0 *)
 apply (simp add: gr0_conv_Suc)
 apply clarsimp
 (*
 apply (subgoal_tac "(n0 - Suc 0) + Suc 0 \<le> n0")
  apply simp
  apply simp
 *)
 (* rs ... *)
 apply (subgoal_tac "rs = [Ref r'h1, Ref rv]")
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule ConstrInd.cases)
 apply simp_all
 apply (subgoal_tac "E1' h1var = R (Ref r'h1)")
 apply clarsimp
 apply (rotate_tac 7)
 apply (erule PConInd.cases)
 apply simp_all
 apply (case_tac "freshloc (fmap_dom h'a) \<in> Uv")
      apply (rotate_tac -9)
      apply (frule mListDom)
      apply (subgoal_tac " freshloc (fmap_dom h'a) \<in> fmap_dom h'a")
       prefer 2 apply fastsimp
      apply (subgoal_tac "finite (fmap_dom h'a)")
       prefer 2 apply (rule finite_dom_fmap)
      apply (frule freshloc)
      apply simp
   apply simp
 (*
 apply (subgoal_tac "E1' h0var = R (Ref r'h1)")
 apply clarsimp
 apply (rotate_tac 8)
 apply (erule PConInd.cases)
 apply simp_all
 *)
 defer 1
 apply (subgoal_tac "E1' h0var = E0' h0var \<and> E0' h0var = R (Ref r'h0)")
 apply clarsimp
 apply (rule conjI)
 apply (rotate_tac 8)
 apply (drule env_env_notin)
 prefer 2
 apply (subgoal_tac " E0' h0var =  E1' h0var")
 apply simp
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp
 apply (erule PConInd.cases)
 apply simp_all
 apply clarsimp?
done


(* Example finished *)

(* check that the important unproven lemma can be proven with copy0Ass 

constdefs list_in :: "Heap \<Rightarrow> HoVal \<Rightarrow> nat \<Rightarrow> bool"
"list_in h w n \<equiv> (\<exists> r U . w = (R (Ref r)) \<and> (n,r,U,h) \<in> mList)"

lemma proven1: "\<lbrakk>list_in h v n; (a,h',c) = exe (emptyEnv(xin := v)) h (CallFunExp copy [xin])\<rbrakk> \<Longrightarrow> list_in h' v n"
apply (insert copy0_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
apply (erule_tac x="n" in allE)
apply (erule_tac x="U" in allE)
apply (erule impE)
apply simp
apply clarsimp
apply (rule_tac x="U" in exI)
apply clarsimp
done

lemma none: "True"
sorry
*)

end

(* ============================================================================= *)
