theory Listcopy_workin = ListCopy:

(*
   box copy
   in  (w1)
   out (w2,w3)

   box fuse
   in  (w2,w3)
   out (w1)
*)

syntax 
 xarg  :: "VarName"
 x0arg  :: "VarName"
 x1arg  :: "VarName"

translations
"xarg"  == "(VN ''xarg'')"
"x0arg"  == "(VN ''x0arg'')"
"x1arg"  == "(VN ''x1arg'')"

consts locs :: "Locn \<Rightarrow> Heap \<Rightarrow> Locn set"
defs
  locs: "locs l h \<equiv> \<Union>{x. \<exists> n. (n,l,x,h) \<in> mList}"

consts
  heap_restrict :: "Heap \<Rightarrow> (Locn set) \<Rightarrow> Heap"
  loc_restrict :: "Heap \<Rightarrow> Locn \<Rightarrow> Heap"
  HoVal_restrict :: "Heap \<Rightarrow> HoVal \<Rightarrow> Heap"

defs
  heap_restrict: "heap_restrict h ls \<equiv> fimap ((themap h) |` ls)"
  loc_restrict: "loc_restrict h l \<equiv> heap_restrict h (locs l h)"

lemma vdmexe3: "\<forall> E h h' v p e P. ({} \<rhd> e : P) \<and> ((v,h',p) = exe E h e) \<longrightarrow>  P E h h' v p"
  by (simp add: vdmexe)

constdefs shallowCopy :: "Heap \<Rightarrow> Locn set \<Rightarrow> Heap"
"shallowCopy h s == heap_restrict h s"

consts
 copyExe :: "HoVal \<Rightarrow> Locn set \<Rightarrow> Locn set \<Rightarrow> Heap \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"
defs
  copyExe: "copyExe v' Uin Uout hwire hbox \<equiv>
    let hbox = shallowCopy hwire Uin in 
    let (v,h',rho) = exe (emptyEnv(xarg := v')) hbox copyBody in
    (v, shallowCopy h' (Uin \<union> Uout), rho)"

consts 
 mkHeap :: "Locn set \<Rightarrow> Heap \<Rightarrow> Heap"
defs
 mkHeap: "mkHeap U h \<equiv> fimap (\<lambda> l . if (l:U) then fmap_lookup h l else fmap_lookup emptyfinmap l)"

constdefs copyAss :: "[VarName list] => Assertion"
"copyAss zs == (\<lambda> E h h' v p. 
                                 (\<forall> n r U . (E (hd zs) = (R (Ref r)) \<and> (n,r,U,h):mList  
                                             \<longrightarrow> (card (fmap_dom h') \<le> card (fmap_dom h) + n + 1) \<and> (\<exists> rv Uv. v = R (Ref rv) \<and> (n,rv,Uv,h'):mList))))"

constdefs fuseAss :: "[VarName list] => Assertion"
"fuseAss zs == \<lambda>E h h' v p.
          			   \<forall>n0 r0 U0 n1.
          			      E (hd zs) = R (Ref r0) \<and> (n0, r0, U0, h) \<in> mList \<and> (\<exists>r1. E (hd (tl zs)) = R (Ref r1) \<and> (\<exists>U1. (n1, r1, U1, h) \<in> mList) \<and> n0 = n1) \<longrightarrow>
          			      card (fmap_dom h') \<le> card (fmap_dom h) + n0 + 1 \<and> (\<exists> nv rv Uv. v = R (Ref rv) \<and> (nv, rv, Uv, h') \<in> mList \<and> nv \<le> n0 ) "


consts
  fuseExe :: "HoVal \<Rightarrow> HoVal \<Rightarrow> Locn set \<Rightarrow> Locn set \<Rightarrow> Locn set \<Rightarrow> Heap \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"
defs
  fuseExe: "fuseExe v1 v2 Uin1 Uin2 Uout hwire hbox \<equiv>
    let hbox = shallowCopy (shallowCopy hwire Uin1) Uin2 in 
    let (v,h',rho) = exe ((emptyEnv(x0arg := v1))(x1arg := v2)) hbox fuseBody in
    (v, shallowCopy h' Uout, rho)"

(* function to check if output can be asserted *)
consts 
  assertOut :: "HoVal \<Rightarrow> HoVal \<Rightarrow> bool"
defs
  asserOut: "assertOut x y \<equiv> (x = BOTTOM) \<or> (y = BOTTOM)"

(* function to copy output values, i.e. pick the non-BOTTOM version *)
consts
  nwire :: "HoVal \<Rightarrow> HoVal \<Rightarrow> HoVal"
defs
  nwire: "nwire x y \<equiv> (if x = BOTTOM then y else x)"

datatype PC = pc_fuse | pc_copy | pc_env

consts
  pc :: "PC statefun" (* the program counter - sequential execution of boxes *)
  fuse_st :: "BoxState statefun" (* state/mode of box *)
  copy_st :: "BoxState statefun" (* state/mode of box *)
  fuse_res :: "HoVal statefun"
  copy_res1_old :: "(HoVal*HoVal) statefun"
  copy_res1 :: "HoVal statefun" (* see my comment on HACK below -- HWL *)
  copy_res2 :: "HoVal statefun"
  heap :: "Heap statefun"
  w1 :: "HoVal statefun" (* input copy *)
  w2 :: "HoVal statefun" (* input fuse *) 
  w3 :: "HoVal statefun" (* input fuse *)
  w4 :: "HoVal statefun" (* output fuse *)
  rs :: "Rescomp statefun"

consts
  fuseTemp :: "temporal"
  copyTemp :: "temporal"
  fuse_sup :: "temporal"
  fuse_exe :: "temporal"
  copy_sup :: "temporal"
  copy_exe :: "temporal"
  super :: "temporal"
  env :: "temporal"
  init :: "temporal"
  program :: "temporal"

(* the link to expression layer *)
consts
 simplefuseExe :: "HoVal \<Rightarrow> HoVal \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"
 simplecopyExe2 :: "HoVal \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"
 (* also returns old value *)
 simplecopyExe :: "HoVal \<Rightarrow> Heap \<Rightarrow> HoVal*HoVal*Heap*Rescomp"

constdefs checkAvail1 :: "HoVal \<Rightarrow> bool" 
[simp]: "checkAvail1 w == (w = BOTTOM)"

(* binds a value *)
constdefs bindVar1 :: "Env \<Rightarrow> HoVal \<Rightarrow> Env" 
"bindVar1 E w == E(xarg := w)"

(* checks against a value *)
constdefs checkAvail2 :: "HoVal \<Rightarrow> HoVal \<Rightarrow> bool" 
[simp]: "checkAvail2 wire1 wire2 == (wire1 = BOTTOM) \<or> (wire2 = BOTTOM)"

(* binds a value *)
constdefs bindVar2 :: "Env \<Rightarrow> HoVal \<Rightarrow> HoVal \<Rightarrow> Env" 
"bindVar2 E wire1 wire2 == E(x0arg := wire1, x1arg := wire2)"

constdefs checkAvailN :: "HoVal list \<Rightarrow> bool" 
"checkAvailN xs == list_all checkAvail1 xs"

consts upds :: "(VarName \<times> HoVal) list \<Rightarrow> Env \<Rightarrow> Env" 
primrec 
"upds []          E = E"
"upds (xw#xws) E = (upds xws (E((fst xw):=(snd xw))))"

constdefs bindVarN :: "Env \<Rightarrow> VarName list \<Rightarrow> HoVal list \<Rightarrow> Env" 
"bindVarN E xs ws == upds (zip xs ws) E"

(* pick-up variable names from function specification table *)
constdefs bindVarFunN :: "FunName \<Rightarrow> HoVal list \<Rightarrow> Env" 
"bindVarFunN fx ws == bindVarN emptyEnv (fst (funTab fx)) ws"

defs
 simplefuseExe: "simplefuseExe v1 v2 h \<equiv> exe (bindVarFunN fuse0 [v1,v2]) h (snd (funTab fuse0))"

defs
 simplecopyExe2: "simplecopyExe2 v h \<equiv> exe (emptyEnv(xarg := v)) h copyBody"
 simplecopyExe: "simplecopyExe v h \<equiv> (v,simplecopyExe2 v h)"

defs 
 fuse_exe: "fuse_exe \<equiv> TEMP
(if $pc \<noteq> #pc_fuse
 then Unchanged (w1,w2,fuse_res,fuse_st)
 else pc$ = #pc_copy \<and> 
 (if $fuse_st = #Blocked
   then Unchanged (w1,w2,fuse_res,fuse_st,heap)
   else 
 (if checkAvail2<$w1,$w2>
   then Unchanged (w1,w2,fuse_res,fuse_st,heap)
     \<and>  fuse_res$ = #BOTTOM 
     \<and>  fuse_st$ = #Matchfail
   else (w1,w2)$ = #(BOTTOM,BOTTOM)
    \<and> (fuse_res,heap,rs)$ = simplefuseExe<$w1,$w2,$heap>)))"
defs
 copy_exe: "copy_exe \<equiv> TEMP
(if $pc \<noteq> #pc_copy
 then Unchanged (w3,copy_res1,copy_res2,copy_st)
 else pc$ = #pc_copy \<and> 
 (if $copy_st = #Blocked
   then Unchanged (w3,copy_res1,copy_res2,copy_st,heap)
   else 
 (if checkAvail1<$w3>
   then Unchanged (w3,copy_res1,copy_res2,copy_st,heap)
     \<and>  (copy_res1,copy_res2)$ = #(BOTTOM,BOTTOM) 
     \<and>  fuse_st$ = #Matchfail
   else w3$ = #BOTTOM
    \<and> (copy_res1,copy_res2,heap,rs)$ = simplecopyExe<$w3,$heap>)))"

defs
 fuse_sup: "fuse_sup \<equiv> TEMP
   (if assertOut<$fuse_res,$w4>
    then fuse_st$ = #Blocked \<and> Unchanged (w4,fuse_res)  
    else fuse_st$ = #Runnable \<and>  fuse_res$ = #BOTTOM
     \<and> w4$ = nwire<$fuse_res,$w4>)"
defs
 copy_sup: "copy_sup \<equiv> TEMP
   (if assertOut<$copy_res1,$w1> \<and> assertOut<$copy_res2,$w2> 
    then copy_st$ = #Blocked \<and> Unchanged (w1,w2,copy_res1,copy_res2)  
    else copy_st$ = #Runnable \<and> (copy_res1,copy_res2)$ = #(BOTTOM,BOTTOM)
     \<and> w1$ = nwire<$copy_res1,$w1>
     \<and> w2$ = nwire<$copy_res2,$w2>)"
(* WAS:
     \<and> w1$ = nwire<fst<$copy_res>,$w1>
     \<and> w2$ = nwire<fst<$copy_res>,$w2>)"
*)
defs
  fuseTemp: "fuseTemp \<equiv> TEMP ($sch = #Execute \<longrightarrow> fuse_exe) \<and> ($sch = #Super \<longrightarrow> fuse_sup)"
  copyTemp: "copyTemp \<equiv> TEMP ($sch = #Execute \<longrightarrow> copy_exe) \<and> ($sch = #Super \<longrightarrow> copy_sup)"


constdefs
  S' :: "temporal"
  "S' \<equiv> TEMP (sch$ = (if isExe \<and> $pc = #pc_copy then #Super else #Execute))
           \<and> (isSup \<longrightarrow> pc$ = #pc_env)"

consts
  H :: "Heap"
  L :: "Locn"
  N :: "nat"  (* renamed this for convenience -- HWL *)
  Locs :: "Locn set"

axioms
  initHeap: "(N,L,Locs,H) \<in> mList"
  initHeapEmpty: "card (fmap_dom H) = N+1" (* or that remaining part of heap is empty??? *)
 
defs
  init: "init \<equiv> TEMP
    $sch = #Execute 
  \<and> $pc = #pc_env
  \<and> $heap = #H
  \<and> $(fuse_st,copy_st) = #(Runnable,Runnable)
  \<and> $(fuse_res,copy_res1,copy_res2) = #(BOTTOM,BOTTOM,BOTTOM)
  \<and> $(w1,w2,w3,w4) = #(BOTTOM,BOTTOM,BOTTOM,BOTTOM)"

consts
  env2 :: "temporal"

defs
 env: "env \<equiv> TEMP
  ($sch = #Super \<longrightarrow> heap$ = $heap \<and> w3$=$w3) \<and>
  ($sch = #Execute \<longrightarrow>
    (if $pc \<noteq> #pc_env
      then w4$ = $w4
      else w4$ = #BOTTOM \<and> heap$=$heap \<and> pc$ = #pc_fuse))"

 env2: "env2 \<equiv> TEMP
  ($sch = #Super \<longrightarrow> 
    (if (w1,w2,w3,w4,fuse_res,copy_res1,copy_res2)$ = #(BOTTOM,BOTTOM,BOTTOM,BOTTOM,BOTTOM,BOTTOM,BOTTOM)
      then w3$ = #(R(Ref L)) \<and>  heap$ = #H
      (* then (\<exists> l h n locs. w3$ = #(R(Ref l)) \<and>  heap$ = #h \<and> (n,l,locs,h) \<in> mList *)
      else w3$ = $w3 \<and> heap$ = $heap)) \<and>
  ($sch = #Execute \<longrightarrow>
    (if $pc \<noteq> #pc_env
      then w4$ = $w4
      else w4$ = #BOTTOM))"

defs
  program: "program \<equiv> TEMP init 
  \<and> \<box>[S' \<and> fuseTemp \<and> copyTemp \<and> env]_(sch,pc,w1,w2,w3,w4,heap,fuse_st,fuse_res,copy_st,copy_res1,copy_res2,rs)"

lemmas hume_rews =
  intensional_rews all_after_eq all_before_eq
  after_before next_all_eq toBeforeVal toAfterVal
  isExe_def isSup_def

lemma vdmexe2: "\<lbrakk>{} \<rhd> e : P; (v,h',p) = exe E h e\<rbrakk> 
  \<Longrightarrow> P E h h' v p"
  by (simp add: vdmexe)

lemma vdmexe3: "\<forall> E h h' v p e P. ({} \<rhd> e : P) \<and> ((v,h',p) = exe E h e) \<longrightarrow>  P E h h' v p"
  by (simp add: vdmexe)

constdefs list_in :: "Heap \<Rightarrow> HoVal \<Rightarrow> nat \<Rightarrow> bool"
"list_in h w n \<equiv> (\<exists> r U . w = (R (Ref r)) \<and> (n,r,U,h) \<in> mList)"

lemma list_in_initHeap: "list_in H (R (Ref L)) N"
  apply (unfold list_in_def)
  apply (rule_tac x=L in exI)
  apply (rule_tac x="Locs" in exI)
  apply simp
  apply (rule initHeap)
  done

lemma nw1: "nwire BOTTOM BOTTOM = BOTTOM"
  apply (simp add: nwire)
  done

lemma w3bot: "\<turnstile> program \<longrightarrow> \<box>($w3 \<noteq> #BOTTOM \<longrightarrow> 
  $w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> 
  $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM \<and>
  $fuse_res = #BOTTOM)"
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI)
  apply (elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
     apply (case_tac "w \<Turnstile> $pc")
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: env)
     apply (simp add: isExe_def isSup_def copyTemp copy_sup env fuseTemp fuse_sup)
      apply (split split_if_asm)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply (simp add: nwire)
   apply (split split_if_asm)
   apply simp
   apply (simp only: unch_eq1, simp add: unch_def nwire)
   apply (simp only: hume_rews)
   apply (simp add: unch_def nwire)
   done

lemma w3list_in: "\<turnstile> program \<longrightarrow> \<box>($w3 = #BOTTOM \<or> list_in<$heap,$w3,#N>)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule w3bot[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI,elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
   prefer 2
  apply (simp add: isExe_def isSup_def env copyTemp copy_sup fuseTemp fuse_sup)
    apply (case_tac "w \<Turnstile> $pc")
    apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
    apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
    apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
    apply (simp only: unch_eq1, simp add: unch_def)
  done

lemma copy_res1bot: "\<turnstile> program \<longrightarrow> \<box>($copy_res1 \<noteq> #BOTTOM \<longrightarrow> 
  $w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM \<and> $fuse_res = #BOTTOM)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule w3bot[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI)
  apply (elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
     apply (case_tac "w \<Turnstile> $pc")
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: env)
     apply (simp add: isExe_def isSup_def copyTemp copy_sup env fuseTemp fuse_sup)
      apply (split split_if_asm)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply (simp add: nwire)
   apply (split split_if_asm)
   apply simp
   apply (simp only: unch_eq1, simp add: unch_def nwire)
   apply (simp only: hume_rews)
   apply (simp add: unch_def nwire)
   done

lemma copy_res2bot: "\<turnstile> program \<longrightarrow> \<box>($copy_res2 \<noteq> #BOTTOM \<longrightarrow> 
  $w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM \<and> $fuse_res = #BOTTOM)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule w3bot[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI)
  apply (elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
     apply (case_tac "w \<Turnstile> $pc")
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: env)
     apply (simp add: isExe_def isSup_def copyTemp copy_sup env fuseTemp fuse_sup)
      apply (split split_if_asm)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply (simp add: nwire)
   apply (split split_if_asm)
   apply simp
   apply (simp only: unch_eq1, simp add: unch_def nwire)
   apply (simp only: hume_rews)
   apply (simp add: unch_def nwire)
   done

theorem inv_join: "\<lbrakk>\<turnstile> P \<longrightarrow> \<box>Q; \<turnstile> P \<longrightarrow> \<box>T\<rbrakk> \<Longrightarrow> \<turnstile> P \<longrightarrow> \<box>(Q \<and> T)"
 apply (int_unlift,rule impI)
 apply (simp only: STL5[THEN inteq_reflection])
 apply simp
 thm intD
 apply (drule_tac w="w" in intD)
 apply (drule_tac w="w" in intD)
 apply auto
 done


(*
lemma toProve1: "\<lbrakk>list_in h v n; (a,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' v n"
  sorry 
*)

lemma toProve1: "\<lbrakk>list_in h v n; (a,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' v n"
apply (insert copy0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
apply (erule_tac x="n" in allE)
apply (erule_tac x="U" in allE)
apply (erule impE)
apply simp
apply clarsimp
apply (rule_tac x="U" in exI)
apply clarsimp
done


lemma copy_res1list_in: "\<turnstile> program \<longrightarrow> \<box>($copy_res1 = #BOTTOM \<or> list_in<$heap,$copy_res1,#N>)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule inv_join)
  apply (rule copy_res1bot[unfolded program])
  apply (rule  w3list_in[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI,elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
   prefer 2
  apply (simp add: isExe_def isSup_def env copyTemp copy_sup fuseTemp fuse_sup)
   apply (split split_if_asm)
    apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
   apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
    apply (case_tac "w \<Turnstile> $pc")
    apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   prefer 2
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
  apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply (simp only: hume_rews)
    apply (elim conjE)
    apply (simp only: simplecopyExe simplecopyExe2)
    apply simp
    apply (elim conjE)
    apply (rule toProve1)
    apply assumption
    apply assumption
    done

(*
lemma toProve2: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' nv n"
  sorry
*)

lemma toProve2: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' nv n"
apply (insert copy0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
apply (erule_tac x="n" in allE)
apply (erule_tac x="U" in allE)
apply (erule impE)
apply simp
apply clarsimp
apply (rule_tac x="Uv" in exI)
apply clarsimp
done

lemma copy_res2list_in: "\<turnstile> program \<longrightarrow> \<box>($copy_res2 = #BOTTOM \<or> list_in<$heap,$copy_res2,#N>)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule inv_join)
  apply (rule copy_res2bot[unfolded program])
  apply (rule  w3list_in[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI,elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
   prefer 2
  apply (simp add: isExe_def isSup_def env copyTemp copy_sup fuseTemp fuse_sup)
   apply (split split_if_asm)
    apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
   apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
    apply (case_tac "w \<Turnstile> $pc")
    apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   prefer 2
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
  apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply (simp only: hume_rews)
    apply (elim conjE)
    apply (simp only: simplecopyExe simplecopyExe2)
    apply simp
    apply (elim conjE)
    apply (rule disjI2)
    apply (rule toProve2)
    apply assumption
    apply assumption
    done

(*
lemma toProve3: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> nv \<noteq> BOTTOM"
  sorry
*)

lemma toProve3: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> nv \<noteq> BOTTOM"
apply (insert copy0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
done


lemma copy_res1_res2_boteq: "\<turnstile> program \<longrightarrow> \<box>(($copy_res1 = #BOTTOM)=($copy_res2 = #BOTTOM))"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule w3list_in[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI)
  apply (elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
     apply (case_tac "w \<Turnstile> $pc")
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: simplecopyExe simplecopyExe2)
     apply (simp only: hume_rews)
     apply simp
     apply (rule toProve3)
     apply assumption
     apply (elim conjE)
     apply assumption
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: env)
     apply (simp add: isExe_def isSup_def copyTemp copy_sup env fuseTemp fuse_sup)
      apply (split split_if_asm)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply (simp add: nwire)
   apply (split split_if_asm)
   apply simp
   apply (simp only: unch_eq1, simp add: unch_def nwire)
   apply (simp only: hume_rews)
   apply (simp add: unch_def nwire)
   done

lemma copy_w1bot: "\<turnstile> program \<longrightarrow> \<box>($w1 \<noteq> #BOTTOM \<longrightarrow> 
  $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM \<and> $w3 = #BOTTOM \<and> $fuse_res = #BOTTOM)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule inv_join)
  apply (rule copy_res1bot[unfolded program])
  apply (rule copy_res1_res2_boteq[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI)
  apply (elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
     apply (case_tac "w \<Turnstile> $pc")
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: hume_rews)
     apply simp
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: env)
     apply (simp add: isExe_def isSup_def copyTemp copy_sup env fuseTemp fuse_sup)
      apply (split split_if_asm)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply (simp add: nwire)
   apply (split split_if_asm)
   apply simp
   apply (simp only: unch_eq1, simp add: unch_def nwire)
   apply (simp only: hume_rews)
   apply (simp add: unch_def nwire)
   apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (elim conjE)
   apply (simp add: nwire)
   apply (case_tac "(w \<Turnstile> $copy_res2) = BOTTOM")
   apply simp
   apply simp
   done

lemma w1list_in: "\<turnstile> program \<longrightarrow> \<box>($w1 = #BOTTOM \<or> list_in<$heap,$w1,#N>)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule inv_join)
  apply (rule copy_res1list_in[unfolded program])
  apply (rule copy_w1bot[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI,elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
   prefer 2
  apply (simp add: isExe_def isSup_def env copyTemp copy_sup fuseTemp fuse_sup)
   apply (split split_if_asm)
    apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
  apply (elim conjE)
  apply (simp only: nwire)
   apply (case_tac "(w \<Turnstile> $copy_res1)=BOTTOM")
   apply simp
   apply simp
   apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
  apply (simp only: nwire)
  apply simp
  apply (case_tac "(w \<Turnstile> $copy_res1) = BOTTOM")
  apply simp
  apply simp
    apply (case_tac "w \<Turnstile> $pc")
    apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply simp
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
  apply (simp only: unch_eq1, simp add: unch_def)
  apply (elim conjE)
  apply (split split_if_asm)
    apply ( simp add: unch_def)
    apply simp
  apply (split split_if_asm)
    apply ( simp add: unch_def)
    apply simp
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
    apply (simp only: unch_eq1, simp add: unch_def)
done

lemma copy_w2bot: "\<turnstile> program \<longrightarrow> \<box>($w2 \<noteq> #BOTTOM \<longrightarrow> 
  $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM \<and> $w3 = #BOTTOM \<and> $fuse_res = #BOTTOM)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule inv_join)
  apply (rule copy_res2bot[unfolded program])
  apply (rule copy_res1_res2_boteq[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI)
  apply (elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
     apply (case_tac "w \<Turnstile> $pc")
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: hume_rews)
     apply simp
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: env)
     apply (simp add: isExe_def isSup_def copyTemp copy_sup env fuseTemp fuse_sup)
      apply (split split_if_asm)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply (simp add: nwire)
   apply (split split_if_asm)
   apply simp
   apply (simp only: unch_eq1, simp add: unch_def nwire)
   apply (simp only: hume_rews)
   apply (simp add: unch_def nwire)
   apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (elim conjE)
   apply (simp add: nwire)
   apply (case_tac "(w \<Turnstile> $copy_res2) = BOTTOM")
   apply simp
   apply simp
   done

lemma w2list_in: "\<turnstile> program \<longrightarrow> \<box>($w2 = #BOTTOM \<or> list_in<$heap,$w2,#N>)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule inv_join)
  apply (rule copy_res2list_in[unfolded program])
  apply (rule copy_w2bot[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI,elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
   prefer 2
  apply (simp add: isExe_def isSup_def env copyTemp copy_sup fuseTemp fuse_sup)
   apply (split split_if_asm)
    apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
  apply (elim conjE)
  apply (simp only: nwire)
   apply (case_tac "(w \<Turnstile> $copy_res2)=BOTTOM")
   apply simp
   apply simp
   apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
  apply (simp only: nwire)
  apply simp
  apply (case_tac "(w \<Turnstile> $copy_res2) = BOTTOM")
  apply simp
  apply simp
    apply (case_tac "w \<Turnstile> $pc")
    apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply simp
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
  apply (simp only: unch_eq1, simp add: unch_def)
  apply (elim conjE)
  apply (split split_if_asm)
    apply ( simp add: unch_def)
    apply simp
  apply (split split_if_asm)
    apply ( simp add: unch_def)
    apply simp
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
    apply (simp only: unch_eq1, simp add: unch_def)
done

lemma fuse_resbot: "\<turnstile> program \<longrightarrow> \<box>($fuse_res \<noteq> #BOTTOM \<longrightarrow> 
  $w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM \<and> $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule copy_w1bot[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI)
  apply (elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
     apply (case_tac "w \<Turnstile> $pc")
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: hume_rews)
     apply simp
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: env)
     apply (simp add: isExe_def isSup_def copyTemp copy_sup env fuseTemp fuse_sup)
      apply (split split_if_asm)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply (simp add: nwire)
   apply (split split_if_asm)
   apply simp
   apply simp
   done

lemma w1w2_list_in: "\<turnstile> program \<longrightarrow> \<box>(
  ($w1 = #BOTTOM \<or> list_in<$heap,$w1,#N>)
\<and> ($w2 = #BOTTOM \<or> list_in<$heap,$w2,#N>))"
  apply (rule inv_join)
  apply (rule w1list_in)
  apply (rule w2list_in)
  done

(*
lemma toProve4: "\<lbrakk>list_in h v1 n; list_in h v2 n; 
  (v,h',x) = exe (emptyEnv(x0arg := v1, x1arg := v2)) h fuseBody \<rbrakk>
  \<Longrightarrow> list_in h' v n"
  sorry
*)

lemma toProve4: "\<lbrakk>list_in h v1 n; list_in h v2 n; 
                 (v,h',x) = exe (bindVarFunN fuse0 [v1,v2]) h (snd (funTab fuse0)) \<rbrakk>
  \<Longrightarrow> list_in h' v n"
apply (insert fuse0Body_resources_ok)
apply (drule vdmexe)
apply (simp add: fuse0_funTabBody)
apply (simp add: fuse0Ass_def list_in_def)
apply clarsimp
apply (erule_tac x=n in allE)
apply (erule impE)
apply (rule conjI)
apply (rule_tac x=r in exI)
apply (rule conjI)
apply (simp add: bindVarFunN_def  bindVarN_def fuse0_funTabArgs)
apply (rule_tac x=U in exI)
apply simp
apply (rule_tac x=ra in exI)
apply (rule conjI)
apply (simp add: bindVarFunN_def  bindVarN_def fuse0_funTabArgs)
apply (rule_tac x=Ua in exI)
apply simp
apply clarsimp
done

lemma toProve4old: "\<lbrakk>list_in h v1 n; list_in h v2 n; 
                 (v,h',x) = exe (emptyEnv(x0arg := v1, x1arg := v2)) h fuse0Body \<rbrakk>
  \<Longrightarrow> list_in h' v n"
apply (insert fuse0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 2)
apply (erule thin_rl)
apply (simp add: fuse0Ass_def list_in_def)
done

lemma fuse_reslist_in: "\<turnstile> program \<longrightarrow> \<box>($fuse_res = #BOTTOM \<or> list_in<$heap,$fuse_res,#N>)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule inv_join)
  apply (rule w1w2_list_in[unfolded program])
  apply (rule fuse_resbot[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI,elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
   prefer 2
  apply (simp add: isExe_def isSup_def env copyTemp copy_sup fuseTemp fuse_sup)
   apply (split split_if_asm)
    apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
    apply (case_tac "w \<Turnstile> $pc")
    apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   prefer 2
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
  apply (simp only: unch_eq1, simp add: unch_def)
  apply (elim conjE)
  apply (split split_if_asm)
  apply (simp add: unch_def)
  apply simp
  apply (split split_if_asm)
  apply (simp add: unch_def)
  apply simp
  apply (elim conjE)
  defer 1
  apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
  apply (simp only: unch_eq1, simp add: unch_def)
  apply (simp add: simplefuseExe)
  apply (rule disjI2)
  apply (rule toProve4)
  apply simp
  apply simp
  back
  apply (simp only: hume_rews)
  done

lemma w4bot: "\<turnstile> program \<longrightarrow> \<box>($w4 \<noteq> #BOTTOM \<longrightarrow> 
  $w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM \<and> $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM
  \<and> $fuse_res = #BOTTOM)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule fuse_resbot[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI)
  apply (elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
     apply (case_tac "w \<Turnstile> $pc")
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
     apply (simp add: isExe_def isSup_def copyTemp copy_sup env fuseTemp fuse_sup)
      apply (split split_if_asm)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply (simp add: nwire)
   apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp add: nwire)
   apply (case_tac "(w \<Turnstile> $fuse_res)=BOTTOM")
   apply simp
   apply simp
   apply simp
   apply (simp add: nwire)
   apply (case_tac "(w \<Turnstile> $fuse_res)=BOTTOM")
   apply simp
   apply (simp only: hume_rews)
   apply simp
   apply simp
   apply (simp only: hume_rews)
   apply simp
   done


lemma w4list_in: "\<turnstile> program \<longrightarrow> \<box>($w4 = #BOTTOM \<or> list_in<$heap,$w4,#N>)"
  apply (unfold program)
  apply (rule spec_inv2_mono)
  apply (rule inv_join)
  apply (rule fuse_reslist_in[unfolded program])
  apply (rule w4bot[unfolded program])
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI,elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
   prefer 2
  apply (simp add: isExe_def isSup_def env copyTemp copy_sup fuseTemp fuse_sup)
   apply (split split_if_asm)
    apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
  apply simp
  apply (elim conjE)
  apply (simp only: nwire)
   apply (case_tac "(w \<Turnstile> $copy_res2)=BOTTOM")
   apply simp
   apply (split split_if_asm)
   apply (split split_if_asm)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply simp
  apply simp
  apply (split split_if_asm)
  apply (split split_if_asm)
  apply simp
  apply simp
  apply simp
    apply (case_tac "w \<Turnstile> $pc")
    apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
     apply (elim conjE,split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
    apply simp
     apply (split split_if_asm)
    apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
  apply (simp only: unch_eq1, simp add: unch_def)
  apply (elim conjE)
  apply (split split_if_asm)
    apply ( simp add: unch_def)
    apply simp
  apply (split split_if_asm)
    apply ( simp add: unch_def)
    apply simp
   apply (simp add: isExe_def isSup_def env copyTemp copy_exe fuseTemp fuse_exe)
done

(* I think we should something like this as top-level coord-level theorem: -- HWL *)

lemma sepconj: 
  assumes h1: "\<turnstile> Prog \<longrightarrow> P" and h2: "\<turnstile> Prog \<longrightarrow> Q"
  shows "\<turnstile> Prog \<longrightarrow> P \<and> Q"
  proof (int_unlift,rule impI,rule conjI)
    fix w
    assume "w \<Turnstile> Prog"
    with h1[THEN intD] show "w \<Turnstile> P" by auto 
  next
    fix w
    assume "w \<Turnstile> Prog"
    with h2[THEN intD] show "w \<Turnstile> Q" by auto 
  qed

theorem main_resources_ok:
  "\<turnstile> program \<longrightarrow> \<box>( ($w1 = #BOTTOM \<or> list_in<$heap,$w1,#N>)
                  \<and> ($w2 = #BOTTOM \<or> list_in<$heap,$w2,#N>)
                  \<and> ($w3 = #BOTTOM \<or> list_in<$heap,$w3,#N>)
                  \<and> ($w4 = #BOTTOM \<or> list_in<$heap,$w4,#N>))"
  apply (simp only: STL5[THEN inteq_reflection])
  apply (rule sepconj)
  apply (rule w1list_in)
  apply (rule sepconj)
  apply (rule w2list_in)
  apply (rule sepconj)
  apply (rule w3list_in)
  apply (rule w4list_in)
  done

theorem main_resources_ok_all:
  "\<turnstile> program \<longrightarrow> \<box>( ($w1 = #BOTTOM \<or> list_in<$heap,$w1,#N>)
                  \<and> ($w2 = #BOTTOM \<or> list_in<$heap,$w2,#N>)
                  \<and> ($w3 = #BOTTOM \<or> list_in<$heap,$w3,#N>)
                  \<and> ($w4 = #BOTTOM \<or> list_in<$heap,$w4,#N>)
                  \<and> ($copy_res1 = #BOTTOM \<or> list_in<$heap,$copy_res1,#N>)
                  \<and> ($copy_res2 = #BOTTOM \<or> list_in<$heap,$copy_res2,#N>)
                  \<and> ($fuse_res = #BOTTOM \<or> list_in<$heap,$fuse_res,#N>))"
  apply (simp only: STL5[THEN inteq_reflection])
  apply (rule sepconj)
  apply (rule w1list_in)
  apply (rule sepconj)
  apply (rule w2list_in)
  apply (rule sepconj)
  apply (rule w3list_in)
  apply (rule sepconj)
  apply (rule w4list_in)
  apply (rule sepconj)
  apply (rule copy_res1list_in)
  apply (rule sepconj)
  apply (rule copy_res2list_in)
  apply (rule fuse_reslist_in)
  done

lemmas w4_allempty =  w4bot

lemma w3_allempty: "\<turnstile> program \<longrightarrow> \<box>($w3 \<noteq> #BOTTOM \<longrightarrow> 
  $w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> $w4 = #BOTTOM \<and>
  $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM \<and>
  $fuse_res = #BOTTOM)"
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
  apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
  apply (intro impI)
  apply (elim conjE)
   apply (case_tac "w \<Turnstile> $sch")
     apply (case_tac "w \<Turnstile> $pc")
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply simp
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: isExe_def isSup_def copyTemp copy_exe env fuseTemp fuse_exe)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp add: env)
     apply (simp add: isExe_def isSup_def copyTemp copy_sup env fuseTemp fuse_sup)
      apply (split split_if_asm)
      apply (split split_if_asm)
     apply (simp only: unch_eq1, simp add: unch_def)
     apply (simp only: unch_eq1, simp add: unch_def)
   apply (simp only: hume_rews)
   apply (simp add: nwire)
   done



end

