theory T = ListCopy:

(*
   box copy
   in  (w1)
   out (w2,w3)

   box fuse
   in  (w2,w3)
   out (w1)
*)

syntax 
 xarg  :: "VarName"
 x0arg  :: "VarName"
 x1arg  :: "VarName"

translations
"xarg"  == "(VN ''xarg'')"
"x0arg"  == "(VN ''x0arg'')"
"x1arg"  == "(VN ''x1arg'')"

(* -------------------------------------------------- *)
(* modelling LISTS 
   COMMENTED OUT since all this is defined in ListCopy
syntax 
 CONS :: ConName
 NIL  :: ConName
translations
 "CONS" == "(CN ''CONS'')"
 "NIL" == "(CN ''NIL'')"

consts mList::"(nat \<times> Locn \<times> (Locn set) \<times> Heap) set"
inductive mList intros
mListNIL: "\<lbrakk> fmap_lookup h a = Some (C NIL []) \<rbrakk> \<Longrightarrow> (0,a,{a},h) : mList"
mListCONS:"\<lbrakk> fmap_lookup h a = Some (C CONS [rx,rxs]) ; a \<notin> U; (Ref b) = rxs ; (n,b,U,h):mList\<rbrakk>
          \<Longrightarrow> (Suc n, a, U \<union> {a}, h): mList"


lemma mListDom: "\<lbrakk>(n, a, U, h) \<in> mList\<rbrakk> \<Longrightarrow>  U \<subseteq> fmap_dom h"
apply (erule mList.induct)
apply (simp_all add:fmap_lookup_def fmap_dom_def dom_def)
done

lemma mListFresh: "(L, a, U, h) \<in> mList \<Longrightarrow> freshloc (fmap_dom h) \<notin> U"
apply (subgoal_tac "U \<subseteq> fmap_dom h")
apply (subgoal_tac "finite (fmap_dom h)")
apply (drule freshloc)
apply fastsimp
apply fastsimp
apply (erule mListDom)
done

(* TODO: move this to Basics.thy or thereabouts *)
constdefs sameH::"Locn set \<Rightarrow> Heap \<Rightarrow> Heap \<Rightarrow> bool"
"sameH U h hh \<equiv> (\<forall> l. (l \<in> U \<longrightarrow> (fmap_lookup h l = fmap_lookup hh l)))"


lemma mListSameH[rule_format]:
"(n,a,U,h) \<in> mList \<Longrightarrow> (\<forall> hh. sameH U h hh \<longrightarrow> (n,a,U,hh) \<in> mList)"
apply (erule mList.induct, safe)
apply (rule mListNIL)
apply (simp add: sameH_def)
apply (simp add: sameH_def)
apply (erule_tac x="hh" in allE)
apply (erule impE)
apply clarsimp
apply (subgoal_tac "(Suc n, a, U \<union> {a}, hh) \<in> mList")
apply simp
apply (rule mListCONS)
apply (erule_tac x="a" in allE)
apply ( simp add: sameH_def)
apply (subgoal_tac "fmap_lookup hh a = Some (C CONS [rx, Ref b])") 
apply (rotate_tac -2)
apply (erule thin_rl)
apply ( simp add: sameH_def)
apply (rule conjI)
apply simp+
done
*)
(* -------------------------------------------------- *)
(* Variant 1: full pattern matching, as in expression level, encoding failure as BOTTOM
              this needs unwinding of MATCH definition from expression level *)

consts boxMatch :: "Env \<Rightarrow> Heap \<Rightarrow> Pattern \<Rightarrow> Locn \<Rightarrow> (Env \<times> HoVal)" 
axioms
 boxMatch: "((E', v') = boxMatch E h b l) == (MATCH E, h \<turnstile> b @ l \<Down> (E', v'))"

consts pattMatch1 :: "Env \<Rightarrow> Heap \<Rightarrow> Pattern \<Rightarrow> Locn \<Rightarrow> HoVal" 
axioms
 pattMatch1: "(v' = pattMatch1 E h b l) == (\<exists> E' .MATCH E, h \<turnstile> b @ l \<Down> (E', v'))"

consts makeEnv1 :: "Env \<Rightarrow> Heap \<Rightarrow> Pattern \<Rightarrow> Locn \<Rightarrow> Env" 
axioms
 makeEnv1: "(E' = makeEnv1 E h b l) == (\<exists> v' .MATCH E, h \<turnstile> b @ l \<Down> (E', v'))"

(* Variant 2: check only availability on this level, and if so bind the reference to a fixed argument name;
              no pattern matching is being done here  
              the names of the arguments must match the ones in funTab (a global table used at the expression level)

*)

(* checks against a location on the heap
constdefs checkAvail1 :: "Env \<Rightarrow> Heap \<Rightarrow> Locn \<Rightarrow> bool" 
"checkAvail1 E h l == (fmap_lookup h l \<noteq> Some BOTTOM)"

constdefs bindVar1 :: "Env \<Rightarrow> Heap \<Rightarrow> Locn \<Rightarrow> Env" 
"bindVar1 E h l == E(xarg := R (Ref l))"
*)

(* checks against a value *)
constdefs checkAvail1 :: "HoVal \<Rightarrow> bool" 
"checkAvail1 w == (w \<noteq> BOTTOM)"

(* binds a value *)
constdefs bindVar1 :: "Env \<Rightarrow> HoVal \<Rightarrow> Env" 
"bindVar1 E w == E(xarg := w)"

(* checks against a location on the heap
constdefs checkAvail2 :: "Env \<Rightarrow> Heap \<Rightarrow> Locn \<Rightarrow> Locn \<Rightarrow> bool" 
"checkAvail2 E h l1 l2 == (fmap_lookup h l1 \<noteq> Some BOTTOM) \<and> (fmap_lookup h l2 \<noteq> Some BOTTOM)"

constdefs bindVar2 :: "Env \<Rightarrow> Heap \<Rightarrow> Locn \<Rightarrow> Locn \<Rightarrow> Env" 
"bindVar2 E h l1 l2 == E(x0arg := R (Ref l1), x1arg := R (Ref l2))"
*)

(* checks against a value *)
constdefs checkAvail2 :: "HoVal \<Rightarrow> HoVal \<Rightarrow> bool" 
"checkAvail2 w1 w2 == (w1 \<noteq> BOTTOM) \<and> (w2 \<noteq> BOTTOM)"

(* binds a value *)
constdefs bindVar2 :: "Env \<Rightarrow> HoVal \<Rightarrow> HoVal \<Rightarrow> Env" 
"bindVar2 E w1 w2 == E(x0arg := w1, x1arg := w2)"

(* general versions of the above:
   . checkAvailN: fold (\<noteq> BOTTOM) all_wires
   . binVarN:     multi-update of a list of vars with a list of values
*)

constdefs checkAvailN :: "HoVal list \<Rightarrow> bool" 
"checkAvailN xs == list_all checkAvail1 xs"

consts upds :: "(VarName \<times> HoVal) list \<Rightarrow> Env \<Rightarrow> Env" 
primrec 
"upds []          E = E"
"upds (xw#xws) E = (upds xws (E((fst xw):=(snd xw))))"

constdefs bindVarN :: "Env \<Rightarrow> VarName list \<Rightarrow> HoVal list \<Rightarrow> Env" 
"bindVarN E xs ws == upds (zip xs ws) E"

(* pick-up variable names from function specification table *)
constdefs bindVarFunN :: "FunName \<Rightarrow> HoVal list \<Rightarrow> Env" 
"bindVarFunN fx ws == bindVarN emptyEnv (fst (funTab fx)) ws"

(* ----------------------------------------------------------------------------- *)
(* example: spec for fuse using the above functions *)
datatype PC = pc_fuse | pc_copy | pc_env

consts
  pc :: "PC statefun" (* the program counter - sequential execution of boxes *)
  fuse_st :: "BoxState statefun" (* state/mode of box *)
  copy_st :: "BoxState statefun" (* state/mode of box *)
  fuse_res :: "HoVal statefun"
  copy_res1_old :: "(HoVal*HoVal) statefun"
  copy_res1 :: "HoVal statefun" (* see my comment on HACK below -- HWL *)
  copy_res2 :: "HoVal statefun"
  heap :: "Heap statefun"
  w1 :: "HoVal statefun" (* input copy *)
  w2 :: "HoVal statefun" (* input fuse *) 
  w3 :: "HoVal statefun" (* input fuse *)
  w4 :: "HoVal statefun" (* output fuse *)
  rs :: "Rescomp statefun"

consts
  fuseTemp :: "temporal"
  copyTemp :: "temporal"
  fuse_sup :: "temporal"
  fuse_exe :: "temporal"
  copy_sup :: "temporal"
  copy_exe :: "temporal"
  super :: "temporal"
  env :: "temporal"
  init :: "temporal"
  program :: "temporal"

consts ex_fuse0Exe :: "HoVal \<Rightarrow> HoVal \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"
defs
 ex_fuse0Exe: "ex_fuse0Exe v1 v2 h \<equiv> exe (bindVarFunN fuse0 [v1,v2]) h (snd (funTab fuse0))"

consts ex_fuse0_exe :: "temporal"
defs 
 ex_fuse0_exe: "ex_fuse0_exe \<equiv> TEMP
(if $pc \<noteq> #pc_fuse
 then Unchanged (w1,w2,fuse_res,fuse_st)
 else pc$ = #pc_copy \<and> 
 (if $fuse_st = #Blocked
   then Unchanged (w1,w2,fuse_res,fuse_st,heap)
   else 
 (if checkAvail2<$w1,$w2>
   then (w1,w2)$ = #(BOTTOM,BOTTOM)
    \<and> (fuse_res,heap,rs)$ = ex_fuse0Exe<$w1,$w2,$heap>
   else Unchanged (w1,w2,fuse_res,fuse_st,heap)
     \<and>  fuse_res$ = #BOTTOM 
     \<and>  fuse_st$ = #Matchfail)))"

(* ----------------------------------------------------------------------------- *)
(* test checkAvail/bindVar with one branch of the copy function *)

constdefs copyTrueAss :: "VarName list \<Rightarrow> Assertion"
 "copyTrueAss zs == (\<lambda> E h h' v p. (\<forall> n r0 r U . (E (hd zs) = (R (Ref r0)) \<and> E (hd (tl zs)) = (R (Ref r)) \<and> (n,r,U,h):mList  
                                             \<longrightarrow> (\<exists> rv Uv. v = R (Ref rv) \<and> (Suc n,rv,Uv,h'):mList))))"

constdefs copyTrue9Ass :: "VarName list \<Rightarrow> Env \<Rightarrow> Assertion"
 "copyTrue9Ass zs E' == (\<lambda> E h h' v p. (\<forall> n r0 r U . E=E' \<and> (E (hd zs) = (R (Ref r0)) \<and> E (hd (tl zs)) = (R (Ref r)) \<and> (n,r,U,h):mList  
                                             \<longrightarrow> (\<exists> rv Uv. v = R (Ref rv) \<and> (Suc n,rv,Uv,h'):mList))))"

constdefs copyTrueBody :: "VarName list \<Rightarrow> Exp"
"copyTrueBody zs == (ConstrExp CONS zs)"

(* expression-level assertion *)
lemma copyTrueAss_ok: "{} \<rhd> copyTrueBody [hvar,tvar] : copyTrueAss [hvar,tvar]"
apply (simp add: copy_funTabBody copyTrueBody_def copyTrueAss_def)
apply (rule vdmConseq)
prefer 2
apply (rule vdmConstr)
apply clarsimp
apply (subgoal_tac "h = h'a")
 prefer 2 apply (rule ConstrInd_h_same) apply assumption
apply clarsimp
(* apply (subgoal_tac "\<exists> rr0. rs = [rr0, Ref r]") *)
apply (rule_tac x="U \<union> {freshloc (fmap_dom h'a)}" in exI)
apply (rule mListCONS)
apply clarsimp
prefer 4
apply (rule mListSameH)
apply assumption
apply (simp add: sameH_def)
apply clarsimp
apply (case_tac "l \<noteq> freshloc (fmap_dom h'a)")
 apply (simp add: FMAPlookup1)
 (* OTW *)
 apply simp
 apply (frule mListDom)
 apply (subgoal_tac "l \<in> fmap_dom h'a")
  prefer 2 apply fastsimp
 apply (subgoal_tac "finite (fmap_dom h'a)")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply simp
prefer 2
 apply (frule mListDom)
 apply (subgoal_tac "finite (fmap_dom h'a)")
  prefer 2 apply (rule finite_dom_fmap)
 apply (frule freshloc)
 apply fastsimp
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
apply (erule ConstrInd.cases)
apply simp_all
done

lemma copyTrue9Ass_ok: "\<lbrakk> (E', R (Ref r)) = boxMatch E h (PCon CONS [hvar,tvar]) r \<rbrakk> 
                        \<Longrightarrow> {} \<rhd> copyTrueBody [hvar,tvar] : copyTrue9Ass [hvar,tvar] E'"
sorry

lemma aux11: "\<lbrakk> (E', R (Ref r)) = boxMatch E h (PCon CONS [hvar,tvar]) r ; (n,r,U,h):mList \<rbrakk>
              \<Longrightarrow> (\<exists> n r0 r U . (E' hvar = (R r0) \<and> E' tvar = (R (Ref r)) \<and> (n,r,U,h):mList))"
apply (simp only: boxMatch pattMatch1 makeEnv1)
apply clarsimp
apply (erule semmatch.cases)
apply simp_all
apply clarsimp
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply clarsimp
oops

lemma testV1: "\<lbrakk> (E', R (Ref r)) = boxMatch E h (PCon CONS [hvar,tvar]) r ; (n,r,U,h):mList ; (v,h',p) = exe E' h (copyTrueBody [hvar,tvar]) ; E' tvar = R (Ref r) ; {} \<rhd> copyTrueBody [hvar,tvar] : copyTrue9Ass [hvar,tvar] E' \<rbrakk>
               \<Longrightarrow> (\<exists> rv Uv. v = R (Ref rv) \<and> (Suc n,rv,Uv,h'):mList)"

apply (simp only: boxMatch pattMatch1 makeEnv1)
apply clarsimp
apply (erule semmatch.cases)
apply simp_all
apply clarsimp
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply (erule PConInd.cases)
apply simp_all
apply clarsimp

apply (insert copyTrue9Ass_ok)
apply (rule vdmexe)
apply simp
apply (rule vdmConseq)
prefer 2
apply assumption
apply clarsimp

apply (rotate_tac 1)
apply (erule thin_rl) (* exe *)
apply (rotate_tac -3)
apply (erule thin_rl) (* \<rhd> *)
apply (rotate_tac -2)
apply (erule thin_rl) (* \<rhd> *)

apply (simp add: copyTrue9Ass_def)
apply clarsimp
apply (erule_tac x="n" in allE)
apply (rotate_tac -1)
apply (erule impE)
apply (rule conjI)
apply (rule_tac x="r" in exI)

oops

lemma testV2: "\<lbrakk> E' = bindVarFunN copy [R (Ref r)] ; checkAvailN [(R (Ref r))]; (v,h',p) = exe E' h (copyTrueBody [hvar,tvar])  \<rbrakk>
               \<Longrightarrow> (\<forall> r0 r U . (E' hvar = (R (Ref r0)) \<and> E' tvar = (R (Ref r)) \<and> (n,r,U,h):mList \<longrightarrow> 
                       (\<exists> rv Uv. v = R (Ref rv) \<and> (Suc n,rv,Uv,h'):mList)))"
(* apply (simp add:  pattMatch01_def makeEnv01_def) *)
apply (insert copyTrueAss_ok)
apply (drule vdmexe)
apply assumption
(*
apply (rule vdmConseq)
prefer 2
apply assumption
apply clarsimp
*)
apply (rotate_tac 2)
(* apply (erule thin_rl) (* exe *) *)
apply (erule thin_rl) (* \<rhd> *)
apply (simp only: copyTrueAss_def)
apply clarsimp
done

(* --------------------------------------------------------------------------- *)
(* unproven lemmas from ST.thy for coordination-level of ListCopy *)

constdefs list_in :: "Heap \<Rightarrow> HoVal \<Rightarrow> nat \<Rightarrow> bool"
"list_in h w n \<equiv> (\<exists> r U . w = (R (Ref r)) \<and> (n,r,U,h) \<in> mList)"

lemma proven1_as_in_toProve1: "\<lbrakk>list_in h v n; (a,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' v n"
apply (insert copy0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
apply (erule_tac x="n" in allE)
apply (erule_tac x="U" in allE)
apply (erule impE)
apply simp
apply clarsimp
apply (rule_tac x="U" in exI)
apply clarsimp
done

lemma proven2_as_in_toProve2: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' nv n"
apply (insert copy0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
apply (erule_tac x="n" in allE)
apply (erule_tac x="U" in allE)
apply (erule impE)
apply simp
apply clarsimp
apply (rule_tac x="Uv" in exI)
apply clarsimp
done

lemma proven3_as_in_toProve3: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> nv \<noteq> BOTTOM"
apply (insert copy0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copy0Ass_def  list_in_def)
apply clarsimp
done

(* NB: this has to use fuse0, otw the size of the result will not be exactly n! *)
lemma proven4_as_in_toProve4: "\<lbrakk>list_in h v1 n; list_in h v2 n; 
                 (v,h',x) = exe (emptyEnv(x0arg := v1, x1arg := v2)) h fuse0Body \<rbrakk>
  \<Longrightarrow> list_in h' v n"
apply (insert fuse0Body_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 2)
apply (erule thin_rl)
apply (simp add: fuse0Ass_def list_in_def)
done


(*
lemma copyBody_resources_ok: "{}  \<rhd> copyBody : copyAss [xarg]"
apply (simp add: copyAss_def)
apply (insert copy_resources_ok)
sorry

lemma toProve1: "\<lbrakk>list_in h v n; (a,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' v n"
sorry

(* NB: this is the preferred version, because it matches the already proven copy_resources_ok (a function call) *)
lemma proven2: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xin := v)) h (CallFunExp copy [xin]) \<rbrakk> \<Longrightarrow> list_in h' nv n"
apply (insert copy_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copyAss_def)
apply (erule_tac x="n" in allE)
apply (erule impE)
apply (simp add: copyAss_def list_in_def)
apply (simp add: copyAss_def list_in_def)
done

lemma toProve2: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> list_in h' nv n"
apply (insert copyBody_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copyAss_def list_in_def)
done

lemma toProve3: "\<lbrakk>list_in h v n; (nv,h',c) = exe (emptyEnv(xarg := v)) h copyBody\<rbrakk> \<Longrightarrow> nv \<noteq> BOTTOM"
apply (insert copyBody_resources_ok)
apply (drule vdmexe)
apply simp
apply (rotate_tac 1)
apply (erule thin_rl)
apply (simp add: copyAss_def list_in_def)
apply clarsimp
done
*)

(*
lemma fuseBody_resources_ok: "{}  \<rhd> fuseBody : fuse0Ass [x0arg,x1arg]"
apply (simp add: fuse0Ass_def)
apply (insert copy_resources_ok)
sorry
*)


(* -------------------------------------------------- *)

consts locs :: "Locn \<Rightarrow> Heap \<Rightarrow> Locn set"
defs
  locs: "locs l h \<equiv> \<Union>{x. \<exists> n. (n,l,x,h) \<in> mList}"

consts
  heap_restrict :: "Heap \<Rightarrow> (Locn set) \<Rightarrow> Heap"
  loc_restrict :: "Heap \<Rightarrow> Locn \<Rightarrow> Heap"
  HoVal_restrict :: "Heap \<Rightarrow> HoVal \<Rightarrow> Heap"

defs
  heap_restrict: "heap_restrict h ls \<equiv> fimap ((themap h) |` ls)"
  loc_restrict: "loc_restrict h l \<equiv> heap_restrict h (locs l h)"

lemma "fmap_lookup h l = Some (I n) \<longrightarrow> fmap_dom (loc_restrict h l) = {}"
  oops

lemma "fmap_lookup h l = Some (R (Ref r)) \<longrightarrow> fmap_dom (loc_restrict h l) \<noteq> {}"
  oops

(* show a property for pre- and post-heap on box level *)
lemma vdmexe3: "\<forall> E h h' v p e P. ({} \<rhd> e : P) \<and> ((v,h',p) = exe E h e) \<longrightarrow>  P E h h' v p"
  by (simp add: vdmexe)

(*
lemma exebody1: "(v',h',r) = exebody (I n) h \<Longrightarrow> v' = I (n+1)"
  apply (auto simp add: exebody)
  apply (insert vdmexe2[OF body1])
  apply auto
  done
*)

constdefs shallowCopy :: "Heap \<Rightarrow> Locn set \<Rightarrow> Heap"
"shallowCopy h s == heap_restrict h s"

consts copyBody :: Exp
consts fuseBody :: Exp

consts
 copyExe :: "HoVal \<Rightarrow> Locn set \<Rightarrow> Locn set \<Rightarrow> Heap \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"
(* now using a shallowcopy *)
defs
  copyExe: "copyExe v' Uin Uout hwire hbox \<equiv>
    let hbox = shallowCopy hwire Uin in 
    let (v,h',rho) = exe (emptyEnv(xarg := v')) hbox copyBody in
    (v, shallowCopy h' (Uin \<union> Uout), rho)"

consts 
 mkHeap :: "Locn set \<Rightarrow> Heap \<Rightarrow> Heap"
defs
 mkHeap: "mkHeap U h \<equiv> fimap (\<lambda> l . if (l:U) then fmap_lookup h l else fmap_lookup emptyfinmap l)"


(* show a property for the copy box; this consists of 
   . defining heap-change from wire to box-heap
   . proving a property for the expresion level, over pre- and post-heap on box level
   . defining heap-change for box- to wire-heap
*)

(* copied from ListCopy-NEW.thy *)
constdefs copyAss :: "[VarName list] => Assertion"
"copyAss zs == (\<lambda> E h h' v p. 
                                 (\<forall> n r U . (E (hd zs) = (R (Ref r)) \<and> (n,r,U,h):mList  
                                             \<longrightarrow> (card (fmap_dom h') \<le> card (fmap_dom h) + n + 1) \<and> (\<exists> rv Uv. v = R (Ref rv) \<and> (n,rv,Uv,h'):mList))))"

constdefs fuseAss :: "[VarName list] => Assertion"
"fuseAss zs == \<lambda>E h h' v p.
          			   \<forall>n0 r0 U0 n1.
          			      E (hd zs) = R (Ref r0) \<and> (n0, r0, U0, h) \<in> mList \<and> (\<exists>r1. E (hd (tl zs)) = R (Ref r1) \<and> (\<exists>U1. (n1, r1, U1, h) \<in> mList) \<and> n0 = n1) \<longrightarrow>
          			      card (fmap_dom h') \<le> card (fmap_dom h) + n0 + 1 \<and> (\<exists> nv rv Uv. v = R (Ref rv) \<and> (nv, rv, Uv, h') \<in> mList \<and> nv \<le> n0 ) "


(* HWL: current working version of encoding the restriction of the heap to just
        the live data at the end of the box;
        wh, wh' are the wireheaps, seen at coordination level
        h, h' are the box heaps, used in the expression level proof
*)

lemma vdmexe4[rule_format]: "\<forall> E wh wh' wv f x P. ({} \<rhd> CallFunExp f [x] : (\<lambda> E h h' v p . \<forall> n1 r1 U1 n2 r2 U2 . E x = (R (Ref r1)) \<and> (n1,r1,U1,h):mList \<and> h = mkHeap U1 wh \<and> v = R (Ref r2) \<and> (n2,r2,U2,h):mList) \<and> wh' = mkHeap U2 h' \<and> v = wv \<and> wp = p \<longrightarrow> P E h h' wv wp)
                \<longrightarrow>  P E wh wh' wv wp"
sorry

consts
  fuseExe :: "HoVal \<Rightarrow> HoVal \<Rightarrow> Locn set \<Rightarrow> Locn set \<Rightarrow> Locn set \<Rightarrow> Heap \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"
(* now using a shallowcopy *)
defs
  fuseExe: "fuseExe v1 v2 Uin1 Uin2 Uout hwire hbox \<equiv>
    let hbox = shallowCopy (shallowCopy hwire Uin1) Uin2 in 
    let (v,h',rho) = exe ((emptyEnv(x0arg := v1))(x1arg := v2)) hbox fuseBody in
    (v, shallowCopy h' Uout, rho)"

(* trying to phrase copy/fuse properties on coordination level *)
(* structure for each box: \<exists> r n \<dots> s.t. input is well-formed and of length n \<and>  output is bounded by n' *)
(* the 2 clauses need to be wrapped into conditionals checking for pc and availability of input;
   if both are true than the property below should hold *)

(* GG comment: I think too much of the VDM stuff is copied into the coordination *)
(*    will be extremely hard to reason of properties intended for VDM in TLA *)


(* a simpler version of coordination layer *)
(* here one heap (think we agreed to use one heap only???) *)
(* all update on heap in the VDM bit *)
(* output buffer of box and wires represented as HoVal. These
   references are the only ones to be changed in coordination layer *)

(* function to check if output can be asserted *)
consts 
  assertOut :: "HoVal \<Rightarrow> HoVal \<Rightarrow> bool"
defs
  asserOut: "assertOut x y \<equiv> (x = BOTTOM) \<or> (y = BOTTOM)"

(* function to copy output values, i.e. pick the non-BOTTOM version *)
consts
  nwire :: "HoVal \<Rightarrow> HoVal \<Rightarrow> HoVal"
defs
  nwire: "nwire x y \<equiv> (if x = BOTTOM then y else x)"

datatype PC = pc_fuse | pc_copy

consts
  pc :: "PC statefun" (* the program counter - sequential execution of boxes *)
  fuse_st :: "BoxState statefun" (* state/mode of box *)
  copy_st :: "BoxState statefun" (* state/mode of box *)
  fuse_res :: "HoVal statefun"
  copy_res1_old :: "(HoVal*HoVal) statefun"
  copy_res1 :: "HoVal statefun" (* see my comment on HACK below -- HWL *)
  copy_res2 :: "HoVal statefun"
  heap :: "Heap statefun"
  w1 :: "HoVal statefun" (* input copy *)
  w2 :: "HoVal statefun" (* input fuse *) 
  w3 :: "HoVal statefun" (* input fuse *)
  rs :: "Rescomp statefun"

consts
  fuse :: "temporal"
  copy :: "temporal"
  fuse_sup :: "temporal"
  fuse_exe :: "temporal"
  copy_sup :: "temporal"
  copy_exe :: "temporal"
  super :: "temporal"
  init :: "temporal"
  program :: "temporal"

(* the link to expression layer *)
consts
 simplefuseExe :: "HoVal \<Rightarrow> HoVal \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"
 simplecopyExe2 :: "HoVal \<Rightarrow> Heap \<Rightarrow> HoVal*Heap*Rescomp"
 (* also returns old value *)
 simplecopyExe :: "HoVal \<Rightarrow> Heap \<Rightarrow> HoVal*HoVal*Heap*Rescomp"

(* HWL: I expect simplecopyExe to produce just the copied list; the input should
        be directly forwarded from the input-wire to the other output wire
        as sketched in the HACK in super at the end of the file *)

(* For the InvOneHeap invariant below, I expect an instance of exe as interface:
    it should require only local reasoning (that something is a list), but no
    reasoning over the entire heap; for that, I think vdmexe3 should be sufficient
*)
defs
 simplefuseExe: "simplefuseExe v1 v2 h \<equiv> exe ((emptyEnv(x0arg := v1))(x1arg := v2)) h fuseBody"
 simplecopyExe2: "simplecopyExe2 v h \<equiv> exe (emptyEnv(xarg := v)) h copyBody"
 simplecopyExe: "simplecopyExe v h \<equiv> (v,simplecopyExe2 v h)"

(*  For the InvMultiHeap invariant below, a boxexe function will be needed, which also
    models the shallow copying of input and ouput; I'm still working on this version
    (there is a first instance,from ListCopy-NEW.thy, below

axioms
 boxexe: "(boxexe E wh e = (v,wh',p)) = (\<exists> h h' n1 r1 U1 n2 r2 U2 . E xin = (R (Ref r1)) \<and> (n1,r1,U1,h):mList \<and> h = mkHeap U1 wh \<and> v = R (Ref r2) \<and> (n2,r2,U2,h'):mList \<and> wh' = mkHeap U2 h' \<and> (E, h \<turnstile> e \<Down> (v,h',p)))"
*)

defs 
 fuse_exe: "fuse_exe \<equiv> TEMP
(if $pc \<noteq> #pc_fuse
 then Unchanged (w2,w3,fuse_res,fuse_st)
 else pc$ = #pc_copy \<and> 
 (if $fuse_st = #Blocked
   then Unchanged (w2,w3,fuse_res,fuse_st,heap)
   else 
 (if $w2 = #BOTTOM \<or> $w3 = #BOTTOM
   then Unchanged (w2,w3,fuse_res,fuse_st,heap)
     \<and>  fuse_res$ = #BOTTOM 
     \<and>  fuse_st$ = #Matchfail
   else (w2,w3)$ = #(BOTTOM,BOTTOM)
    \<and> (fuse_res,heap,rs)$ = simplefuseExe<$w2,$w3,$heap>)))"
defs
 copy_exe: "copy_exe \<equiv> TEMP
(if $pc \<noteq> #pc_copy
 then Unchanged (w1,copy_res1,copy_res2,copy_st)
 else pc$ = #pc_copy \<and> 
 (if $copy_st = #Blocked
   then Unchanged (w1,copy_res1,copy_res2,copy_st,heap)
   else 
 (if $w1 = #BOTTOM
   then Unchanged (w1,copy_res1,copy_res2,copy_st,heap)
     \<and>  (copy_res1,copy_res2)$ = #(BOTTOM,BOTTOM) 
     \<and>  fuse_st$ = #Matchfail
   else w1$ = #BOTTOM
    \<and> (copy_res1,copy_res2,heap,rs)$ = simplecopyExe<$w1,$heap>)))"

defs
 super: "super \<equiv> TEMP $sch = #Super \<longrightarrow> heap$ = $heap"
defs
 fuse_sup: "fuse_sup \<equiv> TEMP
   (if assertOut<$fuse_res,$w3>
    then fuse_st$ = #Blocked \<and> Unchanged (w3,fuse_res)  
    else fuse_st$ = #Runnable \<and>  fuse_res$ = #BOTTOM
     \<and> w3$ = nwire<$fuse_res,$w3>)"
defs
 copy_sup: "copy_sup \<equiv> TEMP
   (if assertOut<$copy_res1,$w1> \<and> assertOut<$copy_res2,$w2> 
    then copy_st$ = #Blocked \<and> Unchanged (w1,w2,copy_res1,copy_res2)  
    else copy_st$ = #Runnable \<and> (copy_res1,copy_res2)$ = #(BOTTOM,BOTTOM)
     \<and> w1$ = nwire<$copy_res1,$w1>
     \<and> w2$ = nwire<$copy_res2,$w2>)"
(* WAS:
     \<and> w1$ = nwire<fst<$copy_res>,$w1>
     \<and> w2$ = nwire<fst<$copy_res>,$w2>)"
*)
(* FIX this
defs
  fuseTemp: "fuseTemp \<equiv> TEMP ($sch = #Execute \<longrightarrow> fuse_exe) \<and> ($sch = #Super \<longrightarrow> fuse_sup)"
  copyTemp: "copyTemp \<equiv> TEMP ($sch = #Execute \<longrightarrow> copy_exe) \<and> ($sch = #Super \<longrightarrow> copy_sup)"
*)
consts
 copyTemp :: "temporal"
 fuseTemp :: "temporal"

constdefs
  S' :: "temporal"
  "S' \<equiv> TEMP (sch$ = (if isExe \<and> $pc = #pc_copy then #Super else #Execute))
           \<and> (isSup \<longrightarrow> pc$ = #pc_fuse)"

consts
  H :: "Heap"
  L :: "Locn"
  N :: "nat"  (* renamed this for convenience -- HWL *)
  Locs :: "Locn set"

axioms
  initHeap: "(N,L,Locs,H) \<in> mList"
  initHeapEmpty: "card (fmap_dom H) = N+1" (* or that remaining part of heap is empty??? *)
 
defs
  init: "init \<equiv> TEMP
    $sch = #Execute 
  \<and> $pc = #pc_fuse
  \<and> $heap = #H
  \<and> $(fuse_st,copy_st) = #(Runnable,Runnable)
  \<and> $(fuse_res,copy_res1,copy_res2) = #(BOTTOM,BOTTOM,BOTTOM)
  \<and> $(w1,w2,w3) = #(BOTTOM,BOTTOM,R(Ref L))"

defs
  program: "program \<equiv> TEMP init 
  \<and> \<box>[S' \<and> fuseTemp \<and> copyTemp \<and> super]_(sch,pc,w1,w2,w3,heap,fuse_st,fuse_res,copy_st,copy_res1,copy_res2,rs)"

lemmas hume_rews =
  intensional_rews all_after_eq all_before_eq
  after_before next_all_eq toBeforeVal toAfterVal
  isExe_def isSup_def

lemma vdmexe2: "\<lbrakk>{} \<rhd> e : P; (v,h',p) = exe E h e\<rbrakk> 
  \<Longrightarrow> P E h h' v p"
  by (simp add: vdmexe)

lemma vdmexe3: "\<forall> E h h' v p e P. ({} \<rhd> e : P) \<and> ((v,h',p) = exe E h e) \<longrightarrow>  P E h h' v p"
  by (simp add: vdmexe)

constdefs list_in :: "Heap \<Rightarrow> HoVal \<Rightarrow> nat \<Rightarrow> bool"
"list_in h w n \<equiv> (\<exists> r U . w = (R (Ref r)) \<and> (n,r,U,h) \<in> mList)" 

consts InvOneHeap :: "temporal"
defs 
  InvOneHeap: "InvOneHeap \<equiv> TEMP 
    (list_in<$heap,$w3,#N> \<and> $w1 = #BOTTOM \<and> $w2 = #BOTTOM 
     \<and> $fuse_res = #BOTTOM \<and> $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM)
\<or>  ($w3 = #BOTTOM \<and> list_in<$heap,$w1,#N>  \<and> list_in<$heap,$w2,#N>
     \<and> $fuse_res = #BOTTOM \<and> $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM)
\<or>  ($w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM 
     \<and> list_in<$heap,$fuse_res,#N> \<and>  $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM)
\<or>  ($w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM 
     \<and> $fuse_res = #BOTTOM \<and> list_in<$heap,$copy_res1,#N> \<and> list_in<$heap,$copy_res2,#N>)"


consts InvOneHeapImp :: "temporal"
defs 
  InvOneHeapImp: "InvOneHeapImp \<equiv> TEMP 
    ($w3 \<noteq> #BOTTOM \<longrightarrow> list_in<$heap,$w3,#N> \<and> $w1 = #BOTTOM \<and> $w2 = #BOTTOM 
     \<and> $fuse_res = #BOTTOM \<and> $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM)
\<and>  ($w1 \<noteq> #BOTTOM \<and> $w2 \<noteq> #BOTTOM \<longrightarrow> $w3 = #BOTTOM \<and> list_in<$heap,$w1,#N>  \<and> list_in<$heap,$w2,#N>
     \<and> $fuse_res = #BOTTOM \<and> $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM)
\<and>  ($fuse_res \<noteq> #BOTTOM \<longrightarrow> $w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM 
     \<and> list_in<$heap,$fuse_res,#N> \<and>  $copy_res1 = #BOTTOM \<and> $copy_res2 = #BOTTOM)
\<and>  ($copy_res1 \<noteq> #BOTTOM \<and> $copy_res2 \<noteq> #BOTTOM \<longrightarrow> $w1 = #BOTTOM \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM 
     \<and> $fuse_res = #BOTTOM \<and> list_in<$heap,$copy_res1,#N> \<and> list_in<$heap,$copy_res2,#N>)"


lemma list_in_initHeap: "list_in H (R (Ref L)) N"
  apply (unfold list_in_def)
  apply (rule_tac x=L in exI)
  apply (rule_tac x="Locs" in exI)
  apply simp
  apply (rule initHeap)
  done

lemma inv1: "\<turnstile> program \<longrightarrow> \<box>InvOneHeapImp" 
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply (simp only: InvOneHeapImp)
   apply int_unlift
   apply (rule impI)
   apply (simp only: hume_rews)
   apply simp
   apply (rule list_in_initHeap)
   oops

lemma inv2: "\<turnstile> program \<longrightarrow> \<box>InvOneHeap" 
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply (simp only: InvOneHeap)
   apply int_unlift
   apply (rule impI)
   apply (elim conjE)
   apply (rule disjI1)
   apply (simp only: hume_rews)
   apply simp
   apply (rule list_in_initHeap)

sorry





(* OLD STUFF *)

lemma aux1:  "\<turnstile> program \<longrightarrow> \<box>(($w1 = #BOTTOM)=($w2=#BOTTOM))"
  apply (unfold program)
   apply (rule invmono)
   apply (unfold init)
   apply int_unlift
   apply (simp only: hume_rews)
   apply simp
   apply (rule preimpsplit)
   apply (simp_all only: hume_rews)
   prefer 2
  apply (simp only: unch_eq1, simp add: unch_def)
  apply pref_unlift
   apply (case_tac "w \<Turnstile> $sch")
   prefer 2
   apply (simp add: isExe_def isSup_def super copy copy_sup)
   apply (case_tac "(assertOut (w \<Turnstile> $copy_res) (w \<Turnstile> $w1)) \<and> (assertOut (w \<Turnstile> $copy_res) (w \<Turnstile> $w2))")
  apply (simp only: unch_eq1, simp add: unch_def)
  apply (simp only: unch_eq1, simp add: unch_def)
  (* need nwire rules and result buffer rules *)
  (* need two result buffers!!! *)
  oops

  apply simp
  apply 
  apply (simp add: isExe_def isSup_def)
   apply (case_tac "w \<Turnstile> $pc")
   apply (simp add: odd, simp only: unch_eq1, simp add: unch_def)
   (* even step *)
   apply (simp only: even)
    apply (case_tac "(w \<Turnstile> $even_st) = Blocked")
   apply (simp only: unch_eq1, simp add: unch_def)
    apply (case_tac "(w \<Turnstile> $w2) = BOTTOM")
   apply (simp only: unch_eq1, simp add: unch_def)
   apply simp
   apply (intro impI,elim conjE)
   apply (simp only: evenoddExe Let_def)
   apply (simp only: hume_rews)
   apply clarsimp
   apply (subgoal_tac "\<forall> i r . ((emptyEnv(xin := I n)) xin = (I i))  \<longrightarrow> (w \<Turnstile> even_res$) = (I (i+1))")
   prefer 2
   apply (insert vdmexe3)
   apply (drule_tac x = "emptyEnv(xin := I n)" in spec)
   apply (drule_tac x = "w \<Turnstile> $evenheap" in spec)
   apply (drule_tac x = "w \<Turnstile> evenheap$" in spec)
   apply (drule_tac x = "w \<Turnstile> even_res$" in spec)
   apply (drule_tac x = "w \<Turnstile> rso1$" in spec)
   apply (drule_tac x = "incbodyI" in spec)
   apply (drule_tac x = "(\<lambda> E h hh v p. (\<forall> i r . (E xin = (I i))  \<longrightarrow> v = (I (i+1))))" in spec)
   apply (insert incbodyI_spec)
   apply simp
   apply simp
   (* odd step !!! *)
   apply (simp only: even odd)
    apply (case_tac "(w \<Turnstile> $odd_st) = Blocked")
   apply (simp only: unch_eq1, simp add: unch_def)
    apply (case_tac "(w \<Turnstile> $w1) = BOTTOM")
   apply (simp only: unch_eq1, simp add: unch_def)
   apply simp
   apply (intro impI,elim conjE)
   apply (simp only: evenoddExe Let_def)
   apply (simp only: hume_rews)
   apply clarsimp
   apply (subgoal_tac "\<forall> i r . ((emptyEnv(xin := I n)) xin = (I i))  \<longrightarrow> (w \<Turnstile> odd_res$) = (I (i+1))")
   prefer 2
   apply (insert vdmexe3)
   apply (drule_tac x = "emptyEnv(xin := I n)" in spec)
   apply (drule_tac x = "w \<Turnstile> $oddheap" in spec)
   apply (drule_tac x = "w \<Turnstile> oddheap$" in spec)
   apply (drule_tac x = "w \<Turnstile> odd_res$" in spec)
   apply (drule_tac x = "w \<Turnstile> rso2$" in spec)
   apply (drule_tac x = "incbodyI" in spec)
   apply (drule_tac x = "(\<lambda> E h hh v p. (\<forall> i r . (E xin = (I i))  \<longrightarrow> v = (I (i+1))))" in spec)
   apply (insert incbodyI_spec)
   apply simp
   apply (simp only: unch_eq1, simp add: unch_def)
   done



oops


(* maybe these need to be proved at the same time + with result buffers!!! *)
lemma aux2: "\<turnstile> program \<longrightarrow> \<box>(($w2 \<noteq> #BOTTOM) \<longrightarrow> ($w1 = #BOTTOM))"
oops
lemma aux2: "\<turnstile> program \<longrightarrow> \<box>(($w1 \<noteq> #BOTTOM) \<longrightarrow> ($w2 = #BOTTOM))"
oops

(* think the main theorem (heap usage) would be something like this.
   Obviously, several auxiliary lemmas are required *)


theorem main: "\<turnstile> program \<longrightarrow> \<box>(card<fmap_dom<$heap>> \<le> #(2*N+2))"

oops


(* GG note: syntax error when I run this *)
(* HWL: I just sketched the properties, and didn't type check it, sorry *)
(*
defs
 copyTemp: "copyTemp \<equiv> TEMP \<dots> *)
                 (* spec for copy: if input available and Runnable then \<dots>*)
 (*                (\<exists> r n U. $w1 = #(R (Ref r)) \<and> (n,r,U,$wireheap) \<in> mList \<and>
                   (copy_res,copyheap,rso1)$ = copyExe<$w1,$U,$Uv,$wireheap,$copyheap>)"
 fuseTemp: "fuseTemp \<equiv> TEMP \<dots> *)
                 (* spec for fuse: if input available and Runnable then \<dots>*)
     (*            ((\<exists> n0 r0 U0 n1 r1 U1. $w2 = #(R (Ref r0)) \<and> (n0, r0, U0, $wireheap) \<in> mList \<and> $w3 = #(R (Ref r1)) \<and> (n1, r1, U1, $wireheap) \<in> mList) \<and> n0 = n1) \<and> 
                  (fuse_res,fuseheap,rso2)$ = fuseExe<$w2,$w3,$U0,$U1,$Uv,$wireheap,$fuseheap>) "

defs
 super: "super \<equiv> TEMP \<dots> *)
            (* 1 case of transfer: copy \<rightarrow> fuse *)
      (*      \<and> (if $w2 = #BOTTOM \<and> $w3 = #BOTTOM \<and> $copy_res \<noteq> #BOTTOM
                 then fuse_st$ = #Runnable \<and> w2$ = $w1 \<and> copy_res$ = #BOTTOM \<and> w3$ = $copy_res \<and> copy_res$ = #BOTTOM ) *)
            (* HACK: 1st output, w2$, is encoded as pass-through of input, 2nd output, w3$, is result
                     of the expression level copy function *) (*" ) *)


(* GG note: the invariant must be a state invariant. All variables must be on the form $x (i.e. not e.g. y$).
   Thus, pre-post conditions cannot be expressed. This must be modelled as actions \<box>[Pre \<and> funcall \<longrightarrow> Post]_v,
   but we should focus on invariants. 
*)

(* list-predicate, parameterised with r as root location of the list *)
constdefs list_at :: "Heap \<Rightarrow> Locn \<Rightarrow> nat \<Rightarrow> bool"
"list_at h r n \<equiv> (\<exists> U . (n,r,U,h) \<in> mList)" 

(* list-predicate, parameterised with the (expression-level) variable *)
constdefs list_var :: "Env \<Rightarrow> Heap \<Rightarrow> VarName \<Rightarrow> nat \<Rightarrow> bool"
"list_var E h x n \<equiv> (\<exists> r U .E x = (R (Ref r)) \<and> (n,r,U,h) \<in> mList)" 

syntax
 "_list_var" :: "Env \<Rightarrow> Heap \<Rightarrow> VarName \<Rightarrow> nat \<Rightarrow> bool" ("(_, _ [\<turnstile>] _, _)" [40,40,40,40] 90)

translations
 "_list_var"     \<rightleftharpoons> "list_var" 

(* list-predicate, parameterised with the (coordination-level) variable *)
constdefs list_in :: "state \<Rightarrow> Heap \<Rightarrow> HoVal statefun \<Rightarrow> nat \<Rightarrow> bool"
"list_in s h w n \<equiv> (\<exists> r U . w s = (R (Ref r)) \<and> (n,r,U,h) \<in> mList)" 

(* list-predicate, parameterised with the (coordination-level) variable *)
constdefs list_in2 :: "Heap \<Rightarrow> HoVal \<Rightarrow> nat \<Rightarrow> bool"
"list_in2 h w n \<equiv> (\<exists> r U . w = (R (Ref r)) \<and> (n,r,U,h) \<in> mList)" 

syntax
 "_list_in" :: "state \<Rightarrow> Heap \<Rightarrow> HoVal statefun \<Rightarrow> nat \<Rightarrow> bool" ("(_, _ [\<turnstile>] _, _)" [40,40,40,40] 90)

translations
 "_list_in"     \<rightleftharpoons> "list_in" 

(* slightly rewritten, to use the defs above; use either these, or the versions in ListCopy-NEW.thy for the proof *)
constdefs copyAssAgain :: "[VarName list] \<Rightarrow> Assertion"
"copyAssAgain zs == (\<lambda> E h h' v p. 
                  (\<forall> n . E, h [\<turnstile>] (hd zs), n
                           \<longrightarrow> (* (card (fmap_dom h') \<le> card (fmap_dom h) + n + 1) \<and> NOT USED *)
                               (\<exists> rv . v = R (Ref rv) \<and> list_at h' rv n)))"

constdefs fuseAssAgain :: "[VarName list] \<Rightarrow> Assertion"
"fuseAssAgain zs == \<lambda>E h h' v p.
	   \<forall>n .
              (E, h [\<turnstile>] (hd zs), n \<and> E, h [\<turnstile>] (hd (tl zs)), n ) \<longrightarrow>
              (* card (fmap_dom h') \<le> card (fmap_dom h) + n + 1 \<and> NOT USED *) 
              (\<exists> rv. v = R (Ref rv) \<and> list_at h' rv n) "

constdefs fuse0Ass :: "[VarName list] => Assertion"
"fuse0Ass zs == \<lambda>E h h' v p.
          		\<forall>n0 n1.
          		   ((\<exists> r0 U0. E (hd zs) = R (Ref r0) \<and> (n0, r0, U0, h) \<in> mList) \<and> (\<exists> r1 U1. E (hd (tl zs)) = R (Ref r1) \<and> (n1, r1, U1, h) \<in> mList) \<and> n0 = n1) \<longrightarrow>
          		    card (fmap_dom h') \<le> card (fmap_dom h) + n0 + 1 \<and> (\<exists> nv rv Uv. v = R (Ref rv) \<and> (nv, rv, Uv, h') \<in> mList \<and> nv = n0 ) "

lemma copyAgain_resources_ok: "{} \<rhd> (CallFunExp copy [xin]) : copyAssAgain [xin]"
sorry

(* uses ineq over list sizes *)
lemma fuseAgain_resources_ok: "{} \<rhd> (CallFunExp fuse [x0in,x1in]) : fuseAssAgain [x0in,x1in]"
sorry

(* equal list sizes; PREFERRED version *)
lemma fuse0Again_resources_ok: "{} \<rhd> (CallFunExp fuse [x0in,x1in]) : fuse0Ass [x0in,x1in]"
sorry

consts InvOneHeap :: "temporal"
defs 
  InvOneHeap: "InvOneHeap \<equiv> TEMP 
                  (list_in2<$heap,$w1,#N> \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM)
               \<or>  ($w1 = #BOTTOM \<and> list_in2<$heap,$w2,#N>  \<and> list_in2<$heap,$w3,#N>)"

(*
                  ((currState, $wireheap [\<turnstile>] w1, N) \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM) 
                    ($w1 = #BOTTOM \<and> (currState, $wireheap [\<turnstile>] w2, N) \<and> (currState, $wireheap [\<turnstile>] w3, N)) ) ) "
*)

lemma inv1: "\<turnstile> program \<longrightarrow> \<box>InvOneHeap" 
oops

consts InvMultiHeap :: "temporal"
defs 
 InvMultiHeap: "InvMultiHeap \<equiv> TEMP (card<fmap_dom<$heap>> \<le> ((# 2)*(#N))+ # 2)"

lemma inv: "\<turnstile> program \<longrightarrow> \<box>InvMultiHeap"
oops


(* ngoq ngo':
                (((* case: copy *)
                  (currState, $wireheap \<turnstile>[] w1, N \<and> $w2 = #BOTTOM \<and> $w3 = #BOTTOM)
                   card (fmap_dom $wireheap) \<le> (n + 1) ))
                \<or>
                ( (* case: fuse *)
                   ($w1 = #BOTTOM \<and>
                   (\<exists> r0 r1. $w2 = #(R (Ref r0)) \<and> list_at h r0 $wireheap \<and> $w3 = #(R (Ref r1)) \<and> list_at h r1 $wireheap) \<and> 
                   card (fmap_dom $wireheap) \<le> 2*(n + 1) ))"
*)


(*
very rough sketch of a proof of bounded heap:

P1 h    = (\<exists> r n U. $w1 = #(R (Ref r)) \<and> (n,r,U,h) \<in> mList)
P2 h h' = (card (fmap_dom h') \<le> card (fmap_dom h) + n + 1) \<and> (\<exists> rv Uv. v = R (Ref rv) \<and> (n,rv,Uv,h'):mList)))

init:
initially the wireheap is s.t. P1 $wireheap  i.e. the heap contains a list structure of length n in r

case copy:
copyExe is defined s.t. it passes just this portion of the heap to the copy box, so the heap at start of the copy fct, $copyheap, 
 fulfills P1 $copyheap, which is the precondition for the VDM-level assertion;
 furthermore, the size of $copyheap is n+1, by definition of mList and $copyheap = shallowCopy $wireheap U
 by the expression-level assertion, the copyheap$ fulfills the post-condition: P2 $copyheap copyheap$
 since the size of $copyheap is n+1,
 by the def of copyExe, wireheap$ is shallowCopy copyheap$ (U \<union> Uv);
 by the def of mList and post-condition, size U = size Uv = n+1;
 therefore, size wireheap$ \<le> 2 (n+1) 

the fact that, (card (fmap_dom h') \<le> card (fmap_dom h) + n + 1), 
 is an expression level resource bound and not used on coordination level;
*)
