header "Axiomatisation of TLA* "

theory Rules
imports PreFormulas
begin


section "The Basic Axioms"

theorem fmp: assumes H1: "\<turnstile> F" and H2: "\<turnstile> F \<longrightarrow> G" shows "\<turnstile> G"
proof (unfold Valid_def,rule allI)
  fix w
  from H1 have "w \<Turnstile> F" by auto
  also from H2 have "w \<Turnstile> F \<Longrightarrow> w \<Turnstile> G" by auto
  ultimately show "w \<Turnstile> G" by auto
qed

theorem pmp: assumes H1: "|~ F" and H2: "|~ F \<longrightarrow> G" shows "|~ G"
proof (unfold PreValid_def,rule allI)
  fix w
  from H1 have "w \<Turnstile> F" by auto
  also from H2 have "w \<Turnstile> F \<Longrightarrow> w \<Turnstile> G" by auto
  ultimately show "w \<Turnstile> G" by auto
qed

theorem sq: assumes h1: "|~ P" shows "\<turnstile> \<box>[P]_v"
proof (int_unlift,unfold action_def,rule allI,rule disjI1)
  fix w i
  from h1 have "\<forall> w. w \<Turnstile> P" by auto
  thus "(w |\<^sub>s i) \<Turnstile> P" ..
qed

theorem "pre": assumes h1: "\<turnstile> F" shows "|~ F"
using h1 by auto

theorem nex: assumes h1: "\<turnstile> F" shows "|~ \<circ>F"
proof (unfold PreValid_def, unfold nexts_def,rule allI)
  fix w
  from h1 have "\<forall> w. w \<Turnstile> F" by auto
  thus "tail w \<Turnstile> F" ..
qed

theorem ax0: "\<turnstile> # True"
by auto

theorem ax1: "\<turnstile> \<box>F \<longrightarrow> F"
proof (int_unlift,unfold always_def,rule impI)
  fix w
  assume a1: "\<forall> n. (w |\<^sub>s n) \<Turnstile> F"
  hence "(w |\<^sub>s 0) \<Turnstile> F" ..
  thus "w \<Turnstile> F" by (simp add: suffix_zero)
qed

theorem ax2: "\<turnstile> \<box>F \<longrightarrow> \<box>[\<box>F]_v"
proof (int_unlift, rule impI,unfold action_def always_def,rule allI)
  fix w i
  assume a1: "\<forall> n. (w |\<^sub>s n) \<Turnstile> F"
  have "\<forall> n. (w |\<^sub>s i |\<^sub>s n) \<Turnstile> F"
  proof
    fix n
    from a1 have "(w |\<^sub>s (n+i)) \<Turnstile> F" ..     
    thus "(w |\<^sub>s i |\<^sub>s n) \<Turnstile> F" by (auto simp add: suffix_plus)
  qed
  thus "(\<forall> n. (w |\<^sub>s i |\<^sub>s n) \<Turnstile> F) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)"
    by auto
qed

theorem ax3: assumes H: "|~ F \<and> Unchanged v \<longrightarrow> \<circ>F" 
    shows "\<turnstile> \<box>[F \<longrightarrow> \<circ>F]_v \<longrightarrow> (F \<longrightarrow> \<box>F)"
proof (int_unlift,(rule impI)+)
  fix w
  assume a1: "w \<Turnstile> \<box>[F \<longrightarrow> \<circ>F]_v"
    and  a2: "w \<Turnstile> F"
  show "w \<Turnstile> \<box>F"
  proof (unfold always_def,rule allI)
    fix n
    show "(w |\<^sub>s n) \<Turnstile> F"
    proof (induct n)
      from a2 show "w |\<^sub>s 0 \<Turnstile> F" by (simp add: suffix_zero) 
    next
      fix m
      assume a3: "w |\<^sub>s m \<Turnstile> F" (is "?w \<Turnstile> F")
      from a1 have "\<forall>i. ((w |\<^sub>s i) \<Turnstile> F \<longrightarrow> \<circ>F) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)" 
	by (unfold action_def)
      hence "((w |\<^sub>s m) \<Turnstile> F \<longrightarrow> \<circ>F) \<or> ((w |\<^sub>s m) \<Turnstile> Unchanged v)" ..
      thus "w |\<^sub>s (Suc m) \<Turnstile> F" 
      proof
	assume a4: "?w \<Turnstile> Unchanged v"
	from H have "?w \<Turnstile> F \<and> Unchanged v \<longrightarrow> \<circ>F" by auto
	hence "(?w \<Turnstile> F) \<and> (?w \<Turnstile> Unchanged v) \<longrightarrow> (?w \<Turnstile> \<circ>F)" by auto
	with a3 a4 have "(?w \<Turnstile> \<circ>F)" by auto
	hence "tail ?w \<Turnstile> F" by (unfold nexts_def)
	thus ?thesis by (simp add: tail_suffix_suc)
      next
	assume a4: "(w |\<^sub>s m) \<Turnstile> F \<longrightarrow> \<circ>F"
	hence "((w |\<^sub>s m) \<Turnstile> F) \<longrightarrow> ((tail (w |\<^sub>s m)) \<Turnstile> F)" by (auto simp add: nexts_def)
	hence "((w |\<^sub>s m) \<Turnstile> F) \<longrightarrow> (((w |\<^sub>s (Suc m))) \<Turnstile> F)" by (auto simp add: tail_def suffix_def)
	with a3 show ?thesis by auto
      qed
    qed
  qed
qed

theorem ax4: "\<turnstile> \<box>[P \<longrightarrow> Q]_v \<longrightarrow> (\<box>[P]_v \<longrightarrow> \<box>[Q]_v)"
proof (int_unlift,(rule impI)+)
  fix w
  assume a1: "w \<Turnstile> \<box>[P \<longrightarrow> Q]_v"
    and  a2: "w \<Turnstile> \<box>[P]_v"
  show "w \<Turnstile> \<box>[Q]_v"
  proof (unfold action_def,rule allI)
    fix i 
    from a1 have "\<forall> i. ((w |\<^sub>s i) \<Turnstile> P \<longrightarrow> Q) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)" 
      by (simp add: action_def)
    hence a1': "((w |\<^sub>s i) \<Turnstile> P \<longrightarrow> Q) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)" (is "?imp \<or> ?Unch") ..
    from a2 have "\<forall> i. ((w |\<^sub>s i) \<Turnstile> P) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)" 
      by (simp add: action_def)
    hence "((w |\<^sub>s i) \<Turnstile> P ) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)" (is "?P \<or> ?Unch") ..   
    with a1' have "(?imp \<and> ?P) \<or> ?Unch" by auto
    thus "((w |\<^sub>s i) \<Turnstile> Q) \<or> ((w |\<^sub>s i) \<Turnstile> (Unchanged v))"
    proof 
      assume "(w |\<^sub>s i) \<Turnstile> Unchanged v"
      thus ?thesis by auto
    next
      assume "((w |\<^sub>s i) \<Turnstile> P \<longrightarrow> Q) \<and> ((w |\<^sub>s i) \<Turnstile> P)"
      hence " ((w |\<^sub>s i) \<Turnstile> Q)" by auto
      thus ?thesis by auto
    qed
  qed
qed

theorem ax5: "\<turnstile> \<box>[v` \<noteq> $v]_v"
  by (int_unlift)(auto simp add:action_def unch_def)

theorem pax0: "|~ # True"
  by auto

theorem pax1: "|~ (\<circ>\<not>F) = (\<not>\<circ>F)"
proof (unfold PreValid_def,rule allI)
  fix w i
  have "((tail w) \<Turnstile> \<not>F) = (\<not>((tail w) \<Turnstile> F))" by auto
  hence "(w \<Turnstile> (\<circ>\<not> F)) = (w \<Turnstile> (\<not> \<circ>F))" by (auto simp add: nexts_def)
  thus "w \<Turnstile> (\<circ>\<not> F) = (\<not> \<circ>F)" by auto
qed

theorem pax2: "|~ \<circ>(F \<longrightarrow> G) \<longrightarrow> (\<circ>F \<longrightarrow> \<circ>G)"
proof (pref_unlift,(rule impI)+)
  fix w
  assume a1: "w \<Turnstile> \<circ>(F \<longrightarrow> G)"
    and  a2: "w \<Turnstile> \<circ>F"
  from a1 have "(tail w) \<Turnstile> (F \<longrightarrow> G)" by (auto simp add: nexts_def)
  also from a2 have "(tail w) \<Turnstile> F" by (auto simp add: nexts_def)
  ultimately have "(tail w) \<Turnstile> G" by auto
  thus "w \<Turnstile> \<circ>G" by (auto simp add: nexts_def)
qed

theorem pax3: "|~ \<box>F \<longrightarrow> \<circ>\<box>F"
proof (pref_unlift,rule impI,unfold always_def nexts_def, rule allI)
  fix w n
  assume a1: "\<forall> n. (w |\<^sub>s n) \<Turnstile> F" 
  hence "(w |\<^sub>s 1+n) \<Turnstile> F" ..
  hence "((w |\<^sub>s 1) |\<^sub>s n) \<Turnstile> F" by (auto simp add: suffix_def)
  thus "((tail w) |\<^sub>s n) \<Turnstile> F" by (auto simp add: tail_def)
qed

theorem pax4: "|~ (\<box>[P]_v) = ([P]_v \<and> \<circ>\<box>[P]_v)"
proof (pref_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>[P]_v"
  show "(w \<Turnstile> [P]_v) \<and> (w \<Turnstile> \<circ>\<box>[P]_v)"
  proof
    from a1 have "\<forall> i. ((w |\<^sub>s i) \<Turnstile> P) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)"
      by (unfold action_def)
    hence "((w |\<^sub>s 0) \<Turnstile> P) \<or> ((w |\<^sub>s 0) \<Turnstile> Unchanged v)" ..
    hence "w \<Turnstile> P \<or> Unchanged v" by auto
    thus "w \<Turnstile> [P]_v" by (fold actrans_def)
  next
    have "\<forall> i. (((tail w) |\<^sub>s i) \<Turnstile> P) \<or> (((tail w) |\<^sub>s i) \<Turnstile> Unchanged v)"
    proof  
      fix i
      from a1 have "\<forall> i. ((w |\<^sub>s i) \<Turnstile> P) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)"
	by (unfold action_def)
      hence "((w |\<^sub>s (i+1)) \<Turnstile> P) \<or> ((w |\<^sub>s (i+1)) \<Turnstile> Unchanged v)" ..
      hence "(((w |\<^sub>s 1)|\<^sub>s i) \<Turnstile> P) \<or> (((w |\<^sub>s 1)|\<^sub>s i) \<Turnstile> Unchanged v)"
	by (auto simp add: suffix_def)
      thus "(((tail w) |\<^sub>s i) \<Turnstile> P) \<or> (((tail w) |\<^sub>s i) \<Turnstile> Unchanged v)"
	by (unfold tail_def)
    qed
    thus "(w \<Turnstile> \<circ>\<box>[P]_v)" by (auto simp add: nexts_def action_def)
  qed
next
  fix w
  assume a: "(w \<Turnstile> [P]_v) \<and> (w \<Turnstile> \<circ>\<box>[P]_v)"
  from a have a1: "w \<Turnstile> [P]_v" by simp
  from a have a2: "w \<Turnstile> \<circ>\<box>[P]_v" by simp
  show "w \<Turnstile> \<box>[P]_v"
  proof (unfold action_def,rule allI)
    fix i
    show "((w |\<^sub>s i) \<Turnstile> P) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)"
    proof (cases "i")
      assume "i = 0"
      with a1 show ?thesis
	by (simp add: actrans_def first_def second_def)
    next
      fix j
      assume hy: "i = Suc j"
      from a2 have "\<forall> i. (((tail w) |\<^sub>s i) \<Turnstile> P) \<or> (((tail w) |\<^sub>s i) \<Turnstile> Unchanged v)"
	by (unfold nexts_def action_def)
      hence "(((tail w) |\<^sub>s j) \<Turnstile> P) \<or> (((tail w) |\<^sub>s j) \<Turnstile> Unchanged v)" ..
      hence "((w |\<^sub>s (Suc j)) \<Turnstile> P) \<or> ((w |\<^sub>s (Suc j)) \<Turnstile> Unchanged v)" 
	by (auto simp add: tail_def suffix_def)
      with hy show "((w |\<^sub>s i) \<Turnstile> P) \<or> ((w |\<^sub>s i) \<Turnstile> Unchanged v)"
	by simp
    qed
  qed
qed

theorem pax5: "|~ \<circ>\<box>F \<longrightarrow> \<box>[\<circ>F]_v"
proof (pref_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> \<circ>\<box>F"
  show "w \<Turnstile> \<box>[\<circ>F]_v"
  proof (unfold action_def,rule allI,rule disjI1)
    fix i
    from a1 have "tail w \<Turnstile> \<box>F" by (unfold nexts_def)
    hence "\<forall> n. ((tail w) |\<^sub>s n) \<Turnstile> F" by (unfold always_def)
    hence "((tail w) |\<^sub>s i) \<Turnstile> F" ..
    hence "(w |\<^sub>s 1 |\<^sub>s i) \<Turnstile> F" by (auto simp add: tail_def)
    hence "(w |\<^sub>s i |\<^sub>s 1) \<Turnstile> F" by (simp add: suffix_commute)
    hence "tail (w |\<^sub>s i) \<Turnstile> F" by (simp add: tail_def)
    thus "(w |\<^sub>s i) \<Turnstile> \<circ>F" by (simp add: nexts_def)
  qed
qed

text {* The deduction theorem of TLA cannot be proved with the 
 current formalisation of @{text \<turnstile>}. This is since it has to be
 inductively defined and proof of it is on the induction on the
 assumed derivations of it. The theorem has theorem just been 
 assumed as an axiom. *}

axioms Deduct: "(\<turnstile> \<box>F \<longrightarrow> G) = ((\<turnstile> F) \<longrightarrow> (\<turnstile> G))"

text {* Theorem to show that universal quantification distributes
over \<box>. Since TLA* paper only addresses the propositional fragment
this must be added. *}

theorem allT:  "\<turnstile> (\<forall> x. \<box>(F x)) = (\<box>(\<forall> x. F x))"
by (int_unlift) (auto simp add: always_def)

theorem allActT:  "\<turnstile> (\<forall> x. \<box>[(F x)]_v) = (\<box>[(\<forall> x. F x)]_v)"
by (int_unlift) (auto simp add: action_def)

section "Derived Theorems (From TLA* paper)"

text{* This sections includes some derived theorems based on the axioms *}

text {* This theorem cannot be derived currently. Need to be able to access
  the set of FV of a formula, or know when vars are unchanged. *}

theorem alw: assumes H1: "\<turnstile> F" shows "\<turnstile> \<box>F"
proof (unfold Valid_def,rule allI)
  fix w
  from H1 have h: "\<forall> w. w \<Turnstile> F" by auto
  show "w \<Turnstile> \<box>F"
  proof (unfold always_def,rule allI)
    fix n 
    from h show "(w |\<^sub>s n) \<Turnstile> F" ..
  qed
qed

theorem alw2: assumes h1: "\<turnstile> F" 
                and   h2: "|~ F \<and> Unchanged v \<longrightarrow> \<circ>F"
               shows      "\<turnstile> \<box>F"
proof -
  from h1 have g2: "|~ \<circ>F" by (rule nex)
  hence g3: "|~ F \<longrightarrow> \<circ>F" by auto
  hence g4:"\<turnstile> \<box>[(F \<longrightarrow> \<circ>F)]_v" by (rule sq)
  from h2 have "\<turnstile> \<box>[(F \<longrightarrow> \<circ>F)]_v \<longrightarrow> F \<longrightarrow> \<box>F" by (rule ax3)
  from this[THEN intD] g4[THEN intD]
     have g5: "\<turnstile> F \<longrightarrow> \<box>F" by auto
  from this[THEN intD] h1[THEN intD] show ?thesis by auto
qed

theorem alw3: assumes h1: "\<turnstile> F" 
                and   h2: "stutinv F"
               shows      "\<turnstile> \<box>F"
proof -
  from h1 have g2: "|~ \<circ>F" by (rule nex)
  hence g3: "|~ F \<longrightarrow> \<circ>F" by auto
  hence g4:"\<turnstile> \<box>[(F \<longrightarrow> \<circ>F)]_id" by (rule sq)
  from h2 have "|~ F \<and> Unchanged id \<longrightarrow> \<circ>F"
    by (rule pre_id_unch)
  hence  "\<turnstile> \<box>[(F \<longrightarrow> \<circ>F)]_id \<longrightarrow> F \<longrightarrow> \<box>F" by (rule ax3)
  from this[THEN intD] g4[THEN intD]
     have g5: "\<turnstile> F \<longrightarrow> \<box>F" by auto
  from this[THEN intD] h1[THEN intD] show ?thesis by auto
qed

theorem T1: "\<turnstile> (\<box>F) = (\<box>\<box>F)"
proof int_unlift
  fix w
  show "(w \<Turnstile> \<box>F) = (w \<Turnstile> \<box>\<box>F)" 
  proof
    assume a1: "w \<Turnstile> \<box>F"
    have "(\<turnstile> F) \<longrightarrow> (\<turnstile> \<box>F)" by (simp add: alw)
    hence "(\<turnstile> F) \<longrightarrow> (\<turnstile> \<box>\<box>F)" by (simp add: alw)
    hence "\<turnstile> \<box>F \<longrightarrow> \<box>\<box>F" by (auto simp add: Deduct)
    from a1 this[THEN intD] show "w \<Turnstile> \<box>\<box>F" by auto
  next
    assume "w \<Turnstile> \<box>\<box>F"
    with ax1[THEN intD] show "w \<Turnstile> \<box>F" by auto
  qed
qed

theorem T2: "\<turnstile> (\<box>[P]_v) = (\<box>\<box>[P]_v)" 
proof (int_unlift)
  fix w 
  have h1: "|~ \<box>[P]_v \<and> Unchanged v \<longrightarrow> \<circ>\<box>[P]_v"
  proof (pref_unlift,rule impI,erule conjE)
    fix s
    assume a1: "s \<Turnstile> \<box>[P]_v"
    have "s \<Turnstile> (\<box>[P]_v) = ([P]_v \<and> \<circ>\<box>[P]_v)" by (rule pax4[THEN prefD])
    with a1 show "s \<Turnstile> \<circ>\<box>[P]_v" by auto
  qed
  show "(w \<Turnstile> (\<box>[P]_v)) = (w \<Turnstile> \<box>\<box>[P]_v)" 
  proof 
    assume a1: "w \<Turnstile> \<box>[P]_v"
    have "|~ (\<box>[P]_v \<longrightarrow> \<circ>\<box>[P]_v)" 
    proof (pref_unlift)
      fix w
      have "w \<Turnstile> (\<box>[P]_v) = ([P]_v \<and> \<circ>\<box>[P]_v)" by (rule pax4[THEN prefD])
      thus "(w \<Turnstile> \<box>[P]_v) \<longrightarrow> (w \<Turnstile> \<circ>\<box>[P]_v)" by auto
    qed
    hence g1: "w \<Turnstile> \<box>[(\<box>[P]_v \<longrightarrow> \<circ>\<box>[P]_v)]_v" by (rule sq[THEN intD])
    from h1 have "w \<Turnstile> \<box>[(\<box>[P]_v \<longrightarrow> \<circ>\<box>[P]_v)]_v \<longrightarrow> \<box>[P]_v \<longrightarrow> \<box>\<box>[P]_v"
      by (rule ax3[THEN intD])
    with g1 a1 show "w \<Turnstile> \<box>\<box>[P]_v" by auto  
  next
    assume a1: "w \<Turnstile> \<box>\<box>[P]_v"
    with ax1[THEN intD] have "w \<Turnstile> \<box>\<box>[P]_v \<longrightarrow> \<box>[P]_v" by auto
    with a1 show "w \<Turnstile> \<box>[P]_v" by auto
  qed
qed

theorem T3: "\<turnstile> (\<box>[[P]_v]_v) = (\<box>[P]_v)"
proof (int_unlift, rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>[P]_v"
  have "|~ P \<longrightarrow> [P]_v"
  proof (pref_unlift,rule impI)
    fix w
    assume "w \<Turnstile> P"
    thus "w \<Turnstile> [P]_v" by (auto simp add: actrans_def)
  qed
  hence g1: "\<turnstile> \<box>[(P \<longrightarrow> [P]_v)]_v" by (rule sq)
  have g2: "\<turnstile> \<box>[P]_v \<longrightarrow> \<box>[[P]_v]_v"
  proof (int_unlift,rule impI)
    fix w
    assume g3: "w \<Turnstile> \<box>[P]_v"
    have "w \<Turnstile> \<box>[(P \<longrightarrow> [P]_v)]_v \<longrightarrow> \<box>[P]_v \<longrightarrow> \<box>[[P]_v]_v"
      by (rule ax4[THEN intD])
    with g1[THEN intD] g3 show "w \<Turnstile> \<box>[[P]_v]_v" by auto
  qed
  from this[THEN intD] a1 show "w \<Turnstile> \<box>[[P]_v]_v" by auto
next
  fix w
  assume a1: "w \<Turnstile> \<box>[[P]_v]_v"
  have "|~ ([P]_v \<longrightarrow> (v`\<noteq> $v \<longrightarrow> P))"
    by (auto simp add: unch_def actrans_def)
  hence g1: "\<turnstile> \<box>[([P]_v \<longrightarrow> (v`\<noteq> $v \<longrightarrow> P))]_v" by (rule sq)
  from ax4 have "\<turnstile> \<box>[([P]_v \<longrightarrow> (v`\<noteq> $v \<longrightarrow> P))]_v \<longrightarrow> \<box>[[P]_v]_v \<longrightarrow> \<box>[(v`\<noteq> $v \<longrightarrow> P)]_v" by auto
  from this[THEN intD] g1[THEN intD] have
    "w \<Turnstile> \<box>[[P]_v]_v \<longrightarrow> \<box>[(v`\<noteq> $v \<longrightarrow> P)]_v" by auto
  with a1 have g2: "w \<Turnstile> \<box>[(v`\<noteq> $v \<longrightarrow> P)]_v" by auto
  from ax4 have "\<turnstile> \<box>[(v`\<noteq> $v \<longrightarrow> P)]_v \<longrightarrow> \<box>[v`\<noteq> $v]_v \<longrightarrow> \<box>[P]_v" by auto
  from this[THEN intD] g2 have "(w \<Turnstile> \<box>[v`\<noteq> $v]_v) \<longrightarrow> (w \<Turnstile> \<box>[P]_v)" by auto
  with ax5[THEN intD] show " (w \<Turnstile> \<box>[P]_v)" by auto
qed

theorem T4: "\<turnstile> \<box>[P]_v \<longrightarrow> \<box>[[P]_v]_w"
proof (int_unlift,rule impI)
  fix s
  assume a1: "s \<Turnstile> \<box>[P]_v"
  have g1: "\<turnstile> \<box>[P]_v \<longrightarrow> \<box>\<box>[P]_v"
  proof (int_unlift,rule impI)
    fix w
    assume a2: "w \<Turnstile> \<box>[P]_v"
    have "w \<Turnstile> (\<box>[P]_v) = (\<box>\<box>[P]_v)" by (rule T2[THEN intD])
    with a2 show "w \<Turnstile> \<box>\<box>[P]_v" by auto
  qed
  have g2: "\<turnstile> \<box>\<box>[P]_v \<longrightarrow> \<box>[\<box>\<box>[P]_v]_w" 
  proof (int_unlift,rule impI)
    fix w1
    assume a2: "w1 \<Turnstile> \<box>\<box>[P]_v"
    with T1[THEN intD] have g1: "w1 \<Turnstile> \<box>\<box>[P]_v" by auto
    have "w1 \<Turnstile> \<box>\<box>[P]_v \<longrightarrow> \<box>[\<box>\<box>[P]_v]_w" by (rule ax2[THEN intD])
    with g1 show "w1 \<Turnstile> \<box>[\<box>\<box>[P]_v]_w" by auto
  qed
  have g3: "|~ \<box>\<box>[P]_v \<longrightarrow> \<box>[P]_v" by (rule ax1[THEN pre])
  have g4: "|~ \<box>[P]_v \<longrightarrow> [P]_v"
  proof (pref_unlift,rule impI)
    fix w
    assume a1: "w \<Turnstile> \<box>[P]_v"
    have "w \<Turnstile> (\<box>[P]_v) = ([P]_v \<and> \<circ>\<box>[P]_v)" by (rule pax4[THEN prefD])
    with a1 show "w \<Turnstile> [P]_v" by auto
  qed
  have g5: "\<turnstile> \<box>[(\<box>\<box>[P]_v \<longrightarrow> [P]_v)]_w"
  proof -
    fix s
    have "|~ \<box>\<box>[P]_v \<longrightarrow> [P]_v" 
    proof (pref_unlift,rule impI)
      fix w
      assume "w \<Turnstile> \<box>\<box>[P]_v"
      with g3[THEN prefD] g4[THEN prefD] 
          show "w \<Turnstile> [P]_v" by auto  
    qed
    thus ?thesis by (rule sq)
  qed
  have g6: "\<turnstile> \<box>[\<box>\<box>[P]_v]_w \<longrightarrow> \<box>[[P]_v]_w"
  proof int_unlift
    fix s
    have "s \<Turnstile> \<box>[(\<box>\<box>[P]_v \<longrightarrow> [P]_v)]_w \<longrightarrow> \<box>[\<box>\<box>[P]_v]_w \<longrightarrow> \<box>[[P]_v]_w"
      by (rule ax4[THEN intD])
    with g5[THEN intD] show "(s \<Turnstile> \<box>[\<box>\<box>[P]_v]_w) \<longrightarrow> (s \<Turnstile> \<box>[[P]_v]_w)" by auto
  qed
  from g1[THEN intD] g2[THEN intD] this[THEN intD] a1 show "s \<Turnstile> \<box>[[P]_v]_w" by auto
qed

theorem T5: "\<turnstile> \<box>[[P]_w]_v \<longrightarrow> \<box>[[P]_v]_w"
proof int_unlift
  fix s
  have g1: "\<turnstile> \<box>[[P]_w]_v \<longrightarrow> \<box>[[[P]_w]_v]_w" by (rule T4)
  have "|~ [[P]_w]_v \<longrightarrow> [[P]_v]_w"
    by (pref_unlift) (auto simp add: actrans_def)
  hence g2: "\<turnstile> \<box>[([[P]_w]_v \<longrightarrow> [[P]_v]_w)]_w" by (rule sq)
  have g4: "\<turnstile> \<box>[[[P]_w]_v]_w \<longrightarrow> \<box>[[[P]_v]_w]_w"
  proof (int_unlift)
    fix s
    have "\<turnstile> \<box>[([[P]_w]_v \<longrightarrow> [[P]_v]_w)]_w \<longrightarrow> \<box>[[[P]_w]_v]_w \<longrightarrow> \<box>[[[P]_v]_w]_w"
      by (rule ax4)
    from this[THEN intD] g2[THEN intD]
    show "(s \<Turnstile> \<box>[[[P]_w]_v]_w) \<longrightarrow> (s \<Turnstile> \<box>[[[P]_v]_w]_w)" by auto
  qed
  have g5: "\<turnstile> \<box>[[[P]_v]_w]_w \<longrightarrow> \<box>[[P]_v]_w"
  proof (rule intI)
    fix s
    have "\<turnstile> (\<box>[[[P]_v]_w]_w) = (\<box>[[P]_v]_w)" by (rule T3)
    from this[THEN intD] show "s \<Turnstile> \<box>[[[P]_v]_w]_w \<longrightarrow> \<box>[[P]_v]_w" by auto
  qed
  from g1[THEN intD] g4[THEN intD] 
  have "s \<Turnstile> \<box>[[P]_w]_v \<longrightarrow> \<box>[[[P]_v]_w]_w" by auto
  with g5[THEN intD] show "(s \<Turnstile> \<box>[[P]_w]_v) \<longrightarrow> (s \<Turnstile> \<box>[[P]_v]_w)" by auto
qed

theorem T6: "\<turnstile> \<box>F \<longrightarrow> \<box>[\<circ>F]_v"
proof (rule intI)
  fix w
  have g1: "\<turnstile> \<box>F \<longrightarrow> \<box>[\<box>F]_v" by (rule ax2)
  have g2: "|~ \<box>F \<longrightarrow> \<circ>\<box>F" by (rule pax3)
  have g3: "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  hence g4: "|~ \<circ>(\<box>F \<longrightarrow> F)" by (rule nex)
  from this[THEN prefD] pax2[THEN prefD]
     have g5: "|~ \<circ>\<box>F \<longrightarrow> \<circ>F" by auto
  from this[THEN prefD] g2[THEN prefD]
     have g6: "|~ \<box>F \<longrightarrow> \<circ>F" by auto 
  hence g7: "\<turnstile> \<box>[(\<box>F \<longrightarrow> \<circ>F)]_v" by (rule sq)
  have g8: "\<turnstile> \<box>[\<box>F]_v \<longrightarrow> \<box>[\<circ>F]_v"
  proof (rule intI)
    fix w
    have "\<turnstile> \<box>[(\<box>F \<longrightarrow> \<circ>F)]_v \<longrightarrow> \<box>[\<box>F]_v \<longrightarrow> \<box>[\<circ>F]_v" by (rule ax4)  
    from this[THEN intD] g7[THEN intD]
      show "w \<Turnstile> \<box>[\<box>F]_v \<longrightarrow> \<box>[\<circ>F]_v" by auto
  qed
  from this[THEN intD] g1[THEN intD] show "w \<Turnstile> \<box>F \<longrightarrow> \<box>[\<circ>F]_v" 
    by auto
qed

theorem T7: assumes h1: "|~ F \<and> Unchanged v \<longrightarrow> \<circ>F"
              shows "|~ (\<box>F) = (F \<and> \<circ>\<box>F)"
proof (pref_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>F"
  have g1: "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  hence g2: "|~ \<box>F \<longrightarrow> F" by (rule pre)
  have g3: "|~ \<box>F \<longrightarrow> \<circ>\<box>F" by (rule pax3)
  from a1 g1[THEN intD] g3[THEN prefD]
     show "(w \<Turnstile> F) \<and> (w \<Turnstile> \<circ>\<box>F)" by auto
next
  fix w
  assume a: "(w \<Turnstile> F) \<and> (w \<Turnstile> \<circ>\<box>F)"
  from a have a1: "w \<Turnstile> F" by simp
  from a have a2: "w \<Turnstile> \<circ>\<box>F" by simp
  have g5: "|~ \<circ>\<box>F \<longrightarrow> \<box>[\<circ>F]_v" by (rule pax5)
  have g6: "|~ \<circ>F \<longrightarrow> (F \<longrightarrow> \<circ>F)" by auto
  hence g7: "\<turnstile> \<box>[(\<circ>F \<longrightarrow> (F \<longrightarrow> \<circ>F))]_v" by (rule sq)
  have "\<turnstile> \<box>[(\<circ>F \<longrightarrow> (F \<longrightarrow> \<circ>F))]_v \<longrightarrow> \<box>[\<circ>F]_v \<longrightarrow> \<box>[(F \<longrightarrow> \<circ>F)]_v"
    by (rule ax4)
  with this[THEN intD] g7[THEN intD] 
    have g8: "\<turnstile> \<box>[\<circ>F]_v \<longrightarrow> \<box>[(F \<longrightarrow> \<circ>F)]_v" by auto
  have g9: "\<turnstile> \<box>[\<circ>F]_v \<longrightarrow> (F \<longrightarrow> \<box>F)"
  proof (int_unlift,(rule impI)+)
    fix w
    assume aa1: "w \<Turnstile> \<box>[\<circ>F]_v"
      and  aa2: "w \<Turnstile> F"
    from h1 have h1': "w \<Turnstile> \<box>[(F \<longrightarrow> \<circ>F)]_v \<longrightarrow> F \<longrightarrow> \<box>F"
      by (rule ax3[THEN intD])
    from g8[THEN intD] aa1 have "w \<Turnstile> \<box>[(F \<longrightarrow> \<circ>F)]_v" by auto
    with h1' aa2 show "w \<Turnstile> \<box>F" by auto
  qed
  hence "|~ \<box>[\<circ>F]_v \<longrightarrow> (F \<longrightarrow> \<box>F)" by (rule pre)
  from this[THEN prefD] g5[THEN prefD]
     have "|~ \<circ>\<box>F \<longrightarrow> (F \<longrightarrow> \<box>F)" by auto 
  with this[THEN prefD] a1 a2 show "w \<Turnstile> \<box>F" by auto
qed

theorem T8: "|~ (\<circ>(F \<and> G)) = (\<circ>F \<and> \<circ>G)"
proof (pref_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<circ>(F \<and> G)"
  have "\<turnstile> F \<and> G \<longrightarrow> F" by auto
  hence g1: "|~ \<circ>(F \<and> G \<longrightarrow> F)" by (rule nex)
  have "|~ \<circ>((F \<and> G) \<longrightarrow> F) \<longrightarrow> \<circ>(F \<and> G) \<longrightarrow> \<circ>F" by (rule pax2)
  from a1 g1[THEN prefD] this[THEN prefD]
    have mg1: "w \<Turnstile> \<circ>F" by auto
  have "\<turnstile> F \<and> G \<longrightarrow> G" by auto
  hence g1': "|~ \<circ>(F \<and> G \<longrightarrow> G)" by (rule nex)
  have "|~ \<circ>((F \<and> G) \<longrightarrow> G) \<longrightarrow> \<circ>(F \<and> G) \<longrightarrow> \<circ>G" by (rule pax2)
  from a1 g1'[THEN prefD] this[THEN prefD]
    have mg2: "w \<Turnstile> \<circ>G" by auto
  from mg1 mg2 show "(w \<Turnstile> \<circ>F) \<and> (w \<Turnstile> \<circ>G)" by simp
next
  fix w
  assume a: "(w \<Turnstile> \<circ>F) \<and> (w \<Turnstile> \<circ>G)"
  from a have a1: "(w \<Turnstile> \<circ>F)" by simp
  from a have a2: "(w \<Turnstile> \<circ>G)" by simp
  have g5: "|~ \<circ>(F \<longrightarrow> \<not>G) \<longrightarrow> \<circ>F \<longrightarrow> \<not>\<circ>G"
  proof (pref_unlift,(rule impI)+)
    fix s
    assume aa1: "s \<Turnstile> \<circ>(F \<longrightarrow> \<not>G)"
      and  aa2: "s \<Turnstile> \<circ>F"
    have "s \<Turnstile> \<circ>(F \<longrightarrow> \<not>G) \<longrightarrow> \<circ>F \<longrightarrow> \<circ>\<not>G" by (rule pax2[THEN prefD])
    with aa1 aa2 have "s \<Turnstile> \<circ>\<not>G" by simp
    with pax1[THEN prefD] show  "\<not>(s \<Turnstile> \<circ>G)" by auto
  qed
  have g6: "|~ \<circ>F \<and> \<circ>G \<longrightarrow> \<circ>\<not>(F \<longrightarrow> \<not>G)"
  proof (pref_unlift,rule impI,erule conjE)
    fix s
    assume aa1: "s \<Turnstile> \<circ>F"
      and  aa2: "s \<Turnstile> \<circ>G"
    from g5 have "|~ \<circ>F \<and> \<circ>G \<longrightarrow> \<not>\<circ>(F \<longrightarrow> \<not>G)" by auto
    with this[THEN prefD] aa1 aa2
       have "s \<Turnstile> \<not>\<circ>(F \<longrightarrow> \<not>G)" by auto
    with pax1[THEN prefD] show "s \<Turnstile> \<circ>\<not>(F \<longrightarrow> \<not>G)" by auto
  qed
  have g7: "\<turnstile> \<not>(F \<longrightarrow> \<not>G) \<longrightarrow> F \<and> G" by auto
  have g8: "|~ \<circ>\<not>(F \<longrightarrow> \<not>G) \<longrightarrow> \<circ>(F \<and> G)"
  proof (pref_unlift,rule impI)
    fix s
    assume aa1: "s \<Turnstile> \<circ>\<not>(F \<longrightarrow> \<not>G)"
    from g7 have "s \<Turnstile> \<circ>(\<not>(F \<longrightarrow> \<not>G) \<longrightarrow> F \<and> G)" by (rule nex[THEN prefD])
    with aa1 pax2[THEN prefD] show "s \<Turnstile> \<circ>(F \<and> G)" by auto
  qed
  from a1 a2 g6[THEN prefD] this[THEN prefD] show "w \<Turnstile> \<circ>(F \<and> G)"
    by auto
qed

theorem H1: assumes h1: "\<turnstile> \<box>[P]_v" and h2: "\<turnstile> \<box>[P \<longrightarrow> Q]_v" 
   shows "\<turnstile> \<box>[Q]_v"
proof int_unlift
  fix w
  have "w \<Turnstile> \<box>[(P \<longrightarrow> Q)]_v \<longrightarrow> (\<box>[P]_v \<longrightarrow> \<box>[Q]_v)" by (rule intD[OF ax4])
  hence "(w \<Turnstile> \<box>[(P \<longrightarrow> Q)]_v) \<longrightarrow> ((w \<Turnstile> \<box>[P]_v) \<longrightarrow> (w \<Turnstile> \<box>[Q]_v))" by auto
  with h1[THEN intD] h2[THEN intD] show "w \<Turnstile> \<box>[Q]_v" by auto
qed

theorem H2: assumes h1: "\<turnstile> F" shows "\<turnstile> \<box>[F]_v"
proof -
  from h1 have "|~ F" by (rule pre)
  thus ?thesis by (rule sq)
qed

theorem H3: assumes h1: "\<turnstile> \<box>[P \<longrightarrow> Q]_v" and h2: "\<turnstile> \<box>[Q \<longrightarrow> R]_v"
              shows "\<turnstile> \<box>[P \<longrightarrow> R]_v"
proof -
  have "|~ (P \<longrightarrow> Q) \<longrightarrow> ((Q \<longrightarrow> R) \<longrightarrow> (P \<longrightarrow> R))" by auto
  hence "\<turnstile> \<box>[((P \<longrightarrow> Q) \<longrightarrow> ((Q \<longrightarrow> R) \<longrightarrow> (P \<longrightarrow> R)))]_v" by (rule sq)
  with h1 have "\<turnstile> \<box>[(((Q \<longrightarrow> R) \<longrightarrow> (P \<longrightarrow> R)))]_v" by (rule H1)
  with h2 show ?thesis by (rule H1)
qed

theorem H4: "\<turnstile> \<box>[[P]_v \<longrightarrow> P]_v"
proof -
  have "|~ v` \<noteq> $v \<longrightarrow> ([P]_v \<longrightarrow> P)"
    by (auto simp add: unch_def actrans_def)
  hence "\<turnstile> \<box>[v` \<noteq> $v \<longrightarrow> ([P]_v \<longrightarrow> P)]_v" by (rule sq)
  with ax5 show ?thesis by (rule H1)
qed

theorem H5: "\<turnstile> \<box>[\<box>F \<longrightarrow> \<circ>\<box>F]_v"
  by (rule pax3[THEN sq])

lemmas hx5unl = T5[THEN intD]

lemmas Deductrefl = Deduct[THEN eq_reflection]

section "Some Other Useful Derived Theorems"

theorem P1: "|~ \<box>F \<longrightarrow> \<circ>F"
proof - 
  have "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  hence t1: "|~ \<circ>(\<box>F \<longrightarrow> F)" by (rule nex)
  have "|~ \<circ>(\<box>F \<longrightarrow> F) \<longrightarrow> \<circ>\<box>F \<longrightarrow> \<circ>F" by (rule pax2)
  with t1 have g1:"|~ \<circ>\<box>F \<longrightarrow> \<circ>F" by auto
  with pax3[of F] show "|~ \<box>F \<longrightarrow> \<circ>F" by auto
qed

theorem P2: "|~ \<box>F \<longrightarrow> F \<and> \<circ>F"
proof (pref_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> \<box>F"
  show "(w \<Turnstile> F) \<and> (w \<Turnstile> \<circ>F)"
  proof
    have "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
    from this[THEN intD] a1 show "w \<Turnstile> F"
      by auto
  next
    have "|~ \<box>F \<longrightarrow> \<circ>F" by (rule P1)
    from this[THEN prefD] a1 show "w \<Turnstile> \<circ>F"
      by auto
  qed
qed 

theorem P3: "|~ \<box>F \<longrightarrow> F \<longrightarrow> \<circ>F"
proof -
  have "|~ \<box>F \<longrightarrow> F \<and> \<circ>F" by (rule P2)
  thus ?thesis by auto
qed

theorem P4: "\<turnstile> \<box>F \<longrightarrow> \<box>[F]_v"
proof -
  have g1: "\<turnstile> \<box>F \<longrightarrow> \<box>[\<box>F]_v" by (rule ax2)
  have "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  hence "|~ \<box>F \<longrightarrow> F" by (rule pre)
  hence g2: "\<turnstile> \<box>[(\<box>F \<longrightarrow> F)]_v" by (rule sq)
  have "\<turnstile> \<box>[(\<box>F \<longrightarrow> F)]_v \<longrightarrow> \<box>[\<box>F]_v \<longrightarrow> \<box>[F]_v"
    by (rule ax4)
  from this[THEN intD] g2[THEN intD]
    have "\<turnstile> \<box>[\<box>F]_v \<longrightarrow> \<box>[F]_v" by auto
  from g1[THEN intD] this[THEN intD] show ?thesis
    by auto
qed

theorem P5: "\<turnstile> \<box>[P]_v \<longrightarrow> \<box>[\<box>[P]_v]_w"  
proof -
  have "\<turnstile> (\<box>[P]_v) = (\<box>\<box>[P]_v)" by (rule T2)
  from this[THEN intD] 
    have g1:"\<turnstile> \<box>[P]_v \<longrightarrow> \<box>\<box>[P]_v" by auto
  have "\<turnstile> \<box>\<box>[P]_v \<longrightarrow> \<box>[\<box>[P]_v]_w"
    by (rule P4) 
  with g1 show ?thesis by (rule lift_imp_trans)
qed

theorem M0: "\<turnstile> \<box>F \<longrightarrow> \<box>[(F \<longrightarrow> \<circ>F)]_v" 
proof -
  have "|~ \<box>F \<longrightarrow> F \<longrightarrow> \<circ>F" by (rule P3)
  hence g1: "\<turnstile> \<box>[(\<box>F \<longrightarrow> F \<longrightarrow> \<circ>F)]_v" by (rule sq)
  have "\<turnstile> \<box>[(\<box>F \<longrightarrow> F \<longrightarrow> \<circ>F)]_v \<longrightarrow> \<box>[\<box>F]_v \<longrightarrow> \<box>[(F \<longrightarrow> \<circ>F)]_v"
    by (rule ax4)
  from this[THEN intD] g1[THEN intD] have g2:"\<turnstile> \<box>[\<box>F]_v \<longrightarrow> \<box>[(F \<longrightarrow> \<circ>F)]_v"
    by auto
  have "\<turnstile> \<box>F \<longrightarrow> \<box>[\<box>F]_v" by (rule ax2)
  with this[THEN intD] g2[THEN intD] show ?thesis
    by auto
qed

theorem M1: "\<turnstile> \<box>F \<longrightarrow> \<box>[(F \<and> \<circ>F)]_v"
proof -
  have "|~ \<box>F \<longrightarrow> F \<and> \<circ>F" by (rule P2)
  hence g1: "\<turnstile> \<box>[(\<box>F \<longrightarrow> F \<and> \<circ>F)]_v" by (rule sq)
  have "\<turnstile> \<box>[(\<box>F \<longrightarrow> F \<and> \<circ>F)]_v \<longrightarrow> \<box>[\<box>F]_v \<longrightarrow> \<box>[(F \<and> \<circ>F)]_v"
    by (rule ax4)
  from this[THEN intD] g1[THEN intD] have g2:"\<turnstile> \<box>[\<box>F]_v \<longrightarrow> \<box>[(F \<and> \<circ>F)]_v"
    by auto
  have "\<turnstile> \<box>F \<longrightarrow> \<box>[\<box>F]_v" by (rule ax2)
  with this[THEN intD] g2[THEN intD] show ?thesis
    by auto
qed

theorem M2: assumes h1: "\<turnstile> F \<longrightarrow> G"
            shows "\<turnstile> \<box>[F]_v \<longrightarrow> \<box>[G]_v"
proof (int_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> \<box>[F]_v"
  from h1 have "\<turnstile> \<box>[F \<longrightarrow> G]_v" by (rule sq[OF pre])
  from this[THEN intD] ax4[of F G v,THEN intD]
    have "\<turnstile> \<box>[F]_v \<longrightarrow> \<box>[G]_v" by auto
  from this[THEN intD] a1 show "w \<Turnstile> \<box>[G]_v" by auto
qed

theorem M3: assumes h1: "\<turnstile> F" shows "\<turnstile> \<box>[\<circ>F]_v"
proof int_unlift
  fix w
  have "\<turnstile> \<box>F \<longrightarrow> \<box>[\<circ>F]_v" by (rule T6)
  from this[THEN intD] h1[THEN alw[THEN intD]]
    show "w \<Turnstile> \<box>[\<circ>F]_v" by auto
qed

lemma M4: "\<turnstile> \<box>[(\<circ>(F \<and> G)) = (\<circ>F \<and> \<circ>G)]_v" 
proof - 
  have "|~ (\<circ>(F \<and> G)) = (\<circ>F \<and> \<circ>G)" by (rule T8)
  thus ?thesis by (rule sq)
qed

theorem M5: "\<turnstile> \<box>[(\<box>[P]_v \<longrightarrow> \<circ>\<box>[P]_v)]_w"
proof -
  from pax4[THEN prefD] have "|~ \<box>[P]_v \<longrightarrow> \<circ>\<box>[P]_v" 
    by auto
  thus ?thesis by (rule sq)
qed

theorem M6: "\<turnstile> \<box>[F \<and> G]_v \<longrightarrow> \<box>[F]_v \<and> \<box>[G]_v"
proof -
  have "|~ F \<and> G \<longrightarrow> F" by simp
  hence g1: "\<turnstile>  \<box>[(F \<and> G \<longrightarrow> F)]_v" by (rule sq)
  have "\<turnstile>  \<box>[(F \<and> G \<longrightarrow> F)]_v \<longrightarrow> \<box>[(F \<and> G)]_v \<longrightarrow> \<box>[F]_v"
    by (rule ax4)
  from this[THEN intD] g1[THEN intD]
    have G1: "\<turnstile> \<box>[(F \<and> G)]_v \<longrightarrow> \<box>[F]_v" by auto
  have "|~ F \<and> G \<longrightarrow> G" by simp
  hence g2: "\<turnstile>  \<box>[(F \<and> G \<longrightarrow> G)]_v" by (rule sq)
  have "\<turnstile>  \<box>[(F \<and> G \<longrightarrow> G)]_v \<longrightarrow> \<box>[(F \<and> G)]_v \<longrightarrow> \<box>[G]_v"
    by (rule ax4)
  from this[THEN intD] g2[THEN intD]
    have G2: "\<turnstile> \<box>[(F \<and> G)]_v \<longrightarrow> \<box>[G]_v" by auto
  from G1[THEN intD] G2[THEN intD] show ?thesis by auto
qed

theorem M7: "\<turnstile> \<box>[F]_v \<and> \<box>[G]_v \<longrightarrow> \<box>[F \<and> G]_v"
proof (int_unlift,rule impI,erule conjE)
  fix w
  assume a1: "w \<Turnstile> \<box>[F]_v"
    and  a2: "w \<Turnstile> \<box>[G]_v"
  from a1 T3[THEN intD] have a1': "w \<Turnstile> \<box>[[F]_v]_v" by auto
  from a2 T3[THEN intD] have a2': "w \<Turnstile> \<box>[[G]_v]_v" by auto
  have "|~ \<box>[F]_v \<and> \<box>[G]_v \<longrightarrow> \<box>[F]_v" by auto
  have "|~ [F]_v \<longrightarrow> [G]_v \<longrightarrow> [(F \<and> G)]_v"
    by (auto simp add: actrans_def)   
  hence g1: "\<turnstile> \<box>[([F]_v \<longrightarrow> [G]_v \<longrightarrow> [(F \<and> G)]_v)]_v" by (rule sq)
  have "\<turnstile> \<box>[([F]_v \<longrightarrow> [G]_v \<longrightarrow> [(F \<and> G)]_v)]_v \<longrightarrow> \<box>[[F]_v]_v \<longrightarrow> \<box>[([G]_v \<longrightarrow> [(F \<and> G)]_v)]_v"
    by (rule ax4)
  from this[THEN intD] g1[THEN intD]
    have g2: "\<turnstile> \<box>[[F]_v]_v \<longrightarrow> \<box>[([G]_v \<longrightarrow> [(F \<and> G)]_v)]_v" by auto
  have "\<turnstile> \<box>[([G]_v \<longrightarrow> [(F \<and> G)]_v)]_v \<longrightarrow> \<box>[[G]_v]_v \<longrightarrow> \<box>[[(F \<and> G)]_v]_v"
    by (rule ax4)
  from this[THEN intD] g2[THEN intD]  
    have "\<turnstile> \<box>[[F]_v]_v \<longrightarrow> \<box>[[G]_v]_v \<longrightarrow> \<box>[[(F \<and> G)]_v]_v" by auto
  hence "(w \<Turnstile> \<box>[[F]_v]_v) \<longrightarrow> (w \<Turnstile> \<box>[[G]_v]_v) \<longrightarrow> (w \<Turnstile> \<box>[[(F \<and> G)]_v]_v)" by auto
  with a1' a2' have "w \<Turnstile> \<box>[[(F \<and> G)]_v]_v" by simp
  with T3[THEN intD] show "w \<Turnstile> \<box>[(F \<and> G)]_v" by auto
qed
    
theorem M8: "\<turnstile> (\<box>[F]_v \<and> \<box>[G]_v) = (\<box>[F \<and> G]_v)"
proof (int_unlift,rule iffI)
  fix w
  assume a: "(w \<Turnstile> \<box>[F]_v) \<and> (w \<Turnstile> \<box>[G]_v)"
  from M7[THEN intD] have "(w \<Turnstile> \<box>[F]_v) \<and> (w \<Turnstile> \<box>[G]_v) \<longrightarrow> (w \<Turnstile> \<box>[(F \<and> G)]_v)"
    by auto
  with a show "w \<Turnstile> \<box>[(F \<and> G)]_v" by auto
next
  fix w
  assume a: "w \<Turnstile> \<box>[(F \<and> G)]_v"
  from M6[THEN intD] have "(w \<Turnstile> \<box>[(F \<and> G)]_v) \<longrightarrow> (w \<Turnstile> \<box>[F]_v) \<and> (w \<Turnstile> \<box>[G]_v)" 
    by auto
  with a show "(w \<Turnstile> \<box>[F]_v) \<and> (w \<Turnstile> \<box>[G]_v)" by auto
qed

theorem M9: "|~ \<box>F \<longrightarrow> F \<and> \<circ>\<box>F"
proof -
  have "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  hence g1: "|~ \<box>F \<longrightarrow> F" by (rule pre)
  have "|~ \<box>F \<longrightarrow> \<circ>\<box>F" by (rule pax3)
  from this[THEN prefD] g1[THEN prefD]
    show ?thesis by auto
qed

theorem M10: 
   assumes h1: "|~ F \<and> Unchanged v \<longrightarrow> \<circ>F"
     shows     "|~ F \<and> \<circ>\<box>F \<longrightarrow> \<box>F"
proof (pref_unlift,rule impI,erule conjE)
  fix w
  assume a1: "w \<Turnstile> F"
   and   a2: "w \<Turnstile> \<circ>\<box>F"  
  have "|~ \<circ>\<box>F \<longrightarrow> \<box>[\<circ>F]_v" by (rule pax5)
  from this[THEN prefD] a2 have a2': "w \<Turnstile> \<box>[\<circ>F]_v" 
    by auto
  have "|~ \<circ>F \<longrightarrow> F \<longrightarrow> \<circ>F" by auto
  hence g1: "\<turnstile> \<box>[(\<circ>F \<longrightarrow> F \<longrightarrow> \<circ>F)]_v" by (rule sq)
  have "\<turnstile> \<box>[(\<circ>F \<longrightarrow> F \<longrightarrow> \<circ>F)]_v \<longrightarrow> \<box>[\<circ>F]_v \<longrightarrow> \<box>[(F \<longrightarrow> \<circ>F)]_v"
    by (rule ax4)
  from this[THEN intD] g1[THEN intD]
    have g2: "\<turnstile> \<box>[\<circ>F]_v \<longrightarrow> \<box>[(F \<longrightarrow> \<circ>F)]_v" by auto
  from h1 have " \<turnstile> \<box>[(F \<longrightarrow> \<circ>F)]_v \<longrightarrow> F \<longrightarrow> \<box>F" 
    by (rule ax3)
  from this[THEN intD] g2[THEN intD]
    have "\<turnstile> \<box>[\<circ>F]_v \<longrightarrow> F \<longrightarrow> \<box>F" by auto
  from this[THEN intD] a1 a2'
    show "w \<Turnstile> \<box>F" by auto
qed

theorem STL2: "\<turnstile> \<box>F \<longrightarrow> F"
  by (rule ax1)

theorem STL3: "\<turnstile> (\<box>F) = (\<box>\<box>F)"
  by (rule T1)

lemma STL4: assumes h: "\<turnstile> F \<longrightarrow> G" shows "\<turnstile> \<box>F \<longrightarrow> \<box>G"
proof (int_unlift, rule impI, unfold always_def)
  fix w
  assume a1: "\<forall> n. (w |\<^sub>s n) \<Turnstile> F"
  show  "\<forall> n. (w |\<^sub>s n) \<Turnstile> G"
  proof
    fix n
    {from a1 have "(w |\<^sub>s n) \<Turnstile> F" ..}
    moreover
    { from h have "((w |\<^sub>s n) \<Turnstile> F) \<longrightarrow> ((w |\<^sub>s n) \<Turnstile> G)" by auto }
    ultimately show "(w |\<^sub>s n) \<Turnstile> G" by auto
  qed
qed

theorem STL4_2: assumes h1: "\<turnstile> F \<longrightarrow> G"
              and   h2: "|~ G \<and> Unchanged v \<longrightarrow> \<circ>G"
            shows "\<turnstile> \<box>F \<longrightarrow> \<box>G"
proof (int_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> \<box>F"
  with T6[of F v,THEN intD] have a1': "w \<Turnstile> \<box>[\<circ>F]_v" by auto
  from a1 ax1[THEN intD] have "w \<Turnstile> F" by auto
  with h1[THEN intD] have a2: "w \<Turnstile> G" by auto
  from h1 have "|~ \<circ>(F \<longrightarrow> G)" by (rule nex)
  from this pax2[of F G] have g1: "|~ \<circ>F \<longrightarrow> \<circ>G" by auto
  hence "|~ \<circ>F \<longrightarrow> G \<longrightarrow> \<circ>G" by auto
  hence g1': "\<turnstile> \<box>[(\<circ>F \<longrightarrow> G \<longrightarrow> \<circ>G)]_v" by (rule sq)
  have "\<turnstile> \<box>[(\<circ>F \<longrightarrow> G \<longrightarrow> \<circ>G)]_v \<longrightarrow> \<box>[\<circ>F]_v \<longrightarrow> \<box>[(G \<longrightarrow> \<circ>G)]_v" 
    by (rule ax4)
  with this[THEN intD] g1'[THEN intD]
    have "\<turnstile> \<box>[\<circ>F]_v \<longrightarrow> \<box>[(G \<longrightarrow> \<circ>G)]_v" by auto
  from this[THEN intD] a1' have g2: "w \<Turnstile> \<box>[(G \<longrightarrow> \<circ>G)]_v" 
    by auto
  from h2 have "\<turnstile> \<box>[(G \<longrightarrow> \<circ>G)]_v \<longrightarrow> G \<longrightarrow> \<box>G" by (rule ax3)
  from this[THEN intD] g2 a2 show "w \<Turnstile> \<box>G" by auto
qed

lemma STL4_3:  assumes h1: "\<turnstile> F \<longrightarrow> G"
  and   h2: "STUTINV G"
  shows "\<turnstile> \<box>F \<longrightarrow> \<box>G"
using h1 h2[THEN pre_id_unch] by (rule STL4_2)

lemma STL5: "\<turnstile> (\<box>(F \<and> G)) = (\<box>F \<and> \<box>G)" 
proof (int_unlift,rule iffI,unfold always_def)
  fix w 
  assume a1: "\<forall> n. (w |\<^sub>s n) \<Turnstile> F \<and> G"
  show "(\<forall> n. (w |\<^sub>s n) \<Turnstile> F) \<and> (\<forall> n. (w |\<^sub>s n) \<Turnstile> G)"
  proof (rule conjI,rule allI)
    fix n
    from a1 have "(w |\<^sub>s n) \<Turnstile> F \<and> G" ..
    thus "(w |\<^sub>s n) \<Turnstile> F" by auto
  next
    show "\<forall> n. (w |\<^sub>s n) \<Turnstile> G"
    proof
      fix n
      from a1 have "(w |\<^sub>s n) \<Turnstile> F \<and> G" ..
      thus "(w |\<^sub>s n) \<Turnstile> G" by auto
    qed
  qed
next
  fix w
  assume a: "(\<forall> n. (w |\<^sub>s n) \<Turnstile> F) \<and> (\<forall> n. (w |\<^sub>s n) \<Turnstile> G)"
  from a have a1: "(\<forall> n. (w |\<^sub>s n) \<Turnstile> F)" by simp
  from a have a2: "(\<forall> n. (w |\<^sub>s n) \<Turnstile> G)" by simp
  show "\<forall> n. (w |\<^sub>s n) \<Turnstile> F \<and> G"
  proof 
    fix n
    { from a1 have "(w |\<^sub>s n) \<Turnstile> F" .. }
    moreover
    { from a2 have "(w |\<^sub>s n) \<Turnstile> G" .. }
    ultimately show "(w |\<^sub>s n) \<Turnstile> F \<and> G" by auto
  qed
qed  

theorem STL5_2: assumes h1: "|~ F \<and> Unchanged v1 \<longrightarrow> \<circ>F"
                and   h2: "|~ G \<and> Unchanged v2 \<longrightarrow> \<circ>G"
                and   h3: "|~ (F \<and> G) \<and> Unchanged v3 \<longrightarrow> \<circ>(F \<and> G)"
               shows      "\<turnstile> (\<box>(F \<and> G)) = (\<box>F \<and> \<box>G)" 
proof (int_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>(F \<and> G)"
  show "(w \<Turnstile> \<box>F) \<and> (w \<Turnstile> \<box>G)"
  proof
    have f1:"\<turnstile> \<box>(F \<and> G) \<longrightarrow> (F \<and> G)" by (rule ax1)
    have "\<turnstile> (F \<and> G) \<longrightarrow> F" by auto
    from this[THEN intD] f1[THEN intD]
       have f2: "\<turnstile> \<box>(F \<and> G) \<longrightarrow> F" by auto
    have p1: "|~ \<box>(F \<and> G) \<longrightarrow> \<circ>(F \<and> G)" by (rule P1)   
    have "\<turnstile> (F \<and> G) \<longrightarrow> F" by simp
    hence p2: "|~ \<circ>((F \<and> G) \<longrightarrow> F)" by (rule nex)
    have "|~ \<circ>((F \<and> G) \<longrightarrow> F) \<longrightarrow> \<circ>(F \<and> G) \<longrightarrow> \<circ>F" by (rule pax2)
    from this[THEN prefD] p2[THEN prefD]
       have "|~ \<circ>(F \<and> G) \<longrightarrow> \<circ>F" by auto
    from this[THEN prefD] p1[THEN prefD]
       have p3: "|~ \<box>(F \<and> G) \<longrightarrow> \<circ>F" by auto
    hence "|~  \<box>(F \<and> G) \<longrightarrow> F \<longrightarrow> \<circ>F" by auto
    hence c1: "\<turnstile> \<box>[(\<box>(F \<and> G) \<longrightarrow> F \<longrightarrow> \<circ>F)]_v1" by (rule sq)
    have "\<turnstile> \<box>[(\<box>(F \<and> G) \<longrightarrow> F \<longrightarrow> \<circ>F)]_v1 \<longrightarrow> \<box>[\<box>(F \<and> G)]_v1 \<longrightarrow> \<box>[(F \<longrightarrow> \<circ>F)]_v1"
      by (rule ax4)
    from this[THEN intD] c1[THEN intD]
      have c2: "\<turnstile>  \<box>[\<box>(F \<and> G)]_v1 \<longrightarrow> \<box>[(F \<longrightarrow> \<circ>F)]_v1" by auto
    from h1 have "\<turnstile> \<box>[(F \<longrightarrow> \<circ>F)]_v1 \<longrightarrow> F \<longrightarrow> \<box>F" by (rule ax3)
    from this[THEN intD] c2[THEN intD]
      have c3: "(w \<Turnstile> \<box>[\<box>(F \<and> G)]_v1) \<longrightarrow> (w \<Turnstile> F) \<longrightarrow> (w \<Turnstile> \<box>F)" by auto
    have "\<turnstile> \<box>(F \<and> G) \<longrightarrow> \<box>[\<box>(F \<and> G)]_v1" by (rule ax2)
    from this[THEN intD] a1 have c4: "w \<Turnstile> \<box>[\<box>(F \<and> G)]_v1" by auto
    from a1 f2[THEN intD] have "w \<Turnstile> F" by auto
    with c3 c4 show "w \<Turnstile> \<box>F" by auto
  next
    have f1:"\<turnstile> \<box>(F \<and> G) \<longrightarrow> (F \<and> G)" by (rule ax1)
    have "\<turnstile> (F \<and> G) \<longrightarrow> G" by auto
    from this[THEN intD] f1[THEN intD]
       have f2: "\<turnstile> \<box>(F \<and> G) \<longrightarrow> G" by auto
    have p1: "|~ \<box>(F \<and> G) \<longrightarrow> \<circ>(F \<and> G)" by (rule P1)   
    have "\<turnstile> (F \<and> G) \<longrightarrow> G" by simp
    hence p2: "|~ \<circ>((F \<and> G) \<longrightarrow> G)" by (rule nex)
    have "|~ \<circ>((F \<and> G) \<longrightarrow> G) \<longrightarrow> \<circ>(F \<and> G) \<longrightarrow> \<circ>G" by (rule pax2)
    from this[THEN prefD] p2[THEN prefD]
       have "|~ \<circ>(F \<and> G) \<longrightarrow> \<circ>G" by auto
    from this[THEN prefD] p1[THEN prefD]
       have p3: "|~ \<box>(F \<and> G) \<longrightarrow> \<circ>G" by auto
    hence "|~  \<box>(F \<and> G) \<longrightarrow> G \<longrightarrow> \<circ>G" by auto
    hence c1: "\<turnstile> \<box>[(\<box>(F \<and> G) \<longrightarrow> G \<longrightarrow> \<circ>G)]_v2" by (rule sq)
    have "\<turnstile> \<box>[(\<box>(F \<and> G) \<longrightarrow> G \<longrightarrow> \<circ>G)]_v2 \<longrightarrow> \<box>[\<box>(F \<and> G)]_v2 \<longrightarrow> \<box>[(G \<longrightarrow> \<circ>G)]_v2"
      by (rule ax4)
    from this[THEN intD] c1[THEN intD]
      have c2: "\<turnstile>  \<box>[\<box>(F \<and> G)]_v2 \<longrightarrow> \<box>[(G \<longrightarrow> \<circ>G)]_v2" by auto
    from h2 have "\<turnstile> \<box>[(G \<longrightarrow> \<circ>G)]_v2 \<longrightarrow> G \<longrightarrow> \<box>G" by (rule ax3)
    from this[THEN intD] c2[THEN intD]
      have c3: "(w \<Turnstile> \<box>[\<box>(F \<and> G)]_v2) \<longrightarrow> (w \<Turnstile> G) \<longrightarrow> (w \<Turnstile> \<box>G)" by auto
    have "\<turnstile> \<box>(F \<and> G) \<longrightarrow> \<box>[\<box>(F \<and> G)]_v2" by (rule ax2)
    from this[THEN intD] a1 have c4: "w \<Turnstile> \<box>[\<box>(F \<and> G)]_v2" by auto
    from a1 f2[THEN intD] have "w \<Turnstile> G" by auto
    with c3 c4 show "w \<Turnstile> \<box>G" by auto
  qed
next
  fix w
  assume a: "(w \<Turnstile> \<box>F) \<and> (w \<Turnstile> \<box>G)"
  from a have a1: "w \<Turnstile> \<box>F" by simp
  from a have a2: "w \<Turnstile> \<box>G" by simp
  from a1 a2 have a3: "w \<Turnstile> \<box>F \<and> \<box>G" by simp
  have f1: "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  have f2: "\<turnstile> \<box>G \<longrightarrow> G" by (rule ax1)
  from f1[THEN intD] f2[THEN intD]
    have f12: "\<turnstile> \<box>F \<and> \<box>G \<longrightarrow> F \<and> G" by auto
  have p1: "|~ \<box>F \<longrightarrow> \<circ>F" by (rule P1)
  have p2: "|~ \<box>G \<longrightarrow> \<circ>G" by (rule P1)
  from p1[THEN prefD] p2[THEN prefD]
    have p3: "|~ \<box>F \<and> \<box>G \<longrightarrow> \<circ>F \<and> \<circ>G" by auto  
  have "|~ (\<circ>(F \<and> G)) = (\<circ>F \<and> \<circ>G)" by (rule T8)
  hence "|~ \<circ>F \<and> \<circ>G \<longrightarrow> \<circ>(F \<and> G)" by auto
  from this[THEN prefD] p3[THEN prefD]
    have "|~ \<box>F \<and> \<box>G \<longrightarrow> \<circ>(F \<and> G)" by auto
  hence "|~ \<box>F \<and> \<box>G \<longrightarrow> (F \<and> G) \<longrightarrow> \<circ>(F \<and> G)" by auto
  hence f3: "\<turnstile> \<box>[(\<box>F \<and> \<box>G \<longrightarrow> (F \<and> G) \<longrightarrow> \<circ>(F \<and> G))]_v3" 
    by (rule sq)
  have "\<turnstile> \<box>[(\<box>F \<and> \<box>G \<longrightarrow> (F \<and> G) \<longrightarrow> \<circ>(F \<and> G))]_v3 \<longrightarrow> \<box>[(\<box>F \<and> \<box>G)]_v3 \<longrightarrow> \<box>[((F \<and> G) \<longrightarrow> \<circ>(F \<and> G))]_v3"
    by (rule ax4)
  from this[THEN intD] f3[THEN intD]
    have f4: "\<turnstile> \<box>[(\<box>F \<and> \<box>G)]_v3 \<longrightarrow> \<box>[((F \<and> G) \<longrightarrow> \<circ>(F \<and> G))]_v3" by auto
  from h3 have "\<turnstile> \<box>[((F \<and> G) \<longrightarrow> \<circ>(F \<and> G))]_v3 \<longrightarrow> (F \<and> G) \<longrightarrow> \<box>(F \<and> G)"
    by (rule ax3)
  from this[THEN intD] f4[THEN intD]
    have "\<turnstile> \<box>[(\<box>F \<and> \<box>G)]_v3 \<longrightarrow> (F \<and> G) \<longrightarrow> \<box>(F \<and> G)" by auto
  hence f5: "(w \<Turnstile> \<box>[(\<box>F \<and> \<box>G)]_v3) \<longrightarrow> (w \<Turnstile> F \<and> G) \<longrightarrow> (w \<Turnstile> \<box>(F \<and> G))" 
    by auto
  from a1 ax2[THEN intD] have a1': "w \<Turnstile> \<box>[\<box>F]_v3" by auto
  from a2 ax2[THEN intD] have a2': "w \<Turnstile> \<box>[\<box>G]_v3" by auto
  from a1' a2' M7[THEN intD] have g1: "w \<Turnstile> \<box>[(\<box>F \<and> \<box>G)]_v3" by auto
  from a3 f12[THEN intD] have g2: "w \<Turnstile> F \<and> G" by auto
  from f5 g1 g2 show "w \<Turnstile> \<box>(F \<and> G)" by auto
qed

theorem STL5_21: assumes h1: "stutinv F"
                   and   h2: "stutinv G"
               shows      "\<turnstile> (\<box>(F \<and> G)) = (\<box>F \<and> \<box>G)" 
proof -
  from h1 h2 have "stutinv LIFT(F \<and> G)"
    by (rule stut_and)
  from h1[THEN pre_id_unch]
       h2[THEN pre_id_unch]
       this[THEN pre_id_unch]
  show ?thesis by (rule STL5_2)
qed

lemma MM7: assumes h1: "s \<Turnstile> \<box>F"
             and   h2: "s \<Turnstile> \<box>G"
             and   h3: "s \<Turnstile> \<box>(F \<and> G) \<Longrightarrow> PROP R"
            shows      "PROP R"
proof -
  thm STL5
  have "\<turnstile> (\<box>(F \<and> G)) = (\<box>F \<and> \<box>G)" by (rule STL5)
  from this[THEN intD] have "(s \<Turnstile> \<box>F) \<and> (s \<Turnstile> \<box>G) \<longrightarrow> (s \<Turnstile> \<box>(F \<and> G))"
    by auto
  with h1 h2 have "s \<Turnstile> \<box>(F \<and> G)" by simp
  with h3 show "PROP R" by auto
qed (* box_conjE *)

lemmas box_conjE = MM7

lemma box_thin: assumes h1: "s \<Turnstile> \<box>F"
                  and   h2: "PROP W"
                 shows      "PROP W"
using h2 by simp

lemma linalw: assumes h1: "a \<le> b"
               and    h2: "(w |\<^sub>s a) \<Turnstile> \<box>A"
              shows       "(w |\<^sub>s b) \<Turnstile> \<box>A"
proof (unfold always_def,rule allI)
  fix n
  from h1 have "\<exists> k. b = a + k" by (simp add: le_iff_add)
  then obtain k where g1: "b = a + k" ..
  from h2 have "\<forall> n. ((w |\<^sub>s a) |\<^sub>s n) \<Turnstile> A" by (simp add: always_def)
  (* for some weird reason Isabelle did allow the use of suffix_plus(_com) *)
  hence "\<forall> n. (\<lambda> m. w (m+n+a)) \<Turnstile> A" by (auto simp add: suffix_def)
  hence "(\<lambda> m. w (m+(n+k)+a)) \<Turnstile> A" ..
  hence "(\<lambda> m. w (m+n+(a+k))) \<Turnstile> A" by (auto simp add: nat_add_commute nat_add_left_commute)
  with g1 have "(\<lambda> m. w (m+n+b)) \<Turnstile> A" by auto
  thus "(w |\<^sub>s b |\<^sub>s n) \<Turnstile> A" by (auto simp add: suffix_def)
qed

lemma STL6: "\<turnstile> ((\<diamond>\<box>F) \<and> (\<diamond>\<box>G)) = (\<diamond>\<box>(F \<and> G))"
proof (int_unlift,rule iffI,erule conjE)
  fix w
  assume a1: "w \<Turnstile> \<diamond>\<box>F"
    and  a2: "w \<Turnstile> \<diamond>\<box>G"
  from a1 have a1': "\<exists> n. \<forall>na. ((w |\<^sub>s n |\<^sub>s na) \<Turnstile> F)" by (simp add: eventually_defs always_def)
  from a2 have a2': "\<exists> n. \<forall>na. ((w |\<^sub>s n |\<^sub>s na) \<Turnstile> G)" by (simp add: eventually_defs always_def)
  show  "w \<Turnstile> \<diamond>\<box>(F \<and> G)"
  proof (auto simp add: eventually_def always_def)
    from a1' obtain n1 where g1: "\<forall>na. ((w |\<^sub>s n1 |\<^sub>s na) \<Turnstile> F)" ..
    hence g1': "((w |\<^sub>s n1) \<Turnstile> \<box>F)" by (simp add: always_def)
    from a2' obtain n2 where g2: "\<forall>na. ((w |\<^sub>s n2 |\<^sub>s na) \<Turnstile> G)" ..
    hence g2': "((w |\<^sub>s n2) \<Turnstile> \<box>G)" by (simp add: always_def)
    show "\<exists>n. \<forall>na. ((w |\<^sub>s n |\<^sub>s na) \<Turnstile> F) \<and> ((w |\<^sub>s n |\<^sub>s na) \<Turnstile> G)"
    proof (cases "n1 \<le> n2")
      assume aa1: "n1 \<le> n2"
      have "\<forall>na. ((w |\<^sub>s n2 |\<^sub>s na) \<Turnstile> F) \<and> ((w |\<^sub>s n2 |\<^sub>s na) \<Turnstile> G)"
      proof
	fix na
	from aa1 g1' have "((w |\<^sub>s n2) \<Turnstile> \<box>F)" by (rule linalw)
	hence "\<forall>na. ((w |\<^sub>s n2 |\<^sub>s na) \<Turnstile> F)" by (simp add: always_def)
	hence tg1: "(w |\<^sub>s n2 |\<^sub>s na) \<Turnstile> F" ..
	from g2 have "(w |\<^sub>s n2 |\<^sub>s na) \<Turnstile> G" ..
	with tg1 show "((w |\<^sub>s n2 |\<^sub>s na) \<Turnstile> F) \<and> ((w |\<^sub>s n2 |\<^sub>s na) \<Turnstile> G)" by auto
      qed
      thus ?thesis ..
    next
      assume "\<not>(n1 \<le> n2)"
      hence aa1: "n2 \<le> n1" by auto
      have "\<forall>na. ((w |\<^sub>s n1 |\<^sub>s na) \<Turnstile> F) \<and> ((w |\<^sub>s n1 |\<^sub>s na) \<Turnstile> G)"
      proof
	fix na
	from aa1 g2' have "((w |\<^sub>s n1) \<Turnstile> \<box>G)" by (rule linalw)
	hence "\<forall>na. ((w |\<^sub>s n1 |\<^sub>s na) \<Turnstile> G)" by (simp add: always_def)
	hence tg1: "(w |\<^sub>s n1 |\<^sub>s na) \<Turnstile> G" ..
	from g1 have "(w |\<^sub>s n1 |\<^sub>s na) \<Turnstile> F" ..
	with tg1 show "((w |\<^sub>s n1 |\<^sub>s na) \<Turnstile> F) \<and> ((w |\<^sub>s n1 |\<^sub>s na) \<Turnstile> G)" by auto
      qed
      thus ?thesis ..      
    qed
  qed
next
  fix w
  assume "w \<Turnstile> \<diamond>\<box>(F \<and> G)"
  hence a1: "\<exists> n. \<forall>na. ((w |\<^sub>s n |\<^sub>s na) \<Turnstile> F \<and> G)" by (simp add: eventually_defs always_def) 
  show "(w \<Turnstile> \<diamond>\<box>F) \<and> (w \<Turnstile> \<diamond>\<box>G)"
  proof (simp add: eventually_def always_def,rule conjI)
    show "\<exists> n. \<forall> na. (w |\<^sub>s n |\<^sub>s na) \<Turnstile> F" 
    proof -
      from a1 obtain n where g1: "\<forall>na. ((w |\<^sub>s n |\<^sub>s na) \<Turnstile> F \<and> G)" ..
      have "\<forall> na. (w |\<^sub>s n |\<^sub>s na) \<Turnstile> F"
      proof
	fix m
	from g1 have "(w |\<^sub>s n |\<^sub>s m) \<Turnstile> F \<and> G" ..
	thus "(w |\<^sub>s n |\<^sub>s m) \<Turnstile> F" by simp
      qed      
      thus "\<exists> n. \<forall> na. (w |\<^sub>s n |\<^sub>s na) \<Turnstile> F" ..
    qed
  next
    show "\<exists> n. \<forall> na. (w |\<^sub>s n |\<^sub>s na) \<Turnstile> G" 
    proof -
      from a1 obtain n where g1: "\<forall>na. ((w |\<^sub>s n |\<^sub>s na) \<Turnstile> F \<and> G)" ..
      have "\<forall> na. (w |\<^sub>s n |\<^sub>s na) \<Turnstile> G"
      proof
	fix m
	from g1 have "(w |\<^sub>s n |\<^sub>s m) \<Turnstile> F \<and> G" ..
	thus "(w |\<^sub>s n |\<^sub>s m) \<Turnstile> G" by simp
      qed      
      thus "\<exists> n. \<forall> na. (w |\<^sub>s n |\<^sub>s na) \<Turnstile> G" ..
    qed
  qed    
qed

lemma MM1: assumes h1: "\<turnstile> F = G"
             shows "\<turnstile> (\<box>F) = (\<box>G)"
proof (int_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>F"
  from intD[OF h1] have "\<turnstile> F \<longrightarrow> G" by auto
  hence "\<turnstile> \<box>F \<longrightarrow> \<box>G" by (rule STL4)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>G"
    by auto
next
  fix w
  assume a1: "w \<Turnstile> \<box>G"
  from intD[OF h1] have "\<turnstile> G \<longrightarrow> F" by auto
  hence "\<turnstile> \<box>G \<longrightarrow> \<box>F" by (rule STL4)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>F"
    by auto
qed

theorem MM2: "\<turnstile> \<box>A \<and> \<box>(B \<longrightarrow> C) \<longrightarrow> \<box>(A \<and> B \<longrightarrow> C)"
proof -
  have "\<turnstile> A \<and> (B \<longrightarrow> C) \<longrightarrow> (A \<and> B \<longrightarrow> C)" by auto
  hence g1: "\<turnstile> \<box>(A \<and> (B \<longrightarrow> C)) \<longrightarrow> \<box>(A \<and> B \<longrightarrow> C)" by (rule STL4)
  have "\<turnstile> (\<box>(A \<and> (B \<longrightarrow> C))) = (\<box>A \<and> \<box>(B \<longrightarrow> C))" by (rule STL5)
  from this[THEN inteq_reflection] g1 show ?thesis by auto
qed

theorem MM3: "\<turnstile> \<box>(\<not>A) \<longrightarrow> \<box>(A \<and> B \<longrightarrow> C)"
proof -
  have "\<turnstile> \<not>A \<longrightarrow> (A \<and> B \<longrightarrow> C)" by auto
  thus ?thesis by (rule STL4) 
qed

theorem MM4: "\<turnstile> (#F) = (\<box>(#F))"
proof (rule intI,cases "F")
  assume a1: "F"
  fix w :: "'a seq"
  have g1: "F \<longrightarrow> (\<turnstile> #F)" by auto
  with a1 have g1: "w \<Turnstile> #F" by auto
  have "(\<turnstile> (#F)) \<longrightarrow> (\<turnstile> \<box>(#F))" by (auto simp add: alw)
  with g1 have "F \<longrightarrow> (\<turnstile> \<box>(#F))" by auto
  with a1 have "w \<Turnstile> \<box>(#F)" by auto
  with g1 show "w \<Turnstile> #F = (\<box>#F)" by auto
next  
  assume a1: "\<not>F"
  fix w :: "'a seq"
  have "\<not>F \<longrightarrow> \<not>(\<turnstile> #(F))" by auto
  with a1 have g1: "w \<Turnstile> \<not>(#F)" by auto
  have "\<turnstile> \<box>(#F) \<longrightarrow> (#F)" by (rule ax1)
  hence "\<turnstile> \<not> (#F) \<longrightarrow> \<not>(\<box>(#F))" by (rule lift_imp_neg)
  from this[THEN intD] g1 show "w \<Turnstile> #F = (\<box>#F)" by auto
qed

theorem MM4b: "\<turnstile> (\<not>#F) = (\<box>(\<not>#F))"
proof (rule intI,cases "F")
  assume a1: "\<not>F"
  fix w :: "'a seq"
  have g1: "\<not>F \<longrightarrow> (\<turnstile> \<not>#F)" by auto
  with a1 have g1: "w \<Turnstile> \<not>#F" by auto
  have "(\<turnstile> (\<not>#F)) \<longrightarrow> (\<turnstile> \<box>(\<not>#F))" by (auto simp add: alw)
  with g1 have "\<not>F \<longrightarrow> (\<turnstile> \<box>(\<not>#F))" by auto
  with a1 have "w \<Turnstile> \<box>(\<not>#F)" by auto
  with g1 show "w \<Turnstile> (\<not>#F) = (\<box>\<not>#F)" by auto
next  
  assume a1: "F"
  fix w :: "'a seq"
  have "\<not>\<not>F \<longrightarrow> \<not>(\<turnstile> (\<not>#F))" by auto
  with a1 have g1: "w \<Turnstile> \<not>(\<not>#F)" by auto
  have "\<turnstile> \<box>(\<not>#F) \<longrightarrow> (\<not>#F)" by (rule ax1)
  hence "\<turnstile> \<not> (\<not>#F) \<longrightarrow> \<not>(\<box>(\<not>#F))" by (rule lift_imp_neg)
  from this[THEN intD] g1 show "w \<Turnstile> (\<not>#F) = (\<box>\<not>#F)" by auto
qed

theorem MM5: "\<turnstile> \<box>F \<or> \<box>G \<longrightarrow> \<box>(F \<or> G)"
proof -
  have "\<turnstile> F \<longrightarrow> F \<or> G" by auto
  hence g1: "\<turnstile> \<box>F \<longrightarrow> \<box>(F \<or> G)" by (rule STL4) 
  have "\<turnstile> G \<longrightarrow> F \<or> G" by auto
  hence g2: "\<turnstile> \<box>G \<longrightarrow> \<box>(F \<or> G)" by (rule STL4) 
  from g1[THEN intD] g2[THEN intD] show ?thesis
    by auto
qed

theorem MM6: "\<turnstile> \<box>F \<or> \<box>G \<longrightarrow> \<box>(\<box>F \<or> \<box>G)"
proof -
  have tg1: "\<turnstile> (\<box>F) = (\<box>\<box>F)" by (rule T1)
  have tg2: "\<turnstile> (\<box>G) = (\<box>\<box>G)" by (rule T1)  
  have "\<turnstile> \<box>F \<longrightarrow> \<box>F \<or> \<box>G" by auto
  hence "\<turnstile> \<box>\<box>F \<longrightarrow> \<box>(\<box>F \<or> \<box>G)" by (rule STL4) 
  from this[THEN intD] tg1[THEN inteq_reflection]
    have g1: "\<turnstile> \<box>F \<longrightarrow> \<box>(\<box>F \<or> \<box>G)" by auto
  have "\<turnstile> \<box>G \<longrightarrow> \<box>F \<or> \<box>G" by auto
  hence "\<turnstile> \<box>\<box>G \<longrightarrow> \<box>(\<box>F \<or> \<box>G)" by (rule STL4) 
  from this[THEN intD] tg2[THEN inteq_reflection]
    have g2: "\<turnstile> \<box>G \<longrightarrow> \<box>(\<box>F \<or> \<box>G)" by auto
  from g1[THEN intD] g2[THEN intD] show ?thesis
    by auto
qed

theorem MM7: "\<turnstile> \<box>(\<box>F \<longrightarrow> G) \<and> \<box>F \<longrightarrow> \<box>G"
proof -
  have "\<turnstile> ((\<box>F \<longrightarrow> G) \<and> \<box>F) \<longrightarrow> G" by auto
  hence g1:"\<turnstile> \<box>((\<box>F \<longrightarrow> G) \<and> \<box>F) \<longrightarrow> \<box>G" by (rule STL4)
  have "\<turnstile> (\<box>((\<box>F \<longrightarrow> G) \<and> \<box>F)) = (\<box>(\<box>F \<longrightarrow> G) \<and> \<box>\<box>F)" by (rule STL5)
  from this[THEN intD] have
    "\<turnstile> (\<box>(\<box>F \<longrightarrow> G) \<and> \<box>\<box>F) \<longrightarrow> \<box>((\<box>F \<longrightarrow> G) \<and> \<box>F)" by auto
  from this g1 have g2: "\<turnstile> \<box>(\<box>F \<longrightarrow> G) \<and> \<box>\<box>F \<longrightarrow> \<box>G" by (rule lift_imp_trans)
  have "\<turnstile> (\<box>F) = (\<box>\<box>F)" by (rule STL3)
  from this[THEN inteq_reflection] g2
    show "\<turnstile> \<box>(\<box>F \<longrightarrow> G) \<and> \<box>F \<longrightarrow> \<box>G" by auto
qed

theorem MM8: assumes h1: "|~ F \<longrightarrow> G"
              shows      "\<turnstile> \<box>[F]_v \<longrightarrow> \<box>[G]_v"
proof -
  from h1 have g1: "\<turnstile> \<box>[(F \<longrightarrow> G)]_v" by (rule sq)
  have "\<turnstile> \<box>[(F \<longrightarrow> G)]_v \<longrightarrow> \<box>[F]_v \<longrightarrow> \<box>[G]_v"
    by (rule ax4)
  with g1 show ?thesis by (rule fmp)
qed

lemma MM9: assumes h1: "\<turnstile> F = G"
             shows "\<turnstile> (\<box>[F]_v) = (\<box>[G]_v)"
proof (int_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>[F]_v"
  from intD[OF h1] have "\<turnstile> F \<longrightarrow> G" by auto
  hence "\<turnstile> \<box>[F]_v \<longrightarrow> \<box>[G]_v" by (rule M2)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>[G]_v"
    by auto
next
  fix w
  assume a1: "w \<Turnstile> \<box>[G]_v"
  from intD[OF h1] have "\<turnstile> G \<longrightarrow> F" by auto
  hence "\<turnstile> \<box>[G]_v \<longrightarrow> \<box>[F]_v" by (rule M2)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>[F]_v"
    by auto
qed

lemma MM10: assumes h1: "|~ F = G"
             shows "\<turnstile> (\<box>[F]_v) = (\<box>[G]_v)"
proof (int_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>[F]_v"
  from prefD[OF h1] have "|~ F \<longrightarrow> G" by auto
  hence "\<turnstile> \<box>[F]_v \<longrightarrow> \<box>[G]_v" by (rule MM8)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>[G]_v"
    by auto
next
  fix w
  assume a1: "w \<Turnstile> \<box>[G]_v"
  from prefD[OF h1] have "|~ G \<longrightarrow> F" by auto
  hence "\<turnstile> \<box>[G]_v \<longrightarrow> \<box>[F]_v" by (rule MM8)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>[F]_v"
    by auto
qed

theorem MM11: "\<turnstile> \<box>[\<not>(P \<and> Q)]_v \<longrightarrow> \<box>[P]_v \<longrightarrow> \<box>[(P \<and> \<not>Q)]_v"
proof -
  have "|~ \<not>(P \<and> Q) \<longrightarrow> P \<longrightarrow> (P \<and> \<not>Q)" by auto
  hence g1: "\<turnstile> \<box>[\<not>(P \<and> Q)]_v \<longrightarrow> \<box>[(P \<longrightarrow> (P \<and> \<not>Q))]_v" 
    by (rule MM8)
  have "\<turnstile> \<box>[(P \<longrightarrow> (P \<and> \<not>Q))]_v \<longrightarrow> (\<box>[P]_v \<longrightarrow> \<box>[(P \<and> \<not>Q)]_v)"
    by (rule ax4)
  with g1 show ?thesis by (rule lift_imp_trans)
qed


theorem MM12: "\<turnstile> (\<box>[\<box>[P]_v]_v) = (\<box>[P]_v)"
proof (rule int_iffI)
  show "\<turnstile> \<box>[\<box>[P]_v]_v \<longrightarrow> \<box>[P]_v"
  proof -
    from pax4[of P v,THEN prefD]
    have "|~ \<box>[P]_v \<longrightarrow> [P]_v" by auto
    hence g1: "\<turnstile> \<box>[\<box>[P]_v]_v \<longrightarrow> \<box>[[P]_v]_v"
      by (rule MM8)
    have "\<turnstile> (\<box>[[P]_v]_v) = (\<box>[P]_v)"
      by (rule T3)
    from this[THEN inteq_reflection] g1
    show ?thesis by auto
  qed
next
  show " \<turnstile> \<box>[P]_v \<longrightarrow> \<box>[\<box>[P]_v]_v"
  proof -
    have g1: "\<turnstile> \<box>\<box>[P]_v \<longrightarrow> \<box>[\<box>\<box>[P]_v]_v"
      by (rule ax2)
    have "\<turnstile> (\<box>[P]_v) = (\<box>\<box>[P]_v)"
      by (rule T2)
    from this[THEN inteq_reflection] g1
    show ?thesis by auto
  qed
qed

section "Theorems about the @{text \<diamond>} operator"

theorem E1: "\<turnstile> (\<diamond>(F \<or> G)) = (\<diamond>F \<or> \<diamond>G)"
proof (int_unlift,rule iffI) 
  fix w
  assume a1: "(w \<Turnstile> \<diamond>F) \<or> (w \<Turnstile> \<diamond>G)"
  have "\<turnstile> \<not>(F \<or> G) \<longrightarrow> \<not>F \<and> \<not>G" by auto
  hence g1: "\<turnstile> \<box>\<not>(F \<or> G) \<longrightarrow> \<box>(\<not>F \<and> \<not>G)" by (rule STL4)
  have "\<turnstile> (\<box>(\<not>F \<and> \<not>G)) = ((\<box>\<not>F) \<and> (\<box>\<not>G))" by (rule STL5)
  from this[THEN intD] have "\<turnstile> \<box>(\<not>F \<and> \<not>G) \<longrightarrow> (\<box>\<not>F) \<and> (\<box>\<not>G)"
    by auto
  from this[THEN intD] g1[THEN intD]
    have "\<turnstile> \<box>\<not>(F \<or> G) \<longrightarrow> (\<box>\<not>F) \<and> (\<box>\<not>G)" by auto
  from this[THEN intD]
    have "\<turnstile> \<not>((\<box>\<not>F) \<and> (\<box>\<not>G)) \<longrightarrow> \<not>\<box>\<not>(F \<or> G)"
      by (auto simp add: Valid_def)
  from this[THEN intD]
    have "\<turnstile> \<not>\<box>\<not>F \<or> \<not>\<box>\<not>G \<longrightarrow> \<not>\<box>\<not>(F \<or> G)" by auto
  hence "\<turnstile> \<diamond>F \<or> \<diamond>G \<longrightarrow> \<diamond>(F \<or> G)" by (unfold eventually_def)
  from this[THEN intD] a1 show "w \<Turnstile> \<diamond>(F \<or> G)" by auto
next
  fix w
  assume a1: "w \<Turnstile> \<diamond>(F \<or> G)"
  have "\<turnstile> \<not>F \<and> \<not>G \<longrightarrow>  \<not>(F \<or> G)" by auto
  hence g1: "\<turnstile> \<box>(\<not>F \<and> \<not>G) \<longrightarrow>  \<box>\<not>(F \<or> G)" by (rule STL4)
  have "\<turnstile> (\<box>(\<not>F \<and> \<not>G)) = ((\<box>\<not>F) \<and> (\<box>\<not>G))" by (rule STL5)
  from this[THEN intD] have "\<turnstile> (\<box>\<not>F) \<and> (\<box>\<not>G) \<longrightarrow> \<box>(\<not>F \<and> \<not>G)"
    by auto
  from this[THEN intD] g1[THEN intD]
    have "\<turnstile> (\<box>\<not>F) \<and> (\<box>\<not>G) \<longrightarrow> \<box>\<not>(F \<or> G)" by auto
  from this[THEN intD]
    have "\<turnstile> \<not>\<box>\<not>(F \<or> G) \<longrightarrow> \<not>((\<box>\<not>F) \<and> (\<box>\<not>G))"
      by (auto simp add: Valid_def)
  from this[THEN intD]
    have "\<turnstile> \<not>\<box>\<not>(F \<or> G) \<longrightarrow> \<not>\<box>\<not>F \<or> \<not>\<box>\<not>G" by auto
  hence "\<turnstile> \<diamond>(F \<or> G) \<longrightarrow> \<diamond>F \<or> \<diamond>G" by (unfold eventually_def)
  from this[THEN intD] a1 show "(w \<Turnstile> \<diamond>F) \<or> (w \<Turnstile> \<diamond>G)" by auto
qed

theorem E2: "\<turnstile> (\<box>F) = (\<not>\<diamond>\<not>F)"
proof -
  have "\<turnstile> F = (\<not>\<not>F)" by auto
  hence "\<turnstile> (\<box>F) = (\<box>\<not>\<not>F)" by (rule MM1)
  from this[THEN intD] have "\<turnstile> (\<box>F) = (\<not>(\<not>\<box>\<not>(\<not>F)))" 
    by auto
  thus "\<turnstile> (\<box>F) = (\<not>\<diamond>\<not>F)" 
    by (unfold eventually_def)
qed

theorem E3: "\<turnstile> F \<longrightarrow> \<diamond>F"
proof (int_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> F"
  show "w \<Turnstile> \<diamond>F"
  proof (rule ccontr)
    assume "\<not>(w \<Turnstile> \<diamond>F)"
    hence "w \<Turnstile> \<not>\<diamond>F" by auto
    hence g1: "w \<Turnstile> \<box>\<not>F" by (simp add: eventually_def)
    have "\<turnstile> \<box>\<not>F \<longrightarrow> \<not>F" by (rule ax1)
    from this[THEN intD] g1
      have "w \<Turnstile> \<not>F" by auto
    hence "\<not>(w \<Turnstile> F)" by auto
    with a1 show False by simp
  qed
qed

theorem E4: "\<turnstile> \<box>F \<longrightarrow> \<diamond>F"
proof -
  have g1: "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  have "\<turnstile> F \<longrightarrow> \<diamond>F" by (rule E3)
  from this[THEN intD] g1[THEN intD] 
    show ?thesis by auto
qed

theorem E5: "\<turnstile> \<box>F \<longrightarrow> \<box>\<diamond>F"
proof -
  have "\<turnstile> \<box>F \<longrightarrow> \<diamond>F" by (rule E4)
  hence g1: "\<turnstile> \<box>\<box>F \<longrightarrow> \<box>\<diamond>F" by (rule STL4)
  from T1[THEN intD] have "\<turnstile> \<box>F \<longrightarrow> \<box>\<box>F" 
    by (auto simp add: Valid_def)
  from this[THEN intD] g1[THEN intD]
    show ?thesis by auto
qed

theorem E6:  "\<turnstile> \<box>F \<longrightarrow> \<diamond>\<box>F"
proof -
  have g1: "\<turnstile> \<box>\<box>F \<longrightarrow> \<diamond>\<box>F" by (rule E4)
  from T1[THEN intD] have "\<turnstile> \<box>F \<longrightarrow> \<box>\<box>F" 
    by (auto simp add: Valid_def)
  from this[THEN intD] g1[THEN intD]
    show ?thesis by auto
qed

theorem E7: 
  assumes h1: "|~ \<not>F \<and> Unchanged v \<longrightarrow> \<circ>\<not>F"
   shows      "|~ \<diamond>F \<longrightarrow> F \<or> \<circ>\<diamond>F"
proof (pref_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> \<diamond>F"
  show "(w \<Turnstile> F) \<or> (w \<Turnstile> \<circ>\<diamond>F)" 
  proof (cases "w \<Turnstile> F")
    assume "w \<Turnstile> F" thus ?thesis by auto
  next
    assume a2: "\<not>(w \<Turnstile> F)"
    have "w \<Turnstile> \<circ>\<diamond>F"
    proof -
      from h1 have "|~ \<not>F \<and> \<circ>\<box>\<not>F \<longrightarrow> \<box>\<not>F" by (rule M10)
      from this[THEN prefD]
        have "|~ \<not>\<box>\<not>F \<longrightarrow> \<not>(\<not>F \<and> \<circ>\<box>\<not>F)" by auto
      from this[THEN prefD]
	have "|~ \<not>\<box>\<not>F \<longrightarrow> F \<or> \<not>\<circ>\<box>\<not>F " by auto
      from this[THEN prefD] a2
	have g1: "w \<Turnstile> \<not>\<box>\<not>F \<longrightarrow> \<not>\<circ>\<box>\<not>F" by auto
      have "|~ (\<circ>\<not>\<box>\<not>F) = (\<not>\<circ>\<box>\<not>F)" by (rule pax1)
      from this[THEN prefD]
        have "|~ (\<not>\<circ>\<box>\<not>F) \<longrightarrow> (\<circ>\<not>\<box>\<not>F)" by auto
      with this[THEN prefD] g1
        have "w \<Turnstile> \<not>\<box>\<not>F \<longrightarrow> \<circ>\<not>\<box>\<not>F" by auto   
      hence "w \<Turnstile> \<diamond>F \<longrightarrow> \<circ>\<diamond>F" by (simp add: eventually_def)
      with a1 show "w \<Turnstile> \<circ>\<diamond>F" by simp
    qed
    thus ?thesis by simp
  qed
qed

theorem E8: "\<turnstile> \<diamond>(F \<longrightarrow> G) \<longrightarrow> \<box>F \<longrightarrow> \<diamond>G"
proof -
  have "\<turnstile> (F \<and> \<not>G) \<longrightarrow> \<not>(F \<longrightarrow> G)"
    by auto
  hence g1: "\<turnstile> \<box>(F \<and> \<not>G) \<longrightarrow> \<box>\<not>(F \<longrightarrow> G)"
    by (rule STL4)
  have "\<turnstile> (\<box>(F \<and> \<not>G)) = (\<box>F \<and> \<box>\<not>G)"
    by (rule STL5)
  from this[THEN intD]
    have "\<turnstile> (\<box>F \<and> \<box>\<not>G) \<longrightarrow> \<box>(F \<and> \<not>G)" by auto
  from this[THEN intD] g1[THEN intD]
    have g2: "\<turnstile> (\<box>F \<and> \<box>\<not>G) \<longrightarrow> \<box>\<not>(F \<longrightarrow> G)" by auto
  have "\<turnstile> ((\<not>\<box>\<not>(F \<longrightarrow> G)) \<longrightarrow> (\<not>(\<box>F \<and> \<box>\<not>G))) = ((\<box>F \<and> \<box>\<not>G) \<longrightarrow> \<box>\<not>(F \<longrightarrow> G))"
    by (rule int_simps32)
  from this[THEN intD] 
    have "\<turnstile> ((\<box>F \<and> \<box>\<not>G) \<longrightarrow> \<box>\<not>(F \<longrightarrow> G)) \<longrightarrow> ((\<not>\<box>\<not>(F \<longrightarrow> G)) \<longrightarrow> (\<not>(\<box>F \<and> \<box>\<not>G)))"
      by auto
  from this[THEN intD] g2[THEN intD]
    have "\<turnstile> \<not>\<box>\<not>(F \<longrightarrow> G) \<longrightarrow> \<not>(\<box>F \<and> \<box>\<not>G)" by auto
  from this[THEN intD]
    have "\<turnstile> \<not>\<box>\<not>(F \<longrightarrow> G) \<longrightarrow> \<not>\<box>F \<or> \<not>\<box>\<not>G" by auto
  hence "\<turnstile> \<diamond>(F \<longrightarrow> G) \<longrightarrow> \<not>\<box>F \<or> \<diamond>G" by (unfold eventually_def)
  from this[THEN intD] show "\<turnstile> \<diamond>(F \<longrightarrow> G) \<longrightarrow> \<box>F \<longrightarrow> \<diamond>G" by auto
qed

theorem E9: "\<turnstile> \<box>(F \<longrightarrow> G) \<longrightarrow> \<diamond>F \<longrightarrow> \<diamond>G"
proof -
  have "\<turnstile> \<not>G \<and> (F \<longrightarrow> G) \<longrightarrow> \<not>F"
    by (auto simp add: Valid_def)
  hence g1: "\<turnstile> \<box>(\<not>G \<and> (F \<longrightarrow> G)) \<longrightarrow> \<box>\<not>F"
    by (rule STL4)
  have "\<turnstile> (\<box>(\<not>G \<and> (F \<longrightarrow> G))) = (\<box>\<not>G \<and> \<box>(F \<longrightarrow> G))"
    by (rule STL5)
  from this[THEN intD]
    have "\<turnstile> \<box>\<not>G \<and> \<box>(F \<longrightarrow> G) \<longrightarrow>  \<box>(\<not>G \<and> (F \<longrightarrow> G))" by auto
  from this[THEN intD] g1[THEN intD]
    have "\<turnstile> \<box>\<not>G \<and> \<box>(F \<longrightarrow> G) \<longrightarrow> \<box>\<not>F" by auto
  from this[THEN intD]
    have g2: "\<turnstile> \<box>(F \<longrightarrow> G) \<longrightarrow> (\<box>\<not>G \<longrightarrow> \<box>\<not>F)" by auto
  have "\<turnstile> (\<not>\<box>\<not>F \<longrightarrow> \<not>\<box>\<not>G) = (\<box>\<not>G \<longrightarrow> \<box>\<not>F)" by (rule int_simps32)
  from this[THEN intD]
    have "\<turnstile> (\<box>\<not>G \<longrightarrow> \<box>\<not>F) \<longrightarrow> (\<not>\<box>\<not>F \<longrightarrow> \<not>\<box>\<not>G)" by auto
  with g2
    have "\<turnstile> \<box>(F \<longrightarrow> G) \<longrightarrow> (\<not>\<box>\<not>F \<longrightarrow> \<not>\<box>\<not>G)" by (rule lift_imp_trans)
  thus "\<turnstile> \<box>(F \<longrightarrow> G) \<longrightarrow> \<diamond>F \<longrightarrow> \<diamond>G" by (unfold eventually_def)
qed

theorem E10: "\<turnstile> (\<diamond>F) = (\<diamond>\<diamond>F)"
proof (int_unlift,rule iffI)
  fix w 
  assume a1: "w \<Turnstile> \<diamond>F"
  have "\<turnstile> \<diamond>F \<longrightarrow> \<diamond>\<diamond>F" by (rule E3)
  from this[THEN intD] a1
    show "w \<Turnstile> \<diamond>\<diamond>F" by auto
next
  fix w
  assume a1: "w \<Turnstile> \<diamond>\<diamond>F"
  have "\<turnstile> \<box>\<not>F \<longrightarrow> \<not>\<not>\<box>\<not>F" by auto  
  hence "\<turnstile> \<box>\<box>\<not>F \<longrightarrow> \<box>\<not>\<not>\<box>\<not>F" by (rule STL4)
  hence g1: "\<turnstile> \<not>\<box>\<not>\<not>\<box>\<not>F \<longrightarrow> \<not>\<box>\<box>\<not>F" by (rule lift_imp_neg)
  have "\<turnstile> (\<box>\<not>F) = (\<box>\<box>\<not>F)" by (rule T1)
  from this[THEN intD] have "\<turnstile> \<box>\<not>F \<longrightarrow> \<box>\<box>\<not>F" by auto
  hence "\<turnstile> \<not>\<box>\<box>\<not>F \<longrightarrow> \<not>\<box>\<not>F" by (rule lift_imp_neg)
  with g1 have "\<turnstile> \<not>\<box>\<not>\<not>\<box>\<not>F \<longrightarrow> \<not>\<box>\<not>F" by (rule lift_imp_trans)
  hence "\<turnstile> \<diamond>\<diamond>F \<longrightarrow> \<diamond>F" by (unfold eventually_def)
  from this[THEN intD] a1 show "w \<Turnstile> \<diamond>F" by auto
qed

theorem E11: "\<turnstile> \<box>\<diamond>F \<longrightarrow> \<diamond>F"
by (rule ax1)

theorem E12: "\<turnstile> (\<not>\<diamond>\<box>F) = (\<box>\<diamond>\<not>F)"
proof (unfold eventually_def)
 have "\<turnstile> F = (\<not>\<not>F)" by auto
 hence "\<turnstile> (\<box>F) = (\<box>\<not>\<not>F)" by (rule MM1)
 from this[THEN intD]
   have "\<turnstile> (\<not>\<box>F) = (\<not>\<box>\<not>\<not>F)" by auto
 hence "\<turnstile> (\<box>\<not>\<box>F) = (\<box>\<not>\<box>\<not>\<not>F)" by (rule MM1)
  from this[THEN intD]
    show "\<turnstile> (\<not>\<not>\<box>\<not>\<box>F) = (\<box>\<not>\<box>\<not>\<not>F)" by auto
qed

theorem E13: "\<turnstile> (\<diamond>\<box>\<not>F) = (\<not>\<box>\<diamond>F)"
by (unfold eventually_def) (auto)

theorem E14: assumes h1: "\<turnstile> F \<longrightarrow> G"
             shows "\<turnstile> \<diamond>F \<longrightarrow> \<diamond>G"
proof -
  from h1 have "\<turnstile> \<not>G \<longrightarrow> \<not>F" by (rule lift_imp_neg)
  hence "\<turnstile> \<box>\<not>G \<longrightarrow> \<box>\<not>F" by (rule STL4)
  hence "\<turnstile> \<not>\<box>\<not>F \<longrightarrow> \<not>\<box>\<not>G" by (rule lift_imp_neg)
  thus ?thesis by (unfold eventually_def)
qed

theorem E15: "\<turnstile> (#F) = (\<diamond>(#F))"
proof (rule intI,cases "F")
  assume a1: "F"
  fix w :: "'a seq"
  have g1: "F \<longrightarrow> (\<turnstile> #F)" by auto
  with a1 have g1: "w \<Turnstile> #F" by auto
  have "(\<turnstile> (#F)) \<longrightarrow> (\<turnstile> \<box>(#F))" by (auto simp add: alw)
  with g1 have "F \<longrightarrow> (\<turnstile> \<box>(#F))" by auto
  with a1 have tg1: "w \<Turnstile> \<box>(#F)" by auto
  have "\<turnstile>  \<box>(#F) \<longrightarrow> \<diamond>(#F)" by (rule E4)
  from this[THEN intD] tg1 have "w \<Turnstile> \<diamond>(#F)"
    by auto
  with g1 show "w \<Turnstile> #F = (\<diamond>#F)" by auto
next  
  assume a1: "\<not>F"
  fix w :: "'a seq"
  have "\<not>F \<longrightarrow> \<not>(\<turnstile> #(F))" by auto
  with a1 have g1: "w \<Turnstile> \<not>(#F)" by auto
  have tg1: "(\<turnstile> \<not>(#F)) \<longrightarrow> (\<turnstile> \<box>\<not>(#F))" by (auto simp add: alw)  
  have "\<turnstile> (\<box>\<not>#F) = (\<not>\<diamond>\<not>\<not>#F)" by (rule E2)
  from this[THEN intD]
    have tg2: "\<turnstile> (\<box>\<not>#F) \<longrightarrow> (\<not>\<diamond>\<not>\<not>#F)" by (auto simp add: Valid_def)
  have "\<turnstile> \<not>\<diamond>\<not>\<not>#F \<longrightarrow> \<not>\<diamond>#F"
  proof -
    have "\<turnstile> #F \<longrightarrow> \<not>\<not>#F" by auto
    hence "\<turnstile> \<diamond>#F \<longrightarrow> \<diamond>\<not>\<not>#F" by (rule E14)
    thus ?thesis by (rule lift_imp_neg)
  qed
  from tg2 this have tg3: "\<turnstile> (\<box>\<not>#F) \<longrightarrow> \<not>\<diamond>#F" by (auto simp add: Valid_def)
  from tg1 have "\<not>F \<longrightarrow>  (\<turnstile> \<box>\<not>(#F))" by auto
  with a1 have "\<turnstile> \<box>\<not>(#F)" by auto
  from this tg3 have "\<turnstile> \<not>\<diamond>#F" by (rule fmp)
  from this[THEN intD] g1 show "w \<Turnstile> #F = (\<diamond>#F)" by auto
qed

theorem E15b: "\<turnstile> (\<not>#F) = (\<diamond>(\<not>#F))"
proof (rule intI,cases "F")
  assume a1: "\<not>F"
  fix w :: "'a seq"
  have g1: "\<not>F \<longrightarrow> (\<turnstile> \<not>#F)" by auto
  with a1 have g1: "w \<Turnstile> \<not>#F" by auto
  have "(\<turnstile> (\<not>#F)) \<longrightarrow> (\<turnstile> \<box>(\<not>#F))" by (auto simp add: alw)
  with g1 have "\<not>F \<longrightarrow> (\<turnstile> \<box>(\<not>#F))" by auto
  with a1 have tg1: "w \<Turnstile> \<box>(\<not>#F)" by auto
  have "\<turnstile>  \<box>(\<not>#F) \<longrightarrow> \<diamond>(\<not>#F)" by (rule E4)
  from this[THEN intD] tg1 have "w \<Turnstile> \<diamond>(\<not>#F)"
    by auto
  with g1 show "w \<Turnstile> (\<not>#F) = (\<diamond>\<not>#F)" by auto
next  
  assume a1: "F"
  fix w :: "'a seq"
  have "\<not>\<not>F \<longrightarrow> \<not>(\<turnstile> \<not>#(F))" by auto
  with a1 have g1: "w \<Turnstile> \<not>(\<not>(#F))" by auto
  have tg1: "(\<turnstile> \<not>(\<not>#F)) \<longrightarrow> (\<turnstile> \<box>\<not>(\<not>#F))" by (auto simp add: alw)  
  have "\<turnstile> (\<box>\<not>\<not>#F) = (\<not>\<diamond>\<not>\<not>\<not>#F)" by (rule E2)
  from this[THEN intD]
    have tg2: "\<turnstile> (\<box>\<not>\<not>#F) \<longrightarrow> (\<not>\<diamond>\<not>\<not>\<not>#F)" by (auto simp add: Valid_def)
  have "\<turnstile> \<not>\<diamond>\<not>\<not>\<not>#F \<longrightarrow> \<not>\<diamond>\<not>#F"
  proof -
    have "\<turnstile> \<not>#F \<longrightarrow> \<not>\<not>\<not>#F" by auto
    hence "\<turnstile> \<diamond>\<not>#F \<longrightarrow> \<diamond>\<not>\<not>\<not>#F" by (rule E14)
    thus ?thesis by (rule lift_imp_neg)
  qed
  from tg2 this have tg3: "\<turnstile> (\<box>\<not>(\<not>#F)) \<longrightarrow> \<not>\<diamond>\<not>#F" by (auto simp add: Valid_def)
  from tg1 have "F \<longrightarrow>  (\<turnstile> \<box>\<not>(\<not>#F))" by auto
  with a1 have "\<turnstile> \<box>\<not>(\<not>#F)" by auto
  from this tg3 have "\<turnstile> \<not>\<diamond>\<not>#F" by (rule fmp)
  from this[THEN intD] g1 show "w \<Turnstile> (\<not>#F) = (\<diamond>\<not>#F)" by auto
qed

theorem E16: "\<turnstile> \<diamond>\<box>F \<longrightarrow> \<diamond>F"
proof -
  have "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  thus ?thesis by (rule E14)
qed

theorem E17: "\<turnstile> \<box>\<diamond>\<box>F \<longrightarrow> \<box>\<diamond>F"
proof -
  have "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  hence "\<turnstile> \<diamond>\<box>F \<longrightarrow> \<diamond>F" by (rule E14)
  thus ?thesis by (rule STL4)
qed

  
lemma STL6_act: "\<turnstile> ((\<diamond>\<box>[F]_v) \<and> (\<diamond>\<box>[G]_w)) = (\<diamond>(\<box>[F]_v \<and> \<box>[G]_w))"
proof -
  have g1: "\<turnstile> (\<diamond>\<box>\<box>[F]_v \<and> \<diamond>\<box>\<box>[G]_w) = (\<diamond>\<box>(\<box>[F]_v \<and> \<box>[G]_w))"
    by (rule STL6)
  show ?thesis
  proof (int_unlift,rule iffI)
    fix s
    assume a1: "(s \<Turnstile> \<diamond>\<box>[F]_v) \<and> (s \<Turnstile> \<diamond>\<box>[G]_w)"
    have "\<turnstile> (\<box>[F]_v) = (\<box>\<box>[F]_v)" by (rule T2)
    from this[THEN intD] have "\<turnstile> \<box>[F]_v \<longrightarrow> \<box>\<box>[F]_v"
      by auto
    hence tg1: "\<turnstile> \<diamond>\<box>[F]_v \<longrightarrow> \<diamond>\<box>\<box>[F]_v" by (rule E14)
    have "\<turnstile> (\<box>[G]_w) = (\<box>\<box>[G]_w)" by (rule T2)
    from this[THEN intD] have "\<turnstile> \<box>[G]_w \<longrightarrow> \<box>\<box>[G]_w"
      by auto
    hence tg2: "\<turnstile> \<diamond>\<box>[G]_w \<longrightarrow> \<diamond>\<box>\<box>[G]_w" by (rule E14)
    from tg1[THEN intD] tg2[THEN intD] a1
    have tg3: "s \<Turnstile> \<diamond>\<box>\<box>[F]_v \<and> \<diamond>\<box>\<box>[G]_w" by auto
    from g1[THEN intD] have "s \<Turnstile> \<diamond>\<box>\<box>[F]_v \<and> \<diamond>\<box>\<box>[G]_w \<longrightarrow> \<diamond>\<box>(\<box>[F]_v \<and> \<box>[G]_w)"
      by auto
    with tg3 have "s \<Turnstile> \<diamond>\<box>(\<box>[F]_v \<and> \<box>[G]_w)" by auto
    with E16[THEN intD] show "s \<Turnstile> \<diamond>(\<box>[F]_v \<and> \<box>[G]_w)"
      by auto
  next
    fix s
    assume a1: "s \<Turnstile> \<diamond>(\<box>[F]_v \<and> \<box>[G]_w)"
    have "\<turnstile> (\<box>[F]_v) = (\<box>\<box>[F]_v)" by (rule T2)
    from this[THEN intD] have tg1: "\<turnstile> \<box>[F]_v \<longrightarrow> \<box>\<box>[F]_v"
      by auto
    have "\<turnstile> (\<box>[G]_w) = (\<box>\<box>[G]_w)" by (rule T2)
    from this[THEN intD] have tg2: "\<turnstile> \<box>[G]_w \<longrightarrow> \<box>\<box>[G]_w"
      by auto
    from tg1[THEN intD] tg2[THEN intD]
    have tg3: "\<turnstile> \<box>[F]_v \<and> \<box>[G]_w \<longrightarrow> \<box>\<box>[F]_v \<and> \<box>\<box>[G]_w"
      by (auto simp add: Valid_def)
    have "\<turnstile> (\<box>(\<box>[F]_v \<and> \<box>[G]_w))=(\<box>\<box>[F]_v \<and> \<box>\<box>[G]_w)"
      by (rule STL5)
    from this[THEN intD] have "\<turnstile> \<box>\<box>[F]_v \<and> \<box>\<box>[G]_w \<longrightarrow> \<box>(\<box>[F]_v \<and> \<box>[G]_w)"
      by auto
    with tg3 have "\<turnstile> \<box>[F]_v \<and> \<box>[G]_w \<longrightarrow> \<box>(\<box>[F]_v \<and> \<box>[G]_w)"
      by (rule lift_imp_trans)
    hence "\<turnstile> \<diamond>(\<box>[F]_v \<and> \<box>[G]_w) \<longrightarrow> \<diamond>\<box>(\<box>[F]_v \<and> \<box>[G]_w)"
      by (rule E14)
    from this[THEN intD] a1 have tg4: "s \<Turnstile> \<diamond>\<box>(\<box>[F]_v \<and> \<box>[G]_w)"
      by auto
    from g1[THEN intD] have "\<turnstile> \<diamond>\<box>(\<box>[F]_v \<and> \<box>[G]_w) \<longrightarrow> \<diamond>\<box>\<box>[F]_v \<and> \<diamond>\<box>\<box>[G]_w"
      by auto
    from this[THEN intD] tg4 have tg5: "s \<Turnstile> \<diamond>\<box>\<box>[F]_v \<and> \<diamond>\<box>\<box>[G]_w"
      by auto
    have "\<turnstile> (\<box>[F]_v) = (\<box>\<box>[F]_v)" by (rule T2)
    from this[THEN intD] have tg1: "\<turnstile> \<box>\<box>[F]_v \<longrightarrow> \<box>[F]_v"
      by auto
    hence tg6: "\<turnstile> \<diamond>\<box>\<box>[F]_v \<longrightarrow> \<diamond>\<box>[F]_v" by (rule E14)
    have "\<turnstile> (\<box>[G]_w) = (\<box>\<box>[G]_w)" by (rule T2)
    from this[THEN intD] have tg2: "\<turnstile> \<box>\<box>[G]_w \<longrightarrow> \<box>[G]_w"
      by auto
    hence tg7: "\<turnstile> \<diamond>\<box>\<box>[G]_w \<longrightarrow> \<diamond>\<box>[G]_w" by (rule E14)    
    from tg5 tg6[THEN intD] tg7[THEN intD] 
      show "(s \<Turnstile> \<diamond>\<box>[F]_v) \<and> (s \<Turnstile> \<diamond>\<box>[G]_w)"
      by auto
  qed
qed

lemma STL6_E1: "\<turnstile> \<box>F \<and> \<diamond>G \<longrightarrow> \<diamond>(\<box>F \<and> G)"
proof -
  have "\<turnstile> \<box>\<not>(\<box>F \<and> G) \<longrightarrow> (\<box>F \<longrightarrow> \<box>\<not>G)"
  proof (int_unlift,intro impI,rule ccontr) 
    fix w
    assume a1: "w \<Turnstile> \<box>\<not>(\<box>F \<and> G)"
      and  a2: "w \<Turnstile> \<box>F"
      and  a3: "\<not>(w \<Turnstile> \<box>\<not> G)"
    from a3 have a3': "w \<Turnstile> \<not>\<box>\<not> G" by auto
    have "\<turnstile> \<not>(\<box>F \<and> G) \<longrightarrow> \<box>F \<longrightarrow> \<not>G" by auto
    hence "\<turnstile> \<box>\<not>(\<box>F \<and> G) \<longrightarrow> \<box>(\<box>F \<longrightarrow> \<not>G)"
      by (rule STL4)
    from this[THEN intD] a1 
      have a1': "w \<Turnstile> \<box>(\<box>F \<longrightarrow> \<not>G)" by auto
    thm "MM7"
    have "\<turnstile> \<box>(\<box>F \<longrightarrow> \<not>G) \<and> \<box>F \<longrightarrow> \<box>\<not>G" by (rule MM7)
    from this[THEN intD] a2 a1' have "w \<Turnstile> \<box>\<not>G"
      by auto
    with a3 show False by simp
  qed
  from this[THEN intD]
    have "\<turnstile> \<box>\<not>(\<box>F \<and> G) \<longrightarrow> \<not>\<box>F \<or> \<box>\<not>G" by auto
  from this[THEN intD]
    have "\<turnstile> \<box>F \<and> \<not>\<box>\<not>G \<longrightarrow> \<not>\<box>\<not>(\<box>F \<and> G)" 
      by (auto simp add: Valid_def)
  thus ?thesis by (unfold eventually_def)
qed

lemma STL6_E2: "\<turnstile> \<box>F \<and> \<diamond>G \<longrightarrow> \<diamond>(F \<and> G)"
proof -
  have g1: "\<turnstile> \<box>F \<and> \<diamond>G \<longrightarrow> \<diamond>(\<box>F \<and> G)" by (rule STL6_E1)
  have "\<turnstile> \<box>F \<longrightarrow> F" by (rule ax1)
  from this[THEN intD]
   have "\<turnstile> \<box>F \<and> G \<longrightarrow> F \<and> G" by auto
  hence "\<turnstile> \<diamond>(\<box>F \<and> G) \<longrightarrow> \<diamond>(F \<and> G)" by (rule E14)
  from g1 this show ?thesis by (rule lift_imp_trans)
qed 

lemma STL6_E3: "\<turnstile> \<box>F \<and> \<diamond>G \<longrightarrow> \<diamond>(G \<and> F)"
proof -
  have g1: "\<turnstile> \<box>F \<and> \<diamond>G \<longrightarrow> \<diamond>(F \<and> G)" by (rule STL6_E2)
   have "\<turnstile> F \<and> G \<longrightarrow> G \<and> F" by auto
  hence "\<turnstile> \<diamond>(F \<and> G) \<longrightarrow> \<diamond>(G \<and> F)" by (rule E14)
  from g1 this show ?thesis by (rule lift_imp_trans)
qed

lemma STL6_E4: assumes h1: "s \<Turnstile> \<box>A"
                  and  h2: "s \<Turnstile> \<diamond>F"
                  and  h3: "\<turnstile> \<box>A \<and> F \<longrightarrow> G"
                 shows     "s \<Turnstile> \<diamond>G"
proof -
  have g1: "\<turnstile> \<box>A \<and> \<diamond>F \<longrightarrow> \<diamond>(\<box>A \<and> F)" by (rule STL6_E1)
  from this[THEN intD] h1 h2
    have g2: "s \<Turnstile> \<diamond>(\<box>A \<and> F)" by auto
  from h3 have "\<turnstile> \<diamond>(\<box>A \<and> F) \<longrightarrow> \<diamond>G"
    by (rule E14)
  from this[THEN intD] g2 show ?thesis
    by auto
qed

theorem E18: "\<turnstile> \<box>\<diamond>\<box>F \<longrightarrow> \<diamond>\<box>F"
by (rule ax1)

theorem E19: "\<turnstile> \<diamond>\<box>F \<longrightarrow> \<box>\<diamond>\<box>F"
proof (int_unlift,rule impI)
  fix w 
  assume a1: "w \<Turnstile> \<diamond>\<box>F"
  have "\<turnstile> (\<box>F) = (\<box>\<box>F)" by (rule T1)
  from this[THEN intD] have
   "\<turnstile> \<box>F \<longrightarrow> \<box>\<box>F" by auto
  hence "\<turnstile> \<diamond>\<box>F \<longrightarrow> \<diamond>\<box>\<box>F" by (rule E14)
  from this[THEN intD] a1 have a1': "w \<Turnstile> \<diamond>\<box>\<box>F"
    by auto
  show "w \<Turnstile> \<box>\<diamond>\<box>F"
  proof (rule ccontr)
    assume a2: "\<not>(w \<Turnstile> \<box>\<diamond>\<box>F)"
    hence a2': "w \<Turnstile> \<not>\<box>\<diamond>\<box>F" by auto
    have "\<turnstile> (\<diamond>\<box>\<not>\<box>F) = (\<not>\<box>\<diamond>\<box>F)" by (rule E13)
    from this[THEN intD] have "\<turnstile> \<not>\<box>\<diamond>\<box>F \<longrightarrow> \<diamond>\<box>\<not>\<box>F"
      by auto
    from this[THEN intD]  a2' have a2'': "w \<Turnstile> \<diamond>\<box>\<not>\<box>F"
      by auto
    have "\<turnstile> (\<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<not>\<box>F) = (\<diamond>\<box>(\<box>F \<and> \<not>\<box>F))"
      by (rule STL6)
    from this[THEN intD]
      have "w \<Turnstile> \<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<not>\<box>F \<longrightarrow> \<diamond>\<box>(\<box>F \<and> \<not>\<box>F)"
	by auto
    with a1' a2'' have g1: "w \<Turnstile> \<diamond>\<box>(\<box>F \<and> \<not>\<box>F)"
      by auto
    have "\<turnstile> \<box>F \<and> \<not>\<box>F \<longrightarrow> #False" by auto
    hence "\<turnstile> \<box>(\<box>F \<and> \<not>\<box>F) \<longrightarrow> \<box>#False" by (rule STL4)
    hence g1': "\<turnstile> \<diamond>\<box>(\<box>F \<and> \<not>\<box>F) \<longrightarrow> \<diamond>\<box>#False" by (rule E14)
    have "\<turnstile> (#False) = (\<box>#False)" by (rule MM4)
    from this[THEN intD] have "\<turnstile> \<box>#False \<longrightarrow> #False" 
      by (auto simp add: Valid_def)
    hence g2: "\<turnstile> \<diamond>\<box>#False \<longrightarrow> \<diamond>#False" by (rule E14)
    have "\<turnstile> (#False) = (\<diamond>#False)" by (rule E15)
    from this[THEN intD] have "\<turnstile> \<diamond>#False \<longrightarrow> #False" 
      by (auto simp add: Valid_def)
    with g2 have "\<turnstile> \<diamond>\<box>#False \<longrightarrow> #False" by (rule lift_imp_trans)
    with g1' have "\<turnstile> \<diamond>\<box>(\<box>F \<and> \<not>\<box>F) \<longrightarrow> #False" by (rule lift_imp_trans)
    from this[THEN intD] g1 have "w \<Turnstile> #False" by auto
    thus "False" by auto
  qed
qed

theorem E20: "\<turnstile> \<diamond>\<box>F \<longrightarrow> \<box>\<diamond>F"
proof -
  have "\<turnstile> \<diamond>\<box>F \<longrightarrow> \<box>\<diamond>\<box>F" by (rule E19)
  moreover  have "\<turnstile> \<box>\<diamond>\<box>F \<longrightarrow> \<box>\<diamond>F" by (rule E17)
  ultimately show ?thesis by (rule lift_imp_trans)
qed

theorem E21: "\<turnstile> (\<box>\<diamond>\<box>F) = (\<diamond>\<box>F)"
proof (int_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<box>\<diamond>\<box>F"
  have "\<turnstile> \<box>\<diamond>\<box>F \<longrightarrow> \<diamond>\<box>F" by (rule E18)
  from this[THEN intD] a1 show "w \<Turnstile> \<diamond>\<box>F"
    by auto
next
  fix w
  assume a1: "w \<Turnstile> \<diamond>\<box>F"
  have "\<turnstile> \<diamond>\<box>F \<longrightarrow> \<box>\<diamond>\<box>F" by (rule E19)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>\<diamond>\<box>F"
    by auto
qed

theorem E22: assumes h1: "\<turnstile> F = G"
             shows "\<turnstile> (\<diamond>F) = (\<diamond>G)"
proof(int_unlift,intro iffI)
  fix w
  assume a1: "w \<Turnstile> \<diamond>F"
  from h1[THEN intD] have "\<turnstile> F \<longrightarrow> G" by auto
  hence "w \<Turnstile> \<diamond>F \<longrightarrow> \<diamond>G" by (intro E14[THEN intD])
  thus "w \<Turnstile> \<diamond>G" by (auto simp add: a1)
next
  fix w
  assume a1: "w \<Turnstile> \<diamond>G"
  from h1[THEN intD] have "\<turnstile> G \<longrightarrow> F" by auto
  hence "w \<Turnstile> \<diamond>G \<longrightarrow> \<diamond>F" by (intro E14[THEN intD])
  thus "w \<Turnstile> \<diamond>F" by (auto simp add: a1)
qed

lemma E23: "|~ \<circ>F \<longrightarrow> \<diamond>F"
proof -
  have g1: "|~ \<box>\<not>F \<longrightarrow> \<circ>\<not>F" by (rule P1)
  have "|~ (\<circ>\<not>F) = (\<not>\<circ>F)" by (rule pax1)
  from this[THEN prefD]
    have "|~ \<circ>\<not>F \<longrightarrow> \<not>\<circ>F" by auto
  from this[THEN prefD] g1[THEN prefD]
    have "|~ \<box>\<not>F \<longrightarrow> \<not>\<circ>F" by auto
  from this[THEN prefD]
    have "|~ \<circ>F \<longrightarrow> \<not>\<box>\<not>F" by auto
  thus ?thesis by (unfold eventually_def)
qed

lemma E24: "\<turnstile> \<diamond>\<box>Q \<longrightarrow> \<box>[\<diamond>Q]_v"
proof -
  have g1: "\<turnstile> \<diamond>\<box>Q \<longrightarrow> \<box>\<diamond>Q" by (rule E20)
  have "\<turnstile> \<box>\<diamond>Q \<longrightarrow> \<box>[\<diamond>Q]_v" by (rule P4)
  with g1 show ?thesis by (rule lift_imp_trans)
qed

lemma E25: "\<turnstile> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>A"
proof -
  have "\<turnstile> \<box>\<not>A \<longrightarrow> \<box>[\<not>A]_v" by (rule P4)
  from this[THEN intD] have "\<turnstile> \<not>\<box>[\<not>A]_v \<longrightarrow> \<not>\<box>\<not>A"
    by (auto simp add: Valid_def)
  thus ?thesis by (unfold eventually_def angle_action_def)
qed

lemma E26: "\<turnstile> \<box>\<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<box>\<diamond>A" 
by (rule STL4[OF E25])

lemma allBox: "(s \<Turnstile> \<box>(\<forall> x. F x)) = (\<forall> x. s \<Turnstile> \<box>(F x))"
proof -
  have "\<turnstile> (\<forall>x. \<box>(F x)) = (\<box>(\<forall>x. (F x)))" by (rule allT)
  from this[THEN intD]
    have "(s \<Turnstile> (\<forall>x. \<box>(F x))) = (s \<Turnstile> \<box>(\<forall>x. (F x)))" by auto
  hence "(s \<Turnstile> \<box>(\<forall>x. (F x))) = (s \<Turnstile> (\<forall>x. \<box>(F x)))" by (rule sym)
  thus ?thesis by auto
qed

theorem E27: "\<turnstile> (\<diamond>\<box>\<diamond>F) = (\<box>\<diamond>F)"
proof (int_unlift,rule iffI)
  fix w
  assume a1: "w \<Turnstile> \<diamond>\<box>\<diamond>F"
  have "\<turnstile> (\<diamond>F) = (\<diamond>\<diamond>F)" by (rule E10)
  from this[THEN intD] have "\<turnstile> \<diamond>\<diamond>F \<longrightarrow> \<diamond>F"
    by auto
  hence g1: "\<turnstile> \<box>\<diamond>\<diamond>F \<longrightarrow> \<box>\<diamond>F" by (rule STL4)
  have "\<turnstile> \<diamond>\<box>\<diamond>F \<longrightarrow> \<box>\<diamond>\<diamond>F" by (rule E20)
  from lift_imp_trans[OF this g1,THEN intD] a1
    show "w \<Turnstile> \<box>\<diamond>F" by auto
next
  fix w
  assume a1: "w \<Turnstile> \<box>\<diamond>F"
  have "\<turnstile> \<box>\<diamond>F \<longrightarrow> \<diamond>\<box>\<diamond>F" by (rule E3)
  from this[THEN intD] a1 show "w \<Turnstile> \<diamond>\<box>\<diamond>F"
    by auto
qed

lemma E28: "\<turnstile> \<diamond>\<box>F \<and> \<box>\<diamond>G \<longrightarrow> \<box>\<diamond>(F \<and> G)"
proof -
  have "\<turnstile> \<box>F \<and> \<diamond>G \<longrightarrow> \<diamond>(F \<and> G)" by (rule STL6_E2)
  hence "\<turnstile> \<box>(\<box>F \<and> \<diamond>G) \<longrightarrow> \<box>\<diamond>(F \<and> G)" by (rule STL4)
  hence g1: "\<turnstile> \<diamond>\<box>(\<box>F \<and> \<diamond>G) \<longrightarrow> \<diamond>\<box>\<diamond>(F \<and> G)" by (rule E14)
  thm STL6
  have "\<turnstile> (\<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<diamond>G)=(\<diamond>\<box>(\<box>F \<and> \<diamond>G))" by (rule STL6)
  from this[THEN intD] have "\<turnstile> \<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<diamond>G \<longrightarrow> \<diamond>\<box>(\<box>F \<and> \<diamond>G)"
    by auto
  from this g1 have g2: "\<turnstile> \<diamond>\<box>\<box>F \<and> \<diamond>\<box>\<diamond>G \<longrightarrow> \<diamond>\<box>\<diamond>(F \<and> G)"
    by (rule lift_imp_trans)
  have "\<turnstile> (\<box>F) = (\<box>\<box>F)" by (rule STL3)
  from this[THEN intD] have "\<turnstile> \<box>F \<longrightarrow> \<box>\<box>F"
    by auto
  hence tg1: "\<turnstile> \<diamond>\<box>F \<longrightarrow> \<diamond>\<box>\<box>F" by (rule E14)
  have "\<turnstile> (\<diamond>\<box>\<diamond>G)=(\<box>\<diamond>G)" by (rule E27)
  from this[THEN intD] have tg2: "\<turnstile> \<box>\<diamond>G \<longrightarrow> \<diamond>\<box>\<diamond>G"
    by auto
  have "\<turnstile> (\<diamond>\<box>\<diamond>(F \<and> G))=(\<box>\<diamond>(F \<and> G))" by (rule E27)
  from this[THEN intD] have tg3: "\<turnstile> \<diamond>\<box>\<diamond>(F \<and> G) \<longrightarrow> \<box>\<diamond>(F \<and> G)"
    by auto 
  from g2[THEN intD] tg1[THEN intD] tg2[THEN intD] tg3[THEN intD]
    show ?thesis by auto
qed

lemma E29: "|~ \<circ>\<diamond>F \<longrightarrow> \<diamond>F"
proof (unfold eventually_def)
  have "|~ \<box>\<not>F \<longrightarrow> \<circ>\<box>\<not>F" by (rule pax3)
  hence g1: "|~ \<not>\<circ>\<box>\<not>F \<longrightarrow> \<not>\<box>\<not>F"
    by auto
  have "|~ (\<circ>\<not>\<box>\<not>F) = (\<not>\<circ>\<box>\<not>F)"
    by (rule pax1)
  from g1 this[THEN prefeq_reflection,symmetric]
  show "|~ \<circ>\<not>\<box>\<not>F \<longrightarrow> \<not>\<box>\<not>F" by auto
qed

lemma allActBox: "(s \<Turnstile> \<box>[(\<forall> x. F x)]_v) = (\<forall> x. s \<Turnstile> \<box>[(F x)]_v)"
proof -
  have "\<turnstile> (\<forall>x. \<box>[(F x)]_v) = (\<box>[(\<forall>x. (F x))]_v)" by (rule allActT)
  from this[THEN intD]
    have "(s \<Turnstile> (\<forall>x. \<box>[(F x)]_v)) = (s \<Turnstile> \<box>[(\<forall>x. (F x))]_v)" by auto
  hence "(s \<Turnstile> \<box>[(\<forall>x. (F x))]_v) = (s \<Turnstile> (\<forall>x. \<box>[(F x)]_v))" by (rule sym)
  thus ?thesis by auto
qed

theorem exEE: "\<turnstile> (\<exists> x. \<diamond>(F x)) = (\<diamond>(\<exists> x. F x))"
proof (int_unlift,unfold eventually_def)
  fix w
  from Not_Rex[THEN intD]
  have "\<turnstile> (\<not> (\<exists> x. F x)) = (\<forall>x. \<not> F x)" by auto
  from this[THEN inteq_reflection]
    have "(w \<Turnstile> \<not> \<box>\<not> (\<exists> x. F x)) = (w \<Turnstile> \<not> \<box>(\<forall>x. \<not> F x))" 
      by auto
  hence g1: "(w \<Turnstile> \<not> \<box>\<not> (\<exists> x. F x)) = (\<not>(w \<Turnstile> \<box>(\<forall>x. \<not> F x)))" 
    by auto
  have "(w \<Turnstile> \<box>(\<forall> x. \<not> F x)) = (\<forall> x. w \<Turnstile> \<box>\<not>(F x))" by (rule allBox)
  from this[THEN eq_reflection] g1
    have "(w \<Turnstile> \<not> \<box>\<not> (\<exists> x. F x)) = (\<not>(\<forall> x. w \<Turnstile> \<box>\<not>(F x)))" by auto
  hence "(w \<Turnstile> \<not> \<box>\<not> (\<exists> x. F x)) = (\<exists> x. w \<Turnstile> \<not>\<box>\<not>(F x))" by auto
  thus "(\<exists>x. (w \<Turnstile> \<not> \<box>\<not> F x)) = (w \<Turnstile> \<not> \<box>\<not> (\<exists> x. F x))" by (rule sym)
qed

theorem exActE: "\<turnstile> (\<exists> x. \<diamond>\<langle>F x\<rangle>_v) = (\<diamond>\<langle>(\<exists> x. F x)\<rangle>_v)"
proof (int_unlift,unfold angle_action_def)
  fix w
  from Not_Rex[THEN intD]
  have "\<turnstile> (\<not> (\<exists> x. F x)) = (\<forall>x. \<not> F x)" by auto
  from this[THEN inteq_reflection]
    have "(w \<Turnstile> \<not> \<box>[(\<not> (\<exists> x. F x))]_v) = (w \<Turnstile> \<not> \<box>[(\<forall>x. \<not> F x)]_v)" 
      by auto
  hence g1: "(w \<Turnstile> \<not> \<box>[(\<not>(\<exists> x. F x))]_v) = (\<not>(w \<Turnstile> \<box>[(\<forall>x. \<not> F x)]_v))" 
    by auto
  have "(w \<Turnstile> \<box>[(\<forall> x. \<not> F x)]_v) = (\<forall> x. w \<Turnstile> \<box>[(\<not>(F x))]_v)" by (rule allActBox)
  from this[THEN eq_reflection] g1
    have "(w \<Turnstile> \<not> \<box>[(\<not> (\<exists> x. F x))]_v) = (\<not>(\<forall> x. w \<Turnstile> \<box>[(\<not>(F x))]_v))" by auto
  hence "(w \<Turnstile> \<not> \<box>[(\<not> (\<exists> x. F x))]_v) = (\<exists> x. w \<Turnstile> \<not>\<box>[(\<not>(F x))]_v)" by auto
  thus "(\<exists>x. (w \<Turnstile> \<not> \<box>[(\<not> F x)]_v)) = (w \<Turnstile> \<not> \<box>[(\<not> (\<exists> x. F x))]_v)" by (rule sym)
qed

section "Theorems about the @{text \<leadsto>} operator"

theorem LT1: "\<turnstile> F \<leadsto> F"
proof (unfold leadsto_def)
  have "\<turnstile> F \<longrightarrow> \<diamond>F" by (rule E3)
  thus "\<turnstile> \<box>(F \<longrightarrow> \<diamond>F)" by (rule alw) 
qed

theorem LT2: assumes h: "\<turnstile> F \<longrightarrow> G" shows "\<turnstile> F \<longrightarrow> \<diamond>G"
proof -
  have "\<turnstile> G \<longrightarrow> \<diamond>G" by (rule E3)
  from h[THEN intD] this[THEN intD]
    show ?thesis by auto
qed

theorem LT3: assumes h: "\<turnstile> F \<longrightarrow> G" shows "\<turnstile> F \<leadsto> G"
proof -
  from h have "\<turnstile> F \<longrightarrow> \<diamond>G" by (rule LT2)
  hence "\<turnstile> \<box>(F \<longrightarrow> \<diamond>G)" by (rule alw)
  thus ?thesis by (unfold leadsto_def)
qed

theorem LT4: "\<turnstile> F \<and> (F \<leadsto> G) \<longrightarrow> \<diamond>G"
proof (int_unlift,rule impI,erule conjE)
  fix w
  assume a1: "w \<Turnstile> F"
    and  a2: "w \<Turnstile> F \<leadsto> G"
   hence a2': "w \<Turnstile> \<box>(F \<longrightarrow> \<diamond>G)" by (unfold leadsto_def)
  with ax1 have "\<turnstile> \<box>(F \<longrightarrow> \<diamond>G) \<longrightarrow> (F \<longrightarrow> \<diamond>G)" by auto
  from this[THEN intD] a2' have "w \<Turnstile> (F \<longrightarrow> \<diamond>G)" by auto
  with a1 show "w \<Turnstile> \<diamond>G" by auto
qed

theorem LT5: "\<turnstile> \<box>(F \<longrightarrow> \<diamond>G) \<longrightarrow> \<diamond>F \<longrightarrow> \<diamond>G"
proof -
  have "\<turnstile> (\<diamond>G) = (\<diamond>\<diamond>G)" by (rule E10)
  from this[THEN intD] have g1: "\<turnstile> \<diamond>\<diamond>G \<longrightarrow> \<diamond>G"
    by auto
  have "\<turnstile> \<box>(F \<longrightarrow> \<diamond>G) \<longrightarrow> \<diamond>F \<longrightarrow> \<diamond>\<diamond>G"
    by (rule E9)
  from this[THEN intD] g1[THEN intD] show ?thesis
    by auto
qed

theorem LT6: "\<turnstile> \<diamond>F \<and> (F \<leadsto> G) \<longrightarrow> \<diamond>G"
proof -
  have "\<turnstile> \<box>(F \<longrightarrow> \<diamond>G) \<longrightarrow> \<diamond>F \<longrightarrow> \<diamond>G" by (rule LT5)
  hence "\<turnstile> (F \<leadsto> G) \<longrightarrow> \<diamond>F \<longrightarrow> \<diamond>G" by (unfold leadsto_def)
  from this[THEN intD] have g1: "\<turnstile> (F \<leadsto> G) \<and> \<diamond>F \<longrightarrow> \<diamond>G" 
    by auto
  from this[THEN intD] show ?thesis by auto
qed

theorem LT7: "\<turnstile> \<box>\<diamond>F \<and> (F \<leadsto> G) \<longrightarrow> \<box>\<diamond>G"
proof -
  have "\<turnstile> \<diamond>F \<and> (F \<leadsto> G) \<longrightarrow> \<diamond>G" by (rule LT6)
  hence g1: "\<turnstile> \<box>(\<diamond>F \<and> (F \<leadsto> G)) \<longrightarrow> \<box>\<diamond>G" by (rule STL4)
  have "\<turnstile> (\<box>(\<diamond>F \<and> (F \<leadsto> G))) = (\<box>\<diamond>F \<and> \<box>(F \<leadsto> G))" by (rule STL5)
  from this[THEN intD] have "\<turnstile> \<box>\<diamond>F \<and> \<box>(F \<leadsto> G) \<longrightarrow> \<box>(\<diamond>F \<and> (F \<leadsto> G))"
    by auto
  from this g1 have g2: "\<turnstile> \<box>\<diamond>F \<and> \<box>(F \<leadsto> G) \<longrightarrow> \<box>\<diamond>G"
    by (rule lift_imp_trans)
  have g3: "\<turnstile>  \<box>\<diamond>F \<and> (F \<leadsto> G) \<longrightarrow>  \<box>\<diamond>F \<and> \<box>(F \<leadsto> G)" 
  proof (unfold leadsto_def)
    have "\<turnstile> (\<box>(F \<longrightarrow> \<diamond>G)) = (\<box>\<box>(F \<longrightarrow> \<diamond>G))" by (rule T1)
    from this[THEN intD] have "\<turnstile> \<box>(F \<longrightarrow> \<diamond>G) \<longrightarrow> \<box>\<box>(F \<longrightarrow> \<diamond>G)"
      by auto
    from this[THEN intD] 
      show "\<turnstile> \<box>\<diamond>F \<and> \<box>(F \<longrightarrow> \<diamond>G) \<longrightarrow> \<box>\<diamond>F \<and> \<box>\<box>(F \<longrightarrow> \<diamond>G)"
	by auto
  qed
  from this g2 show ?thesis
    by (rule lift_imp_trans)
qed

theorem LT8: "\<turnstile> \<box>\<diamond>G \<longrightarrow> (F \<leadsto> G)"
proof (unfold leadsto_def)
  have "\<turnstile> \<diamond>G \<longrightarrow> (F \<longrightarrow> \<diamond>G)" by auto
  thus "\<turnstile> \<box>\<diamond>G \<longrightarrow> \<box>(F \<longrightarrow> \<diamond>G)" by (rule STL4)
qed

(* FIXME: insert a (LT9) LT10 theorem *)

theorem LT11: "\<turnstile> (F \<leadsto> G) \<longrightarrow>  (F \<leadsto> (G \<or> H))"
proof -
  have "\<turnstile> G \<longrightarrow> (G \<or> H)" by auto
  hence "\<turnstile> \<diamond>G \<longrightarrow> \<diamond>(G \<or> H)" by (rule E14)
  from this[THEN intD]
    have "\<turnstile> (F \<longrightarrow> \<diamond>G) \<longrightarrow> F \<longrightarrow> \<diamond>(G \<or> H)"
      by auto
  hence "\<turnstile> \<box>(F \<longrightarrow> \<diamond>G) \<longrightarrow> \<box>(F \<longrightarrow> \<diamond>(G \<or> H))"
    by (rule STL4)
  thus ?thesis by (unfold leadsto_def)
qed

theorem LT12: "\<turnstile> (F \<leadsto> H) \<longrightarrow>  (F \<leadsto> (G \<or> H))"
proof -
  have "\<turnstile> H \<longrightarrow> (G \<or> H)" by auto
  hence "\<turnstile> \<diamond>H \<longrightarrow> \<diamond>(G \<or> H)" by (rule E14)
  from this[THEN intD]
    have "\<turnstile> (F \<longrightarrow> \<diamond>H) \<longrightarrow> F \<longrightarrow> \<diamond>(G \<or> H)"
      by auto
  hence "\<turnstile> \<box>(F \<longrightarrow> \<diamond>H) \<longrightarrow> \<box>(F \<longrightarrow> \<diamond>(G \<or> H))"
    by (rule STL4)
  thus ?thesis by (unfold leadsto_def)
qed

theorem LT13: "\<turnstile> (G \<leadsto> H) \<and> (F \<leadsto> G) \<longrightarrow> (F \<leadsto> H)"
proof -
  have "\<turnstile> \<box>(G \<longrightarrow> \<diamond>H) \<and> (F \<longrightarrow> \<diamond>G) \<longrightarrow> (F \<longrightarrow> \<diamond>H)"
  proof (int_unlift,intro impI,elim conjE)
    fix w
    assume a1: "w \<Turnstile> F"
     and   a2: "w \<Turnstile> \<box>(G \<longrightarrow> \<diamond>H)"
     and   a3: "(w \<Turnstile> F) \<longrightarrow> (w \<Turnstile> \<diamond>G)"
    from a1 a3 have g1: "w \<Turnstile> \<diamond>G" by simp
    have "\<turnstile> \<diamond>G \<and> (G \<leadsto> H) \<longrightarrow> \<diamond>H" by (rule LT6)
    hence "\<turnstile> \<diamond>G \<and> \<box>(G \<longrightarrow> \<diamond>H) \<longrightarrow> \<diamond>H" by (unfold leadsto_def)
    from this[THEN intD] g1 have "w \<Turnstile> \<box>(G \<longrightarrow> \<diamond>H) \<longrightarrow> \<diamond>H" 
      by auto
    with a2 show "w \<Turnstile> \<diamond>H" by auto
  qed
  hence g1: "\<turnstile> \<box>(\<box>(G \<longrightarrow> \<diamond>H) \<and> (F \<longrightarrow> \<diamond>G)) \<longrightarrow> \<box>(F \<longrightarrow> \<diamond>H)" by (rule STL4)
  have "\<turnstile> (\<box>(\<box>(G \<longrightarrow> \<diamond>H) \<and> (F \<longrightarrow> \<diamond>G))) = (\<box>\<box>(G \<longrightarrow> \<diamond>H) \<and> \<box>(F \<longrightarrow> \<diamond>G))"
    by (rule STL5)
  from this[THEN inteq_reflection] g1[THEN intD]
    have g2: "\<turnstile> (\<box>\<box>(G \<longrightarrow> \<diamond>H) \<and> \<box>(F \<longrightarrow> \<diamond>G)) \<longrightarrow> \<box>(F \<longrightarrow> \<diamond>H)" by auto
  have "\<turnstile> (\<box>(G \<longrightarrow> \<diamond>H)) = (\<box>\<box>(G \<longrightarrow> \<diamond>H))" by (rule T1)
  from this[THEN inteq_reflection] g2[THEN intD]
    have "\<turnstile>  \<box>(G \<longrightarrow> \<diamond>H) \<and> \<box>(F \<longrightarrow> \<diamond>G) \<longrightarrow> \<box>(F \<longrightarrow> \<diamond>H)" by auto  
  thus ?thesis by (unfold leadsto_def)
qed

theorem LT14: "\<turnstile> ((F \<or> G) \<leadsto> H) \<longrightarrow> (F \<leadsto> H)"
proof -
  have "\<turnstile> ((F \<or> G) \<longrightarrow> \<diamond>H) \<longrightarrow> (F \<longrightarrow> \<diamond>H)" by auto
  hence "\<turnstile> \<box>((F \<or> G) \<longrightarrow> \<diamond>H) \<longrightarrow> \<box>(F \<longrightarrow> \<diamond>H)" by (rule STL4)
  thus ?thesis by (unfold leadsto_def)
qed

theorem LT15: "\<turnstile> ((F \<or> G) \<leadsto> H) \<longrightarrow> (G \<leadsto> H)"
proof -
  have "\<turnstile> ((F \<or> G) \<longrightarrow> \<diamond>H) \<longrightarrow> (G \<longrightarrow> \<diamond>H)" by auto
  hence "\<turnstile> \<box>((F \<or> G) \<longrightarrow> \<diamond>H) \<longrightarrow> \<box>(G \<longrightarrow> \<diamond>H)" by (rule STL4)
  thus ?thesis by (unfold leadsto_def)
qed

theorem LT16: "\<turnstile> (F \<leadsto> H) \<and> (G \<leadsto> H) \<longrightarrow> ((F \<or> G) \<leadsto> H)"
proof -
  have "\<turnstile> (F \<longrightarrow> \<diamond>H) \<and> (G \<longrightarrow> \<diamond>H) \<longrightarrow> (F \<or> G \<longrightarrow> \<diamond>H)" 
    by auto
  hence g1: "\<turnstile> \<box>((F \<longrightarrow> \<diamond>H) \<and> (G \<longrightarrow> \<diamond>H)) \<longrightarrow> \<box>(F \<or> G \<longrightarrow> \<diamond>H)" 
    by (rule STL4)
  have "\<turnstile> (\<box>((F \<longrightarrow> \<diamond>H) \<and> (G \<longrightarrow> \<diamond>H))) = (\<box>(F \<longrightarrow> \<diamond>H) \<and> \<box>(G \<longrightarrow> \<diamond>H))"
    by (rule STL5)
  from this[THEN inteq_reflection] g1[THEN intD]
    have "\<turnstile> \<box>(F \<longrightarrow> \<diamond>H) \<and> \<box>(G \<longrightarrow> \<diamond>H) \<longrightarrow> \<box>((F \<or> G) \<longrightarrow> \<diamond>H)" by auto
  thus ?thesis by (unfold leadsto_def)  
qed

theorem LT17: "\<turnstile> ((F \<or> G) \<leadsto> H) = ((F \<leadsto> H) \<and> (G \<leadsto> H))"
proof (int_unlift, rule iffI)
  fix w
  assume a1: "w \<Turnstile> (F \<or> G) \<leadsto> H"
  have g1: "\<turnstile> ((F \<or> G) \<leadsto> H) \<longrightarrow> (F \<leadsto> H)" by (rule LT14)
  have g2: "\<turnstile> ((F \<or> G) \<leadsto> H) \<longrightarrow> (G \<leadsto> H)" by (rule LT15)
  from g1[THEN intD] g2[THEN intD]
    have "\<turnstile> ((F \<or> G) \<leadsto> H) \<longrightarrow> (F \<leadsto> H) \<and> (G \<leadsto> H)" by auto
  with a1 show "(w \<Turnstile> (F \<leadsto> H)) \<and> (w \<Turnstile> (G \<leadsto> H))" by auto
next
  fix w
  assume a1: "(w \<Turnstile> (F \<leadsto> H)) \<and> (w \<Turnstile> (G \<leadsto> H))"
  have "\<turnstile> (F \<leadsto> H) \<and> (G \<leadsto> H) \<longrightarrow> ((F \<or> G) \<leadsto> H)" by (rule LT16)
  from this[THEN intD] a1 show "w \<Turnstile> (F \<or> G) \<leadsto> H" by auto
qed

theorem LT18: "\<turnstile> (A \<leadsto> (B \<or> C)) \<and> (B \<leadsto> D) \<and> (C \<leadsto> D) \<longrightarrow> (A \<leadsto> D)"
proof -
  have "\<turnstile> ((B \<or> C) \<leadsto> D) = ((B \<leadsto> D) \<and> (C \<leadsto> D))" by (rule LT17)
  from this[THEN inteq_reflection]
    have g1: "\<turnstile> ((A \<leadsto> (B \<or> C)) \<and> (B \<leadsto> D) \<and> (C \<leadsto> D)) \<longrightarrow> (A \<leadsto> (B \<or> C)) \<and> ((B \<or> C) \<leadsto> D)"
      by auto
  have g2: "\<turnstile> ((B \<or> C) \<leadsto> D) \<and> (A \<leadsto> (B \<or> C)) \<longrightarrow> (A \<leadsto> D)" by (rule LT13)
  have "\<turnstile>  (((B \<or> C) \<leadsto> D) \<and> (A \<leadsto> (B \<or> C))) = ((A \<leadsto> (B \<or> C)) \<and> ((B \<or> C) \<leadsto> D))" 
    by (rule lift_and_com)
  from g2 this[THEN inteq_reflection]
    have "\<turnstile> (A \<leadsto> (B \<or> C)) \<and> ((B \<or> C) \<leadsto> D) \<longrightarrow> (A \<leadsto> D)" by auto
  with g1 show ?thesis by (rule lift_imp_trans)
qed  

theorem LT19: "\<turnstile> (A \<leadsto> (D \<or> B)) \<and> (B \<leadsto> D) \<longrightarrow> (A \<leadsto> D)"
proof (int_unlift,rule impI,erule conjE)
  fix w
  assume a1: "w \<Turnstile> A \<leadsto> (D \<or> B)"
    and  a2: "w \<Turnstile> B \<leadsto> D"
  have g1: "w \<Turnstile> (D \<or> B) \<leadsto> D"
  proof -
    have "\<turnstile> ((D \<or> B) \<leadsto> D) = ((D \<leadsto> D) \<and> (B \<leadsto> D))" by (rule LT17)
    hence tg1: "w \<Turnstile> (D \<leadsto> D) \<and> (B \<leadsto> D) \<longrightarrow> (D \<or> B) \<leadsto> D" by auto
    have "\<turnstile> (D \<leadsto> D)" by (rule LT1)
    from this[THEN intD] tg1 have tg2: "w \<Turnstile> (B \<leadsto> D) \<longrightarrow> (D \<or> B) \<leadsto> D" by auto
    with a2 show "w \<Turnstile> (D \<or> B) \<leadsto> D" by auto
  qed
  have  "\<turnstile> ((D \<or> B) \<leadsto> D) \<and> (A \<leadsto> (D \<or> B)) \<longrightarrow> (A \<leadsto> D)" by (rule LT13)
  from this[THEN intD] a1 g1 show "w \<Turnstile> A \<leadsto> D" by auto
qed

theorem LT20: "\<turnstile> (A \<leadsto> (B \<or> D)) \<and> (B \<leadsto> D) \<longrightarrow> (A \<leadsto> D)"
proof (int_unlift,rule impI,erule conjE)
  fix w
  assume a1: "w \<Turnstile> A \<leadsto> (B \<or> D)"
    and  a2: "w \<Turnstile> B \<leadsto> D"
  have g1: "w \<Turnstile> (B \<or> D) \<leadsto> D"
  proof -
    have "\<turnstile> ((B \<or> D) \<leadsto> D) = ((B \<leadsto> D) \<and> (D \<leadsto> D))" by (rule LT17)
    hence tg1: "w \<Turnstile> (B \<leadsto> D) \<and> (D \<leadsto> D) \<longrightarrow> (B \<or> D) \<leadsto> D" by auto
    with a2 have tg2: "w \<Turnstile>  (D \<leadsto> D) \<longrightarrow> (B \<or> D) \<leadsto> D" by auto
    have "\<turnstile> (D \<leadsto> D)" by (rule LT1)
    from this[THEN intD] tg2 show "w \<Turnstile> (B \<or> D) \<leadsto> D" by auto
  qed
  have  "\<turnstile> ((B \<or> D) \<leadsto> D) \<and> (A \<leadsto> (B \<or> D)) \<longrightarrow> (A \<leadsto> D)" by (rule LT13)
  from this[THEN intD] a1 g1 show "w \<Turnstile> A \<leadsto> D" by auto
qed

theorem LT21: "\<turnstile> ((\<exists> x. F x) \<leadsto> G) = (\<forall> x. (F x \<leadsto> G))"
proof -
  have "\<turnstile> (\<forall> x. (F x \<longrightarrow> \<not>\<box>\<not>G)) =  ((\<exists> x. F x) \<longrightarrow> \<not>\<box>\<not>G)" by auto
  hence g1: "\<turnstile> (\<box>(\<forall> x. (F x \<longrightarrow> \<not>\<box>\<not>G))) = (\<box>((\<exists> x. F x) \<longrightarrow> \<not>\<box>\<not>G))" by (rule MM1)
  have "\<turnstile> (\<forall>x. \<box>(F x \<longrightarrow> \<not>\<box>\<not>G)) = (\<box>(\<forall> x. (F x \<longrightarrow> \<not>\<box>\<not>G)))" by (rule allT)
  from this[THEN inteq_reflection] g1
   have "\<turnstile> (\<forall>x.(\<box>(F x \<longrightarrow> \<not>\<box>\<not>G))) = (\<box>((\<exists> x. F x) \<longrightarrow> \<not>\<box>\<not>G))" by auto
  from this[THEN intD] have "\<turnstile> (\<box>((\<exists> x. F x) \<longrightarrow> \<not>\<box>\<not>G)) = (\<forall>x.(\<box>(F x \<longrightarrow> \<not>\<box>\<not>G)))"
    by auto
  hence "\<turnstile> (\<box>((\<exists> x. F x) \<longrightarrow> \<diamond>G)) = (\<forall>x.(\<box>(F x \<longrightarrow> \<diamond>G)))" by (unfold eventually_def)
  thus ?thesis by (unfold leadsto_def)
qed

theorem LT22: "\<turnstile> (F \<leadsto> (G \<or> H)) \<longrightarrow> \<not>\<diamond>G \<longrightarrow> (F \<leadsto> H)"
proof (int_unlift,intro impI)
  fix w
  assume a1: "w \<Turnstile> (F \<leadsto> (G \<or> H))"
    and  a2: "\<not>(w \<Turnstile> \<diamond>G)"
  have g1: "\<turnstile> \<not>\<diamond>G \<longrightarrow> (G \<leadsto> H)"
  proof -
    have "\<turnstile> \<not>G \<longrightarrow> (G \<longrightarrow> \<diamond>H)" by auto
    hence tg1: "\<turnstile> \<box>\<not>G \<longrightarrow> \<box>(G \<longrightarrow> \<diamond>H)" by (rule STL4)
    have "\<turnstile> (\<box>\<not>G) = (\<not>\<diamond>\<not>\<not>G)" by (intro E2)
    from this[THEN intD] have "\<turnstile> \<not>\<diamond>\<not>\<not>G \<longrightarrow> \<box>\<not>G" by auto
    from this tg1 have tg1':"\<turnstile> \<not>\<diamond>\<not>\<not>G \<longrightarrow> \<box>(G \<longrightarrow> \<diamond>H)" by (rule lift_imp_trans)
    have "\<turnstile> G = (\<not>\<not>G)" by auto
    hence "\<turnstile> (\<diamond>G) = (\<diamond>(\<not>\<not>G))" by (intro E22)
    from this[THEN inteq_reflection] tg1'[THEN intD]
      have "\<turnstile> \<not>\<diamond>G \<longrightarrow> \<box>(G \<longrightarrow> \<diamond>H)" by simp 
    thus ?thesis by (unfold leadsto_def)
  qed 
  from this[THEN intD] a2 have a2':"w \<Turnstile> (G \<leadsto> H)" by auto
  have "\<turnstile> (F \<leadsto> (G \<or> H)) \<and> (G \<leadsto> H) \<longrightarrow> (F \<leadsto> H)" by (intro LT20)
  from this[THEN intD] a1 a2' show "w \<Turnstile> (F \<leadsto> H)" by auto
qed

lemma LT23: "|~ (P \<longrightarrow> \<circ>Q) \<longrightarrow> (P \<longrightarrow> \<diamond>Q)"
proof (pref_unlift, intro impI)
  fix w
  assume a1: "(w \<Turnstile> P) \<longrightarrow> (w \<Turnstile> \<circ>Q)"
    and  a2:  "w \<Turnstile> P"
  from a1 a2 have a: "w \<Turnstile> \<circ>Q" by simp
  have "|~ \<circ>Q \<longrightarrow> \<diamond>Q" by (rule E23)
  from this[THEN prefD] a show "w \<Turnstile> \<diamond>Q" 
    by simp
qed

theorem LT24: "\<turnstile> \<box>I \<and> ((P \<and> I) \<leadsto> Q) \<longrightarrow> P \<leadsto> Q"
proof -
  have "\<turnstile> I \<and> (P \<and> I \<longrightarrow> \<diamond>Q) \<longrightarrow> (P \<longrightarrow> \<diamond>Q)" 
    by auto
  hence g1: "\<turnstile> \<box>(I \<and> (P \<and> I \<longrightarrow> \<diamond>Q)) \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)"
    by (rule STL4)
  have "\<turnstile> (\<box>(I \<and> (P \<and> I \<longrightarrow> \<diamond>Q))) = (\<box>I \<and> \<box>(P \<and> I \<longrightarrow> \<diamond>Q))"
    by (rule STL5)
  from this[THEN intD]
    have "\<turnstile> \<box>I \<and> \<box>(P \<and> I \<longrightarrow> \<diamond>Q) \<longrightarrow> \<box>(I \<and> (P \<and> I \<longrightarrow> \<diamond>Q))"
      by auto
  from this g1 have "\<turnstile> \<box>I \<and> \<box>((P \<and> I) \<longrightarrow> \<diamond>Q) \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)"
    by (rule lift_imp_trans)
  thus ?thesis by (unfold leadsto_def)
qed
  
theorem LT25: "\<turnstile> (F \<leadsto> #False) = (\<box>\<not>F)"
proof (unfold leadsto_def)
  have "\<turnstile> (#False) = (\<diamond>(#False))"
    by (rule E15)
  from this[THEN intD] have "\<turnstile> (F \<longrightarrow> \<diamond>#False) = (F \<longrightarrow> #False)"
    by (auto simp add: Valid_def)
  from this[THEN intD] have "\<turnstile> (F \<longrightarrow> \<diamond>#False) = (\<not>F)"
    by (auto simp add: Valid_def)
  thus "\<turnstile> (\<box>(F \<longrightarrow> \<diamond>#False)) = (\<box>\<not>F)"
    by (rule MM1)
qed

lemma LT26: assumes h1: "\<turnstile> \<diamond>F \<longrightarrow> \<box>\<diamond>F"
            shows "\<turnstile> F \<longrightarrow> \<box>\<diamond>F"
proof -
  have "\<turnstile> F \<longrightarrow> \<diamond>F" by (rule E3)
  from this h1 show ?thesis
    by (rule lift_imp_trans)
qed

lemma LT28: assumes h1: "|~ P \<longrightarrow> \<circ>P \<or> \<circ>Q"
              shows     "|~ (P \<longrightarrow> \<circ>P) \<or> \<diamond>Q"
proof (pref_unlift)
  fix w
  show "((w \<Turnstile> P) \<longrightarrow> (w \<Turnstile> \<circ>P)) \<or> (w \<Turnstile> \<diamond>Q)"
  proof (cases "w \<Turnstile> \<circ>Q")
    assume a1: "w \<Turnstile> \<circ>Q"
    have "|~ \<circ>Q \<longrightarrow> \<diamond>Q" by (rule E23)
    from this[THEN prefD] a1 have "w \<Turnstile> \<diamond>Q" 
      by auto
    thus ?thesis by auto
  next
    assume a1: "\<not>(w \<Turnstile> \<circ>Q)"
    from h1 have "((w \<Turnstile> P) \<longrightarrow> (w \<Turnstile> \<circ>P)) \<or> (w \<Turnstile> \<circ>Q)"
      by auto
    with a1 show ?thesis by auto
  qed
qed

lemmas unless = LT28

lemma LT29: assumes h1: "|~ P \<longrightarrow> \<circ>P \<or> \<circ>Q"
              and   h2: "|~ P \<and> Unchanged v \<longrightarrow> \<circ>P"
              shows     "\<turnstile> P \<longrightarrow> \<box>P \<or> \<diamond>Q"
proof (int_unlift,intro impI)
  fix w
  assume aa1: "w \<Turnstile> P"
  show "(w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q)"
  proof (rule ccontr)
    assume "\<not>((w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q))"
    hence a1: "(w \<Turnstile> \<not>\<box>P) \<and> (w \<Turnstile> \<not>\<diamond>Q)" by auto
    from a1 have a2: "(w \<Turnstile> \<not>\<box>P)" by simp
    from a1 have a3: "(w \<Turnstile> \<box>\<not>Q)"
      by (simp add: eventually_def)
    with ax2[THEN intD] have a3': "w \<Turnstile> \<box>[\<box>\<not>Q]_v" 
      by auto
    from h1 have "|~ (P \<longrightarrow> \<circ>P) \<or> \<diamond>Q"
      by (rule LT28)
    from this[THEN prefD] have "|~ \<not>\<diamond>Q \<longrightarrow> (P \<longrightarrow> \<circ>P)"
      by auto
    from this[THEN prefD] have "|~ \<box>\<not>Q \<longrightarrow> (P \<longrightarrow> \<circ>P)"
      by (auto simp add: eventually_def)
    hence g1: "\<turnstile> \<box>[(\<box>\<not>Q \<longrightarrow> (P \<longrightarrow> \<circ>P))]_v" by (rule sq)
    have "\<turnstile> \<box>[(\<box>\<not>Q \<longrightarrow> (P \<longrightarrow> \<circ>P))]_v \<longrightarrow> \<box>[\<box>\<not>Q]_v \<longrightarrow> \<box>[(P \<longrightarrow> \<circ>P)]_v"
      by (rule ax4)
    with g1 have g2: "\<turnstile> \<box>[\<box>\<not>Q]_v \<longrightarrow> \<box>[(P \<longrightarrow> \<circ>P)]_v" by (rule fmp)
    from h2 have "\<turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_v \<longrightarrow> P \<longrightarrow> \<box>P" by (rule ax3)
    with g2 have g3: "\<turnstile> \<box>[\<box>\<not>Q]_v \<longrightarrow> P \<longrightarrow> \<box>P"
      by (rule lift_imp_trans)
    from this[THEN intD] aa1 a3' have "w \<Turnstile> \<box>P" by auto
    with a2 show False by auto
  qed
qed

lemma LT30: assumes h1: "|~ N \<and> P \<longrightarrow> \<circ>P \<or> \<circ>Q"
              shows     "|~ N \<longrightarrow> (P \<longrightarrow> \<circ>P) \<or> \<diamond>Q"
proof (pref_unlift,rule impI)
  fix w
  assume aa1: "w \<Turnstile> N"
  show "((w \<Turnstile> P) \<longrightarrow> (w \<Turnstile> \<circ>P)) \<or> (w \<Turnstile> \<diamond>Q)"
  proof (cases "w \<Turnstile> \<circ>Q")
    assume a1: "w \<Turnstile> \<circ>Q"
    have "|~ \<circ>Q \<longrightarrow> \<diamond>Q" by (rule E23)
    from this[THEN prefD] a1 have "w \<Turnstile> \<diamond>Q" 
      by auto
    thus ?thesis by auto
  next
    assume a1: "\<not>(w \<Turnstile> \<circ>Q)"
    from h1 have "((w \<Turnstile> N) \<longrightarrow> (w \<Turnstile> P) \<longrightarrow> (w \<Turnstile> \<circ>P)) \<or> (w \<Turnstile> \<circ>Q)"
      by auto
    with a1 aa1 show ?thesis by auto
  qed
qed

lemma LT31: assumes h1: "|~ N \<and> P \<longrightarrow> \<circ>P \<or> \<circ>Q"
              and   h2: "|~ P \<and> Unchanged v \<longrightarrow> \<circ>P"
              shows     "\<turnstile> \<box>N \<longrightarrow> P \<longrightarrow> \<box>P \<or> \<diamond>Q"
proof (int_unlift,intro impI)
  fix w
  assume aa1: "w \<Turnstile> P"
    and  aa2: "w \<Turnstile> \<box>N"
  from aa2 P4[THEN intD] have aa2': "w \<Turnstile> \<box>[N]_v" 
    by auto
  show "(w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q)"
  proof (rule ccontr)
    assume "\<not>((w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q))"
    hence a1: "(w \<Turnstile> \<not>\<box>P) \<and> (w \<Turnstile> \<not>\<diamond>Q)" by auto
    from a1 have a2: "(w \<Turnstile> \<not>\<box>P)" by simp
    from a1 have a3: "(w \<Turnstile> \<box>\<not>Q)"
      by (simp add: eventually_def)
    with ax2[THEN intD] have a3': "w \<Turnstile> \<box>[\<box>\<not>Q]_v" 
      by auto
    from h1 have "|~ N \<longrightarrow> (P \<longrightarrow> \<circ>P) \<or> \<diamond>Q"
      by (rule LT30)
    from this[THEN prefD] have "|~ \<not>\<diamond>Q \<longrightarrow> N \<longrightarrow> (P \<longrightarrow> \<circ>P)"
      by auto
    from this[THEN prefD] have "|~ \<box>\<not>Q \<longrightarrow> N \<longrightarrow> (P \<longrightarrow> \<circ>P)"
      by (auto simp add: eventually_def)
    hence g1: "\<turnstile> \<box>[(\<box>\<not>Q \<longrightarrow> N \<longrightarrow> (P \<longrightarrow> \<circ>P))]_v" by (rule sq)
    have "\<turnstile> \<box>[(\<box>\<not>Q \<longrightarrow> N \<longrightarrow> (P \<longrightarrow> \<circ>P))]_v \<longrightarrow> \<box>[\<box>\<not>Q]_v \<longrightarrow> \<box>[(N \<longrightarrow> (P \<longrightarrow> \<circ>P))]_v"
      by (rule ax4)
    with g1 have g2: "\<turnstile> \<box>[\<box>\<not>Q]_v \<longrightarrow> \<box>[(N \<longrightarrow> (P \<longrightarrow> \<circ>P))]_v" by (rule fmp)
    have "\<turnstile> \<box>[(N \<longrightarrow> (P \<longrightarrow> \<circ>P))]_v \<longrightarrow> \<box>[N]_v \<longrightarrow> \<box>[(P \<longrightarrow> \<circ>P)]_v"
      by (rule ax4)
    with g2 have g2': "\<turnstile> \<box>[\<box>\<not>Q]_v \<longrightarrow> \<box>[N]_v \<longrightarrow> \<box>[(P \<longrightarrow> \<circ>P)]_v"
      by (rule lift_imp_trans)
    from h2 have "\<turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_v \<longrightarrow> P \<longrightarrow> \<box>P" by (rule ax3)
    from g2'[THEN intD] this[THEN intD]
       have g3: "\<turnstile> \<box>[\<box>\<not>Q]_v \<longrightarrow> \<box>[N]_v \<longrightarrow> P \<longrightarrow> \<box>P"
      by (auto simp add: Valid_def)
    from this[THEN intD] aa1 a3' aa2' have "w \<Turnstile> \<box>P" by auto
    with a2 show False by auto
  qed
qed

lemma LT32: assumes h1: "|~ (P \<and> [N]_f) \<longrightarrow> \<circ>P \<or> \<circ>Q"
              and   h2: "|~ P \<and> Unchanged f \<longrightarrow> \<circ>P"
              shows     "\<turnstile> \<box>[N]_f \<longrightarrow> P \<longrightarrow> \<box>P \<or> \<diamond>Q"
proof (int_unlift,intro impI)
  fix w
  assume aa1: "w \<Turnstile> P"
    and  aa2: "w \<Turnstile> \<box>[N]_f"
  have "\<turnstile> (\<box>[[N]_f]_f) = (\<box>[N]_f)" by (rule T3)
  from this[THEN intD] aa2 have aa2': "w \<Turnstile> \<box>[[N]_f]_f"
    by auto
  from h1 have h1': "|~ ([N]_f \<and> P) \<longrightarrow> \<circ>P \<or> \<circ>Q"
    by auto
  show "(w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q)"
  proof (rule ccontr)
    assume "\<not>((w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q))"
    hence a1: "(w \<Turnstile> \<not>\<box>P) \<and> (w \<Turnstile> \<not>\<diamond>Q)" by auto
    from a1 have a2: "(w \<Turnstile> \<not>\<box>P)" by simp
    from a1 have a3: "(w \<Turnstile> \<box>\<not>Q)"
      by (simp add: eventually_def)
    with ax2[THEN intD] have a3': "w \<Turnstile> \<box>[\<box>\<not>Q]_f" 
      by auto
    from h1' have "|~ [N]_f \<longrightarrow> (P \<longrightarrow> \<circ>P) \<or> \<diamond>Q"
      by (rule LT30)
    from this[THEN prefD] have "|~ \<not>\<diamond>Q \<longrightarrow> [N]_f \<longrightarrow> (P \<longrightarrow> \<circ>P)"
      by auto
    from this[THEN prefD] have "|~ \<box>\<not>Q \<longrightarrow> [N]_f \<longrightarrow> (P \<longrightarrow> \<circ>P)"
      by (auto simp add: eventually_def)
    hence g1: "\<turnstile> \<box>[(\<box>\<not>Q \<longrightarrow> [N]_f \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f" by (rule sq)
    have "\<turnstile> \<box>[(\<box>\<not>Q \<longrightarrow> [N]_f \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f \<longrightarrow> \<box>[\<box>\<not>Q]_f \<longrightarrow> \<box>[([N]_f \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f"
      by (rule ax4)
    with g1 have g2: "\<turnstile> \<box>[\<box>\<not>Q]_f \<longrightarrow> \<box>[([N]_f \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f" by (rule fmp)
    have "\<turnstile> \<box>[([N]_f \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f \<longrightarrow> \<box>[[N]_f]_f \<longrightarrow> \<box>[(P \<longrightarrow> \<circ>P)]_f"
      by (rule ax4)
    with g2 have g2': "\<turnstile> \<box>[\<box>\<not>Q]_f \<longrightarrow> \<box>[[N]_f]_f \<longrightarrow> \<box>[(P \<longrightarrow> \<circ>P)]_f"
      by (rule lift_imp_trans)
    from h2 have "\<turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_f \<longrightarrow> P \<longrightarrow> \<box>P" by (rule ax3)
    from g2'[THEN intD] this[THEN intD]
       have g3: "\<turnstile> \<box>[\<box>\<not>Q]_f \<longrightarrow> \<box>[[N]_f]_f \<longrightarrow> P \<longrightarrow> \<box>P"
      by (auto simp add: Valid_def)
    from this[THEN intD] aa1 a3' aa2' have "w \<Turnstile> \<box>P" by auto
    with a2 show False by auto
  qed
qed

lemma AA1: "\<turnstile> \<box>[#False]_v \<longrightarrow> \<not>\<diamond>\<langle>Q\<rangle>_v"
proof (unfold angle_action_def)
  have "|~ [#False]_v \<longrightarrow> [\<not>Q]_v"
    by (auto simp add: actrans_def)
  hence g1: "\<turnstile> \<box>[([#False]_v \<longrightarrow> [\<not>Q]_v)]_v"
    by (rule sq)
  have "\<turnstile>  \<box>[([#False]_v \<longrightarrow> [\<not>Q]_v)]_v \<longrightarrow> \<box>[[#False]_v]_v \<longrightarrow> \<box>[[\<not>Q]_v]_v"
    by (rule ax4)
  with g1 have g2: "\<turnstile> \<box>[[#False]_v]_v \<longrightarrow> \<box>[[\<not>Q]_v]_v"
    by (rule fmp)
  from T3[THEN intD] have g3: "\<turnstile> \<box>[#False]_v \<longrightarrow> \<box>[[#False]_v]_v"
    by (auto simp add: Valid_def)
  from T3[THEN intD] have g4: "\<turnstile> \<box>[[\<not>Q]_v]_v \<longrightarrow> \<box>[\<not>Q]_v"
    by (auto simp add: Valid_def)
  from g3 g2 have "\<turnstile> \<box>[#False]_v \<longrightarrow> \<box>[[\<not>Q]_v]_v"
    by (rule lift_imp_trans)
   from this g4 have "\<turnstile> \<box>[#False]_v \<longrightarrow> \<box>[\<not>Q]_v"
    by (rule lift_imp_trans)
  from this[THEN intD] show "\<turnstile> \<box>[#False]_v \<longrightarrow> \<not>\<not>\<box>[\<not>Q]_v"
    by auto
qed

lemma AA2: "\<turnstile> \<box>[P]_v \<and> \<diamond>\<langle>Q\<rangle>_v \<longrightarrow> \<diamond>\<langle>P\<rangle>_v"
proof (int_unlift,intro impI,elim conjE)
  fix w
  assume a1: "w \<Turnstile> \<box>[P]_v"
   and   a2: "w \<Turnstile> \<diamond>\<langle>Q\<rangle>_v"
  show "w \<Turnstile> \<diamond>\<langle>P\<rangle>_v" 
  proof (unfold angle_action_def,rule ccontr)
    assume "\<not>(w \<Turnstile> \<not>\<box>[\<not>P]_v)"
    hence aa1: "w \<Turnstile> \<box>[\<not>P]_v" by simp
    have "|~ [P]_v \<longrightarrow> [\<not>P]_v \<longrightarrow> [#False]_v"
      by (auto simp add: actrans_def)
    hence g1: "\<turnstile> \<box>[([P]_v \<longrightarrow> [\<not>P]_v \<longrightarrow> [#False]_v)]_v"
      by (rule sq)
    have "\<turnstile> \<box>[([P]_v \<longrightarrow> [\<not>P]_v \<longrightarrow> [#False]_v)]_v \<longrightarrow>  
          \<box>[[P]_v]_v \<longrightarrow>  \<box>[([\<not>P]_v \<longrightarrow> [#False]_v)]_v"
      by (rule ax4)
    with g1 have g2: "\<turnstile> \<box>[[P]_v]_v \<longrightarrow> \<box>[([\<not>P]_v \<longrightarrow> [#False]_v)]_v"
      by (rule fmp)
    have "\<turnstile> \<box>[([\<not>P]_v \<longrightarrow> [#False]_v)]_v \<longrightarrow> \<box>[[\<not>P]_v]_v \<longrightarrow> \<box>[[#False]_v]_v"
      by (rule ax4)
    with g2 have g3: "\<turnstile> \<box>[[P]_v]_v \<longrightarrow> \<box>[[\<not>P]_v]_v \<longrightarrow> \<box>[[#False]_v]_v"
      by (rule lift_imp_trans)
    have eq1: "\<turnstile> (\<box>[[P]_v]_v) = (\<box>[P]_v)" by (rule T3)
    have eq2: "\<turnstile> (\<box>[[\<not>P]_v]_v) = (\<box>[\<not>P]_v)" by (rule T3)    
    have eq3: "\<turnstile> (\<box>[[#False]_v]_v) = (\<box>[#False]_v)" by (rule T3)   
    from g3 eq1[THEN inteq_reflection] 
            eq2[THEN inteq_reflection]
            eq3[THEN inteq_reflection]
    have "\<turnstile> \<box>[P]_v \<longrightarrow> \<box>[\<not>P]_v \<longrightarrow> \<box>[#False]_v" by auto
    from this[THEN intD] a1 aa1 have g4: "w \<Turnstile> \<box>[#False]_v"
      by auto
    have "\<turnstile> \<box>[#False]_v \<longrightarrow> \<not>\<diamond>\<langle>Q\<rangle>_v" by (rule AA1)
    from this[THEN intD] g4 have "w \<Turnstile> \<not>\<diamond>\<langle>Q\<rangle>_v"
      by auto
    with a2 show False by simp
  qed
qed

lemma AA3: "\<turnstile> \<box>[P \<longrightarrow> Q]_v \<and> \<box>P \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>Q"
proof (int_unlift,intro impI,elim conjE)
  fix w
  assume a1: "w \<Turnstile> \<box>[P \<longrightarrow> Q]_v"
    and  a2: "w \<Turnstile> \<box>P"
    and  a3: "w \<Turnstile> \<diamond>\<langle>A\<rangle>_v"
  have "\<turnstile> \<box>[(P \<longrightarrow> Q)]_v \<longrightarrow> \<box>[P]_v \<longrightarrow> \<box>[Q]_v"
    by (rule ax4)
  from this[THEN intD] a1 have a1': "w \<Turnstile> \<box>[P]_v \<longrightarrow> \<box>[Q]_v"
    by auto
  have "\<turnstile> \<box>P \<longrightarrow> \<box>[P]_v" by (rule P4)
  from this[THEN intD] a2 have "w \<Turnstile> \<box>[P]_v"
    by auto
  with a1' have g1: "w \<Turnstile> \<box>[Q]_v" by auto
  have "\<turnstile> \<box>[Q]_v \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>Q\<rangle>_v" by (rule AA2)
  from this[THEN intD] g1 a3 have g2: "w \<Turnstile> \<diamond>\<langle>Q\<rangle>_v"
    by auto
  have "\<turnstile> \<diamond>\<langle>Q\<rangle>_v \<longrightarrow> \<diamond>Q" by (rule E25)
  from this[THEN intD] g2 show "w \<Turnstile> \<diamond>Q"
    by auto
qed

lemma AA5: "\<turnstile> \<box>[P \<longrightarrow> Q]_v \<longrightarrow> \<diamond>\<langle>P\<rangle>_v \<longrightarrow> \<diamond>\<langle>Q\<rangle>_v"
proof (int_unlift,intro impI)
  fix w
  assume a1: "w \<Turnstile> \<box>[(P \<longrightarrow> Q)]_v"
    and  a2: "w \<Turnstile> \<diamond>\<langle>P\<rangle>_v"
  have "|~ (P \<longrightarrow> Q) \<longrightarrow> \<not>Q \<longrightarrow> \<not>P" 
    by auto
  hence g1: "\<turnstile> \<box>[((P \<longrightarrow> Q) \<longrightarrow> \<not>Q \<longrightarrow> \<not>P)]_v"
    by (rule sq)
  have "\<turnstile> \<box>[((P \<longrightarrow> Q) \<longrightarrow> \<not>Q \<longrightarrow> \<not>P)]_v \<longrightarrow> \<box>[(P \<longrightarrow> Q)]_v \<longrightarrow> \<box>[(\<not>Q \<longrightarrow> \<not>P)]_v" 
    by (rule ax4)
  with g1 have g2: "\<turnstile> \<box>[(P \<longrightarrow> Q)]_v \<longrightarrow> \<box>[(\<not> Q \<longrightarrow> \<not> P)]_v" by (rule fmp)
  from this[THEN intD] a1
  have g2: "w \<Turnstile> \<box>[(\<not> Q \<longrightarrow> \<not> P)]_v" by auto
  have "\<turnstile> \<box>[(\<not>Q \<longrightarrow> \<not>P)]_v \<longrightarrow> \<box>[\<not>Q]_v \<longrightarrow> \<box>[\<not>P]_v" 
    by (rule ax4)
  with g2 have "w \<Turnstile> \<box>[\<not>Q]_v \<longrightarrow> \<box>[\<not>P]_v"
    by auto
  hence "w \<Turnstile> \<not>\<box>[\<not>P]_v \<longrightarrow> \<not>\<box>[\<not>Q]_v"
    by auto
  hence "w \<Turnstile> \<diamond>\<langle>P\<rangle>_v \<longrightarrow> \<diamond>\<langle>Q\<rangle>_v"
    by (unfold angle_action_def)
  with a2 show "w \<Turnstile> \<diamond>\<langle>Q\<rangle>_v" by auto
qed

lemma AA6: "\<turnstile> \<box>[P \<longrightarrow> Q]_v \<and> \<diamond>\<langle>P\<rangle>_v \<longrightarrow> \<diamond>\<langle>Q\<rangle>_v"
proof -
  have "\<turnstile> \<box>[(P \<longrightarrow> Q)]_v \<longrightarrow> \<diamond>\<langle>P\<rangle>_v \<longrightarrow> \<diamond>\<langle>Q\<rangle>_v"
    by (rule AA5)
  from this[THEN intD] show ?thesis by
    (auto simp add: Valid_def)
qed

lemma AA7: assumes H1: "|~ P \<longrightarrow> Q"
              shows "\<turnstile> \<diamond>\<langle>P\<rangle>_v \<longrightarrow> \<diamond>\<langle>Q\<rangle>_v"
proof -
  from H1 have g1: "\<turnstile> \<box>[(P \<longrightarrow> Q)]_v" 
    by (rule sq)
  have "\<turnstile> \<box>[(P \<longrightarrow> Q)]_v \<longrightarrow> \<diamond>\<langle>P\<rangle>_v \<longrightarrow> \<diamond>\<langle>Q\<rangle>_v"
    by (rule AA5)
  with g1 show ?thesis by (rule fmp)
qed

lemma AA8: "\<turnstile> \<box>[P]_v \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<box>[P]_v \<and> A\<rangle>_v"
proof -
  have "\<turnstile> \<box>[\<not>(\<box>[P]_v \<and> A)]_v \<longrightarrow> (\<box>[P]_v \<longrightarrow> \<box>[\<not>A]_v)"
  proof (int_unlift,intro impI,rule ccontr) 
    fix w
    assume a1: "w \<Turnstile> \<box>[\<not>(\<box>[P]_v \<and> A)]_v"
      and  a2: "w \<Turnstile> \<box>[P]_v"
      and  a3: "\<not>(w \<Turnstile> \<box>[\<not>A]_v)"
    have "|~ \<not>(\<box>[P]_v \<and> A) \<longrightarrow> \<box>[P]_v \<longrightarrow> \<not>A"
      by auto
    hence g1: "\<turnstile> \<box>[(\<not>(\<box>[P]_v \<and> A) \<longrightarrow> \<box>[P]_v \<longrightarrow> \<not>A)]_v"
      by (rule sq)
    have "\<turnstile> \<box>[(\<not>(\<box>[P]_v \<and> A) \<longrightarrow> \<box>[P]_v \<longrightarrow> \<not>A)]_v 
           \<longrightarrow> \<box>[\<not>(\<box>[P]_v \<and> A)]_v \<longrightarrow> \<box>[(\<box>[P]_v \<longrightarrow> \<not>A)]_v"
      by (rule ax4)
    with g1 have "\<turnstile> \<box>[\<not>(\<box>[P]_v \<and> A)]_v \<longrightarrow> \<box>[(\<box>[P]_v \<longrightarrow> \<not>A)]_v"
      by (rule fmp)
    from this[THEN intD] a1 
      have g2: "w \<Turnstile> \<box>[(\<box>[P]_v \<longrightarrow> \<not>A)]_v" by auto
    have "\<turnstile> \<box>[(\<box>[P]_v \<longrightarrow> \<not>A)]_v \<longrightarrow> \<box>[\<box>[P]_v]_v \<longrightarrow> \<box>[\<not>A]_v"
      by (rule ax4)
    from this[THEN intD] g2 have g3: "w \<Turnstile> \<box>[\<box>[P]_v]_v \<longrightarrow> \<box>[\<not>A]_v"
      by auto
    have "\<turnstile> \<box>[P]_v \<longrightarrow> \<box>[\<box>[P]_v]_v" by (rule P5)
    from this[THEN intD] a2 have "w \<Turnstile> \<box>[\<box>[P]_v]_v"
      by auto
    with g3 have "w \<Turnstile> \<box>[\<not>A]_v" by auto
    with a3 show False by simp
  qed
  from this[THEN intD]
    have "\<turnstile> \<box>[\<not>(\<box>[P]_v \<and> A)]_v \<longrightarrow> \<not>\<box>[P]_v \<or> \<box>[\<not>A]_v" by auto
  from this[THEN intD]
    have "\<turnstile> \<box>[P]_v \<and> \<not>\<box>[\<not>A]_v \<longrightarrow> \<not>\<box>[\<not>(\<box>[P]_v \<and> A)]_v" 
      by (auto simp add: Valid_def)
  thus ?thesis by (unfold angle_action_def)
qed

lemma AA9: "\<turnstile> \<box>[P]_v \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>[P]_v \<and> A\<rangle>_v"
proof -
  have g1: "\<turnstile> \<box>[P]_v \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>(\<box>[P]_v \<and> A)\<rangle>_v"
    by (rule AA8)
  have g2: "|~ \<box>[P]_v \<longrightarrow> [P]_v"
  proof (pref_unlift,rule impI)
    fix w
    assume a1: "w \<Turnstile> \<box>[P]_v"
    have "w \<Turnstile> (\<box>[P]_v) = ([P]_v \<and> \<circ>\<box>[P]_v)" by (rule pax4[THEN prefD])
    with a1 show "w \<Turnstile> [P]_v" by auto
  qed
  from this[THEN prefD] have "|~ (\<box>[P]_v \<and> A) \<longrightarrow> ([P]_v \<and> A)"
    by auto
  hence "\<turnstile> \<diamond>\<langle>(\<box>[P]_v \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>([P]_v \<and> A)\<rangle>_v" 
    by (rule AA7)
  with g1 show ?thesis by (rule lift_imp_trans)
qed

lemma AA10: "\<turnstile> \<not>(\<box>[P]_v \<and> \<diamond>\<langle>\<not>P\<rangle>_v)"
proof (int_unlift,rule ccontr)
  fix w
  assume "\<not>\<not>((w \<Turnstile> \<box>[P]_v) \<and> (w \<Turnstile> \<diamond>\<langle>\<not>P\<rangle>_v))"
  hence A: "(w \<Turnstile> \<box>[P]_v) \<and> (w \<Turnstile> \<diamond>\<langle>\<not>P\<rangle>_v)" 
    by auto
  from A have a1: "w \<Turnstile> \<box>[P]_v" by simp
  from A have a2: "w \<Turnstile> \<diamond>\<langle>\<not>P\<rangle>_v" by simp
  from a2 have a2': "w \<Turnstile> \<not>\<box>[\<not>\<not>P]_v" by (unfold angle_action_def)
  have "|~ P \<longrightarrow> \<not>\<not>P" by auto
  hence g1: "\<turnstile> \<box>[(P \<longrightarrow> \<not>\<not>P)]_v" by (rule sq)
  have "\<turnstile> \<box>[(P \<longrightarrow> \<not>\<not>P)]_v \<longrightarrow> \<box>[P]_v \<longrightarrow> \<box>[\<not>\<not>P]_v"
    by (rule ax4)
  with g1 have "\<turnstile> \<box>[P]_v \<longrightarrow> \<box>[\<not>\<not>P]_v"
    by (rule fmp)
  from this[THEN intD] have "\<turnstile> \<not>\<box>[\<not>\<not>P]_v \<longrightarrow> \<not>\<box>[P]_v"
    by (auto simp add: Valid_def)
  from this[THEN intD] a2' have "w \<Turnstile> \<not>\<box>[P]_v"
    by simp
  with a1 show False by simp
qed

lemma AA11: "\<turnstile> \<not>\<diamond>\<langle>v$ = $v\<rangle>_v"
proof (int_unlift,rule ccontr)
  fix w
  assume "\<not>\<not>(w \<Turnstile> \<diamond>\<langle>v$ = $v\<rangle>_v)" 
  hence a1: "w \<Turnstile> \<diamond>\<langle>v$ = $v\<rangle>_v" by simp
  have "|~  v$ = $v \<longrightarrow> \<not>(v$ \<noteq> $v)"
    by auto
  from a1 intD[OF this[THEN AA7,of v]]
    have g1: "w \<Turnstile> \<diamond>\<langle>\<not>(v$ \<noteq> $v)\<rangle>_v" by auto
  have g2: "\<turnstile> \<box>[v$ \<noteq> $v]_v" by (rule ax5)
  have "\<turnstile> \<not>(\<box>[v$ \<noteq> $v]_v \<and> \<diamond>\<langle>\<not>(v$ \<noteq> $v)\<rangle>_v)"
    by (rule AA10)
  from this[THEN intD]
    have "(w \<Turnstile> \<box>[v$ \<noteq> $v]_v \<and> \<diamond>\<langle>\<not>(v$ \<noteq> $v)\<rangle>_v) \<longrightarrow> False"
    by auto
  with g1 g2[THEN intD, of w] show False by auto
qed

lemma AA12: "\<turnstile> (\<box>[P]_v) = (\<not>\<diamond>\<langle>\<not>P\<rangle>_v)"
proof (unfold angle_action_def)
  have "|~ P = (\<not>\<not>P)" by auto
  hence "\<turnstile> (\<box>[P]_v) = (\<box>[\<not>\<not>P]_v)"
    by (rule MM10)
  from this[THEN intD]
    show "\<turnstile> (\<box>[P]_v) = (\<not>\<not>\<box>[\<not>\<not>P]_v)"
    by auto
qed

lemma AA13: "\<turnstile> \<diamond>\<langle>P\<rangle>_v \<longrightarrow> \<diamond>\<langle>v$ \<noteq> $v\<rangle>_v"
proof (int_unlift,intro impI)
  fix w
  assume a1: "w \<Turnstile> \<diamond>\<langle>P\<rangle>_v"
  have g1: "\<turnstile> \<box>[v$ \<noteq> $v]_v" by (rule ax5)
  have g2: "\<turnstile> \<box>[v$ \<noteq> $v]_v \<and> \<diamond>\<langle>P\<rangle>_v \<longrightarrow> \<diamond>\<langle>v$ \<noteq> $v\<rangle>_v"
    by (rule AA2)
  from a1 g1[THEN intD] g2[THEN intD]
   show "w \<Turnstile> \<diamond>\<langle>v$ \<noteq> $v\<rangle>_v" by auto
qed

lemma AA14: "\<turnstile> (\<diamond>\<langle>P \<or> Q\<rangle>_v) = (\<diamond>\<langle>P\<rangle>_v \<or> \<diamond>\<langle>Q\<rangle>_v)"
proof (int_unlift,rule iffI)
  fix w
  assume a1: "(w \<Turnstile> \<diamond>\<langle>P\<rangle>_v) \<or> (w \<Turnstile> \<diamond>\<langle>Q\<rangle>_v)"
  have "|~ \<not>(P \<or> Q) \<longrightarrow> (\<not>P \<and> \<not>Q)" by auto
  hence g1: "\<turnstile> \<box>[\<not>(P \<or> Q)]_v \<longrightarrow> \<box>[(\<not>P \<and> \<not>Q)]_v" 
    by (rule MM8)
  have "\<turnstile> (\<box>[\<not>P]_v \<and> \<box>[\<not>Q]_v) = (\<box>[(\<not>P \<and> \<not>Q)]_v)"
    by (rule M8)
  from this[THEN intD] have
    "\<turnstile> \<box>[(\<not>P \<and> \<not>Q)]_v \<longrightarrow> (\<box>[\<not>P]_v \<and> \<box>[\<not>Q]_v)"
    by auto
  with g1 have "\<turnstile> \<box>[\<not>(P \<or> Q)]_v \<longrightarrow> (\<box>[\<not>P]_v \<and> \<box>[\<not>Q]_v)"
    by (rule lift_imp_trans)
  from this[THEN intD] have "\<turnstile> \<not>\<box>[\<not>P]_v \<or> \<not>\<box>[\<not>Q]_v \<longrightarrow> \<not>\<box>[\<not>(P \<or> Q)]_v"
    by (auto simp add: Valid_def)
  hence "\<turnstile> \<diamond>\<langle>P\<rangle>_v \<or> \<diamond>\<langle>Q\<rangle>_v \<longrightarrow> \<diamond>\<langle>(P \<or> Q)\<rangle>_v"
    by (unfold angle_action_def)
  from this[THEN intD] a1 show "w \<Turnstile> \<diamond>\<langle>(P \<or> Q)\<rangle>_v" by auto
next
  fix w
  assume a1: "w \<Turnstile> \<diamond>\<langle>(P \<or> Q)\<rangle>_v"
  have "|~ \<not>P \<and> \<not>Q \<longrightarrow> \<not>(P \<or> Q)" by auto
  hence g1: "\<turnstile> \<box>[(\<not>P \<and> \<not>Q)]_v \<longrightarrow> \<box>[\<not>(P \<or> Q)]_v" 
    by (rule MM8)
  have "\<turnstile> (\<box>[\<not>P]_v \<and> \<box>[\<not>Q]_v) = (\<box>[(\<not>P \<and> \<not>Q)]_v)"
    by (rule M8)
  from this[THEN intD] have
    "\<turnstile> (\<box>[\<not>P]_v \<and> \<box>[\<not>Q]_v) \<longrightarrow> \<box>[(\<not>P \<and> \<not>Q)]_v"
    by auto
  from this g1 have "\<turnstile> (\<box>[\<not>P]_v \<and> \<box>[\<not>Q]_v) \<longrightarrow> \<box>[\<not>(P \<or> Q)]_v"
    by (rule lift_imp_trans)
  from this[THEN intD] have "\<turnstile> \<not>\<box>[\<not>(P \<or> Q)]_v \<longrightarrow> \<not>\<box>[\<not>P]_v \<or> \<not>\<box>[\<not>Q]_v"
    by (auto simp add: Valid_def)
  hence "\<turnstile> \<diamond>\<langle>(P \<or> Q)\<rangle>_v \<longrightarrow>  \<diamond>\<langle>P\<rangle>_v \<or> \<diamond>\<langle>Q\<rangle>_v" 
    by (unfold angle_action_def)
  from this[THEN intD] a1 show "(w \<Turnstile> \<diamond>\<langle>P\<rangle>_v) \<or> (w \<Turnstile> \<diamond>\<langle>Q\<rangle>_v)" 
    by auto
qed

lemma AA15: "\<turnstile> \<diamond>\<langle>P \<and> Q\<rangle>_v \<longrightarrow> \<diamond>\<langle>P\<rangle>_v"
proof -
  have "|~ P \<and> Q \<longrightarrow> P" by auto
  thus ?thesis by (rule AA7)
qed

lemma AA16: "\<turnstile> \<diamond>\<langle>P \<and> Q\<rangle>_v \<longrightarrow> \<diamond>\<langle>Q\<rangle>_v"
proof -
  have "|~ P \<and> Q \<longrightarrow> Q" by auto
  thus ?thesis by (rule AA7)
qed

lemma AA17: "\<turnstile> \<diamond>\<langle>[P]_v \<and> A\<rangle>_v \<longrightarrow> \<diamond>\<langle>P \<and> A\<rangle>_v"
proof (int_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> \<diamond>\<langle>[P]_v \<and> A\<rangle>_v"
  have "|~ [P]_v \<and> A \<longrightarrow> ((v$ = $v \<and> A) \<or> (P \<and> A))"
    by (auto simp add: actrans_def unch_def)
  hence g1: "\<turnstile> \<diamond>\<langle>([P]_v \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>((v$ = $v \<and> A) \<or> (P \<and> A))\<rangle>_v"
    by (rule AA7)
  have "\<turnstile> (\<diamond>\<langle>((v$ = $v \<and> A) \<or> (P \<and> A))\<rangle>_v) = (\<diamond>\<langle>(v$ = $v \<and> A)\<rangle>_v \<or> \<diamond>\<langle>(P \<and> A)\<rangle>_v)"
    by (rule AA14)
  from this[THEN intD]
    have "\<turnstile> \<diamond>\<langle>((v$ = $v \<and> A) \<or> (P \<and> A))\<rangle>_v \<longrightarrow>
             (\<diamond>\<langle>(v$ = $v \<and> A)\<rangle>_v \<or> \<diamond>\<langle>(P \<and> A)\<rangle>_v)"
      by auto
  with g1 have "\<turnstile> \<diamond>\<langle>([P]_v \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>(v$ = $v \<and> A)\<rangle>_v \<or> \<diamond>\<langle>(P \<and> A)\<rangle>_v"
    by (rule lift_imp_trans)
  from this[THEN intD] a1
    have g2: "w \<Turnstile> \<not>\<diamond>\<langle>(v$ = $v \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>(P \<and> A)\<rangle>_v" 
      by auto
  have "w \<Turnstile> \<not>\<diamond>\<langle>(v$ = $v \<and> A)\<rangle>_v"
  proof (rule ccontr)
    assume a2: "\<not>(w \<Turnstile> \<not>\<diamond>\<langle>(v$ = $v \<and> A)\<rangle>_v)"
    hence a2': "w \<Turnstile> \<diamond>\<langle>(v$ = $v \<and> A)\<rangle>_v" by auto
    have "\<turnstile> \<diamond>\<langle>(v$ = $v \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>v$ = $v\<rangle>_v" 
      by (rule AA15)
    from this[THEN intD] a2' have tg1: "w \<Turnstile> \<diamond>\<langle>v$ = $v\<rangle>_v"
      by auto
    have "\<turnstile> \<not>\<diamond>\<langle>v$ = $v\<rangle>_v" by (rule AA11)
    from this[THEN intD] tg1 show False by auto
  qed
  with g2 show "w \<Turnstile> \<diamond>\<langle>(P \<and> A)\<rangle>_v" by auto
qed

lemma AA18: "\<turnstile> \<box>[P]_v \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>P \<and> A\<rangle>_v"
proof -
  have g1: "\<turnstile> \<box>[P]_v \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>([P]_v \<and> A)\<rangle>_v"
    by (rule AA9)
  have g2: "\<turnstile> \<diamond>\<langle>([P]_v \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>(P \<and> A)\<rangle>_v"
    by (rule AA17)
  with g1 show ?thesis by (rule lift_imp_trans)
qed

lemma AA19: "\<turnstile> \<box>P \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>P \<and> A\<rangle>_v"
proof -
  have g1: "\<turnstile> \<box>P \<longrightarrow> \<box>[P]_v" by (rule P4)
  have "\<turnstile> \<box>[P]_v \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>(P \<and> A)\<rangle>_v"
    by (rule AA18)
  from this[THEN intD] g1[THEN intD]
    show ?thesis by auto
qed

lemma AA20: assumes h1: "|~ P \<longrightarrow> \<circ>P \<or> \<circ>Q"
               and  h2: "|~ P \<and> A \<longrightarrow> \<circ>Q"
               and  h3: "|~ P \<and> Unchanged w \<longrightarrow> \<circ>P" (* replace with stutinv P *)
              shows     "\<turnstile> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> (P \<leadsto> Q)"
proof (unfold leadsto_def)
  have "\<turnstile> (\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> (P \<longrightarrow> \<diamond>Q)"
  proof (int_unlift,intro impI)
    fix w
    assume a1: "(w \<Turnstile> \<box>P) \<longrightarrow> (w \<Turnstile> \<diamond>\<langle>A\<rangle>_v)"
     and   a2: "w \<Turnstile> P"
    from h1 h3 have g1: "\<turnstile> P \<longrightarrow> \<box>P \<or> \<diamond>Q"
      by (rule LT29)
    from this[THEN intD] a2 have "(w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q)"
      by auto
    thus "w \<Turnstile> \<diamond>Q"
    proof
      assume "w \<Turnstile> \<diamond>Q" thus ?thesis .
    next
      assume aa1: "w \<Turnstile> \<box>P"
      with a1 have aa2: "w \<Turnstile> \<diamond>\<langle>A\<rangle>_v" by simp 
      have tg1: "|~ \<circ>Q \<longrightarrow> \<diamond>Q" by (rule E23)
      have "\<turnstile> \<box>P \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>(P \<and> A)\<rangle>_v" 
	by (rule AA19)
      from this[THEN intD] aa1 aa2
        have tg0: "w \<Turnstile> \<diamond>\<langle>(P \<and> A)\<rangle>_v"
	by auto
      from h2[THEN prefD] tg1[THEN prefD]
	have "|~ P \<and> A \<longrightarrow> \<diamond>Q" by auto
      hence tg2: "\<turnstile> \<box>[(P \<and> A \<longrightarrow> \<diamond>Q)]_v"
	by (rule sq)
      have "\<turnstile> \<box>[((P \<and> A) \<longrightarrow> \<diamond>Q)]_v \<longrightarrow> \<diamond>\<langle>(P \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_v"
	by (rule AA5)
      with tg2 have "\<turnstile> \<diamond>\<langle>(P \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_v"
	by (rule fmp)
      from this[THEN intD] tg0 have tg3: "w \<Turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_v"
	by auto
      have "\<turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_v \<longrightarrow> \<diamond>\<diamond>Q" by (rule E25)
      from this[THEN intD] tg3 have tg4: "w \<Turnstile> \<diamond>\<diamond>Q"
	by auto
      with E10[of Q,THEN inteq_reflection]
       show "w \<Turnstile> \<diamond>Q" by auto
    qed
  qed
  thus "\<turnstile> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)" by (rule STL4)
qed

lemma AA21: "|~ \<diamond>\<langle>\<circ>F\<rangle>_v \<longrightarrow> \<circ>\<diamond>F"
proof -
  have "|~ \<circ>\<box>\<not>F \<longrightarrow> \<box>[\<circ>\<not>F]_v" by (rule pax5)
  from this[THEN prefD]
    have g1: "|~ \<not>\<box>[\<circ>\<not>F]_v \<longrightarrow> \<not>\<circ>\<box>\<not>F" by auto
  have "|~ (\<circ>\<not>\<box>\<not>F) = (\<not>\<circ>\<box>\<not>F)" by (rule pax1)
  from this[THEN prefD]
    have "|~ \<not>\<circ>\<box>\<not>F \<longrightarrow> \<circ>\<not>\<box>\<not>F" by auto
  with g1 have g2: "|~ \<not>\<box>[\<circ>\<not>F]_v \<longrightarrow> \<circ>\<not>\<box>\<not>F" by auto
  have "|~ (\<circ>\<not>F) = (\<not>\<circ>F)" by (rule pax1)
  from this[THEN prefD] have "|~ \<circ>\<not>F \<longrightarrow> \<not>\<circ>F" 
    by auto
  hence "|~ \<box>[\<circ>\<not>F]_v \<longrightarrow> \<box>[\<not>\<circ>F]_v" by (rule MM8[THEN pre])
  from this[THEN prefD]
    have "|~ \<not>\<box>[\<not>\<circ>F]_v \<longrightarrow> \<not>\<box>[\<circ>\<not>F]_v" by auto
  from this g2 have "|~ \<not>\<box>[\<not>\<circ>F]_v \<longrightarrow> \<circ>\<not>\<box>\<not>F" by auto
  thus ?thesis by (unfold angle_action_def eventually_def)
qed

lemma AA22: assumes h1: "|~ N \<and> P \<longrightarrow> \<circ>P \<or> \<circ>Q"
               and  h2: "|~ (N \<and> P) \<and> A \<longrightarrow> \<circ>Q"
               and  h3: "|~ P \<and> Unchanged w \<longrightarrow> \<circ>P" (* replace with stutinv P *)
              shows     "\<turnstile> \<box>N \<and> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> (P \<leadsto> Q)"
proof (unfold leadsto_def)
  have "\<turnstile> \<box>N \<and> (\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> (P \<longrightarrow> \<diamond>Q)"
  proof (int_unlift,intro impI,elim conjE)
    fix w
    assume a1: "(w \<Turnstile> \<box>P) \<longrightarrow> (w \<Turnstile> \<diamond>\<langle>A\<rangle>_v)"
     and   a2: "w \<Turnstile> P"
     and   a3: "w \<Turnstile> \<box>N"
     from h1 h3 have g1: "\<turnstile> \<box>N \<longrightarrow> P \<longrightarrow> \<box>P \<or> \<diamond>Q"
      by (rule LT31)
    from this[THEN intD] a2 a3 have "(w \<Turnstile> \<box>P) \<or> (w \<Turnstile> \<diamond>Q)"
      by auto
    thus "w \<Turnstile> \<diamond>Q"
    proof
      assume "w \<Turnstile> \<diamond>Q" thus ?thesis .
    next
      assume aa1: "w \<Turnstile> \<box>P"
      have "\<turnstile> (\<box>(N \<and> P)) = (\<box>N \<and> \<box>P)"
	by (rule STL5)
      from this[THEN intD] have "\<turnstile> \<box>N \<and> \<box>P \<longrightarrow> \<box>(N \<and> P)"
	by auto
      from this[THEN intD] aa1 a3
        have aa3: "w \<Turnstile> \<box>(N \<and> P)" by simp
      from aa1 a1 have aa2: "w \<Turnstile> \<diamond>\<langle>A\<rangle>_v" by simp 
      have tg1: "|~ \<circ>Q \<longrightarrow> \<diamond>Q" by (rule E23)
      have "\<turnstile> \<box>(N \<and> P) \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>((N \<and> P) \<and> A)\<rangle>_v" 
	by (rule AA19)
      from this[THEN intD] aa3 aa2
        have tg0: "w \<Turnstile> \<diamond>\<langle>((N \<and> P) \<and> A)\<rangle>_v"
	by auto
      from h2[THEN prefD] tg1[THEN prefD]
	have "|~ (N \<and> P) \<and> A \<longrightarrow> \<diamond>Q" by auto
      hence tg2: "\<turnstile> \<box>[((N \<and> P) \<and> A \<longrightarrow> \<diamond>Q)]_v"
	by (rule sq)
      have "\<turnstile> \<box>[(((N \<and> P) \<and> A) \<longrightarrow> \<diamond>Q)]_v \<longrightarrow> \<diamond>\<langle>((N \<and> P) \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_v"
	by (rule AA5)
      with tg2 have "\<turnstile> \<diamond>\<langle>((N \<and> P) \<and> A)\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>Q\<rangle>_v"
	by (rule fmp)
      from this[THEN intD] tg0 have tg3: "w \<Turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_v"
	by auto
      have "\<turnstile> \<diamond>\<langle>\<diamond>Q\<rangle>_v \<longrightarrow> \<diamond>\<diamond>Q" by (rule E25)
      from this[THEN intD] tg3 have tg4: "w \<Turnstile> \<diamond>\<diamond>Q"
	by auto
      with E10[of Q,THEN inteq_reflection]
       show "w \<Turnstile> \<diamond>Q" by auto
    qed
  qed
  hence g1: "\<turnstile> \<box>(\<box>N \<and> (\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v)) \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)" by (rule STL4)
  have "\<turnstile> (\<box>(\<box>N \<and> (\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v))) = (\<box>\<box>N \<and> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v))"
    by (rule STL5)
  from this[THEN intD] have "\<turnstile> \<box>\<box>N \<and> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>(\<box>N \<and> (\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v))"
    by auto
  from this g1 have g2: "\<turnstile> \<box>\<box>N \<and> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)"
    by (rule lift_imp_trans)
  have "\<turnstile> (\<box>N) = (\<box>\<box>N)" by (rule STL3)
  from this[THEN inteq_reflection] g2
    show g2: "\<turnstile> \<box>N \<and> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>(P \<longrightarrow> \<diamond>Q)"
      by auto
qed

lemma AA23: assumes h1: "|~ N \<and> P \<longrightarrow> \<circ>P \<or> \<circ>Q"
               and  h2: "|~ (N \<and> P) \<and> A \<longrightarrow> \<circ>Q"
               and  h3: "|~ P \<and> Unchanged w \<longrightarrow> \<circ>P" (* replaced by stutinv P *)
              shows     "\<turnstile> \<box>N \<and> \<box>\<diamond>\<langle>A\<rangle>_v \<longrightarrow> (P \<leadsto> Q)"
proof(int_unlift,intro impI,elim conjE)
  fix w
  assume a1: "w \<Turnstile> \<box>N"
    and  a2: "w \<Turnstile> \<box>\<diamond>\<langle>A\<rangle>_v"
  from h1 h2 h3 have g1: "\<turnstile> \<box>N \<and> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> (P \<leadsto> Q)"
    by (rule AA22)
  have "\<turnstile> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v" by auto
  hence "\<turnstile> \<box>\<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v)"
    by (rule STL4)
  from this[THEN intD] a2 have "w \<Turnstile> \<box>(\<box>P \<longrightarrow> \<diamond>\<langle>A\<rangle>_v)"
    by auto
  with g1[THEN intD] a1 show "w \<Turnstile> (P \<leadsto> Q)"
    by auto
qed

theorem AA24: "\<turnstile> (\<diamond>\<langle>\<langle>P\<rangle>_f\<rangle>_f) = (\<diamond>\<langle>P\<rangle>_f)"
proof (unfold angle_action_def angle_actrans_def)
  have g1: "\<turnstile> (\<box>[\<not>\<not>[\<not>P]_f]_f) = (\<box>[[\<not>P]_f]_f)"
  proof -
    have "|~ (\<not>\<not>[\<not>P]_f) = ([\<not>P]_f)" by auto
    thus ?thesis by (rule MM10)
  qed
  have "\<turnstile> (\<box>[[\<not> P]_f]_f) = (\<box>[\<not> P]_f)" by (rule T3)
  with g1[THEN inteq_reflection]
    have "\<turnstile> (\<box>[\<not> \<not>[\<not> P]_f]_f) = (\<box>[\<not> P]_f)"
     by auto
  from this[THEN intD] 
   show "\<turnstile> (\<not>\<box>[\<not> \<not>[\<not> P]_f]_f) = (\<not>\<box>[\<not> P]_f)"
      by auto
qed

lemma AA25: assumes h1: "|~ \<langle>P\<rangle>_v \<longrightarrow> \<langle>Q\<rangle>_w"
              shows "\<turnstile> \<diamond>\<langle>P\<rangle>_v \<longrightarrow> \<diamond>\<langle>Q\<rangle>_w"
proof -
  from h1 have "|~ \<not>[\<not>P]_v \<longrightarrow> \<not>[\<not>Q]_w"
    by (unfold angle_actrans_def)
  from this[THEN prefD]
    have "|~ [\<not>Q]_w \<longrightarrow> [\<not>P]_v" by auto
  hence g1: "\<turnstile> \<box>[[\<not>Q]_w]_w \<longrightarrow> \<box>[[\<not>P]_v]_w"
    by (rule MM8)
  have "\<turnstile> (\<box>[[\<not>Q]_w]_w) = (\<box>[\<not>Q]_w)"
    by (rule T3)
  from this[THEN intD] have "\<turnstile> \<box>[\<not>Q]_w \<longrightarrow> \<box>[[\<not>Q]_w]_w"
    by auto
  from this g1 have g2: "\<turnstile> \<box>[\<not>Q]_w \<longrightarrow> \<box>[[\<not>P]_v]_w"
    by (rule lift_imp_trans)
  have "\<turnstile> \<box>[[\<not>P]_v]_w \<longrightarrow> \<box>[[\<not>P]_w]_v" by (rule T5)
  with g2 have g3: "\<turnstile> \<box>[\<not>Q]_w \<longrightarrow> \<box>[[\<not>P]_w]_v" 
    by (rule lift_imp_trans)
  from h1[THEN prefD] have "|~ \<not>[\<not>P]_v \<longrightarrow> \<not>[\<not>P]_w"
    by (auto simp add: angle_actrans_def actrans_def)  
  from this[THEN prefD] have "|~ [\<not>P]_w \<longrightarrow> [\<not>P]_v"
    by auto
  hence "\<turnstile> \<box>[[\<not>P]_w]_v \<longrightarrow> \<box>[[\<not>P]_v]_v"
    by (rule MM8)
  with g3 have g4: "\<turnstile> \<box>[\<not>Q]_w \<longrightarrow> \<box>[[\<not>P]_v]_v"
    by (rule lift_imp_trans)
  have "\<turnstile> (\<box>[[\<not>P]_v]_v) = (\<box>[\<not>P]_v)"
    by (rule T3)
  from this[THEN intD] have "\<turnstile> \<box>[[\<not>P]_v]_v \<longrightarrow> \<box>[\<not>P]_v"
    by auto
  with g4 have "\<turnstile> \<box>[\<not>Q]_w \<longrightarrow> \<box>[\<not>P]_v" 
    by (rule lift_imp_trans)  
  from this[THEN intD] have "\<turnstile> \<not>\<box>[\<not>P]_v \<longrightarrow> \<not>\<box>[\<not>Q]_w"
    by (auto simp add: Valid_def)
  thus ?thesis by (unfold angle_action_def)
qed

theorem AA26: "\<turnstile> (\<diamond>\<box>[\<not>P]_v) = (\<not>\<box>\<diamond>\<langle>P\<rangle>_v)"
  by (int_unlift) (simp add: eventually_def angle_action_def)

theorem AA27: "\<turnstile> (\<not>\<diamond>\<box>[\<not>P]_v) = (\<box>\<diamond>\<langle>P\<rangle>_v)"
  by (int_unlift) (simp add: eventually_def angle_action_def)

theorem AA28: "\<turnstile> (\<diamond>\<langle>A\<rangle>_v) = (\<diamond>\<diamond>\<langle>A\<rangle>_v)"
proof (unfold angle_action_def eventually_def,int_unlift,rule iffI)
  fix w
  assume a1: "\<not>(w \<Turnstile> \<box>[\<not> A]_v)"
  have "\<turnstile> \<box>\<not>\<not>\<box>[\<not> A]_v \<longrightarrow> \<not>\<not>\<box>[\<not> A]_v"
    by (rule ax1)
  from this[THEN intD] have "\<turnstile> \<box>\<not>\<not>\<box>[\<not> A]_v \<longrightarrow> \<box>[\<not> A]_v"
    by auto
  from this[THEN intD] have "\<turnstile> \<not>\<box>[\<not> A]_v \<longrightarrow> \<not>\<box>\<not>\<not>\<box>[\<not> A]_v"
    by (auto simp add: Valid_def)
  from this[THEN intD] a1 show "\<not>(w \<Turnstile> \<box>\<not>\<not>\<box>[\<not> A]_v)"
    by auto
next
  fix w
  assume a1: "\<not>(w \<Turnstile> \<box>\<not>\<not>\<box>[\<not> A]_v)"
  have "\<turnstile> \<box>[\<not> A]_v \<longrightarrow> \<not>\<not>\<box>[\<not> A]_v" by auto
  hence g1: "\<turnstile> \<box>\<box>[\<not> A]_v \<longrightarrow> \<box>\<not>\<not>\<box>[\<not> A]_v" by (rule STL4)
  have "\<turnstile> (\<box>[\<not> A]_v) = (\<box>\<box>[\<not> A]_v)" by (rule T2)
  from this[THEN intD] have "\<turnstile> \<box>[\<not> A]_v \<longrightarrow> \<box>\<box>[\<not> A]_v"
    by auto
  from this g1 have  "\<turnstile> \<box>[\<not> A]_v \<longrightarrow> \<box>\<not>\<not>\<box>[\<not> A]_v"
    by (rule lift_imp_trans)
  from this[THEN intD] have "\<turnstile> \<not>\<box>\<not>\<not>\<box>[\<not> A]_v \<longrightarrow> \<not>\<box>[\<not> A]_v"
    by (auto simp add: Valid_def)  
  from this[THEN intD] a1 show "\<not>(w \<Turnstile> \<box>[\<not> A]_v)"
    by auto
qed

theorem AA29: "\<turnstile> \<box>[N]_v \<and> \<box>\<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<box>\<diamond>\<langle>N \<and> A\<rangle>_v"
proof -
  have g1: "\<turnstile> \<box>[N]_v \<and> \<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<diamond>\<langle>(N \<and> A)\<rangle>_v" 
    by (rule AA18)
  hence g2: "\<turnstile> \<box>(\<box>[N]_v \<and> \<diamond>\<langle>A\<rangle>_v) \<longrightarrow> \<box>\<diamond>\<langle>(N \<and> A)\<rangle>_v"
    by (rule STL4)
  have "\<turnstile> (\<box>(\<box>[N]_v \<and> \<diamond>\<langle>A\<rangle>_v)) = (\<box>\<box>[N]_v \<and> \<box>\<diamond>\<langle>A\<rangle>_v)"
    by (rule STL5)
  from this[THEN intD]
    have "\<turnstile> \<box>\<box>[N]_v \<and> \<box>\<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<box>(\<box>[N]_v \<and> \<diamond>\<langle>A\<rangle>_v)"
    by auto
  from this g2 have g3: "\<turnstile> \<box>\<box>[N]_v \<and> \<box>\<diamond>\<langle>A\<rangle>_v \<longrightarrow> \<box>\<diamond>\<langle>(N \<and> A)\<rangle>_v"
    by (rule lift_imp_trans)
  with T2[of N v,THEN inteq_reflection] show ?thesis by auto
qed

theorem AA30: "\<turnstile> (\<diamond>\<langle>\<diamond>\<langle>P\<rangle>_f\<rangle>_f) = (\<diamond>\<langle>P\<rangle>_f)"
proof (unfold angle_action_def angle_actrans_def)
  have g1: "\<turnstile> (\<box>[\<not>\<not>\<box>[\<not>P]_f]_f) = (\<box>[\<box>[\<not>P]_f]_f)"
  proof -
    have "|~ (\<not>\<not>\<box>[\<not>P]_f) = (\<box>[\<not>P]_f)" by auto
    thus ?thesis by (rule MM10)
  qed
  have "\<turnstile> (\<box>[\<box>[\<not> P]_f]_f) = (\<box>[\<not> P]_f)" by (rule MM12)
  with g1[THEN inteq_reflection]
    have "\<turnstile> (\<box>[\<not> \<not>\<box>[\<not> P]_f]_f) = (\<box>[\<not> P]_f)"
     by auto
  from this[THEN intD] 
   show "\<turnstile> (\<not>\<box>[\<not> \<not>\<box>[\<not> P]_f]_f) = (\<not>\<box>[\<not> P]_f)"
      by auto
qed

theorem AA31: "\<turnstile> \<diamond>\<langle>\<circ>F\<rangle>_v \<longrightarrow> \<diamond>F"
proof -
  thm AA21
  have g1: "|~ \<diamond>\<langle>\<circ>F\<rangle>_v \<longrightarrow> \<circ>\<diamond>F" by (rule AA21)
  have g2: "|~ \<circ>\<diamond>F \<longrightarrow> \<diamond>F" by (rule E29)
  from g1[THEN prefD] g2[THEN prefD]
  have "|~ \<diamond>\<langle>\<circ>F\<rangle>_v \<longrightarrow> \<diamond>F" by auto
  hence G1: "\<turnstile> \<diamond>\<langle>\<diamond>\<langle>\<circ>F\<rangle>_v\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>F\<rangle>_v"
    by (rule AA7)
  have "\<turnstile> (\<diamond>\<langle>\<diamond>\<langle>\<circ>F\<rangle>_v\<rangle>_v)= (\<diamond>\<langle>\<circ>F\<rangle>_v)"
    by (rule AA30)
  from this[THEN inteq_reflection,symmetric] G1
  have G2: "\<turnstile> \<diamond>\<langle>\<circ>F\<rangle>_v \<longrightarrow> \<diamond>\<langle>\<diamond>F\<rangle>_v" by auto
  have "\<turnstile> \<diamond>\<langle>\<diamond>F\<rangle>_v \<longrightarrow> \<diamond>\<diamond>F" by (rule E25)
  with E10[of F,THEN inteq_reflection]
  have "\<turnstile> \<diamond>\<langle>\<diamond>F\<rangle>_v \<longrightarrow> \<diamond>F" by auto
  with G2 show ?thesis by (rule lift_imp_trans)
qed

lemma log1: assumes h1: "(\<turnstile> P) = (\<turnstile> Q)" shows "(\<turnstile> \<box>P) = (\<turnstile> \<box>Q)"
proof
    assume a1: "\<turnstile> \<box>P"
    hence "\<turnstile> P"
    proof int_unlift
      fix w
      from ax1[THEN intD] have "w \<Turnstile> \<box>P \<longrightarrow> P" by auto 
      with a1[THEN intD] show "w \<Turnstile> P" by auto
    qed
    hence "\<turnstile> Q" by (simp add: h1)
    thus "\<turnstile> \<box>Q" by (rule alw)
next
    assume a1: "\<turnstile> \<box>Q"
    hence "\<turnstile> Q"
    proof int_unlift
      fix w
      from ax1[THEN intD] have "w \<Turnstile> \<box>Q \<longrightarrow> Q" by auto 
      with a1[THEN intD] show "w \<Turnstile> Q" by auto
    qed
    hence "\<turnstile> P" by (simp add: h1)
    thus "\<turnstile> \<box>P" by (rule alw)
qed

section "Lemmas with @{text \<circ>}"

lemma nex_pax2: assumes H: "\<turnstile> F \<longrightarrow> G"
  shows "|~ \<circ>F \<longrightarrow> \<circ>G"
proof -
  from H have g1: "|~ \<circ>(F \<longrightarrow> G)" 
    by (rule nex)
  have "|~ \<circ>(F \<longrightarrow> G) \<longrightarrow> \<circ>F \<longrightarrow> \<circ>G"
    by (rule pax2)
  with g1 show ?thesis by (rule pmp)
qed

lemma next_and: "|~ (\<circ>(F \<and> G)) = (\<circ>F \<and> \<circ>G)"
proof (rule pref_iffI)
  show "|~ \<circ>(F \<and> G) \<longrightarrow> \<circ>F \<and> \<circ>G"
  proof (rule prefI)
    fix w
    have "w \<Turnstile> (\<circ>(F \<and> G)) = (\<circ>F \<and> \<circ>G)"
      by (rule T8[THEN prefD])
    thus "w \<Turnstile> \<circ>(F \<and> G) \<longrightarrow> \<circ>F \<and> \<circ>G"
      by auto
  qed
next
  show "|~ \<circ>F \<and> \<circ>G \<longrightarrow> \<circ>(F \<and> G)"
  proof (rule prefI)
    fix w
    have "w \<Turnstile> (\<circ>(F \<and> G)) = (\<circ>F \<and> \<circ>G)"
      by (rule T8[THEN prefD])
    thus "w \<Turnstile> \<circ>F \<and> \<circ>G \<longrightarrow> \<circ>(F \<and> G)"
      by auto
  qed
qed

lemma next_or: "|~ (\<circ>(F \<or> G)) = (\<circ>F \<or> \<circ>G)"
proof (rule pref_iffI)
  show "|~ \<circ>(F \<or> G) \<longrightarrow> \<circ>F \<or> \<circ>G"
  proof (pref_unlift,rule impI)
    fix w
    assume a1: "w \<Turnstile> \<circ>(F \<or> G)"
    have "(w \<Turnstile> \<not>\<circ>F) \<longrightarrow> (w \<Turnstile> \<circ>G)"
    proof
      assume a2: "w \<Turnstile> \<not>\<circ>F"
      have "|~ (\<circ>\<not>F) = (\<not> \<circ>F)" by (rule pax1)
      from this[THEN prefD] have "w \<Turnstile> \<not>\<circ>F \<longrightarrow> \<circ>\<not>F"
	by auto
      with a2 have a2': "w \<Turnstile> \<circ>\<not>F" by simp
      have "\<turnstile> (F \<or> G) \<and> \<not>F \<longrightarrow> G" 
	by (auto simp add: Valid_def)
      hence g1: "|~ \<circ>((F \<or> G) \<and> \<not>F \<longrightarrow> G)" 
	by (rule nex)
      have "|~ \<circ>((F \<or> G) \<and> \<not>F \<longrightarrow> G) \<longrightarrow> \<circ>((F \<or> G) \<and> \<not>F) \<longrightarrow> \<circ>G"
	by (rule pax2)
      with g1 have g2: "|~ \<circ>((F \<or> G) \<and> \<not>F) \<longrightarrow> \<circ>G"
	by (rule pmp)
      have "|~ (\<circ>((F \<or> G) \<and> \<not>F)) = (\<circ>(F \<or> G) \<and> \<circ>\<not>F)" 
	by (rule T8)
      from this[THEN prefD]
      have "w \<Turnstile> \<circ>(F \<or> G) \<and> \<circ>\<not>F \<longrightarrow> \<circ>((F \<or> G) \<and> \<not>F)"
	by auto
      with a1 a2' have "w \<Turnstile> \<circ>((F \<or> G) \<and> \<not>F)"
	by auto
      with g2[THEN prefD] show "w \<Turnstile> \<circ>G"
	by auto
    qed
    thus "(w \<Turnstile> \<circ>F) \<or> (w \<Turnstile> \<circ>G)" by auto 
  qed
next
  show "|~ \<circ>F \<or> \<circ>G \<longrightarrow> \<circ>(F \<or> G)"
  proof -
    have "\<turnstile> \<not>(F \<or> G) \<longrightarrow> \<not>F \<and> \<not>G" by auto
    hence g1: "|~ \<circ>\<not>(F \<or> G) \<longrightarrow> \<circ>(\<not>F \<and> \<not>G)"
      by (rule nex_pax2)
    have "|~ (\<circ>(\<not>F \<and> \<not>G)) = (\<circ>\<not>F \<and> \<circ>\<not>G)" 
      by (rule T8)
    from this[THEN prefD] have g2: "|~ \<circ>(\<not>F \<and> \<not>G) \<longrightarrow> \<circ>\<not>F \<and> \<circ>\<not>G"
      by auto
    from pax1[of F,THEN prefD] have tg1: "|~ \<circ>\<not>F \<longrightarrow> \<not>\<circ>F"
      by auto
    from pax1[of G,THEN prefD] have "|~ \<circ>\<not>G \<longrightarrow> \<not>\<circ>G"
      by auto
    from this[THEN prefD] tg1[THEN prefD] g2[THEN prefD]
    have "|~ \<circ>(\<not>F \<and> \<not>G) \<longrightarrow> \<not>\<circ>F \<and> \<not>\<circ>G" by auto
    from this[THEN prefD] have  "|~ \<circ>(\<not>F \<and> \<not>G) \<longrightarrow> \<not>(\<circ>F \<or> \<circ>G)"
      by auto
    from this[THEN prefD] g1[THEN prefD] 
    have g3: "|~ \<circ>\<not>(F \<or> G) \<longrightarrow> \<not>(\<circ>F \<or> \<circ>G)" by auto
    have "|~ (\<circ>\<not>(F \<or> G)) = (\<not>\<circ>(F \<or> G))" by (rule pax1)
    from this[THEN prefD] have "|~ \<not>\<circ>(F \<or> G) \<longrightarrow> \<circ>\<not>(F \<or> G)"
      by auto
    from this[THEN prefD] g3[THEN prefD]
    have "|~ \<not>\<circ>(F \<or> G) \<longrightarrow> \<not>(\<circ>F \<or> \<circ>G)" by auto
    from this[THEN prefD] show ?thesis by auto
  qed
qed

lemma next_imp: "|~ (\<circ>(F \<longrightarrow> G)) = (\<circ>F \<longrightarrow> \<circ>G)"
proof (rule pref_iffI)
  show "|~ \<circ>(F \<longrightarrow> G) \<longrightarrow> \<circ>F \<longrightarrow> \<circ>G"
    by (rule pax2)
next
  show "|~ (\<circ>F \<longrightarrow> \<circ>G) \<longrightarrow> \<circ>(F \<longrightarrow> G)"
  proof -
    have "|~ (\<circ>\<not>F \<or> \<circ>G) \<longrightarrow> \<circ>(\<not>F \<or> G)"
      by (rule next_or[THEN pref_iffD2])
    with pax1[of F,THEN prefeq_reflection]
    have "|~ (\<not>\<circ>F \<or> \<circ>G) \<longrightarrow> \<circ>(\<not>F \<or> G)"
      by auto
    from this[THEN prefD]
    have g1: "|~ (\<circ>F \<longrightarrow> \<circ>G) \<longrightarrow> \<circ>(\<not>F \<or> G)"
      by auto
    have "\<turnstile> \<not>F \<or> G \<longrightarrow> F \<longrightarrow> G"
      by auto
    hence g2: "|~ \<circ>(\<not>F \<or> G \<longrightarrow> F \<longrightarrow> G)"
      by (rule nex)
    have "|~ \<circ>(\<not>F \<or> G \<longrightarrow> F \<longrightarrow> G) \<longrightarrow> \<circ>(\<not>F \<or> G) \<longrightarrow> \<circ>(F \<longrightarrow> G)"
      by (rule pax2)
    with g2 have "|~ \<circ>(\<not>F \<or> G) \<longrightarrow> \<circ>(F \<longrightarrow> G)"
      by auto
    with g1 show ?thesis by auto
  qed
qed

lemma next_not: "|~ (\<circ>\<not>F) = (\<not>\<circ>F)"
  by (rule pax1)

lemma next_eq: "|~ (\<circ>(F = G)) = ((\<circ>F) = (\<circ>G))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_noteq: "|~ (\<circ>(F \<noteq> G)) = ((\<circ>F) \<noteq> (\<circ>G))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_const: "|~ (\<circ>#c) = #c"
  by (pref_unlift)(auto simp add: nexts_def)

lemma next_fun1: "|~ (\<circ>f<x>) = (f<\<circ>x>)"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_fun2: "|~ (\<circ>f<x,y>) = (f<\<circ>x,\<circ>y>)"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_fun3: "|~ (\<circ>f<x,y,z>) = (f<\<circ>x,\<circ>y,\<circ>z>)"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_fun4: "|~ (\<circ>f<x,y,z,zz>) = (f<\<circ>x,\<circ>y,\<circ>z,\<circ>zz>)"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_forall: "|~ (\<circ>(\<forall> x. P x)) = (\<forall> x. \<circ>(P x))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_exists: "|~ (\<circ>(\<exists> x. P x)) = (\<exists> x. \<circ>(P x))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_exists1: "|~ (\<circ>(\<exists>! x. P x)) = (\<exists>! x. \<circ>(P x))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_le: "|~ (\<circ>(x < y)) = ((\<circ>x) < (\<circ>y))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_le_eq: "|~ (\<circ>(x \<le> y)) = ((\<circ>x) \<le> (\<circ>y))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_mem: "|~ (\<circ>(x \<in> X)) = ((\<circ>x) \<in> (\<circ>X))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_notmem: "|~ (\<circ>(x \<notin> X)) = ((\<circ>x) \<notin> (\<circ>X))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_if: "|~ (\<circ>(if T then A else B)) = (if \<circ>T then \<circ>A else \<circ>B)"
  by (pref_unlift)(auto simp add: nexts_def)

lemma next_plus: "|~ (\<circ>(x + y)) = ((\<circ>x) + (\<circ>y))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_minus: "|~ (\<circ>(x - y)) = ((\<circ>x) - (\<circ>y))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_times: "|~ (\<circ>(x * y)) = ((\<circ>x) * (\<circ>y))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_div: "|~ (\<circ>(x div y)) = ((\<circ>x) div (\<circ>y))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_mod: "|~ (\<circ>(x mod y)) = ((\<circ>x) mod (\<circ>y))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_finset: "|~ (\<circ>{X}) = {\<circ>X}"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_pair: "|~ (\<circ>(F,G)) = (\<circ>F,\<circ>G)"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_cons: "|~ (\<circ>(x#xs)) = ((\<circ>x)#(\<circ>xs))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_app: "|~ (\<circ>(xs@ys)) = ((\<circ>xs)@(\<circ>ys))"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_list: "|~ (\<circ>[L]) = [\<circ>L]"
  by (pref_unlift) (auto simp add: nexts_def)

lemma next_always: "|~ \<box>F \<longrightarrow> \<circ>\<box>F" by (rule pax3)


text {*  Theorem @{text next_eventually} should not be used
  "blindly"  *}
lemma next_eventually: assumes h: "stutinv F"
                        shows  "|~ \<diamond>F \<longrightarrow> \<not>F \<longrightarrow> \<circ>\<diamond>F"
proof (pref_unlift, intro impI)
  fix w
  assume a1: "w \<Turnstile> \<diamond>F"
    and  a2: "\<not>(w \<Turnstile> F)"
  show "w \<Turnstile> \<circ>\<diamond>F"
  proof (rule ccontr)
    assume a3: "\<not>(w \<Turnstile> \<circ>\<diamond>F)"
    from h[THEN stut_not[THEN pre_id_unch]] 
      have "|~ (\<box>\<not>F) = (\<not>F \<and> \<circ>\<box>\<not>F)"
        by (rule T7)
    hence "(w \<Turnstile> \<not>F) \<and> (w \<Turnstile> \<circ>\<box>\<not>F) \<longrightarrow> (w \<Turnstile> \<box>\<not>F)"
      by auto
    with a2 have g1: "(w \<Turnstile> \<circ>\<box>\<not>F) \<longrightarrow> (w \<Turnstile> \<box>\<not>F)"
      by auto
    have "|~ (\<circ>\<not>\<box>\<not>F) = (\<not>\<circ>\<box>\<not>F)" by (rule pax1)
    from this[THEN prefD] have "|~ \<not>\<circ>\<box>\<not>F \<longrightarrow> \<circ>\<not>\<box>\<not>F"
      by auto
    from this[THEN prefD] have "|~ \<not>\<circ>\<not>\<box>\<not>F \<longrightarrow> \<not>\<not>\<circ>\<box>\<not>F"
      by auto   
    from this[THEN prefD] have g2: "|~ \<not>\<circ>\<not>\<box>\<not>F \<longrightarrow> \<circ>\<box>\<not>F"
      by auto   
    from a3 have "w \<Turnstile> \<not>\<circ>\<not>\<box>\<not>F" 
      by (simp add: eventually_def)
    with g2[THEN prefD] have "w \<Turnstile> \<circ>\<box>\<not>F"
      by auto
    with g1 have g3: "w \<Turnstile> \<box>\<not>F" by auto
    from a1 have "\<not>(w \<Turnstile> \<box>\<not>F)"
      by (simp add: eventually_def) 
    with g3 show False by simp
  qed
qed

lemmas next_all = next_and next_or next_imp next_not next_eq
  next_noteq next_const next_fun1 next_fun2 next_fun3 next_fun4 next_forall
  next_exists next_exists1 next_le next_le_eq next_mem next_notmem
  next_if next_plus next_minus next_times next_div next_mod next_finset
  next_pair next_cons next_app next_list

lemmas next_all_unl = next_all[THEN prefD]

lemmas next_all_eq = next_all[THEN prefeq_reflection]

lemma next_action: "|~ \<box>[P]_v \<longrightarrow> \<circ>\<box>[P]_v"
proof -
  from pax4[of P v,THEN prefD] show ?thesis by auto
qed

section "Higher Level Derived Rules"

theorem TLA1: assumes H: "|~ P \<and> Unchanged f \<longrightarrow> \<circ>P" shows "\<turnstile> (\<box>P) = (P \<and> \<box>[P \<longrightarrow> \<circ>P]_f)"
proof (int_unlift)
  fix w
  show "(w \<Turnstile> \<box>P) = ((w \<Turnstile> P) \<and> (w \<Turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_f))"
  proof
    assume a1: "(w \<Turnstile> P) \<and> (w \<Turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_f)"
    show "w \<Turnstile> \<box>P"
    proof -
      from H have "w \<Turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_f \<longrightarrow> P \<longrightarrow> \<box>P" by (rule ax3[THEN intD])
      hence "(w \<Turnstile> P) \<and> (w \<Turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_f) \<longrightarrow> (w \<Turnstile> \<box>P)" by auto
      with a1 show ?thesis by auto
    qed
  next
    assume a1: "w \<Turnstile> \<box>P"
    show "(w \<Turnstile> P) \<and> (w \<Turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_f)"
    proof
      from ax1[THEN intD] have "(w \<Turnstile> \<box>P) \<longrightarrow> (w \<Turnstile> P)" by auto
      with a1 show "w \<Turnstile> P" by simp
    next
      from a1 P4[THEN intD] have g1: "w \<Turnstile> \<box>[P]_f" by auto
      from T6[THEN intD] a1 have g2: "w \<Turnstile> \<box>[\<circ>P]_f" by auto
      have "|~ P \<longrightarrow> \<circ>P \<longrightarrow> (P \<longrightarrow> \<circ>P)" by auto
      hence tmp: "\<turnstile> \<box>[(P \<longrightarrow> \<circ>P \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f" by (rule sq)
      from ax4[THEN intD] have
	"(w \<Turnstile> \<box>[(P \<longrightarrow> \<circ>P \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f) \<longrightarrow> (w \<Turnstile> \<box>[P]_f) \<longrightarrow> (w \<Turnstile> \<box>[(\<circ>P \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f)"
	by auto
      with tmp[THEN intD] g1 have tmp: "w \<Turnstile> \<box>[(\<circ>P \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f" by simp
      from ax4[THEN intD] have
	"(w \<Turnstile> \<box>[(\<circ>P \<longrightarrow> (P \<longrightarrow> \<circ>P))]_f) \<longrightarrow> (w \<Turnstile> \<box>[\<circ>P]_f) \<longrightarrow> (w \<Turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_f)"
	by auto      
      with tmp g2 show "w \<Turnstile> \<box>[(P \<longrightarrow> \<circ>P)]_f" by auto
    qed
  qed
qed

theorem TLA2: assumes h1: "\<turnstile> P \<longrightarrow> Q"
                and   h2: "|~ ([A]_f) \<longrightarrow> ([B]_g)"
               shows      "\<turnstile> \<box>P \<and> \<box>[A]_f \<longrightarrow> \<box>Q \<and> \<box>[B]_g" 
proof (int_unlift,intro impI,elim conjE,intro conjI)
  fix w
  assume a1: "w \<Turnstile> \<box>P"
  from h1 have "\<turnstile> \<box>P \<longrightarrow> \<box>Q" by (intro STL4)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>Q" 
    by auto
next
  fix w
  assume a1: "w \<Turnstile> \<box>[A]_f"
  from h2 have g1: "\<turnstile> \<box>[[A]_f]_g \<longrightarrow> \<box>[[B]_g]_g" by (intro MM8)
  have tg2: "\<turnstile> (\<box>[[B]_g]_g) = (\<box>[B]_g)" by (rule T3)
  from this[THEN intD] have "\<turnstile> \<box>[[B]_g]_g \<longrightarrow> \<box>[B]_g"
    by auto
  with g1 have g2: "\<turnstile> \<box>[[A]_f]_g \<longrightarrow> \<box>[B]_g"
    by (rule lift_imp_trans)
  have "\<turnstile> \<box>[[A]_g]_f \<longrightarrow> \<box>[[A]_f]_g" by (rule T5)
  from this g2 have g3: "\<turnstile> \<box>[[A]_g]_f \<longrightarrow> \<box>[B]_g"
    by (rule lift_imp_trans)
  have "|~ A \<longrightarrow> [A]_g"
    by (auto simp add: actrans_def)
  hence "\<turnstile> \<box>[A]_f \<longrightarrow> \<box>[[A]_g]_f" by (rule MM8)
  from this g3 have "\<turnstile> \<box>[A]_f \<longrightarrow> \<box>[B]_g"
    by (rule lift_imp_trans)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>[B]_g"
    by auto
qed

lemma mTLA2: assumes h1: "|~ ([A]_f) \<longrightarrow> [B]_g"
               shows     "\<turnstile> \<box>[A]_f \<longrightarrow> \<box>[B]_g"
proof -
  from h1 have g1: "\<turnstile> \<box>[[A]_f]_g \<longrightarrow> \<box>[[B]_g]_g" by (intro MM8)
  have tg2: "\<turnstile> (\<box>[[B]_g]_g) = (\<box>[B]_g)" by (rule T3)
  from this[THEN intD] have "\<turnstile> \<box>[[B]_g]_g \<longrightarrow> \<box>[B]_g"
    by auto
  with g1 have g2: "\<turnstile> \<box>[[A]_f]_g \<longrightarrow> \<box>[B]_g"
    by (rule lift_imp_trans)
  have "\<turnstile> \<box>[[A]_g]_f \<longrightarrow> \<box>[[A]_f]_g" by (rule T5)
  from this g2 have g3: "\<turnstile> \<box>[[A]_g]_f \<longrightarrow> \<box>[B]_g"
    by (rule lift_imp_trans)
  have "|~ A \<longrightarrow> [A]_g"
    by (auto simp add: actrans_def)
  hence "\<turnstile> \<box>[A]_f \<longrightarrow> \<box>[[A]_g]_f" by (rule MM8)
  from this g3 show "\<turnstile> \<box>[A]_f \<longrightarrow> \<box>[B]_g"
    by (rule lift_imp_trans)
qed

lemma mTLA2_split: assumes H1: "|~ A \<longrightarrow> [B]_g"
  and H2: "|~ Unchanged f \<longrightarrow> [B]_g"
  shows "|~ ([A]_f) \<longrightarrow> [B]_g"
proof (unfold actrans_def,pref_unlift,rule impI,elim disjE)
  fix w
  assume "w \<Turnstile> A"
  with H1[THEN prefD] show "(w \<Turnstile> B) \<or> (w \<Turnstile> Unchanged g)"
    by (auto simp add: actrans_def)
next
  fix w
  assume "w \<Turnstile> Unchanged f"
  with H2[THEN prefD] show "(w \<Turnstile> B) \<or> (w \<Turnstile> Unchanged g)"
    by (auto simp add: actrans_def)
qed

theorem INV1: assumes H: "|~ I \<and> [N]_f \<longrightarrow> \<circ>I" shows "\<turnstile> I \<and> \<box>[N]_f \<longrightarrow> \<box>I" 
proof int_unlift
  fix w 
  show "(w \<Turnstile> I) \<and> (w \<Turnstile> \<box>[N]_f) \<longrightarrow> (w \<Turnstile> \<box>I)" 
  proof (rule impI,erule conjE) 
    assume a1: "w \<Turnstile> I" 
      and  a2: "w \<Turnstile> \<box>[N]_f"
    have "|~ I \<and> Unchanged f \<longrightarrow> \<circ>I" 
    proof (rule prefI)  
      fix w 
      from H[THEN prefD] show "w \<Turnstile> I \<and> Unchanged f \<longrightarrow> \<circ>I"  
	by (auto simp add: actrans_def)
    qed
    hence g1: "\<turnstile> \<box>[(I \<longrightarrow> \<circ>I)]_f \<longrightarrow> I \<longrightarrow> \<box>I" by (rule ax3)
    from H have "|~ [N]_f \<longrightarrow> I \<longrightarrow> \<circ>I" by auto
    hence tg1: "\<turnstile> \<box>[([N]_f \<longrightarrow> I \<longrightarrow> \<circ>I)]_f" by (rule sq)
    have "\<turnstile> \<box>[([N]_f \<longrightarrow> I \<longrightarrow> \<circ>I)]_f \<longrightarrow> \<box>[[N]_f]_f \<longrightarrow> \<box>[(I \<longrightarrow> \<circ>I)]_f"
      by (rule ax4)
    from this[THEN intD] tg1[THEN intD]
      have tg1: "\<turnstile> \<box>[[N]_f]_f \<longrightarrow> \<box>[(I \<longrightarrow> \<circ>I)]_f" by auto
    from a2 T3[THEN intD] have "w \<Turnstile> \<box>[[N]_f]_f" by auto
    with tg1[THEN intD] have "w \<Turnstile> \<box>[(I \<longrightarrow> \<circ>I)]_f" by auto
    with g1 a1 show "w \<Turnstile> \<box>I" by auto
  qed
qed

theorem INV2: "\<turnstile> \<box>I \<longrightarrow> ((\<box>[N]_f) = (\<box>[N \<and> I \<and> \<circ>I]_f))"
proof (int_unlift,rule impI)
  fix w
  assume a1: "w \<Turnstile> \<box>I"
  show "(w \<Turnstile> \<box>[N]_f) = (w \<Turnstile> \<box>[(N \<and> I \<and> \<circ>I)]_f)"
  proof
    assume a2: "w \<Turnstile> \<box>[N]_f"
    have "|~ N \<longrightarrow> I \<longrightarrow> \<circ>I \<longrightarrow> (N \<and> I \<and> \<circ>I)" by auto
    hence "\<turnstile> \<box>[(N \<longrightarrow> I \<longrightarrow> \<circ>I \<longrightarrow> (N \<and> I \<and> \<circ>I))]_f" by (rule sq)
    hence g1: "w \<Turnstile> \<box>[(N \<longrightarrow> I \<longrightarrow> \<circ>I  \<longrightarrow> (N \<and> I \<and> \<circ>I))]_f" by auto
    from ax4[THEN intD] have 
      "(w \<Turnstile> \<box>[(N \<longrightarrow> I \<longrightarrow> \<circ>I \<longrightarrow> (N \<and> I \<and> \<circ>I))]_f) \<longrightarrow> (w \<Turnstile> \<box>[N]_f) 
      \<longrightarrow> (w \<Turnstile> \<box>[(I \<longrightarrow> \<circ>I \<longrightarrow> (N \<and> I \<and> \<circ>I))]_f)"
        by auto
    with g1 a2 have g2: "w \<Turnstile> \<box>[(I \<longrightarrow> \<circ>I \<longrightarrow> (N \<and> I \<and> \<circ>I))]_f" by auto
    from P4[THEN intD] have "(w \<Turnstile> \<box>I) \<longrightarrow> (w \<Turnstile> \<box>[I]_f)" by auto
    with a1 have g3: "w \<Turnstile> \<box>[I]_f" by auto
    from ax4[THEN intD] have 
      "(w \<Turnstile> \<box>[(I \<longrightarrow> \<circ>I \<longrightarrow> (N \<and> I \<and> \<circ>I))]_f) \<longrightarrow> (w \<Turnstile> \<box>[I]_f) \<longrightarrow> (w \<Turnstile> \<box>[(\<circ>I \<longrightarrow> (N \<and> I \<and> \<circ>I))]_f)"
      by auto
    with g2 g3 have g4: "(w \<Turnstile> \<box>[(\<circ>I \<longrightarrow> (N \<and> I \<and> \<circ>I))]_f)" by auto
    from ax4[THEN intD] have 
      "(w \<Turnstile> \<box>[(\<circ>I \<longrightarrow> (N \<and> I \<and> \<circ>I))]_f) \<longrightarrow> (w \<Turnstile> \<box>[\<circ>I]_f) \<longrightarrow> (w \<Turnstile> \<box>[(N \<and> I \<and> \<circ>I)]_f)"
      by auto
    with g4 have g5: "(w \<Turnstile> \<box>[\<circ>I]_f) \<longrightarrow> (w \<Turnstile> \<box>[(N \<and> I \<and> \<circ>I)]_f)" by auto
    from a1 T6[THEN intD] have "w \<Turnstile> \<box>[\<circ>I]_f" by auto
    with g5 show "w \<Turnstile> \<box>[(N \<and> I \<and> \<circ>I)]_f" by simp
  next
    assume a2: "w \<Turnstile> \<box>[(N \<and> I \<and> \<circ>I)]_f"
    have "|~ (N \<and> I \<and> \<circ>I) \<longrightarrow> N" by auto
    hence "\<turnstile> \<box>[((N \<and> I \<and> \<circ>I) \<longrightarrow> N)]_f" by (rule sq)
    from this[THEN intD, of w] ax4[THEN intD]
      have "(w \<Turnstile> \<box>[(N \<and> I \<and> \<circ>I)]_f) \<longrightarrow> (w \<Turnstile> \<box>[N]_f)" by auto
    with a2 show "w \<Turnstile> \<box>[N]_f" by auto
  qed
qed

lemma R1: assumes H: "|~ Unchanged w \<longrightarrow> Unchanged v" 
           shows     "\<turnstile> \<box>[F]_w \<longrightarrow> \<box>[F]_v"
proof -
  from H have "|~ [F]_w \<longrightarrow> [F]_v" 
    by (auto simp add: actrans_def)
  thus ?thesis by (rule mTLA2)
qed

lemma R2: assumes h1: "|~ Unchanged w \<longrightarrow> Unchanged v"
            and   h2: "\<turnstile> \<box>[P]_w" 
            and   h3: "\<turnstile> \<box>[P \<longrightarrow> Q]_w"
           shows      "\<turnstile> \<box>[Q]_v"
proof (int_unlift)
  fix s
  from h2 h3 have tmp: "\<turnstile> \<box>[Q]_w" by (rule H1)
  from h1 have "\<turnstile> \<box>[Q]_w \<longrightarrow> \<box>[Q]_v" by (rule R1)
  from tmp[THEN intD] this[THEN intD] show "s \<Turnstile> \<box>[Q]_v" by auto
qed

lemma act1: assumes h1: "|~ Unchanged w \<longrightarrow> Unchanged v"
             and    h2: "\<turnstile> \<box>[(P \<longrightarrow> [Q]_v)]_w"
            shows       "\<turnstile> \<box>[P]_w \<longrightarrow> \<box>[Q]_v"
proof (int_unlift,rule impI)
  fix s
  assume a1: "s \<Turnstile> \<box>[P]_w"
  from ax4[THEN intD] have 
    "(s \<Turnstile> \<box>[(P \<longrightarrow> [Q]_v)]_w) \<longrightarrow> (s \<Turnstile> \<box>[P]_w) \<longrightarrow> (s \<Turnstile> \<box>[[Q]_v]_w)"
      by auto
    with h2[THEN intD] a1 have g1: "s \<Turnstile> \<box>[[Q]_v]_w" by auto
    from h1 have "\<turnstile> \<box>[[Q]_v]_w \<longrightarrow> \<box>[[Q]_v]_v" by (rule R1)
    from this[THEN intD] g1 have "(s \<Turnstile> \<box>[[Q]_v]_v)" by auto
    with T3[THEN intD] show "s \<Turnstile> \<box>[Q]_v" by auto
qed

theorem invmono: assumes h1: "\<turnstile> I \<longrightarrow> P"
                   and   h2: "|~ P \<and> [N]_f \<longrightarrow> \<circ>P" 
                  shows      "\<turnstile> I \<and> \<box>[N]_f \<longrightarrow> \<box>P"
proof -
  from h2 have "\<turnstile> P \<and> \<box>[N]_f \<longrightarrow> \<box>P" by (rule INV1)
  from this[THEN intD] h1[THEN intD]
  show ?thesis by auto
qed

theorem preimpsplit: assumes h1: "|~ I \<and> N \<longrightarrow> Q"
                       and   h2: "|~ I \<and> Unchanged v \<longrightarrow> Q"
                      shows      "|~ I \<and> ([N]_v) \<longrightarrow> Q"
proof -
  have "|~ (N \<or> Unchanged v) = ([N]_v)"
    by (auto simp add: actrans_def)
  from this[THEN prefD] h1[THEN prefD] h2[THEN prefD]
    show ?thesis by auto
qed

theorem preinvsplit: assumes h1: "|~ I \<and> N \<longrightarrow> \<circ>I"
  and   h2: "|~ I \<and> Unchanged v \<longrightarrow> \<circ>I"
  shows      "|~ I \<and> [N]_v \<longrightarrow> \<circ>I"
  using h1 h2 by (rule preimpsplit)


theorem refinement1: 
  assumes h1: "\<turnstile> P \<longrightarrow> Q"
  and  h2: "|~ ([A]_f) \<longrightarrow> [B]_g"
  shows     "\<turnstile> P \<and> \<box>[A]_f \<longrightarrow> Q \<and> \<box>[B]_g" 
proof (int_unlift,intro impI,elim conjE,intro conjI)
  fix w
  assume a1: "w \<Turnstile> P"
  from h1[THEN intD] a1 show "w \<Turnstile> Q" 
    by auto
next
  fix w
  assume a1: "w \<Turnstile> \<box>[A]_f"
  from h2 have g1: "\<turnstile> \<box>[[A]_f]_g \<longrightarrow> \<box>[[B]_g]_g" by (intro MM8)
  have tg2: "\<turnstile> (\<box>[[B]_g]_g) = (\<box>[B]_g)" by (rule T3)
  from this[THEN intD] have "\<turnstile> \<box>[[B]_g]_g \<longrightarrow> \<box>[B]_g"
    by auto
  with g1 have g2: "\<turnstile> \<box>[[A]_f]_g \<longrightarrow> \<box>[B]_g"
    by (rule lift_imp_trans)
  have "\<turnstile> \<box>[[A]_g]_f \<longrightarrow> \<box>[[A]_f]_g" by (rule T5)
  from this g2 have g3: "\<turnstile> \<box>[[A]_g]_f \<longrightarrow> \<box>[B]_g"
    by (rule lift_imp_trans)
  have "|~ A \<longrightarrow> [A]_g"
    by (auto simp add: actrans_def)
  hence "\<turnstile> \<box>[A]_f \<longrightarrow> \<box>[[A]_g]_f" by (rule MM8)
  from this g3 have "\<turnstile> \<box>[A]_f \<longrightarrow> \<box>[B]_g"
    by (rule lift_imp_trans)
  from this[THEN intD] a1 show "w \<Turnstile> \<box>[B]_g"
    by auto
qed


lemma refstep:   
  assumes h1: "|~ Unchanged v \<longrightarrow> Unchanged w"
  and h2: "|~ P \<longrightarrow> Q \<or> Unchanged w"
  shows  "|~ ([P]_v) \<longrightarrow> [Q]_w"
  using h1[THEN prefD] h2[THEN prefD]
  apply (unfold actrans_def)
  apply int_unlift
  apply auto
  done

theorem spec_inv2_mono:
  assumes h1: "\<turnstile> I \<and> \<box>[N]_v \<longrightarrow> \<box>J"
  and h2: "\<turnstile> I \<and> \<box>[N \<and> J \<and> \<circ>J]_v \<longrightarrow> P"
  shows "\<turnstile> I \<and> \<box>[N]_v \<longrightarrow> P"
proof (int_unlift,intro impI,elim conjE)
  fix w
  assume a1: "w \<Turnstile> I"
  and a2: "w \<Turnstile> \<box>[N]_v"
  from h1[THEN intD,of w] a1 a2 have g1: "w \<Turnstile> \<box>J"
    by simp
  with ax1[of J,THEN intD,of w] have "w \<Turnstile> J"
    by auto
  have "\<turnstile> \<box>J \<longrightarrow> \<box>[N]_v = \<box>[N \<and> J \<and> \<circ>J]_v"
    by (rule INV2)
  from this[THEN intD,of w] g1 have g3: "(w \<Turnstile> \<box>[N]_v)=(w \<Turnstile> \<box>[N \<and> J \<and> \<circ>J]_v)"
    by auto
  with a2 have g4: "w \<Turnstile> \<box>[N \<and> J \<and> \<circ>J]_v" by simp
  with h2[THEN intD] a1 show "w \<Turnstile> P"
    by auto
qed

section "Lemmas left to prove"

lemma linT: "\<turnstile> \<diamond>F \<and> \<diamond>G \<longrightarrow> (\<diamond>(F \<and> \<diamond>G) \<or> \<diamond>(G \<and> \<diamond>F))"
proof (int_unlift,intro impI,elim conjE)
  fix w
  assume a1: "w \<Turnstile> \<diamond>F"
    and  a2: "w \<Turnstile> \<diamond>G"
  show "(w \<Turnstile> \<diamond>(F \<and> \<diamond>G)) \<or> (w \<Turnstile> \<diamond>(G \<and> \<diamond>F))"
oops

lemma discT: "\<turnstile> \<box>(F \<longrightarrow> \<diamond>(\<not>F \<and> \<diamond>F)) \<longrightarrow> (F \<longrightarrow> \<box>\<diamond>F)"
proof (int_unlift,intro impI)
  fix w
  assume a1: "w \<Turnstile> \<box>(F \<longrightarrow> \<diamond>(\<not>F \<and> \<diamond>F))"
    and  a2: "w \<Turnstile> F" 
  show "w \<Turnstile> \<box>\<diamond>F"
oops

(* FIXME: Prove this *)
lemma LT9: "\<turnstile> (\<box>\<diamond>F \<longrightarrow> \<box>\<diamond>G) = (\<diamond>(F \<leadsto> G))"
proof (unfold leadsto_def)
oops

(* not valid
lemma LT27: assumes h1: "|~ \<diamond>F \<longrightarrow> \<circ>\<diamond>F"
              shows     "\<turnstile> F \<longrightarrow> \<box>\<diamond>F"
proof -
  from h1 have "\<turnstile> \<diamond>F \<longrightarrow> \<box>\<diamond>F"
    by (rule ax3_2)
  thus ?thesis by (rule LT26)
qed
*)

(* easily proven by using some lemmas further down *)
lemma AA4: "\<turnstile> \<box>[\<diamond>\<langle>Q\<rangle>_v]_w \<and> \<diamond>\<langle>P\<rangle>_w \<longrightarrow> \<diamond>\<langle>Q\<rangle>_v"
proof (int_unlift,intro impI,elim conjE)
  fix s
  assume a1: "s \<Turnstile> \<box>[\<diamond>\<langle>Q\<rangle>_v]_w"
    and  a2: "s \<Turnstile> \<diamond>\<langle>P\<rangle>_w"
  have "\<turnstile> \<box>[\<diamond>\<langle>Q\<rangle>_v]_w \<and> \<diamond>\<langle>P\<rangle>_w \<longrightarrow> \<diamond>\<langle>\<diamond>\<langle>Q\<rangle>_v\<rangle>_w"
    by (rule AA2)
  from this[THEN intD] a1 a2 have g1: "s \<Turnstile> \<diamond>\<langle>\<diamond>\<langle>Q\<rangle>_v\<rangle>_w"
    by auto
  have "\<turnstile> \<diamond>\<langle>\<diamond>\<langle>Q\<rangle>_v\<rangle>_w \<longrightarrow> \<diamond>\<diamond>\<langle>Q\<rangle>_v" by (rule E25)
oops

lemma assumes h1: "|~ F \<longrightarrow> \<circ>\<diamond>F"
        shows     "\<turnstile> F \<longrightarrow> \<box>\<diamond>F"
proof (int_unlift,intro impI)
  fix w
  assume a1: "w \<Turnstile> F"
  from h1[THEN prefD] have "|~ F \<longrightarrow> \<diamond>F \<longrightarrow> \<circ>\<diamond>F"
    by auto
  hence g1: "\<turnstile> \<box>[(F \<longrightarrow> \<diamond>F \<longrightarrow> \<circ>\<diamond>F)]_v" by (rule sq)
  have "\<turnstile> \<box>[(F \<longrightarrow> \<diamond>F \<longrightarrow> \<circ>\<diamond>F)]_v \<longrightarrow> \<box>[F]_v \<longrightarrow> \<box>[(\<diamond>F \<longrightarrow> \<circ>\<diamond>F)]_v"
    by (rule ax4)
  with g1 have "\<turnstile> \<box>[F]_v \<longrightarrow> \<box>[(\<diamond>F \<longrightarrow> \<circ>\<diamond>F)]_v" by (rule fmp)
oops

(*
subsection "@{text \<circ>} in State Functions"

lemma stfun_next_eq: "|~ F = (\<circ>#G) \<longrightarrow> F = #G"
  by (pref_unlift) (auto simp add: nexts_def)

lemma stfun_next_eq_op: "|~ F = #G \<longrightarrow>  F = (\<circ>#G)"
  by (pref_unlift) (auto simp add: nexts_def)

lemma stfun_next_noteq: "|~ F \<noteq> (\<circ>#G) \<longrightarrow> F \<noteq> #G"
  by (pref_unlift) (auto simp add: nexts_def)

lemma stfun_next_noteq_op: "|~ F \<noteq> #G \<longrightarrow> F \<noteq> (\<circ>#G)"
  by (pref_unlift) (auto simp add: nexts_def)


lemma "|~ (\<circ>(X,Y)) = (\<circ>X,\<circ>Y)"
  apply pref_unlift
  apply (unfold nexts_def)
  apply auto
  done
*)

(* FIXME: also for before/after lemmas
lemmas next_dist = next_imp next_and next_or
  next_not next_const next_fun1 next_fun2 
  next_fun3 next_forall next_exists next_always
  next_action

thm next_dist

lemmas next_dist_op = next_imp_op next_and_op next_or_op
  next_not_op next_const_op next_fun1_op next_fun2_op 
  next_fun3_op next_forall_op next_exists_op

thm next_dist_op
*)

end
